

/**
 * Static class with methods that return a "number" value casted
 * from another type. Cast to and from the same type acts as
 * a JRelix valid range checker.
 *
 * DETAILS NOT YET IMPLEMENTED.
 *
 * @since jRelCaNu 2002-07
 */


import java.util.*;


public class CastToNum implements Constants, TypeConstants {  
    //static private final number RESULT_TYPE_DC = NUMBER_DC;
    //static private final number RESULT_TYPE_DK = NUMBER_DK;
    static private final number RESULT_TYPE_DC = null;
    static private final number RESULT_TYPE_DK = null;
    public static number assertValidValue(double castFromNumber) 
        throws InterpretError {
        return value(castFromNumber);
    };

    /**
     * Checks to see if one JRelix typeNum.is castable to another.
     * We must provide a value() method for each compatible type.
     *
     * @return  a boolean true if the RHS type can be casted.
     * @since jRelCaNu
     */
    public static boolean domCompatible(int typeRHS) {
        switch (typeRHS) {
        case RELIX_VOID:
            return false;

        case BOOLEAN:
        case SHORT:
        case INTEGER:
        case LONG:
        case FLOAT:
        case DOUBLE:
        case NUMERIC:
        case STRING:
        case TEXT:
        case DC:	//added by Yi Zheng Sept 28 2002
        case DK:	//added by Yi Zheng Sept 28, 2002
            return true;

        case STMT:
        case EXPR:
        case IDLIST:
        case COMP:
 //       case DC:
 //       case DK:
        case RELATION:
        case VIEW:
        case COMPUTATION:
        case ATTRIBUTE:
        case UNIVERSAL:
        default:
            return false;
        }
    }

    /**
     * Returns a "number" value from a string.
     *
     * @return  a "number" corresponding to the first numeric
     *          string in the entire castFromStrn.
     *          Uses the standard Java methods to convert the
     *          numeric string to a number.
     * @since jRelCaNu
     */
    public static number value(String castFromStrn) 
        throws InterpretError {
        if (castFromStrn.equals(STRING_DK))
            return RESULT_TYPE_DK;
        else if (castFromStrn.equals(STRING_DC))
            return RESULT_TYPE_DC;
        number strNum = getFirstDecimal(new java.io.StringReader(castFromStrn));

        if (strNum == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DC.");
        if (strNum == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DK.");
        return strNum;
    }

    public static number value(byte castFromBool) 
        throws InterpretError {
        if (castFromBool == (byte) BOOL_DK)
            return RESULT_TYPE_DK;
        else if (castFromBool == (byte) BOOL_DC)
            return RESULT_TYPE_DC;
        else if (castFromBool == (byte) BOOL_FALSE) 
            return value(CastToStrn.value(0));
        else 
            return value(CastToStrn.value(1));
    }

    public static number value(long castFromLong) 
        throws InterpretError {
        if (castFromLong == LONG_DC)
            return RESULT_TYPE_DC;
        else if (castFromLong == LONG_DK)
            return RESULT_TYPE_DK;
        number strNum = value(CastToStrn.value(castFromLong));

        if (strNum == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DC.");
        if (strNum == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DK.");
        return strNum;
    }

    public static number value(int castFromInt) 
        throws InterpretError {
        if (castFromInt == INT_DC)
            return RESULT_TYPE_DC;
        else if (castFromInt == INT_DK)
            return RESULT_TYPE_DK;
        number strNum = value(CastToStrn.value(castFromInt));

        if (strNum == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DC.");
        if (strNum == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DK.");
        return strNum;
    }

    public static number value(short castFromShort) 
        throws InterpretError {
        if (castFromShort == SHORT_DC)
            return RESULT_TYPE_DC;
        else if (castFromShort == SHORT_DK)
            return RESULT_TYPE_DK;
        number strNum = value(CastToStrn.value(castFromShort));

        if (strNum == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DC.");
        if (strNum == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DK.");
        return strNum;
    }

    public static number value(float castFromFloat) 
        throws InterpretError {
        if (castFromFloat == FLOAT_DC)
            return RESULT_TYPE_DC;
        else if (castFromFloat == FLOAT_DK)
            return RESULT_TYPE_DK;
        number strNum = value(CastToStrn.value(castFromFloat));

        if (strNum == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DC.");
        if (strNum == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DK.");
        return strNum;
    }

    public static number value(double castFromDouble)
        throws InterpretError {
        if (castFromDouble == DOUBLE_DC)
            return RESULT_TYPE_DC;
        else if (castFromDouble == DOUBLE_DK)
            return RESULT_TYPE_DK;
        number strNum = value(CastToStrn.value(castFromDouble));

        if (strNum == RESULT_TYPE_DC)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DC.");
        if (strNum == RESULT_TYPE_DK)
            throw new InterpretError(
                    "CastToNum.value: error: Can't cast to RESULT_TYPE_DK.");
        return strNum;
    }

    public static number value(number castFromNumber) 
        throws InterpretError {
        throw new InterpretError(
                "CastToNum.value(): error: Can't cast from castFromNumber: Cast not yet implemented.");
    }

    /**
     MODIFIED FROM the Java Language Specification Second Edition:

     FloatingPointLiteral:
     Digits . Digits(opt) ExponentPart(opt) FloatTypeSuffix(opt)
     . Digits ExponentPart(opt) FloatTypeSuffix(opt)
     Digits ExponentPart FloatTypeSuffix(opt)
     Digits ExponentPart(opt) FloatTypeSuffix

     ExponentPart:
     ExponentIndicator SignedInteger

     ExponentIndicator: one of
     e E

     SignedInteger:
     Signopt Digits

     Sign: one of
     + -

     FloatTypeSuffix: one of
     f F d D l L
     */
    final static int  max_exp_digits = 4;
    final static int  max_num_digits = 8000;
    final static char plus = '+';
    final static char minus = '-';
    final static char exp_indicator = 'E';
    final static char decimalPoint = '.';
    static char buffer[] = null;
    static char exp_digits[] = null;
    private static number getFirstDecimal(java.io.Reader sr)
        throws InterpretError {

        int  cur_char = 0;
        char theSign[] = new char[1];
        int  digits_left_of_decimal = -1;
        int  num_digits = 0;
        int  num_leading_zeros = 0;

        if (buffer == null) 
            buffer = new char[max_num_digits];

        int exp = 0;
        int exp_sign = 1;
        int num_exp_digits = 0;

        // See class number:
        if (exp_digits == null) 
            exp_digits = new char[max_exp_digits];

        try {
            sr.reset();

            theSign[0] = plus;
            cur_char = skipToNextNumToken(sr, theSign);

            digits_left_of_decimal = -1;
            int num_digits_reads = 0;

            num_digits = 0;
                
            while ((cur_char = sr.read()) != -1) { 
                if (!Character.isDigit((char) cur_char)) {
                    if (cur_char == decimalPoint) {
                        if (digits_left_of_decimal != -1) {
                            // Already read a decimal point in this token.
                            // Current token is not a valid decimal number.
                            theSign[0] = plus;
                            if ((cur_char = skipToNextSpace(sr)) != -1)
                                cur_char = skipToNextNumToken(sr, theSign);
                            digits_left_of_decimal = -1;
                            num_digits_reads = 0;
                            num_digits = 0;
                            continue;
                        } else {
                            // Have not read a decimal point yet.
                            digits_left_of_decimal = num_digits;
                            // Always mark the last valid character parsed.
                            sr.mark(2);
                            continue;
                        }
                    } else if (num_digits != 0) {
                        // Already read in some number and encounterred non-digit;
                        // break out of the loop and try to return what we have.
                        break;
                    } else if (num_leading_zeros != 0 && num_digits == 0 && digits_left_of_decimal == -1) {
                        // '0 '
                        // back up to last valid character parsed
                        buffer[num_digits++] = (char) '0';
                        sr.reset();
                        continue; 
                    } else if (num_digits_reads > 0 && num_digits == 0 && digits_left_of_decimal == 0) {
                        // '0. '
                        buffer[num_digits++] = '0';
                        sr.reset();
                        continue; 
                    } else {
                        // ' . '
                        sr.mark(2);
                        theSign[0] = plus;
                        cur_char = skipToNextNumToken(sr, theSign);
                        digits_left_of_decimal = -1;
                        num_digits_reads = 0;
                        num_digits = 0;
                        continue;
                    }
                    // Not digit, not decimal point, haven't read a digit
                    //throw new InterpretError("Internal Parsing Error.");
                } else if ((char) cur_char == '0' && num_digits == 0 && 
                    digits_left_of_decimal == -1) {
                    sr.mark(2);
                    num_leading_zeros++;
                } else {
                    sr.mark(2);
                    buffer[num_digits++] = (char) cur_char;
                }
                num_digits_reads++;

            }

            // We didn't find a number.
            if (num_digits == 0) {
                // We've exhausted the string but did not find a number
                if ((cur_char != -1) && (cur_char = skipToNextSpace(sr)) != -1)
                    return getFirstDecimal(sr);
                throw new InterpretError("Not a castable number string: didn't find a number.");
            }

            if ((cur_char != -1) && (!Character.isWhitespace((char) cur_char))) {
                // non-terminating character
                if (Character.toUpperCase((char) cur_char) == exp_indicator) {
                    // non-terminating character is Exp
                    sr.mark(2);
                    if ((cur_char = sr.read()) == -1)
                        // Invalid terminating character ('e') after digits.
                        throw new InterpretError("Not a castable number string: invalid terminating 'e'.");
                    // We've not exhausted the string and did find a number
                    // We have some string left.
                    else if (cur_char == minus || cur_char == plus) { 
                        // Set a sign flag
                        sr.mark(2);
                        if (cur_char == minus)
                            exp_sign = -1;
                    }
                    sr.reset();
                    // non-terminating Exp {+,-}

                    num_exp_digits = 0; 
                    while ((cur_char = sr.read()) != -1 && num_exp_digits < max_exp_digits) { 
                        if (Character.isDigit((char) cur_char)) { 
                            exp_digits[num_exp_digits++] = (char) cur_char;
                            sr.mark(2);
                        } else { 
                            break;
                        } 
                    }
                    if (num_exp_digits == 0) { 
                        sr.reset();
                        if ((cur_char != -1) && (cur_char = skipToNextSpace(sr)) != -1)
                            return getFirstDecimal(sr);
                        else
                            throw new InterpretError("Not a castable number string: no exp digits.");
                    }
                }
            }

            if ((cur_char != -1) && (!Character.isWhitespace((char) cur_char))) {
                if (!(Character.toUpperCase((char) cur_char) == 'L' 
                        || Character.toUpperCase((char) cur_char) == 'D' 
                        || Character.toUpperCase((char) cur_char) == 'F')) {
                    if ((cur_char != -1) && (cur_char = skipToNextSpace(sr)) != -1)
                        return getFirstDecimal(sr);
                    else
                        throw new InterpretError("Not a castable number string: invalid number terminating character:" + cur_char);
                } else {
                    if (((cur_char = sr.read()) != -1) && (!Character.isWhitespace((char) cur_char))) 
                        if ((cur_char != -1) && (cur_char = skipToNextSpace(sr)) != -1)
                            return getFirstDecimal(sr);
                        else
                            throw new InterpretError("Not a castable number string: invalid terminating character.");
                }
            }

            exp = 0;
            for (int i = 0; i < num_exp_digits; i++) {
                exp *= 10;
                exp += Character.digit(exp_digits[i], 10);
            }
            exp *= exp_sign;

            if (digits_left_of_decimal == -1)
                exp += num_digits;
            else
                exp += digits_left_of_decimal;

            byte digits[] = new byte[num_digits];

            for (int i = 0; i < num_digits; i++) {
                digits[i] = (byte) Character.digit(buffer[i], 10);
            }
            
            return (new number(theSign[0], digits, exp));
        } catch (java.io.IOException e) {

            throw new InterpretError("error: Cannot convert string:" + e);
        }
    }

    static private int skipToNextToken(java.io.Reader sr)
        throws InterpretError {
        int cur_char = 0;

        try {
            while (((cur_char = sr.read()) != -1) 
                && (Character.isWhitespace((char) cur_char))) 
                sr.mark(2);
            return cur_char;
        } catch (java.io.IOException e) {
            throw new InterpretError("skipToNextToken" + e);
        }
    }

    static private int skipToNextSpace(java.io.Reader sr)
        throws InterpretError {
        int cur_char = 0;

        try {
            sr.mark(2);
            while (((cur_char = sr.read()) != -1) 
                && (!Character.isWhitespace((char) cur_char))) 
                sr.mark(2);
            return cur_char;
        } catch (java.io.IOException e) {
            throw new InterpretError("skipToNextToken" + e);
        }
    }

    static private int skipToNextNumToken(java.io.Reader sr, char Sign[])
        throws InterpretError {
        int cur_char = 0;

        try {
            sr.mark(2);
            do {
                // Skip to next token
                cur_char = skipToNextToken(sr);
                if (cur_char == -1)
                    return (-1);
                else {
                    // Start of token and not yet at the end of the string.
                    if (((Character.isDigit((char) cur_char))
                            || ((char) cur_char == '.'))) {
                        sr.reset();
                        break;
                    } else {
                        // Not a decimal start; see if string is signed.
                        if ((((char) cur_char == plus) || ((char) cur_char == minus))) {
                            char the_sign = (char) cur_char;

                            // Not a decimal start; start with sign.
                            sr.mark(2);
                            if ((cur_char = sr.read()) == -1) 
                                // start with sign; end of string encounterred.
                                return (-1);
                            else if (((Character.isDigit((char) cur_char))
                                    || ((char) cur_char == '.'))) {
                                Sign[0] = the_sign; 
                                sr.reset();
                                break;
                                // Not a digit start; sign followed by a digit.
                                // Break out of while loop.
                            }
                            // Not a digit start; sign followed by a non-digit.
                        }
                        // Not a digit start; token doesn't start with a number.
                        // Go to end of token and try next token in while loop.
                    }
                }
                cur_char = skipToNextSpace(sr);
                sr.mark(2);
            }
            while (cur_char != -1);

        } catch (java.io.IOException e) {
            throw new InterpretError("skipToNextToken" + e);
        }
        return (cur_char);
    }

    public static void main(String argv[]) 
        throws Exception {
        java.io.Reader in;

        if (argv.length == 0) {
            in = new java.io.BufferedReader(new java.io.InputStreamReader(System.in));
            while (true) {
                in.mark(2);
                System.out.println(CastToStrn.value(getFirstDecimal(in)));
            }
        } else {
            in = new java.io.BufferedReader(new java.io.FileReader(argv[0]));
            in.mark(2);
            System.out.println(CastToDouble.value(getFirstDecimal(in)));
            in = new java.io.FileReader(argv[0]);
            in.mark(2);
            System.out.println(CastToDouble.value(getFirstDecimal(in)));
        }
    }
}
