/**
 * Class Relation contains all the data/methods for a relation.
 *
 */

/**
 * Bugs :
 * 1.
 * Description :
 * getRelation() when there is no tuple for a surrogate. The
 * data field is not created which creates NullPointerException
 * when swapCollumns is invoked in the project method.
 * Source :
 * In domain algebra where the values for a nested attribute
 * are empty.
 * SOLVED!
 * 
 */
import java.util.*;
import java.io.*;

class Relation implements TypeConstants, Constants
//class Relation extends Relation76
{

//PATCHANGE:
// the system relation table and domain table shared by all relations
//   public static RelTable reltable;
//   public static DomTable domtable;

//debugging Yi Zheng
 public boolean DEBUG = false;
   protected Environment myEnv; //Pointer to the frame of the environment
					   //that this relation (or view or computation)
					   //belongs to. Hence the name myEnv.

   public String name;      // name of the relation
   public int active;       // state of the event handler, active = 0/1
   public int rvc;          // RELATION, VIEW or COMPUTATION
   public int numtuples;    // number of tuples
   public int numattrs;// number of attributes
   public int numsortattrs;      // number of sorted attributes
   public SimpleNode tree;   // reference to the root of AST

   public Domain[] domains;  // array of Domain objects
   public Object[] data; // array of columns, data of relation
   public int capacity;  // capacity of the array of column

   static final int BUFFERSIZE = 32768;   // 32K
   // START Since jRelixNS: optionally turn on Nested-Sigma join code.//Andy, May 12, 2002
   static final public boolean     isSigmaJoinImplemented = true;
   // END Since jRelixNS.



   //Constructor used (implicitly) by Computation
   public Relation()
   {
   }      
   // General constructor
   public Relation(String name, Domain[] domains, int rvc, int tuples,
		int attributes, int sort, SimpleNode root, Environment env)
   {
	 this.name = name;
         this.active = 0;
	 this.rvc = rvc;
	 this.numtuples = tuples;
	 this.numattrs = attributes;
	 this.numsortattrs = sort;
	 this.tree = root;
	 if (domains == null) {
	   this.domains = null;
	 } else {
	   this.domains = new Domain[domains.length];
	   for (int i = 0; i < domains.length; i++)
		 this.domains[i] = domains[i];
	 }
	 myEnv = env;
   }      
   // Constructor used for relation
   public Relation(String name, Domain[] domains, int tuples,
	 int attributes, int sort, Environment env)
   {
	 this(name, domains, RELATION, tuples, attributes,
	   sort, (SimpleNode) null, env);
   }      

   // Constructor for trigger
   public Relation (Environment env)  {
	 this.name = new String("");
	 myEnv = env;
   }
   /**
   * This method appends some tuples in the data array to the relation.
   * The caller of this method MUST make sure that the types of columns
   * of the data array are consistent with the domains of this relation.
   * Note also that the sort status is not changed. The caller should be
   * aware of this. This function is used for relational initialization
   * and assignAdd.
   * 
   * @param   dataArray   the 2D array of the tuple data
   * @param   tuples      the number of tuples to be appended
   */
   public void append(Object[] dataArray, int tuples)
   throws InterpretError
   {
	 // if append to empty relation, use block IO method save()
	 if (numtuples == 0) {
   data = dataArray;
   numtuples = tuples;
   
//   System.out.println("about to save(), numtuples = " + numtuples + " numattrs = " + numattrs);
   save();
   data = null; // reclaim memory //commented out by Yi Zheng June 3, 2002 //restored June 5, 2002
   return;
	 }
//	System.out.println("relation " + name + " has " + numtuples + " tuples before append");
	 RandomAccessFile rfile = null;
	 try {
	   rfile = new RandomAccessFile(name, "rw"); 
   rfile.seek(rfile.length()); // move to the end of file

	   int[] types = toTypes(domains); // domain types
	   String ss = new String("");	
	   String sdk = new String("");
	   String sdc = new String("");
	   for (int i = 0; i < tuples; i++) {
		 for (int j = 0; j < numattrs; j++) {
			switch(types[j]) {
			case INTEGER :
//			  rfile.writeInt(((int[]) dataArray[j])[i]);
//			  rfile.writeBytes(ss.valueOf(((int[])dataArray[j])[i]));		// Weizhong
			  ss = ss.valueOf(((int[])dataArray[j])[i]);
		  	  sdk = sdk.valueOf(INT_DK);
			  sdc = sdc.valueOf(INT_DC);
			  break;

	       //YouHua
			   case NUMERIC :
		          ss = ( ((number[]) dataArray[j])[i].toString());
  //                        sdk = sdk.valueOf(STRING_DK); 
//			  sdc = sdc.valueOf(STRING_DC);
                          sdk = sdk.valueOf(NUMERIC_DK); //from Yiyi
			  sdc = sdc.valueOf(NUMERIC_DC);

			  break;
			   case UNIVERSAL : //from Andrey
			case ATTRIBUTE : //from Andrey
			case STRING : 
			case STMT :
			case EXPR : {
	  		  ss = ((String[]) dataArray[j])[i];
	  		  sdk = sdk.valueOf(STRING_DK);
			  sdc = sdc.valueOf(STRING_DC);
/*
	  if (s.equals(STRING_DK))
//		 rfile.writeShort(Short.MIN_VALUE);
//		 rfile.writeBytes("^B");		// Weizhong
		 rfile.writeByte(2);		// Weizhong
			  else if (s.equals(STRING_DC))
//		 rfile.writeShort(Short.MIN_VALUE + 1);
//		 rfile.writeBytes("^E");		// Weizhong
		 rfile.writeByte(5);		// Weizhong
	  else {
//		 rfile.writeShort(s.length() - Short.MAX_VALUE + 1);
 // Weizhong
		 rfile.writeBytes(s);
	  }
*/
	  break;
			}
			case BOOLEAN :
//			  rfile.writeByte(((byte[]) dataArray[j])[i]);
//			  rfile.writeBytes(ss.valueOf(((byte[])dataArray[j])[i]));		// Weizhong
			  ss = ss.valueOf(((byte[])dataArray[j])[i]);
		  	  sdk = sdk.valueOf(BOOL_DK);
			  sdc = sdc.valueOf(BOOL_DC);
			  break;
			case SHORT :
//			  rfile.writeShort(((short[]) dataArray[j])[i]);
//			  rfile.writeBytes(ss.valueOf(((short[])dataArray[j])[i]));		// Weizhong
			  ss = ss.valueOf(((short[])dataArray[j])[i]);
		  	  sdk = sdk.valueOf(SHORT_DK);
			  sdc = sdc.valueOf(SHORT_DC);
			  break;
			case LONG :
			case IDLIST :
			case COMP: //case added by Yi Zheng May 31, 2002
//			  rfile.writeLong(((long[]) dataArray[j])[i]);
//			  rfile.writeBytes(ss.valueOf(((long[])dataArray[j])[i]));		// Weizhong
			  ss = ss.valueOf(((long[])dataArray[j])[i]);
		  	  sdk = sdk.valueOf(LONG_DK);
			  sdc = sdc.valueOf(LONG_DC);
			  break;
			case TEXT : {
	  		  ss = ((String[]) dataArray[j])[i];
		  	  sdk = sdk.valueOf(STRING_DK);
			  sdc = sdc.valueOf(STRING_DC);
/*
	  String s = ((String[]) dataArray[j])[i];
	  if (s.equals(STRING_DK))
//		 rfile.writeInt(-2);
//		 rfile.writeBytes("^B");		// Weizhong
		 rfile.writeByte(2);		// Weizhong
			  else if (s.equals(STRING_DC))
//		 rfile.writeInt(-1);
//		 rfile.writeBytes("^E");		// Weizhong
		 rfile.writeByte(5);		// Weizhong
	  else {
//		 rfile.writeInt(s.length()); // Weizhong
		 rfile.writeBytes(s);
	  }
*/
	  break;
			}
			case FLOAT :
//			  rfile.writeFloat(((float[]) dataArray[j])[i]);
//			  rfile.writeBytes(ss.valueOf(((float[])dataArray[j])[i]));		// Weizhong
			  ss = ss.valueOf(((float[])dataArray[j])[i]);
		  	  sdk = sdk.valueOf(FLOAT_DK);
			  sdc = sdc.valueOf(FLOAT_DC);
			  break;
			case DOUBLE :
//			  rfile.writeDouble(((double[]) dataArray[j])[i]);
//			  rfile.writeBytes(ss.valueOf(((double[])dataArray[j])[i]));		// Weizhong
			  ss = ss.valueOf(((double[])dataArray[j])[i]);
		  	  sdk = sdk.valueOf(DOUBLE_DK);
			  sdc = sdc.valueOf(DOUBLE_DC);
			  break;
			}
//		 	rfile.writeBytes("^F");	// Weizhong
			if (ss.equals(sdk))
				rfile.writeByte(2);
			else if (ss.equals(sdc))
				rfile.writeByte(5);
			else
				rfile.writeBytes(ss);

		 	rfile.writeByte(6);
		 } // loop on j
		 rfile.writeBytes("\n");
	   } // loop on i
   numtuples += tuples;
	 } catch (IOException e) {
   throw new 
	  InterpretError("IO error in appending tuples to: " + name);
	 } finally {
   try {
	  if (rfile != null) rfile.close();
   } catch (IOException ex) {}
	 }
   }      
   // Append relation relr to relation rell, the duplicate is eliminated.
   // Note that the this is a RAM version, since we assume that all the
   // data for the two relations and their nested domains are loaded. The
   // match for the number and types of the domains of the two relations 
   // must be checked before calling this method.
   public void appendInRAM(Relation relr, Relation rell)
   throws InterpretError
   {
   	//DEBUGING
   	//System.out.println("in appendInRAM:, relr is");
   	//Relation trelr = relr.duplicateLoadAll();
   	//trelr.print();
   	//System.out.println("rell is");
   	//Relation trell = rell.duplicateLoadAll();
   	//trell.print();
   	
	 int ttuples = rell.numtuples + relr.numtuples;
	 Object[] datal = createDataArray(rell.domains, ttuples);

	 // backup for right domains
	 Domain[] dtmp = new Domain[relr.numattrs];
	 for (int j = 0; j < relr.numattrs ; j++)
	   dtmp[j] = relr.domains[j];

	 // rearrange data to the left order so append is ok
	 relr.swapColumns(rell.domains);
	 for (int j = 0; j < rell.numattrs ; j++) {
	   System.arraycopy(rell.data[j], 0, datal[j], 0, rell.numtuples);
	   System.arraycopy(relr.data[j], 0, datal[j], rell.numtuples, 
	  relr.numtuples);
	 }

	 // swap right back to original order
	 relr.swapColumns(dtmp);
	 rell.data = datal;
	 rell.numtuples = ttuples;
	 rell.numsortattrs = 0;
	 rell.project(null); // project on itself to eliminate duplicates
   }      

   // Append relation relr to relation rell, the duplicate is NOT eliminated.
   // --added by YuLing
   // Note that the this is a RAM version, since we assume that all the
   // data for the two relations and their nested domains are loaded. The
   // match for the number and types of the domains of the two relations 
   // must be checked before calling this method.
   public void appendInRAM2(Relation relr, Relation rell)
   throws InterpretError
   {
	 int ttuples = rell.numtuples + relr.numtuples;
	 Object[] datal = createDataArray(rell.domains, ttuples);

	 // backup for right domains
	 Domain[] dtmp = new Domain[relr.numattrs];
	 for (int j = 0; j < relr.numattrs ; j++)
	   dtmp[j] = relr.domains[j];

	 // rearrange data to the left order so append is ok
	 relr.swapColumns(rell.domains);
	 for (int j = 0; j < rell.numattrs ; j++) {
	   System.arraycopy(rell.data[j], 0, datal[j], 0, rell.numtuples);
	   System.arraycopy(relr.data[j], 0, datal[j], rell.numtuples, 
	  relr.numtuples);
	 }

	 // swap right back to original order
	 relr.swapColumns(dtmp);
	 rell.data = datal;
	 rell.numtuples = ttuples;
	 rell.numsortattrs = 0;
   }      

   /**
   * Assign this relation to a new relation with newname, note that null
   * domain array means all domains are involved
   *
   * @param   domsr     array of right side domains
   * @param   domsl     array of left side domains
   * @param   newname   the name of the relation created by the assignment
   */
   public Relation assign(Domain[] domsr, String newname, Domain[] domsl)
   throws InterpretError
   {
	 // check if this relation contains all the domains in domsr
	 if (!isDomainsIn(domsr, domains))
	   throw new InterpretError("invalid right domains in assignment");

	 // check for duplicate domains in domain array
	 if (existDuplicateDomain(domsr) || existDuplicateDomain(domsl))
	   throw new InterpretError("duplicate attributes in assignment");

	 if (domsr != null || domsl != null) {
	   if (domsr == null || domsl == null || domsr.length != domsl.length)
		 throw new 
		 InterpretError("number of domains mismatch in assignment");
	   for (int i = 0; i < domsr.length; i++) {
		 if (domsr[i].type != domsl[i].type)
			throw new InterpretError("domain type mismatch in assignment");
	  if (domsr[i].type == IDLIST) {
		boolean br = domsr[i].name.equals(".id");
		boolean bl = domsl[i].name.equals(".id");
		if ((br && !bl) || (!br && bl)) 
	  throw new 
	  InterpretError("domain type mismatch in assignment : "
		 + domsl[i].name + " vs " + domsr[i].name);
		if (!br && !bl) {
//PATCHANGE:
//        Relation l = reltable.getRel("." + domsl[i].name);
//        Relation r = reltable.getRel("." + domsr[i].name);
	  Relation l = myEnv.lookupRel("." + domsl[i].name, true);
	  Relation r = myEnv.lookupRel("." + domsr[i].name, true);
	  if (!l.isDomainsIn(r.domains) || !r.isDomainsIn(l.domains))
		 throw new InterpretError(domsl[i].name + " and " 
	   + domsr[i].name + " not on same set of attributes");
		}
	  }
	   }
	 }

	 Relation rel = duplicateLoadAll();

	 if (domsl != null) {
	   rel.project(domsr);
	   rel.domains = domsl;
	 }

	 rel.name = newname; // change name, must follow load()
	 rel.save(); // save to create the data file 'newname'

	 if (domsl == null) {
   unloadAll(rel); // reclaim memory
   return rel;
	 }

	 // copy data to each dot files so that the copied surrogates data
	 // in their dot files, note that at this time, all the data needed
	 // are loaded, so we take advantage of this
	 for (int i = 0; i < domsl.length; i++) {
   if (domsl[i].type == IDLIST && domsr[i].type == IDLIST 
	  && !domsl[i].name.equals(".id") && !domsr[i].name.equals(".id")
	  && !domsl[i].name.equals(domsr[i].name)) 
   {
//PATCHANGE:
//       Relation l = reltable.getRel("." + domsl[i].name);
//       Relation r = reltable.getRel("." + domsr[i].name);
	  Relation l = myEnv.lookupRel("." + domsl[i].name, true);
	  Relation r = myEnv.lookupRel("." + domsr[i].name, true);

	  int ttuples = l.numtuples + r.numtuples;
	  Object[] datal = createDataArray(l.domains, ttuples);

	  // backup for right domains
	  Domain[] dtmp = new Domain[r.numattrs];
	  for (int j = 0; j < r.numattrs ; j++)
		dtmp[j] = r.domains[j];

	  // check left is already loaded, this may happen for
	  // R[X,Y <- Y,X]S; here X and Y are nested domains
	  boolean flag = false;
	  if (l.data == null) { // not load
		l.load();
		flag = true;
	  }

	  // rearrange data to the left order so append is ok
	  r.swapColumns(l.domains);
	  for (int j = 0; j < l.numattrs ; j++) {
		System.arraycopy(l.data[j], 0, datal[j], 0, l.numtuples);
		System.arraycopy(r.data[j], 0, datal[j], l.numtuples,
	  r.numtuples);
	  }
	  // swap right back to original order
	  r.swapColumns(dtmp);
	  l.data = datal;
	  l.numtuples = ttuples;
	  l.numsortattrs = 0;
	  l.project(null); // project on itself to eliminate duplicates
	  l.save(); // save a new copy of data
	  if (flag) // reclaim memory
		l.unload(); 
   }
	 }
	 unloadAll(rel); // reclaim memory
	 return rel;
   }      
   /**
   * Assign & add this relation to relation rell, note that relation rell
   * must exist to call this method
   *
   * @param   domsr     array of right side domains
   * @param   domsl     array of left side domains
   * @param   rell      the relation be assign_added
   */
   public void assignAdd(Domain[] domsr, Relation rell, Domain[] domsl)
   throws InterpretError
   {
	 // check if relations contain the domains as IDList
	 if (!isDomainsIn(domsr, domains))
	   throw new InterpretError("invalid right domains in assignment");
	 if (!isDomainsIn(domsl, rell.domains))
	   throw new InterpretError("invalid left domains in assignment");

	 // check for duplicate attributes
	 if (existDuplicateDomain(domsr) || existDuplicateDomain(domsl))
	   throw new InterpretError("duplicate attributes in assignment");

	 if (domsr == null && domsl == null) { // R <+ S;
	   if (!isDomainsIn(domains, rell.domains) || 
		  !isDomainsIn(rell.domains, domains))
		 throw new InterpretError("domains mismatch in assignment");
	   Relation relr = duplicateLoadAll();
   loadAll(rell);
   appendInRAM(relr, rell);
   rell.save();
   unloadAll(relr);
   unloadAll(rell);
	 } else if (domsr != null && domsl != null) {
	   if (!isDomainsIn(rell.domains, domsl)) // all domains are in domsl
		 throw new 
		 InterpretError("not all left domains specified in assignment");
	   if (domsr.length != domsl.length)
		 throw new 
		InterpretError("number of domains mismatch in assignment");
	   for (int i = 0; i < domsr.length; i++) {
		 if (domsr[i].type != domsl[i].type)
			throw new 
			InterpretError("domain type mismatch in assignment");
	  if (domsr[i].type == IDLIST) {
		boolean br = domsr[i].name.equals(".id");
		boolean bl = domsl[i].name.equals(".id");
		if ((br && !bl) || (!br && bl)) 
	  throw new 
	  InterpretError("domain type mismatch in assignment : "
		 + domsl[i].name + " vs " + domsr[i].name);
		if (!br && !bl) {
//PATCHANGE:
//        Relation l = reltable.getRel("." + domsl[i].name);
//        Relation r = reltable.getRel("." + domsr[i].name);
	  Relation l = myEnv.lookupRel("." + domsl[i].name, true);
	  Relation r = myEnv.lookupRel("." + domsr[i].name, true);
	  if (!l.isDomainsIn(r.domains) || !r.isDomainsIn(l.domains))
		 throw new InterpretError(domsl[i].name + " and " 
	   + domsr[i].name + " not on same set of attributes");
		}
	  }
	   }
	   Relation relr = duplicateLoadAll();
	   relr.project(domsr); // project the right side
	   relr.domains = domsl; // rearrange to left attribute order

   loadAll(rell); // load the left relation

   // before append data to the left relation, we must append the
   // data in the right dot relations to the left dot relations
   // so that the relations for some surrogates can be found in
   // the elimination of duplicates in method appendInRAM.
	   for (int i = 0; i < domsl.length; i++) {
	  if (domsl[i].type == IDLIST && domsr[i].type == IDLIST 
		 && !domsl[i].name.equals(".id") && !domsr[i].name.equals(".id")
		 && !domsl[i].name.equals(domsr[i].name)) 
	  {
//PATCHANGE:
//          Relation l = reltable.getRel("." + domsl[i].name);
//          Relation r = reltable.getRel("." + domsr[i].name);
		Relation l = myEnv.lookupRel("." + domsl[i].name, true);
		Relation r = myEnv.lookupRel("." + domsr[i].name, true);

		appendInRAM(r, l);
		l.save();
		 }
	   }

   appendInRAM(relr, rell);
   rell.save();
   unloadAll(relr);
   unloadAll(rell);
	 } else {
	   throw new InterpretError("number of domains mismatch in assignment");
	 }
   }      
   private void buildHeap(int[] atypes, boolean[] flags)
   throws InterpretError
   {
	 for (int i = numtuples/2 - 1; i >= 0; i--)
	   heapify(i, numtuples, atypes, flags);
   }      
   // return the domain array whose elements are common to the two operand
   // domain arrays doma and domb
   protected Domain[] commonDomains(Domain[] doma, Domain[] domb)
   {
	 Domain[] dom = new Domain[doma.length]; 
	 int j = 0; 
	 for (int i = 0; i < doma.length; i++) { 
   if (isDomainIn(doma[i], domb)) dom[j++] = doma[i];
	 }
	 Domain[] tmp = new Domain[j];
	 for (int i = 0; i < j; i++) tmp[i] = dom[i];
	 return tmp;
   }      
   // compare this relation with relation rel
   //    GT   if this > rel
   //    EQ   if this = rel
   //    LT   if this < rel
   // note that the data are assumed to be loaded for the two operand 
   // relations and the two relations are assumed to be defined on same
   // set of attributes
   private int compareRelation(Relation rel)
   throws InterpretError
   {
	 if (numtuples > rel.numtuples) {
   return GT;
	 } else if (numtuples < rel.numtuples) {
   return LT;
	 } else { // same number of tuples
   // sort is on the relations referenced by surrogates 
   boolean[] flags = new boolean[domains.length];
   for (int i = 0; i < domains.length; i++) 
	  flags[i] = false;

   heapSort(domains, flags); // sort this relation
   rel.heapSort(domains, flags); // swap and sort another relation 
   int len = numattrs, result;
   int[] types = toTypes(domains);
   for (int i = 0; i < numtuples; i++) {
	  result = compareTwoRows(data, i, rel.data, i, len, 
		types, flags, domains, rel.domains);
	  if (result != EQ) 
		return result;
   }
   return EQ;
	 }
   }      
   // compare the first len attributes of row m in datal with the first 
   // len attributes of row n in datar, the types of attributes are
   // given by atypes
   //
   //    LT   if ROW(m) < ROW(n)
   //    EQ   if ROW(m) = ROW(n)
   //    GT   if ROW(m) > ROW(n)
   //
   // Note that flags is a boolean array tells if sort is on the surrogate
   // values of nested attributes. They are intended for special handling
   // of such join : R[X : ijoin : .id].X
   public int compareTwoRows(Object[] datal, int m, Object[] datar, int n, 
	 int len, int[] types, boolean[] flags, Domain[] domsl, Domain[] domsr)
	 throws InterpretError
   {
	 for (int j = 0; j < len; j++) {
	   switch(types[j])
	   {
	   case INTEGER : {
		 int bm = ((int[]) datal[j])[m];
		 int bn = ((int[]) datar[j])[n];
		 if (bm > bn) return GT;
		 else if (bm < bn) return LT;
		 else break; 
   }
	  	   case UNIVERSAL ://from Andrey
		 case ATTRIBUTE : //from Andrey
	   case STRING :
	   case TEXT :
	   case STMT :
	   case EXPR : {
		 String bm = ((String[]) datal[j])[m];
		 String bn = ((String[]) datar[j])[n];
		 int result = compareTwoStrings(bm, bn);
		 if (result == GT || result == LT) return result;
		 break; 
   }
   	   case NUMERIC:{ //from Yiyi
	       number bm = ((number[])datal[j])[m];
	       number bn = ((number[])datar[j])[n];
	       int result = bm.compareTo(bn);
	       if(result > 0) return GT;
	       else if(result < 0) return LT;
	       else break;
	   }
		   

	   case BOOLEAN : {
		 byte bm = ((byte[]) datal[j])[m];
		 byte bn = ((byte[]) datar[j])[n];
		 if (bm > bn) return GT;
		 else if (bm < bn) return LT;
		 else break; 
   }
	   case SHORT : {
		 short bm = ((short[]) datal[j])[m];
		 short bn = ((short[]) datar[j])[n];
		 if (bm > bn) return GT;
		 else if (bm < bn) return LT;
		 else break; 
   }
	   case COMP: //REVISIT May1, 2002 Yi Zheng
	   case LONG : {
		 long bm = ((long[]) datal[j])[m];
		 long bn = ((long[]) datar[j])[n];
		 if (bm > bn) return GT;
		 else if (bm < bn) return LT;
		 else break; 
   }
	   case IDLIST : { 
		 long bm = ((long[]) datal[j])[m];
		 long bn = ((long[]) datar[j])[n];
	  if (flags[j] || domsl[j].name.equals(".id") 
	 || domsr[j].name.equals(".id")) {
		if (bm > bn) return GT;
		if (bm < bn) return LT;
		break;
	  }

	  if (bm == bn) // equal if the surrogates are the same  
		break;
	  if (bm == LONG_DK) {
		if (bn != LONG_DK) return LT;
		else break;
		 } else if (bm == LONG_DC) {
		if (bn == LONG_DK) return GT;
		else if (bn == LONG_DC) break;
		else return LT;
	  }

	  if (bn == LONG_DK) {
		if (bm != LONG_DK) return GT;
		else break;
		 } else if (bn == LONG_DC) {
		if (bm == LONG_DK) return LT;
		else if (bm == LONG_DC) break;
		else return GT;
	  }

//PATCHANGE:
//       Relation rl = getRelation(domsl[j].name, bm);
//       Relation dom = lookupRel("."+domsr[j]name, true);
	  Relation doml = myEnv.lookupRel("."+domsl[j].name, true);
	  Relation rl = doml.getRelation(bm);
	  Relation domr = myEnv.lookupRel("."+domsr[j].name, true);
	  Relation rr = domr.getRelation(bn);
	  int result = rl.compareRelation(rr);
	  if (result != EQ) return result;
	  else break; 
   }
	   case FLOAT : {
		 float bm = ((float[]) datal[j])[m];
		 float bn = ((float[]) datar[j])[n];
		 if (bm > bn) return GT;
		 else if (bm < bn) return LT;
		 else break; 
   }
	   case DOUBLE : {
		 double bm = ((double[]) datal[j])[m];
		 double bn = ((double[]) datar[j])[n];
		 if (bm > bn) return GT;
		 else if (bm < bn) return LT;
		 else break; 
   }
	   } // end of switch
	 }
	 return EQ;
   }      
   // compare two strings which are possibly dk and dc.
   // dk < dc < any other strings
   //    GT   if m > n lexicographically
   //    EQ   if m = n lexicographically
   //    LT   if m < n lexicographically
   private int compareTwoStrings(String m, String n)
   {
	 boolean mdk = m.equals(STRING_DK), ndk = n.equals(STRING_DK);
	 boolean mdc = m.equals(STRING_DC), ndc = n.equals(STRING_DC);
	 if (mdk) {
	   if (ndk) return EQ;
	   else return LT;
	 } else if (mdc) {
	   if (ndk) return GT;
	   else if (ndc) return EQ;
	   else return LT;
	 } else {
	   if (ndk || ndc) return GT;
	   else {
		 int result = m.compareTo(n);
		 if (result < 0) return LT;
		 else if (result > 0) return GT;
		 else return EQ;
	   }
	 }
   }      
   // create the data array according to the domain array
   public static Object[] createDataArray(Domain[] doms, int rows)
   {
	 Object[] dataArray = new Object[doms.length];
	 for (int i = 0; i < doms.length; i++) {
	   switch(doms[i].type) {
		 case INTEGER : {
			int[] a = new int[rows];
			dataArray[i] = a; break;
		 }
	   	   case UNIVERSAL ://from Andrey
		  case ATTRIBUTE : //from Andrey
		 case STRING :
		 case TEXT :
		 case STMT :
		 case EXPR : {
			String[] a = new String[rows];
			dataArray[i] = a; break;
		 }
		 case BOOLEAN : {
			byte[] a = new byte[rows];
			dataArray[i] = a; break;
		 }
                 //Youhua
         case NUMERIC : {
		        number[] a = new number[rows];
		        dataArray[i] = a; break;
	  	 } 
		 case SHORT : {
			short[] a = new short[rows];
			dataArray[i] = a; break;
		 }
		 case LONG :
		 case IDLIST :
		 case COMP : {
			long[] a = new long[rows];
			dataArray[i] = a; break;
		 }
		 case FLOAT : {
			float[] a = new float[rows];
			dataArray[i] = a; break;
		 }
		 case DOUBLE : {
			double[] a = new double[rows];
			dataArray[i] = a; break;
		 }
		 //default case added by Yi Zheng March 31, 2002
		 default: Global.println("Warning: unexpected domain type encountered in createDataArray");
		 			break;
	   }
	 }
	 return dataArray;
   }      
   public void createDataArray(int rows)
   {
	 data = createDataArray(domains, rows);
   }      
   /**
   * Duplicate this relation as a working copy for other operations
   * @param   ifload   boolean indicating if load data for the duplicate
   */
   public Relation duplicate(boolean ifload)
   throws InterpretError
   {
//System.out.println("in duplicate:" + name + " " + numtuples + " " + numattrs + " " + numsortattrs);
	 Relation rel = new Relation(name, domains, RELATION, numtuples,
	   numattrs, numsortattrs, (SimpleNode) null, myEnv);
  	 if (ifload) rel.load(true); 
	 return rel;
   }   
    
   /**
   * Duplicate this relation as a working copy for other operations, and
   * load the data into it and all the nested relations it is defined on
   * recursively
   */
   //REVISIT: Should we really use myEnv, or have an other env passed to us
   public Relation duplicateLoadAll()
   throws InterpretError
   {
	 Relation rel = new Relation(name, domains, RELATION, numtuples,
	   numattrs, numsortattrs, (SimpleNode) null, myEnv);
	 rel.data = data;         
	 loadAll(rel); // load this and all its nested relations
	 return rel;
   }      
   // check if domain array doms contains duplicate
//   private boolean existDuplicateDomain(Domain[] doms)
	protected boolean existDuplicateDomain(Domain[] doms)
   {
	 if (doms == null) return false;

	 for (int i = 0; i < doms.length; i++) {
	   for (int j = i + 1; j < doms.length; j++) {
		 if (doms[i] == doms[j]) return true;
	   }
	 }
	 return false;
   }      
   /**
   * Select the rows with corresponding surrogate number, and put the
   * result in the data field of relation provided as an agument, but
   * with the .id field removed
   *
   * @param   surrNo    the surrogate number
   * @param   rel       the relation to add the rows to
   */
   public void extract(long surrNo, Relation rel)
   {  //REVISIT: Is this assumption always true??
	if((rel == null) && (DEBUG))//Yi Zheng April 25, 2002
	{
		try
		{ throw new InterpretError("rel is null in extract()");
		}catch (InterpretError e)
		{ e.printStackTrace();}
	}
		
	if(DEBUG)
	{
		try{
		System.out.println("this relation is");
		this.print();
		}catch (Exception e)
		{ e.printStackTrace();}

		if(data == null)
			System.out.println("data is null");
	}

	//added by Yi Zheng April 25, 2002
	if(data == null) 
	{
		try{
			load(true);
		}catch (Exception e)
		{ Global.processThrowable(e);}
	}
	 long[] surrArray = (long[]) data[0]; // surrogate

	 int start = find(surrArray, surrNo, numtuples);

	 if (start == -1) // surrNo not found in surrArray
	 {  rel.numtuples = 0;
	   return;
	 }
//	if(DEBUG)
	//	System.out.println("start = " + start);

	 int end = start;
	 while (end < numtuples - 1 && surrArray[++end] == surrNo) ;
	 if (surrArray[end] != surrNo) end--;
	 int tuples = end - start + 1;
//	if(DEBUG)
	//	System.out.println("tuples = " + tuples);

	 rel.numtuples = tuples;
//	 if(rel.data == null)
//		System.out.println("rel.data is null");


	 for (int i=0; i<numattrs-1; i++) //recall: rel has no ".id" field
	 {  System.arraycopy(data[i+1], start, rel.data[i], 0, tuples);
	 }
   }      
/*
//PATCHANGE: This is the old version. We should be able to delete it
   // get the relation on nested domain domname with surrogate value be
   // surr, note that the data in relation ".domname" is assumed to be
   // loaded
   public static Relation getRelation2(String domname, long surr)
   {
	 Relation rel = reltable.getRel("." + domname);
	 long[] surArray = (long[]) rel.data[0]; // surrogate

	 Domain[] doms = new Domain[rel.numattrs - 1];
	 for (int i = 0; i < rel.numattrs - 1; i++)
	   doms[i] = rel.domains[i + 1];

	 int start = find(surArray, surr, rel.numtuples);

	 if (start == -1) { // not found 
   Relation r = new Relation("", doms, 0, doms.length, 0, env);
   r.data = createDataArray(doms, 0);
	   return r;
	 }

	 long val = surArray[start];
	 int end = start;
	 
	 while (end < rel.numtuples - 1 && surArray[++end] == val) ;
	 if (surArray[end] != val) end--;

	 int tuples = end - start + 1;
	 Relation r = new Relation("", doms, tuples, doms.length, 0, env);
	 r.data = createDataArray(doms, tuples);
	 for (int i = 0; i < rel.numattrs - 1; i++)
	   System.arraycopy(rel.data[i+1], start, r.data[i], 0, tuples);

	 return r;
   }
*/

   // find the one position of surr in array surArray using binary
   // search, since we assume that the nested relation is sorted 
   // with surroagte in non-decreasing order
   // REVISIT : binary search
   /**
   * @return  one position of surr. -1 if surArray is empty.
   * @since jRelixNS
   */
   private int find(long[] surArray, long surr, int numtuples)
   { // START Since jRelixNS
         // Handle the case where the surArray is empty.
         if (Relation.isSigmaJoinImplemented == true)
           if (numtuples == 0)
              return -1;
         // END Since jRelixNS
    int bottom = 0;
	 int top = numtuples-1; //Yi Zheng June 5,, 2002
    //  int top = surArray.length-1;
//System.out.println("in Relation::find(): surArray.length = " + surArray.length);
	 int middle;

//System.out.println("in Relation::find(): initially top = " + top + " bottom = " + bottom);
	 while(top-bottom > 1)
	 {
//System.out.println("top = " + top + " bottom = " + bottom);
  middle = (top+bottom)/2;
//System.out.println("in Relation::find(): middle = " + middle);
	   if (surArray[middle] >= surr)
	   {  top = middle;
	   }
	   else
	   {  bottom = middle;
	   }
	 }
	 if (surArray[bottom] == surr)
	 {  return bottom;
	 }
	 else if (surArray[top] == surr)
	 {  return top;
	 }
	 else
	 {  return -1;
	 }
/*
//This is the original linear version:
	  int i=0;
	
	  for (int i = 0; i < surArray.length; i++) {
	   if (surArray[i] == surr) return i;
	 }
	 return -1;
*/
   }      
   // get the relation on nested domain domname with surrogate value be
   // surr, note that the data in relation ".domname" is assumed to be
   // loaded
   //REVISIT: Should we really use myEnv, or have an other env passed to us
   public Relation getRelation(long surr)
   throws InterpretError
   { 
//System.out.println("in Relation.getRelation, surr: " + surr);
//System.out.println("in Relation.getRelation, name: " + name + " numtuples: " + numtuples);
//try{
if (data == null) {
//System.out.println("in getRelation, load data");
	load();
}
//} catch (InterpretError e) 
//{};

//System.out.println("domains[0].name" + domains[0].name);
//System.out.println("data[0]: " + data[0]);

        long[] surArray = (long[]) data[0]; // surrogate

//System.out.println("in Relation.getRelation, numattrs: " + numattrs);
	 Domain[] doms = new Domain[numattrs - 1];
	 for (int i = 0; i < numattrs - 1; i++)
	 {  doms[i] = domains[i + 1];
	 }

//	System.out.println("about to call find(surArray, " + surr + ", " + numtuples + ");");
//        System.out.println("relation " + this.name + " with " + this.numtuples + " tuples is: ");
//        this.printdata();
	//this.load(true);

	 int start = find(surArray, surr, numtuples);
	 if (start == -1)
	 {  // surrogate not found in surArray. Return an empty relation.
	   Relation r = new Relation("", doms, 0, doms.length, 0, myEnv);
	   r.data = createDataArray(doms, 0);
	   return r;
	 }

	 long val = surArray[start];
	 int end = start;
	 
	 while (end < numtuples - 1 && surArray[++end] == val)
//	while(end < surArray.length - 1 && surArray[++end] == val) //Yi Zheng June 5, 2002
	   ;
	 if (surArray[end] != val) end--;

	 int tuples = end - start + 1;
	 Relation r = new Relation("", doms, tuples, doms.length, 0, myEnv);
	 r.data = createDataArray(doms, tuples);
	 for (int i = 0; i < numattrs - 1; i++)
	 {  System.arraycopy(data[i+1], start, r.data[i], 0, tuples);
	 }

	 return r;
   }      

   // return the relation with the surrogate number -- Weizhong Sun
   // get the relation on nested domain domname with surrogate value be
   // surr, note that the data in relation ".domname" is assumed to be
   // loaded
   //REVISIT: Should we really use myEnv, or have an other env passed to us
   public Relation getRelwithSurr(long surr)
   { 
//System.out.println("in Relation.getRelwithSurr, surr: " + surr);
        long[] surArray = (long[]) data[0]; // surrogate

	 Domain[] doms = new Domain[numattrs];
	 for (int i = 0; i < numattrs; i++)
	 {  doms[i] = domains[i];
	 }

	 int start = find(surArray, surr, numtuples);

	 if (start == -1)
	 {  // surrogate not found in surArray. Return an empty relation.
	   Relation r = new Relation("", doms, 0, doms.length, 0, myEnv);
	   r.data = createDataArray(doms, 0);
	   return r;
	 }

	 long val = surArray[start];
	 int end = start;
	 
	 while (end < numtuples - 1 && surArray[++end] == val)
	   ;
	 if (surArray[end] != val) end--;

	 int tuples = end - start + 1;
	 Relation r = new Relation("", doms, tuples, doms.length, 0, myEnv);
	 r.data = createDataArray(doms, tuples);
	 for (int i = 0; i < numattrs; i++)
	 {  System.arraycopy(data[i], start, r.data[i], 0, tuples);
	 }

	 return r;
   }      

	/**
	 * get the relation whose .id is not equal to surr, and whose name is relName
	 * @author Yi Zheng June 10, 2002
	 */
   public Relation getRelExceptSurr(long surr, String relName)
   throws InterpretError
   { 
   //first verify that .id is one of the domains of this relation
     Domain iddom = myEnv.lookupDom(".id", true);
     if(iddom == null)
     	throw new InterpretError("Fatal error: .id domain not found");
     int idindex = indexOfDomain(iddom);
     if(idindex == 	ERROR)
     	throw new InterpretError("relation " + name + " has no .id domain");
     	
     //second, make sure .id is domains[0]
     if (data == null) 
     	load();
     Domain[] swd = new Domain[1];
     swd[0] = domains[idindex];
     swapColumns(swd);
     
     //third, get the relation with .id == surr
     Relation withSurr = getRelwithSurr(surr);
     String tname = Interpreter.nextTempName();
     withSurr.name = tname;
     withSurr.save();
     
     //fourth, get this djoin withSurr
     Relation r = mujoin(OP_DJOIN, null, withSurr, null, relName);
     if(r.data != null) r.save();
     
     r.load(true);
     
     //fifth, delete temp relation withSurr
     Interpreter.deleteFile(tname);
     
     //sixth, return the result
     return r;

   }      

   private void heapify(int n, int size, int[] atypes, boolean[] flags)
   throws InterpretError
   {
	 int left, right, largest;
	 left = 2*n + 1; right = left + 1; largest = n;
	 if (left < size && compareTwoRows(data, left, data, n, numsortattrs, 
   atypes, flags, domains, domains) == GT)
	   largest = left;
	 if (right < size && 
	 compareTwoRows(data, right, data, largest, numsortattrs, atypes, flags,
   domains, domains) == GT)
	   largest = right;
	 if (largest != n) {
	   swapTwoRows(n, largest);
	   heapify(largest, size, atypes, flags);
	 }
   }      
   /**
   * <p>
   * heapSort sorts an unnested relation using heap-sort algorithm.
   * Heap-sort is in place sort with worst-case runtime NlogN. See
   * Thomas H. Cormen, "Introduction to algorithms", 1990.
   * </p> <p>
   * The order of attributes of the relation is changed so that
   * the sorted attributes are placed at the beginning.
   * </p> <p>
   * Before invoking this method, the existence of the sorted
   * attributes and the loaded data must be checked.
   * </p>
   * @param   doms   the domains on which the relation is sorted
   *                 if doms == null, then sort on all attributes
   * @param   flags  a boolean array tells if sort is on the surrogate
   *                 values of nested attributes  
   * 
   * Note that if there is a true in flag, the original order which
   * may be on the relations of surrogates does not count so we need 
   * to sort anyway.
   */
   public void heapSort(Domain[] doms, boolean[] flags)
   throws InterpretError
   {
	 boolean flag = false;

	 // if sort is on surrogate values themselves and that attribute
	 // is not .id, the original order is destroyed
	 for (int i = 0; i < flags.length; i++) {
   String s = (doms == null) ? domains[i].name : doms[i].name;
   if (flags[i] && !s.equals(".id"))
	  flag = true;
	 }
	 if (flag) numsortattrs = 0;

	 // step 0. check if doms == null thus sort on itself
	 if (doms == null) {
	   if (numsortattrs == numattrs) return;   // already sorted
	   else numsortattrs = numattrs;
	 } else {
	   // step 1. check if it is sorted
	   if (doms.length <= numsortattrs) {
		 boolean b = true;
		 for (int i = 0; i < doms.length; i++) {
			b = b && (doms[i] == domains[i]);
		 }
		 if (b) return;
	   }
	   // step 2. set sort status and swap columns
	   numsortattrs = doms.length;
	   swapColumns(doms);
	 }

	 if (numtuples <= 1) return;

	 // step 3. heap-sort
	 int[] atypes = toTypes(domains);
	 buildHeap(atypes, flags);
	 for (int i = numtuples - 1; i > 0; i--) {
	   swapTwoRows(0, i);
	   heapify(0, i, atypes, flags);
	 }
   }      
   // find the index of domain dom in the domain array in this relation
   public int indexOfDomain(Domain dom)
   {
	 for (int i = 0; i < domains.length; i++) {
	   // commented and modified by Zongyan -- Aug 04,2002
	   //if (domains[i] == dom) return i;
	   if (domains[i].type == dom.type & domains[i].name.equals( dom.name)) return i;
	   
	 }
	 return ERROR;    // not found
   }      
   /**
   * Add a tuple which is nested relation to this relation using the 
   * provided surrogate number.
   *
   * @param   surrNo    the surrogate number
   * @param   rel       the relation containing the tuple to be added
   */
   public void insert(long surrNo, Relation rel)
   throws InterpretError
   {  int col = -1; //index of .id field
	 int i,j;


	 if(numattrs != rel.numattrs+1)
	 {  throw new
	   InterpretError("Incompatible relation passed to insert routine");
	 }

	 if (data != null)
	 {  save();
	 }
	 if (rel.data == null)
	 {  rel.load();
	 }
	 createDataArray(rel.numtuples);

	 //Make sure the domains of this relation and of that of rel
	 //correspond, and that they are in the same order
	 //Note that rel does not have a .id field, so this must be created
	 //find the index of the .id column
	 for(i=0; i<numattrs; i++)
	 {  if(domains[i].name.equals(".id"))
	   {  col = i;
	   }
	   else
	   {  //look for this domain in rel
		 j=0;
		 while (!rel.domains[j].name.equals(domains[i].name))
		 {  j++;
			if (j == rel.domains.length)  //should never happen
			{  throw new 
	  InterpretError("Imcompatible relation passed to insert routine");
			}
			
		 }
		 data[i] = rel.data[j];
	   }
	 }
	 if(col == -1)
	 {  throw new 
	   InterpretError("Imcompatible relation for insert routine");
	 }
	 //fill in the .id column with the surrogate number
	 for(i=0; i<rel.numtuples; i++)
	 {  ((long[]) data[col])[i] = surrNo;
	 }
	 append(data, rel.numtuples);
	 data = null;
   }      
   // check if dk exists in the first len attributes of tuple index in  
   // dataArray, the types of attributes are given by types
   private boolean isDkIn(Object[] dataArray, int index, int len, int[] types)
   {
	 for (int j = 0; j < len; j++) {
	   switch(types[j]) {
		 case INTEGER :
			if (((int[]) dataArray[j])[index] == INT_DK)
			  return true; 
			break;
	  	   case UNIVERSAL : //from Andrey
		case ATTRIBUTE : //from Andrey
		 case STRING :
		 case STMT :
		 case EXPR :
			if (STRING_DK.equals(((String[]) dataArray[j])[index]))
			  return true;
			break;
		 case BOOLEAN :
			if (((byte[]) dataArray[j])[index] == BOOL_DK) 
			  return true; 
			break;
		 case SHORT :
			if (((short[]) dataArray[j])[index] == SHORT_DK) 
			  return true; 
			break;
		 case LONG :
		 case COMP: //Yi Zheng May 1, 2002
		 case IDLIST :
			if (((long[]) dataArray[j])[index] == LONG_DK)
			  return true;
			break;
		 case TEXT :
			if (STRING_DK.equals(((String[]) dataArray[j])[index]))
			  return true;
			break;
		 case FLOAT :
			if (((float[]) dataArray[j])[index] == FLOAT_DK)
			  return true;
			break;
		 case DOUBLE :
			if (((double[]) dataArray[j])[index] == DOUBLE_DK)
			  return true;
			break;
		case NUMERIC: //added by Yiyi
	       if((((number[]) dataArray[j])[index]).compareTo(NUMERIC_DK) == 0)
		   return true;
	       break;

	   }
	 }
	 return false;
   }      
   // check if domain dom is an element of the domain array of this relation
   public boolean isDomainIn(Domain dom)
   {
	 return isDomainIn(dom, this.domains);
   }      
   // check if one domain dom is in domain array dlist
   public boolean isDomainIn(Domain dom, Domain[] dlist)
   {
	 if (dom == null) return true; // null is considered to be in any list
	 if (dlist == null) return false;
	 for (int i = 0; i < dlist.length; i++) {
	    // commented and modified by Zongyan --   Aug 04, 2002
		// if (dom == dlist[i]) return true; //I wonder if this kind of equlity comparison will cause trouble Yi Zheng March 23, 2002
	     if (dom.name.equals (dlist[i].name) & dom.type == dlist[i].type) return true; 
	 }
	 return false;
   }      
   // check if all the elements in domain array are in the domain array of
   // this relation
   public boolean isDomainsIn(Domain[] doms)
   {
	 return isDomainsIn(doms, this.domains);
   }      
   // check if all the elements in domain array doms are in another domain
   // array dlist
   public boolean isDomainsIn(Domain[] doms, Domain[] dlist)
   {
	 if (doms == null) return true; // null list is considered in any list
	 for (int i = 0; i < doms.length; i++) {
	   if (!isDomainIn(doms[i], dlist)) return false;
	 }
	 return true;
   }      
   public boolean isLoaded()
   {
//	 return (data == null); //modified by Yi Zheng March 22, 2002
	return (data != null);
   }      
   // load data from disk file
   /*
   public void load()
   throws InterpretError
   {
	 BlockInputStream bin = null;

	 int[] types = toTypes(domains); // domain types
	 data = createDataArray(domains, numtuples);

	 if (numtuples == 0) return; // no data to load

	 try {
	   bin = new BlockInputStream(name, BUFFERSIZE);
	 } catch (FileNotFoundException e) {
	   throw new InterpretError("No data file for loading : " + name);
	 } catch (IOException e) {
	   try {
		 if (bin != null) bin.close();
	   } catch (IOException ex) {
	   } finally {
		 throw new InterpretError("IO error in loading : " + name);
	   }
	 }
	 try {
	   for (int i = 0; i < numtuples; i++) {
		 for (int j = 0; j < numattrs; j++) {
			switch(types[j]) {
			  case INTEGER :
				((int[]) data[j])[i] = bin.readInt(); break;
			  case NUMERIC : 
			  ((number[]) data[j])[i] = new number(bin.readString()); break;

			  case STRING :
			  case STMT :
			  case EXPR :
				((String[]) data[j])[i] = bin.readString(); break;
			  case BOOLEAN :
				((byte[]) data[j])[i] = bin.readByte(); break;
			  case SHORT :
				((short[]) data[j])[i] = bin.readShort(); break;
			  case LONG :
			  case IDLIST :
			  case COMP :
				((long[]) data[j])[i] = bin.readLong(); break;
			  case TEXT :
				((String[]) data[j])[i] = bin.readText(); break;
			  case FLOAT :
				((float[]) data[j])[i] = bin.readFloat(); break;
			  case DOUBLE :
				((double[]) data[j])[i] = bin.readDouble(); break;
			}
		 } // loop on j
	   } // loop on i
	 } catch (IOException e) {
	   throw new InterpretError("IO error in loading relation: " + name);
	 } finally {
	   try {
		 if (bin != null) bin.close();
	   } catch (IOException e) {}
	 }
   }      
   */

   // read one field from input file fin.
   // 	used by void load().
   // 	-- created by Weizhong Sun, Jun 9th 1999
   public static String strRead(DataInputStream fin, int Type)
   throws IOException
   {
	BlockInputStream bin = new BlockInputStream(fin);
	return strRead(bin, Type);
/*
	String delim = new String("^F");
	String ret = new String("\n");
   	String readin = new String("");
	int space = 0;
	String temp = new String("");
	byte[] bret = new byte[1];
   	while (space == 0) {
		bret[0] = fin.readByte();
		if (bret[0] == 5) {	// ctrl-E, DC
			readin = new String("^E");
			bret[0] = fin.readByte();	// ctrl-F
			return readin;
		}
		if (bret[0] == 2) {	// ctrl-B, DK
			readin = new String("^B");
			bret[0] = fin.readByte();	// ctrl-F
			return readin;
		}
		if (bret[0] == 6)
			space = 1;	// ctrl-F
		else if (bret[0] != (ret.getBytes())[0]) {
			temp = new String(bret);
			readin = readin + temp;
		}
	}
	return readin;
*/
   }
   // overload the first one. 
   public static String strRead(BlockInputStream fin, int Type)
   throws IOException
   {
	String delim = new String("^F");
	String ret = new String("\n");
   	String readin = new String("");
	int space = 0;
	String temp = new String("");
	byte[] bret = new byte[1];
   	while (space == 0) {
		bret[0] = fin.readByte();
		if (bret[0] == 5) {	// ctrl-E, DC
			switch (Type) {
			case INTEGER :
			  readin = readin.valueOf(INT_DC);
			  break;
	                 //Youhua
	                 case NUMERIC :
			  readin = readin.valueOf(NUMERIC_DC);
			  break;
		   case UNIVERSAL ://from Andrey
			case ATTRIBUTE : //from Andrey
			case STRING : 
			case STMT :
			case EXPR :
			  readin = new String("^E");
			  break;
			case BOOLEAN :
			  readin = readin.valueOf(BOOL_DC);
			  break;
			case SHORT :
			  readin = readin.valueOf(SHORT_DC);
			  break;
			case LONG :
			case COMP: //Yi Zheng May 1, 2002
			case IDLIST :
			  readin = readin.valueOf(LONG_DC);
			  break;
			case TEXT :
			  readin = new String("^E");
			  break;
			case FLOAT :
			  readin = readin.valueOf(FLOAT_DC);
			  break;
			case DOUBLE :
			  readin = readin.valueOf(DOUBLE_DC);
			  break;
 			}
			bret[0] = fin.readByte();	// ctrl-F
			return readin;
		}
		if (bret[0] == 2) {	// ctrl-B, DK
			switch (Type) {
			case INTEGER :
			  readin = readin.valueOf(INT_DK);
			  break;
  			case NUMERIC :
                          readin = readin.valueOf(NUMERIC_DK);
		          break;
			  case UNIVERSAL :
			case ATTRIBUTE :
			case STRING : 
			case STMT :
			case EXPR : 
			  readin = new String("^B");
			  break;
			case BOOLEAN :
			  readin = readin.valueOf(BOOL_DK);
			  break;
			case SHORT :
			  readin = readin.valueOf(SHORT_DK);
			  break;
			case LONG :
			case COMP:
			case IDLIST :
			  readin = readin.valueOf(LONG_DK);
			  break;
			case TEXT :
			  readin = new String("^B");
			  break;
			case FLOAT :
			  readin = readin.valueOf(FLOAT_DK);
			  break;
			case DOUBLE :
			  readin = readin.valueOf(DOUBLE_DK);
			  break;
 			}
			bret[0] = fin.readByte();	// ctrl-F
			return readin;
		}
		if (bret[0] == 6)
			space = 1;	// ctrl-F
/*
		if (bret[0] == (delim.getBytes())[0]) { // ctrl
		   temp = new String(bret);
		   if ((bret[0] = fin.readByte()) != (delim.getBytes())[1]) {// !F
			readin = readin + temp;
			temp = new String(bret);
			readin = readin + temp;
		   }
		   else 
		       space = 1; 	// ctrl-F
		}
		*/
		else if (bret[0] != (ret.getBytes())[0]) {
			temp = new String(bret);
			readin = readin + temp;
		}
	}
	return readin;
   }

   // load data from disk file
   // 	-- modified by Weizhong Sun, Jun 9th 1999
   public void load()
   throws InterpretError
   {
     BlockInputStream ain = null;
     
//     System.out.println("in our load function.");
     int[] types = toTypes(domains); // domain types
     data = createDataArray(domains, numtuples);
    
     if (numtuples == 0) return; // no data to load
     
     try {
//       ain = new DataInputStream(new FileInputStream(name));
       ain = new BlockInputStream(name, BUFFERSIZE);
     } catch (FileNotFoundException e) {
       throw new InterpretError("No data file for loading : " + name);
     } catch (IOException e) {
       try {
             if (ain != null) ain.close();
       } catch (IOException ex) {
       } finally {
             throw new InterpretError("IO error in loading : " + name);
       }
     }

     try {
       for (int i = 0; i < numtuples; i++) {
             for (int j = 0; j < numattrs; j++) {
                    switch(types[j]) {
                      case INTEGER :
		      	    Integer dint = new Integer(strRead(ain, types[j]));
                            ((int[]) data[j])[i] = dint.intValue();
			    break;
	                 //Youhua
	              case NUMERIC :
                            number dnum = new number(strRead(ain, types[j]));
			    ((number[]) data[j])[i] = dnum;
			    break;
			    		  	case UNIVERSAL ://from Andrey
						case ATTRIBUTE ://from Andrey
                      case STRING :
                      case STMT :
                      case EXPR :
			    String ss = strRead(ain, types[j]); 
			    ((String[]) data[j])[i] = ain.readString(ss);
			    break;
                      case BOOLEAN :
		     	    Byte bb = new Byte(strRead(ain, types[j])); 
			   ((byte[]) data[j])[i] = bb.byteValue();		     	
			    break;
                      case SHORT :
		      	    Short dshort = new Short(strRead(ain, types[j]));
                            ((short[]) data[j])[i] = dshort.shortValue();
			    break;
                      case LONG :
                      case IDLIST :
                      case COMP :
		      	    String tempstr = strRead(ain, types[j]);
			    if (tempstr.equals("")) {
//				if(DEBUG)
//					System.out.println("tempstr is empty!!!");
			       tempstr = new String("0");};
			    Long dlong = new Long(tempstr);   
//		      	    Long dlong = new Long(strRead(ain, types[j]));
                            ((long[]) data[j])[i] = dlong.longValue();
			    break;
                      case TEXT :
			    ss = strRead(ain, types[j]); 
			    ((String[]) data[j])[i] = ain.readText(ss);
			    break;
                      case FLOAT :
		      	    Float dfloat = new Float(strRead(ain, types[j]));
                            ((float[]) data[j])[i] = dfloat.floatValue();
			    break;
                      case DOUBLE :
		      	    Double ddouble = new Double(strRead(ain, types[j]));
                            ((double[]) data[j])[i] = ddouble.doubleValue();
			    break;
		}
             }
       }                           
     } catch (IOException ex) {
     } finally {
     	try {
	  if (ain != null) ain.close();
	} catch (IOException e) {}
     }
  }
   // recursive load routine called by loadAll() to load all data
   protected void load(Relation rel)
   throws InterpretError
   {
	 // load all data for all nested domains it is defined on 
	 Domain[] doms = rel.domains;
	 for(int i = 0; i < doms.length; i++) { 
   String s = doms[i].name;
   if(doms[i].type == IDLIST && !s.equals(".id")) {
	  // System.out.println("loading : " + s);
//PATCHANGE:
//       Relation r = reltable.getRel("." + s);
	  Relation r = myEnv.lookupRel("." + s, true);

	  //REVISIT: added temporarily by zhongxia 10/09/97
	  if (r == null) Global.println("null in : " + s);
	  r.load(true); // check load to avoid repeated load
	  load(r);
   }
	 }
   }      
   // If ifcheck is set to true, then load only if not loaded.
   public void load(boolean ifcheck)
   throws InterpretError
   {
	 if (ifcheck)
	 {   if (data == null && rvc == RELATION) 
		{  load();
		}
	 }
	 else load();
   }      
   // load the data for relation rel and all the nested relations
   // it is defined on
   public void loadAll(Relation rel)
   throws InterpretError
   {
	 rel.load(true);
	 load(rel);
   }      
   /**
   * Mu-join this relation with another relation on specified attributes and
   * return the result relation. The two relations participated in the
   * join are not changed. Note that common attributes of the two
   * relations are used for null domain arrays. The existance of
   * domains in the relation is checked.
   *
   * @param   opcode   the join mode
   * @param   relr     the other relation to join with
   * @param   doml     domain array of this relation to be joined on
   * @params  domr     domain array of relation rel to be joined on
   * @params  relname  name the returned relation 
   *
   * Note : Special handling for R[X : mujoin : .id].X is done, since in
   * this case, relation R must be sorted on nested attribute X according
   * to the surrogate values rather than the relations of these surrogates.
   * A special set of flags is used for methods heapSort and  
   * compareTwoRows.
   */
   public Relation mujoin(int opcode, Domain[] doml, Relation relr,
	 Domain[] domr, String relname)
   throws InterpretError
   {
	 Domain[] dlist, dl, dr;

	 // flags for special handling of case : R[X : mujoin : .id].X
	 // true value of flag means special handling is needed
	 boolean[] flag;

	 // check if relations contain the join attributes, this should
	 // be done also in actualizer, so these checks may be removed
	 // later
	 if (!isDomainsIn(doml, domains)){
/*
	 for(int i = 0; i<doml.length; i++)
	   System.out.println("doml :"+doml[i].name);
	 for(int i = 0; i<domains.length; i++)
	   System.out.println("domains :"+domains[i].name);
*/
	   throw new InterpretError("invalid left domains in mujoin");
	 }

	 if (!isDomainsIn(domr, relr.domains))
	   throw new InterpretError("invalid right domains in mujoin");

	 if (existDuplicateDomain(doml) || existDuplicateDomain(domr))
	   throw new InterpretError("duplicate attributes in mujoin");

	 dlist = commonDomains(this.domains, relr.domains);
if(DEBUG)
{
	 System.out.println("list of common domains :");
	 for (int i = 0; i < dlist.length; i++) 
	    System.out.println(dlist[i].name);
}
	 // step 0. determine the join attributes dl and dr, check if there
	 // is any duplicate domain in the result
	 // check for existence of joined attributes ???
	 if (doml == null && domr == null) {
	   dl = new Domain[dlist.length];
	   dr = new Domain[dlist.length];
   flag = new boolean[dlist.length];
	   for (int i = 0; i < dlist.length; i++) {
		 dl[i] = dr[i] = dlist[i];
	  flag[i] = false; // nested attribute join .id not possible
	   }
	 } else if (doml != null && domr != null && doml.length == domr.length) {
   flag = new boolean[doml.length];
	   for (int i = 0; i < doml.length; i++) { // check for type match 
		 if (doml[i].type != domr[i].type)
			throw new InterpretError("attribute type mismatch in mu-join");

	  // default is no special handling for nested attribute join .id
	  flag[i] = false; 

	  // see if special handling for R[X : mujoin .id].X is
	  // necessary and set flag is needed
	  if (doml[i].type == IDLIST && domr[i].type == IDLIST
			&& (doml[i].name.equals(".id") || domr[i].name.equals(".id")))
			flag[i] = true; 

	  // if the joined on nested domains, the nested domains must
	  // defined on same set of attributes, not we don't force such
	  // requirement so that a nested domain can't join to the .id
	  // field in its corresponding dot relation

	  if (doml[i].type == IDLIST && domr[i].type == IDLIST
	 && !doml[i].name.equals(".id") && !domr[i].name.equals(".id")) 
		 {
//PATCHANGE:
//          Relation l = reltable.getRel("." + doml[i].name);
//          Relation r = reltable.getRel("." + domr[i].name);
		Relation l = myEnv.lookupRel("." + doml[i].name, true);
		Relation r = myEnv.lookupRel("." + domr[i].name, true);
		if (!l.isDomainsIn(r.domains) || !r.isDomainsIn(l.domains))
	  throw new InterpretError(doml[i].name + " and " 
		 + domr[i].name + " not on same set of attributes");

		// A further and subtle restriction is that the order of the
		// attributes on which the two nested join attributes are
		// defined must be the same. This restriction is due the
		// sort order implicitly on the two nested join attributes
		// produced by the heapSort before the join process. It
		// is quite difficult to release this restriction if not
		// impossible.
		for (int j = 0; j < l.domains.length; j++) {
	  if (l.domains[j] != r.domains[j])
		 throw new InterpretError(doml[i].name + " and "
	   + domr[i].name + " not have same order of attributes");
		}
		 }
	   }

	   // check if relations are joined on all common attributes (if there 
	   // exist duplicate attributes in the result relation ???)
	   if (!isDomainsIn(dlist, doml) || !isDomainsIn(dlist, domr))
		 throw new 
		 InterpretError("relations not mu-joined on all common attributes");
	   for (int i = 0; i < doml.length; i++) {
		 if (isDomainIn(doml[i], dlist) && doml[i] != domr[i])
			throw new 
		 InterpretError("relations not mu-joined on all common attributes");
	   }
	   dl = doml; dr = domr;
	 } else {
	   throw new InterpretError("attributes number mismatch in mu-join");
	 }

	 // step 1. sort the two duplicated copies of the joined relations,
	 // then create the result relation
	 Relation rl = this.duplicateLoadAll(); 
	 // note flag tells if the sort is on the surrogate values themselves
	 rl.heapSort(dl, flag); 
	 Relation rr = relr.duplicateLoadAll(); 
	 rr.heapSort(dr, flag);
if(DEBUG)
{
	 System.out.println("sorted left relation :"); rl.printdata();
	 System.out.println("sorted right relation :"); rr.printdata();
}

	 int tl = rl.numtuples, tr = rr.numtuples;
	 int dlen = dl.length;      // length of joined attributes
	 
if(DEBUG)
{
	System.out.println("length of joined attributes : " + dlen +
		           "number of tuples of left: " + tl +
			   "number of tuples of right: " + tr);
}
	 int dlenl = rl.domains.length, dlenr = rr.domains.length;
	 Object[] datal = rl.data, datar = rr.data;
	 int[] types = toTypes(dl); // type array of joined attributes
	 int[] typesl = toTypes(rl.domains);
	 int[] typesr = toTypes(rr.domains);

	 // set flags, a false means that the attribute is not present in the 
	 // result
	 boolean[] flagsl = new boolean[dlenl];
	 boolean[] flagsr = new boolean[dlenr];
	 switch (opcode) {
	   case OP_IJOIN :
	   case OP_UJOIN :
	   case OP_SJOIN :
	   case OP_LJOIN :
	   case OP_DLJOIN :
		 for (int i = 0; i < dlen; i++) {
			flagsl[i] = true; 
			flagsr[i] = (dl[i] == dr[i]) ? false : true;
		 }
		 for (int i = dlen; i < dlenl; i++) 
			flagsl[i] = true;
		 for (int i = dlen; i < dlenr; i++)
			flagsr[i] = (opcode == OP_DLJOIN) ? false : true;
		 break;
	   case OP_RJOIN :
	   case OP_DRJOIN :
		 for (int i = 0; i < dlen; i++) { 
			flagsl[i] = (dl[i] == dr[i]) ? false : true;
			flagsr[i] = true;
		 }
		 for (int i = dlen; i < dlenl; i++) 
			flagsl[i] = (opcode == OP_DRJOIN) ? false : true;
		 for (int i = dlen; i < dlenr; i++)
			flagsr[i] = true;
		 break;
	 }

	 // define the domain array for the result relation
	 int length = 0;   // the number of attributes of the result relation
	 for (int i = 0; i < dlenl; i++)
	   if (flagsl[i]) length++;
	 for (int i = 0; i < dlenr; i++)
	   if (flagsr[i]) length++;
if(DEBUG)
	 System.out.println("number of result attributes : " + length);

	 Domain[] doms = new Domain[length];
	 length = 0;
	 for (int i = 0; i < dlenl; i++)
	   if (flagsl[i]) doms[length++] = rl.domains[i];
	 for (int i = 0; i < dlenr; i++)
	   if (flagsr[i]) doms[length++] = rr.domains[i];

	 int ntuples = 0;   // the number of tuples of the result relation
//REVISIT: Should we really use myEnv, or have an other env passed to us
	 Relation rel = new Relation(relname, doms, 0, doms.length, 0, myEnv);
	 BlockOutputStream bout = null;
	 try {
	   bout = new BlockOutputStream(rel.name, BUFFERSIZE);
	 } catch (IOException e) {
	   try {
		 if (bout != null) bout.close();
	   } catch (IOException ex) {
	   } finally {
		 throw new InterpretError("IO error in writing mu-join result");
	   }
	 }

	 // step 2. merge, note that the joined attributes with dk are always 
	 // left out 
	 int result;   // result of comparing two tuples
	 int i = 0, j = 0;

	 //System.out.println(tl); System.out.println(tr); 

	 try {
	 while (i < tl || j < tr) {
	   if (i == tl)       // tuples in left relation are depleted
		 result = GT;
	   else if (j == tr)  // tuples in right relation are depleted
		 result = LT;
   //---------------------------------------------------------------
   // the elimination of the following two lines gives special
   // handling for ijoin, ujoin ... when there are no common
   // join attributes, i.e., 
   //     ijoin = ujoin = ljoin = rjoin = Catesian product
   //     sjoin, dljoin and drjoin are empty on their attributes
   //---------------------------------------------------------------
	   // else if (dlen == 0) // if no join attributes, deplete left first, 
	   //    result = LT;     // then right automatically by the condition 
   //---------------------------------------------------------------
	   else 
		 result = compareTwoRows(datal, i, datar, j, dlen, types, 
		flag, dl, dr);
	   
	   if (result == LT) {  // left is smaller, create left wing
		 switch (opcode) {
			case OP_UJOIN :
			case OP_SJOIN :
			case OP_LJOIN :
			case OP_DLJOIN :
			  writeAttributes(bout, datal, i, typesl, flagsl);
			  writeAttributes(bout, datal, i, types, flagsr);
			  writeDcAttributes(bout, typesr, flagsr, dlen);
			  ntuples++; //System.out.println("tuples : " + ntuples);
			default :
			  i++; //System.out.println("index i : " + i);
			  break;
		 }
	   } else if (result == GT) {  // right is smaller, create right wing
		 switch (opcode) {
			case OP_UJOIN :
			case OP_SJOIN :
			case OP_RJOIN :
			case OP_DRJOIN :
			  writeAttributes(bout, datar, j, types, flagsl);
			  writeDcAttributes(bout, typesl, flagsl, dlen);
			  writeAttributes(bout, datar, j, typesr, flagsr);
			  ntuples++; //System.out.println("tuples : " + ntuples);
			default :
			  j++; //System.out.println("index j : " + j);
			  break;
		 }
	   } else { // equal, create center part
		//if(DEBUG)
		//	System.out.println("\nEQ\n");
		 int nl = 1, nr = 1;   // nl left tuples match nr right tuples
		 while (i + nl < tl && compareTwoRows(datal, i, datal, i + nl, 
		dlen, types, flag, dl, dl) == EQ)
			nl++;
		 while (j + nr < tr && compareTwoRows(datar, j, datar, j + nr, 
		dlen, types, flag, dr, dr) == EQ)
			nr++;
	//	if(DEBUG)
//			System.out.println("nl = " + nl + "  nr = " + nr);
		 switch (opcode) {
			case OP_IJOIN :
			case OP_UJOIN :
			case OP_LJOIN :
			case OP_RJOIN :
			  if (isDkIn(datal, i, dlen, types)) 
				;   // equal tuples with dk's are ignored in Relix
			  else {
				for (int ii = 0; ii < nl; ii++) { 
				  for (int jj = 0; jj < nr; jj++) {
					 writeAttributes(bout, datal, i + ii, typesl, flagsl);
					 writeAttributes(bout, datar, j + jj, typesr, flagsr);
					 ntuples++; 
		 //System.out.println("tuples : " + ntuples);
				  }
				}
			  }
			default :
			  i += nl; j += nr; 
			  //System.out.println("index i : " + i); 
			  //System.out.println("index j : " + j);
			  break;
		 }
	   }
	 } // end of while loop
	 } catch (IOException e) {
	   try {
		 bout.close();
	   } catch (IOException ex) { 
	   }
	   throw new InterpretError("IO error in writing mu-join result");
	 }

	 // step 4. for join on nested attributes, copy data to each dot
	 // files so that the copied surrogates have data in their
	 // dot files, note that in ijoin, there is no need to do it.
	 // note that at this time, all the data needed are loaded, so we
	 // can take advantage of this
	 for (i = 0; i < dl.length; i++) {
	   if (dl[i].type == IDLIST && dr[i].type == IDLIST
		  && !dl[i].name.equals(".id") && !dr[i].name.equals(".id")
	  && !dl[i].name.equals(dr[i].name))
	   {
//PATCHANGE:
//            Relation l = reltable.getRel("." + dl[i].name);
//            Relation r = reltable.getRel("." + dr[i].name);
		 Relation l = myEnv.lookupRel("." + dl[i].name, true);
		 Relation r = myEnv.lookupRel("." + dr[i].name, true);

		 int ttuples = l.numtuples + r.numtuples;

	  switch (opcode) {
		case OP_UJOIN :
		case OP_SJOIN : {
	  datar = createDataArray(r.domains, ttuples);

	  // backup for left domains
	  Domain[] dtmp = new Domain[l.numattrs];
	  for (j = 0; j < l.numattrs ; j++)
		 dtmp[j] = l.domains[j];

	  // rearrange data to the right order so append is ok
	  l.swapColumns(r.domains); 

	  for (j = 0; j < r.numattrs ; j++) {
		 System.arraycopy(r.data[j], 0, datar[j], 0, r.numtuples);
		 System.arraycopy(l.data[j], 0, datar[j], r.numtuples, 
	   l.numtuples);
	  }
	  // save right data
	  r.data = datar;
	  r.numtuples = ttuples;
	  r.numsortattrs = 0;
	  r.project(null); // project on itself to eliminate duplicates
	  r.save(); // save a new copy of data
	  // save the same left data according to left order
	  l.data = datar;
	  l.numtuples = ttuples;
	  l.numsortattrs = 0;
	  // swap left back to original order
	  l.swapColumns(dtmp);
	  l.project(null); // project on itself to eliminate duplicates
	  l.save();
	  break;
		}
		case OP_LJOIN :
		case OP_DLJOIN : {
	  datar = createDataArray(r.domains, ttuples);

	  // backup for left domains
	  Domain[] dtmp = new Domain[l.numattrs];
	  for (j = 0; j < l.numattrs ; j++)
		 dtmp[j] = l.domains[j];

	  // rearrange data to the right order so append is ok
	  l.swapColumns(r.domains); 

	  for (j = 0; j < r.numattrs ; j++) {
		 System.arraycopy(r.data[j], 0, datar[j], 0, r.numtuples);
		 System.arraycopy(l.data[j], 0, datar[j], r.numtuples, 
	   l.numtuples);
	  }
	  // swap left back to original order
	  l.swapColumns(dtmp);
	  r.data = datar;
	  r.numtuples = ttuples;
	  r.numsortattrs = 0;
	  r.project(null); // project on itself to eliminate duplicates
	  r.save(); // save a new copy of data
	  break;
		}
		case OP_RJOIN :
		case OP_DRJOIN : {
	  datal = createDataArray(l.domains, ttuples);

	  // backup for right domains
	  Domain[] dtmp = new Domain[r.numattrs];
	  for (j = 0; j < r.numattrs ; j++)
		 dtmp[j] = r.domains[j];

	  // rearrange data to the left order so append is ok
	  r.swapColumns(l.domains); 

	  for (j = 0; j < l.numattrs ; j++) {
		 System.arraycopy(l.data[j], 0, datal[j], 0, l.numtuples);
		 System.arraycopy(r.data[j], 0, datal[j], l.numtuples, 
	   r.numtuples);
	  }
	  // swap right back to original order
	  r.swapColumns(dtmp);
	  l.data = datal;
	  l.numtuples = ttuples;
	  l.numsortattrs = 0;
	  l.project(null); // project on itself to eliminate duplicates
	  l.save(); // save a new copy of data
	  break;
		}
	  } // end of switch
	   }
	 }

	 // step 5. set fields and return
	 rel.numtuples = ntuples;
	 try {
	   bout.close();
	 } catch (IOException e) {
	   throw new InterpretError("IO error in writing mu-join result");
	 }
	 //System.out.println("tuples : " + ntuples);
	 unloadAll(rl);
	 unloadAll(rr);
	 return rel;
   }      
   /**
   * Return the domains in the result of mujoin
   *
   * @param   opcode   the join mode
   * @param   relr     the other relation to join with
   * @param   doml     domain array of this relation to be joined on
   * @params  domr     domain array of relation rel to be joined on
   *
   * Note that if both doml and domr are null, the join is implicitly
   * on the common attributes of the two relations.
   */
   public Domain[] mujoinType(int opcode, Domain[] doml, Relation relr,
	 Domain[] domr)
   throws InterpretError
   {
	 Domain[] dlist, dl, dr;

	 dlist = commonDomains(this.domains, relr.domains);

	//DEBUGGING
	//for(int i = 0; i < dlist.length; i++)
	//	System.out.println("dlist[" + i + "] = " + dlist[i].name);
		
	 // determine the join attributes dl and dr, check if there
	 // is any duplicate domain in the result
	 // check for existence of joined attributes ???
	 if (doml == null && domr == null) {
	   dl = new Domain[dlist.length];
	   dr = new Domain[dlist.length];
	   for (int i = 0; i < dlist.length; i++)
		 dl[i] = dr[i] = dlist[i];
	 } else if (doml != null && domr != null && doml.length == domr.length) {
	   for (int i = 0; i < doml.length; i++) { // check for type match 
		 if (doml[i].type != domr[i].type)
			throw new InterpretError("attribute type mismatch in mu-join");

	  // if the joined on nested domains, the nested domains must
	  // defined on same set of attributes, not we don't force such
	  // requirement so that a nested domain can't join to the .id
	  // field in its corresponding dot relation

	  if (doml[i].type == IDLIST && domr[i].type == IDLIST
	 && !doml[i].name.equals(".id") && !domr[i].name.equals(".id")) 
		 {
//PATCHANGE:
//          Relation l = reltable.getRel("." + doml[i].name);
//          Relation r = reltable.getRel("." + domr[i].name);
		Relation l = myEnv.lookupRel("." + doml[i].name, true);
		Relation r = myEnv.lookupRel("." + domr[i].name, true);
		if (!l.isDomainsIn(r.domains) || !r.isDomainsIn(l.domains))
	  throw new InterpretError(doml[i].name + " and " 
		 + domr[i].name + " not on same set of attributes");

		// A further and subtle restriction is that the order of the
		// attributes on which the two nested join attributes are
		// defined must be the same. This restriction is due the
		// sort order implicitly on the two nested join attributes
		// produced by the heapSort before the join process. It
		// is quite difficult to release this restriction if not
		// impossible.
		for (int j = 0; j < l.domains.length; j++) {
	  if (l.domains[j] != r.domains[j])
		 throw new InterpretError(doml[i].name + " and "
	   + domr[i].name + " not have same order of attributes");
		}
		 }
	   }

	   // check if relations are joined on all common attributes (if there 
	   // exist duplicate attributes in the result relation ???)
	   if (!isDomainsIn(dlist, doml) || !isDomainsIn(dlist, domr))
		 throw new 
		 InterpretError("relations not mu-joined on all common attributes");
	   for (int i = 0; i < doml.length; i++) {
		 if (isDomainIn(doml[i], dlist) && doml[i] != domr[i])
			throw new 
		 InterpretError("relations not mu-joined on all common attributes");
	   }
	   dl = doml; dr = domr;
	 } else {
	   throw new InterpretError("attributes number mismatch in mu-join");
	 }

	 int dlen = dl.length;      // length of joined attributes
	 int dlenl = this.domains.length, dlenr = relr.domains.length;


	//bug fix by Yi Zheng June 24, 2002
	//note that in order for the following to be correct, we must make sure that
	//the join domains appear first in this.domains and relr.domains
	Domain[] leftDoms = new Domain[dlenl];
	Domain[] rightDoms = new Domain[dlenr];
	
	for(int i = 0; i < dlenl; i++)
		leftDoms[i] = this.domains[i];
	for(int i = 0; i < dlenr; i++)
		rightDoms[i] = relr.domains[i];
		
	//shuffle the domains if necessary
	shuffleDoms(dl, leftDoms);
	shuffleDoms(dr, rightDoms);
	
	//end of bug fix
	 // set flags, a false means that the attribute is not present in the 
	 // result
	 boolean[] flagsl = new boolean[dlenl];
	 boolean[] flagsr = new boolean[dlenr];
	 switch (opcode) {
	   case OP_IJOIN :
	   case OP_UJOIN :
	   case OP_SJOIN :
	   case OP_LJOIN :
	   case OP_DLJOIN :
		 for (int i = 0; i < dlen; i++) {
			flagsl[i] = true; 
			flagsr[i] = (dl[i] == dr[i]) ? false : true;
		 }
		 for (int i = dlen; i < dlenl; i++) 
			flagsl[i] = true;
		 for (int i = dlen; i < dlenr; i++)
			flagsr[i] = (opcode == OP_DLJOIN) ? false : true;
		 break;
	   case OP_RJOIN :
	   case OP_DRJOIN :
		 for (int i = 0; i < dlen; i++) { 
			flagsl[i] = (dl[i] == dr[i]) ? false : true;
			flagsr[i] = true;
		 }
		 for (int i = dlen; i < dlenl; i++) 
			flagsl[i] = (opcode == OP_DRJOIN) ? false : true;
		 for (int i = dlen; i < dlenr; i++)
			flagsr[i] = true;
		 break;
	 }

	 // define the domain array for the result relation
	 int length = 0;   // the number of attributes of the result relation
	 for (int i = 0; i < dlenl; i++)
	   if (flagsl[i]) length++;
	 for (int i = 0; i < dlenr; i++)
	   if (flagsr[i]) length++;
	 Domain[] doms = new Domain[length];
	 length = 0;
//	 for (int i = 0; i < dlenl; i++)
//	   if (flagsl[i]) doms[length++] = this.domains[i];
//	 for (int i = 0; i < dlenr; i++)
//	   if (flagsr[i]) doms[length++] = relr.domains[i];
//Bug fix June 24, 2002
	 for (int i = 0; i < dlenl; i++)
	   if (flagsl[i]) doms[length++] = leftDoms[i];
	 for (int i = 0; i < dlenr; i++)
	   if (flagsr[i]) doms[length++] = rightDoms[i];
//end of bug fix June 24, 2002
	 //System.out.println("number of result attributes : " + length);
	//for(int kk = 0; kk < doms.length; kk++)
	//	System.out.print(doms[kk].name + " ");
	//System.out.println("");
	 return doms;

   }  
   
   
   /**
    * rearrange domains in toShuffle such that those in dj come first
    * @author Yi Zheng June 24, 2002
    */
   private void shuffleDoms(Domain[] dj, Domain[] toShuffle)
   {
   	 int djlen = dj.length;
   	 if (djlen == 0) return;
   	 
   	 int slen = toShuffle.length;
   	 if (slen == 0) return;
   	 
   	 for(int i = 0; i < djlen; i++)
   	 {
   	 	Domain tmp = null;
   	 	int pos = -1;
   	 	for(int j = 0; j < slen; j++)
   	 	{
   	 		if(dj[i].name.compareTo(toShuffle[j].name) != 0)
   	 			continue;
   	 		pos = j;
   	 		break;
   	 	}
   	 	if((i != pos) && (pos != -1))
   	 	{
   	 		tmp = toShuffle[pos];
   	 		toShuffle[pos] = toShuffle[i];
   	 		toShuffle[i] = tmp;
   	 	}
   	 }
   }    
   /**
   * Randomly pick one tuple from this relation and return a relation
   * with that one tuple
   *
   * @param   relname   name of the returned relation
   */
   public Relation pick(String relname)
   throws InterpretError
   {
	 Relation rel = duplicate(true); // duplicate and load only this

	 rel.numsortattrs = rel.numattrs; // one tuple is always sorted

	 if (rel.numtuples == 0) {
   //System.out.println("warning : no tuple to pick in source relation"); modified by Yi Zheng March 19, 2002
   Global.println("warning : no tuple to pick in source relation");
   rel.name = relname;
   rel.save();
   rel.unload();
   return rel;
	 } 

	 int ran = new Random().nextInt();
	 rel.swapTwoRows(0, Math.abs(ran%rel.numtuples));
	 rel.numtuples = 1;

	 rel.name = relname;
	 rel.save();
	 rel.unload();
	 return rel;
   }      
   public void print()
   throws InterpretError
   {
   	//modified by Yi Zheng March 21, 2002 for pretty printing system table
   	 PrettyPrint pp = new PrettyPrint();
   	 if(this.name.compareTo(".rel") == 0)
   	 {
   	 	int[] lens = new int[5];
   	 	lens[0] = pp.lengthOfType(STRING);
   	 	lens[1] = pp.lengthOfType(INTEGER);
   	 	lens[2] = pp.lengthOfType(INTEGER);
   	 	lens[3] = pp.lengthOfType(STRING);
   	 	lens[4] = pp.lengthOfType(INTEGER);
   	 	print(lens, true);
   	 }
   	 else if(this.name.compareTo(".dom") == 0)
   	 {
   	 	int[] lens = new int[4]; //April 29, 2002
   	    lens[0] = pp.lengthOfType(STRING);
   	 	lens[1] = pp.lengthOfType(STRING);
   	 	lens[2] = pp.lengthOfType(INTEGER);
   	 	//Yi Zheng April 29, 2002
   	 	lens[3] = pp.lengthOfType(INTEGER);
   	 	print(lens, true);
  	  }
  	  else
	 	print(null, true);
   }         
   // print the relation onto an OutputStream
   // array lens gives the lenght for each field, if it is null, the
   // default lenght is used.
   // flag indicate if the name of the relation or "expression" 
   // is used for title
   public void print(int[] lens, boolean flag)
   throws InterpretError
   {
	 PrettyPrint pp = new PrettyPrint();
	 
//added by Yi Zheng March 21 , 2002 to handle pretty printing of .rel .dom
 	int[] Lens = null;
   	 if(this.name.compareTo(".rel") == 0)
   	 {
   	 	Lens = new int[5];
   	 	Lens[0] = pp.lengthOfType(STRING);
   	 	Lens[1] = pp.lengthOfType(INTEGER);
   	 	Lens[2] = pp.lengthOfType(INTEGER);
   	 	Lens[3] = pp.lengthOfType(STRING);
   	 	Lens[4] = pp.lengthOfType(INTEGER);
   	 }
   	 else if(this.name.compareTo(".dom") == 0)
   	 {
   	 	Lens = new int[4]; //April 29, 2002
   	    Lens[0] = pp.lengthOfType(STRING);
   	 	Lens[1] = pp.lengthOfType(STRING);
   	 	Lens[2] = pp.lengthOfType(INTEGER);
   	 	
   	 	//April 29, 2002
   	 	Lens[3] = pp.lengthOfType(INTEGER);
  	  }
  	  if(lens == null)
  	  	lens = Lens;

	//added by Yi Zheng for debugging
	if(DEBUG)
	  pp.pprint("<<", 2); 

	 // print head
	 for (int i = 0; i < numattrs; i++) {
	 	//Yi Zheng April 29, 2002
	 	if((!Global.debug) && (domains[i].isStateDom == 1))
	 		continue;
	 	
	   pp.pprint("+-", 2);
	   if (lens == null) 
		 pp.pprintRule(pp.lengthOfType(domains[i].type) + 1);
	   else 
		 pp.pprintRule(lens[i] + 1);
	 }
	 pp.pprint("+", 1); pp.println();

//added by Yi Zheng for debugging
	if(DEBUG) 
		pp.pprint("<<", 2);

	 for (int i = 0; i < numattrs; i++) {
	 	if((!Global.debug) && (domains[i].isStateDom == 1))
	 		continue; //Yi Zheng April 29
	 	
	   pp.pprint("| ", 2);
	   
	   //Yi Zheng April 29, 2002 to handle hidden state domain
	   String domname = domains[i].name;
	   if (domains[i].isStateDom == 1)
	   {
	   	domname = "*" + domname + "*";
	   if (lens == null) 
		 pp.pprint(domname, pp.lengthOfType(domains[i].type));
	   else 
		 pp.pprint(domname, lens[i]);
		}   
		else
		{
	if (lens == null) 
		 pp.pprint(domains[i].name, pp.lengthOfType(domains[i].type));
	   else 
		 pp.pprint(domains[i].name, lens[i]);
		}
	   pp.pprint(" ", 1);
	
	 }
	 pp.pprint("|", 1); pp.println();

//added by Yi Zheng for debugging
	if(DEBUG)
	  pp.pprint("<<", 2);


	 for (int i = 0; i < numattrs; i++) {
	 	//Yi Zheng April 29, 2002
	 	if((!Global.debug) && (domains[i].isStateDom == 1))
	 		continue;

	   pp.pprint("+-", 2);
	   if (lens == null) 
		 pp.pprintRule(pp.lengthOfType(domains[i].type) + 1);
	   else 
		 pp.pprintRule(lens[i] + 1);
	 }
	 pp.pprint("+", 1); pp.println();

	//added by Yi Zheng for handling pr .rel, .dom, .rd
	if(name.compareTo(".rel") == 0)
	{
		numtuples = Global.environment.reltable.getRelTableSize();
		Global.environment.reltable.dump();
		data = null; //force load() of fresh data
	}
	else if (name.compareTo(".dom") == 0)
	{
		numtuples = Global.environment.domtable.getDomTableSize();
		Global.environment.domtable.dump();
		data = null; //force load() of fresh data
	}
	else if (name.compareTo(".rd") == 0)
	{
		numtuples = Global.environment.reltable.getRDTableSize();
		Global.environment.reltable.dump();
		data = null; //force load() of fresh data
	}
	
		
		
	
	 // check and load data
	 if (data == null)
	 {// load(); -- modified by Yi Zheng March 22, 2002
	 	load();
	 	if(DEBUG)
	 	 System.out.println("In Relation:print(): numtuples = " + numtuples);
	 }

	 // relation body
	 int[] types = toTypes2(domains);
	 for (int i = 0; i < numtuples; i++) {
//added by Yi Zheng for debugging
	if(DEBUG)
	  pp.pprint("<<", 2);


	   for (int j = 0; j < numattrs; j++) {
	 	//Yi Zheng April 29, 2002
	 	if((!Global.debug) && (domains[j].isStateDom == 1))
	 		continue;

		 switch(types[j]) {
		 case INTEGER :
			pp.pprint("| ", 2); 
			if (lens == null)
			  pp.pprint(((int[]) data[j])[i]);
			else 
			  pp.pprint(((int[]) data[j])[i], lens[j]);
			pp.pprint("", 1); break;
		 case UNIVERSAL ://from Andrey
		 case ATTRIBUTE ://from Andrey
		 case STRING ://added by Yi Zheng March 21, 2002
		 		pp.pprint("| ", 2);
		 		if(((this.name.compareTo(".rel") == 0) && (domains[j].name.compareTo(".rvc") == 0))
		 		|| ((this.name.compareTo(".dom") == 0) && (domains[j].name.compareTo(".type") == 0)))
		 		{
		 			if (lens == null)
			  			pp.pprint(Utility.type2String(((int[]) data[j])[i]));
					else 
			  			pp.pprint(Utility.type2String(((int[]) data[j])[i]), lens[j]);
			  	}else
			  	{
					if (lens == null)
			  			pp.pprint(((String[]) data[j])[i]);
					else 
			  			pp.pprint(((String[]) data[j])[i], lens[j]);
			  	}
  				pp.pprint("", 1); break;

		 case TEXT :
		 case STMT :
		 case EXPR :
			pp.pprint("| ", 2); 
			if (lens == null)
			  pp.pprint(((String[]) data[j])[i]);
			else 
			  pp.pprint(((String[]) data[j])[i], lens[j]);
			pp.pprint("", 1); break;
		 case NUMERIC://from Yiyi
		     pp.pprint("| ", 2);
		     if(lens == null)
			 pp.pprint((((number[])data[j])[i]).toString());
		     else
			 pp.pprint((((number[])data[j])[i]).toString());
		     pp.pprint("",1);
		     break;


		 case BOOLEAN :
			pp.pprint("| ", 2); 
			if (lens == null)
			  pp.pprint(((byte[]) data[j])[i]);
			else
			  pp.pprint(((byte[]) data[j])[i], lens[j]);
			pp.pprint("", 1); break;
		 case SHORT :
			pp.pprint("| ", 2); 
			if (lens == null)
			  pp.pprint(((short[]) data[j])[i]);
			else
			  pp.pprint(((short[]) data[j])[i], lens[j]);
			pp.pprint("", 1); break;
		 case LONG :
		 case IDLIST :
		 case COMP :
			pp.pprint("| ", 2); 
			if (lens == null)
			  pp.pprint(((long[]) data[j])[i]);
			else
			  pp.pprint(((long[]) data[j])[i], lens[j]);
			pp.pprint("", 1); break;
		 case FLOAT :
			pp.pprint("| ", 2); 
			if (lens == null)
			  pp.pprint(((float[]) data[j])[i]);
			else
			  pp.pprint(((float[]) data[j])[i], lens[j]);
			pp.pprint("", 1); break;
		 case DOUBLE :
			pp.pprint("| ", 2); 
			if (lens == null)
			  pp.pprint(((double[]) data[j])[i]);
			else
			  pp.pprint(((double[]) data[j])[i], lens[j]);
			pp.pprint("", 1); break;
		 }
	   }
	   pp.pprint("|", 1); pp.println();
	 }
	 // end of relation body

	//added by Yi Zheng for debugging
	if (DEBUG)
          pp.pprint("<<", 2);

	 for (int i = 0; i < numattrs; i++) {
	 	//Yi Zheng April 29, 2002
	 	if((!Global.debug) && (domains[i].isStateDom == 1))
	 		continue;

	   pp.pprint("+-", 2);
	   if (lens == null) 
		 pp.pprintRule(pp.lengthOfType(domains[i].type) + 1);
	   else 
		 pp.pprintRule(lens[i] + 1);
	 }
	 pp.pprint("+", 1); pp.println();
	 String stuple = (numtuples > 1) ? " tuples" : " tuple";
	 String stitle = "relation " + name;
	 if (!flag) 
	   stitle = "expression";

	//MODIFIED by Yi Zheng for debugging pp.println(stitle + " has " + numtuples + stuple);
	if(DEBUG) 
		pp.println("<<" + stitle + " has " + numtuples + stuple);
	else
		pp.println(stitle + " has " + numtuples + stuple);

	 // unload data
	 unload();
   }         
   /**
   * Project the relation on an array of domains. Note that the relation 
   * itself is changed, so this method is for use within relation class 
   * for implementing other method, including assign(), external project 
   * method ...
   * Also note that in project, EQDK is regarded as EQ.
   *
   * @param   doms   the doms to project out
   *
   * NOTE: if doms = null, then project on itself
   */
   public void project(Domain[] doms)
   throws InterpretError
   {
	 boolean[] flags;

	 // check if project on itself which is used to eliminate duplicates.
	 if (doms == null || doms == domains) {
   // sort is on the relations referenced by surrogates 
   flags = new boolean[domains.length];
   for (int i = 0; i < domains.length; i++) 
	  flags[i] = false;

	   heapSort(domains, flags);
	 } else {
   // sort is on the relations referenced by surrogates 
   flags = new boolean[doms.length];
   for (int i = 0; i < doms.length; i++) 
	  flags[i] = false;

	   heapSort(doms, flags);
	   numattrs = doms.length;
   // after heapSort, numsortattrs may be larger than doms.length
	   numsortattrs = doms.length; 
   domains = doms; // REVISIT: SEE XXXXXX
	 }

	 if (numtuples <= 1) return;

	 int[] atypes = toTypes(domains);
	 int j = 0;
	 for (int i = 0; i < numtuples; i++) {
	   int result = compareTwoRows(data, i, data, j, numsortattrs, 
	  atypes, flags, domains, domains);
	   if ( result == GT || result == LT)
		 swapTwoRows(i, ++j);
	 }
	 numtuples = ++j;
   }      
   /**
   * Project the relation on an array of domains and return a relation.
   * Note that the relation itself is NOT changed. This method is for
   * use by Interpreter in implementing projection. 
   *
   * @param   doms      the doms to project out
   * @param   relname   name of returned relation
   *
   * NOTE: if doms.length == 1 && doms[0].name == ".bool", then return a 
   * singleton of attribute .bool with true value if there in at least one 
   * tuple in the relation, or false otherwise
   */
   //REVISIT: Should we really use myEnv, or have an other env passed to us
   public Relation project(Domain[] doms, String relname)
   throws InterpretError
   {
	 Relation rel = null;

	 // check if there is duplicate domains in project list
	 if (existDuplicateDomain(doms))
	   throw new InterpretError("duplicate attributes in project");

	 // for empty list in projector, doms contains one domain ".bool"
	 if (doms.length == 1 && doms[0].name.equals(".bool")) {
   rel = new Relation(relname, doms, 1, 1, 1, myEnv);
   rel.data = createDataArray(doms, 1);
   ((byte[]) rel.data[0])[0] 
	  = (numtuples == 0) ? BOOL_FALSE : BOOL_TRUE;
	 } else {
	   rel = duplicateLoadAll(); // duplicate and load data
	   // check if the relation contains all the domains in doms
	   if (!isDomainsIn(doms, rel.domains))
		 throw new InterpretError("invalid attribute list in projection");
	   rel.project(doms);
	   rel.name = relname; // set relation name
   rel.domains = doms; // REVISIT: SEE XXXXXX
	 }
	 rel.save();
	 unloadAll(rel);
	 return rel;
   }      
   public Relation renameTo(String newname)
   throws InterpretError
   {
	 if (name.equals(newname))
   return this;
	 File fsrc = new File(name);
	 File fdest = new File(newname);
	 if (fdest.exists()) fdest.delete();
	 if (fsrc.exists()) fsrc.renameTo(fdest);
	 else Global.println(name + " doesn't exist");//Yi Zheng March 25, 2002
	 name = newname;
	 return this;
   }      

   // save data onto disk file
   /*
   public void save()
   throws InterpretError
   {
	 //System.out.println("in save function.");
	 BlockOutputStream bout = null;
	 try {
	   bout = new BlockOutputStream(name, BUFFERSIZE);
	 } catch (IOException e) {
	   try {
		 if (bout != null) bout.close();
	   } catch (IOException ex) {
  } finally {
	  throw new InterpretError("IO error in saving : " + name);
   }
	 }

	 int[] types = toTypes(domains);  // domain types
	 try {
	   for (int i = 0; i < numtuples; i++) {
		 for (int j = 0; j < numattrs; j++) {
			switch(types[j]) {
			case INTEGER :
			  bout.writeInt(((int[]) data[j])[i]);
			  break;
			case NUMERIC :
			  bout.writeString(((number[]) data[j])[i].toString());
			  break;

			case STRING :
			case STMT :
			case EXPR :
			  bout.writeString(((String[]) data[j])[i]);
			  break;
			case BOOLEAN :
			  bout.writeByte(((byte[]) data[j])[i]);
			  break;
			case SHORT :
			  bout.writeShort(((short[]) data[j])[i]);
			  break;
			case LONG :
			case IDLIST :
			case COMP :
			  bout.writeLong(((long[]) data[j])[i]);
			  break;
			case TEXT :
			  bout.writeText(((String[]) data[j])[i]);
			  break;
			case FLOAT :
			  bout.writeFloat(((float[]) data[j])[i]);
			  break;
			case DOUBLE :
			  bout.writeDouble(((double[]) data[j])[i]);
			  break;
			}
		 } // loop on j
	   } // loop on i
	 } catch (IOException e) {
	   throw new InterpretError("IO error in saving : " + name);
	 } finally {
	   try {
		 if (bout != null) bout.close();
	   } catch (IOException ex) {}
	 }
   }      
   */

   // save data onto disk file
   // 	-- modified by Weizhong Sun, Jun 9th 1999
   public void save()
   throws InterpretError
   {
	 // initialize the pointers
 	 BlockOutputStream tout = null;
   	 try {
//    		tout = new DataOutputStream(new FileOutputStream(name, true));
	   	tout = new BlockOutputStream(name, BUFFERSIZE);
	    } catch (IOException e) {
    		try {
		    if (tout != null) tout.close();
		} catch (IOException ex) {
       		} finally {
       	    		throw new InterpretError("IO error in saving: " + name);
       	 	}
   	 }

//	 ToAscii ObjToAsi = new ToAscii();
	 int[] types = toTypes(domains);  // domain types
	   String space = new String("^F");
	   String ret = new String("\n");
	   String s = new String("");
	try {
	   for (int i = 0; i < numtuples; i++) {
	   	//	System.out.println("in save(), i = " + i);
		 for (int j = 0; j < numattrs; j++) {
		 //	System.out.println("in save(), j = " + j);
			switch(types[j]) {
			case INTEGER :
//			  ObjToAsi.ToAscii(tout, s.valueOf(((int[])data[j])[i]));
			  tout.writeMyBytes(INTEGER, s.valueOf(((int[])data[j])[i]));
			  break;
			case NUMERIC :
//                        ObjToAsi.ToAscii(tout,s.valueOf(((int[])data[j])[i]));
                          tout.writeString(((number[])data[j])[i].toString());
		          break;
			case UNIVERSAL ://from Andrey
			case ATTRIBUTE ://from Andrey
			case STRING :
			case STMT :
			case EXPR :
//System.out.println("before call ToAscii in case String: filename " + name + " String " + ((String[])data[j])[i]);
//			  ObjToAsi.ToAscii(tout, ((String[])data[j])[i]);
			  tout.writeString(((String[])data[j])[i]);
			  break;
			case BOOLEAN :
//			  ObjToAsi.ToAscii(tout, s.valueOf(((byte[])data[j])[i]));
			  tout.writeMyBytes(BOOLEAN, s.valueOf(((byte[])data[j])[i])); 
			  break;
			case SHORT :
//			  ObjToAsi.ToAscii(tout, s.valueOf(((short[]) data[j])[i]));
			  tout.writeMyBytes(SHORT, s.valueOf(((short[])data[j])[i]));
			  break;
			case LONG :
			case IDLIST :
			case COMP :
//			  ObjToAsi.ToAscii(tout, s.valueOf(((long[]) data[j])[i]));
			  tout.writeMyBytes(LONG, s.valueOf(((long[])data[j])[i]));
			  break;
			case TEXT :
//			  ObjToAsi.ToAscii(tout, ((String[]) data[j])[i]);
			  tout.writeText(((String[])data[j])[i]);
			  break;
			case FLOAT :
//System.out.println("before call ToAscii in case Float: filename " + name + " float " + ((float[])data[j])[i]);
//			  ObjToAsi.ToAscii(tout, s.valueOf(((float[]) data[j])[i]));
			  tout.writeMyBytes(FLOAT, s.valueOf(((float[])data[j])[i]));
			  break;
			case DOUBLE :
//			  ObjToAsi.ToAscii(tout, s.valueOf(((double[]) data[j])[i]));
			  tout.writeMyBytes(DOUBLE, s.valueOf(((double[])data[j])[i]));
			  break;
			}
			// insert the space after each field
//			ObjToAsi.ToAscii(tout, space);
//			tout.writeBytes(space);
			tout.writeByte(6);
		 } // loop on j
		// insert the return after each record
//		ObjToAsi.ToAscii(tout, ret);
		tout.writeBytes(ret);
	   } // loop on i
   	  tout.close();
   	  
   	  	  //added by Zongyan Sep 2002
	  String  path = System.getProperty("user.dir");
	  if ( Global.DataFileSet.containsKey(name)) {
		   String mode = Global.DataFileSet.get(name).toString();
	       Runtime rt = Runtime.getRuntime();	
	       Process pr = rt.exec("chmod "+ mode + " " + name);
	   }
	  else if (path.indexOf("public_aldatp") >=0)
	  {
		   Runtime rt = Runtime.getRuntime();	
	       Process pr = rt.exec("chmod 666 " + name);
	  }
	   // end Zongyan's addition

	} catch (IOException e) {}
   }      


   /**
   * <p>
   * This private select is used for internal use, to select on
   * index-th attribute, which must be of boolean type (by byte).
   * It does not check that the index is within the visible
   * attributes. This is particularly useful for implementing
   * selection, e.g.
   * </p> <p>
   * <pre>
   *    where A = 1 and B > 2 in R;
   * </pre>
   * </p> <p>
   * The expression (A = 1 and B > 2) is first evaluated on an
   * internally created boolean attribute, then invoke this method
   * to select the desired tuples. The internally created attribute
   * is not registered thus not visible to outside the relation.
   * This effect is most often desired, since we don't need to
   * bother project the attribute used for evaluate expression.
   * </p> <p>
   * Two special values dk and dc are not equal to true thus not
   * selected.
   * </p>
   */
   private void select(int index)
   {
//System.out.println("in void select(..), numtuples:" + numtuples);
	 if (numtuples == 0) return;

	 byte[] ba = (byte[]) data[index];

/*
// print datas
int[] types = toTypes(domains);  // domain types
for (int i = 0; i < numtuples; i++) {
 for (int j = 0; j < numattrs; j++) {
	switch(types[j]) {
	case INTEGER :
	  System.out.print(((int[])data[j])[i]); break;
	case STRING :
	case STMT :
	case EXPR :
	  System.out.print(((String[])data[j])[i]); break;
	case BOOLEAN :
	  System.out.print(((byte[])data[j])[i]); break;
	case SHORT :
	  System.out.print(((short[]) data[j])[i]); break;
	case LONG :
	case IDLIST :
	case COMP :
	  System.out.print(((long[]) data[j])[i]); break;
	case TEXT :
	  System.out.print(((String[]) data[j])[i]); break;
	case FLOAT :
	  System.out.print(((float[]) data[j])[i]); break;
	case DOUBLE :
	  System.out.print(((double[]) data[j])[i]); break;
	}
	// insert the space after each field
	System.out.print(" ");
 } // loop on j
 System.out.println(";");
}
// end of printing
*/

	 int j = 0;
	 for (int i = 0; i < numtuples; i++) {
//System.out.println("in void select(..), ba[i]:" + ba[i]);
	   if (ba[i] == TRUE)
		 swapTwoRows(i, j++);
	 }
	 numtuples = j;
   }      
   /**
   * Select the rows whose value on a certain domain is true.
   *
   * @param   dom       domain to select, must be of BOOLEAN type
   * @param   num       first num domains remained in the result
   * @param   relname   name of the returned relation
   */
   public Relation select(Domain dom, int num, String relname)
   throws InterpretError
   {
	 Relation rel = null;

	 if (dom.type != BOOLEAN)
	   throw new InterpretError("selected attribute is not of boolean");

	 int index = indexOfDomain(dom);
	 if (index == ERROR)
	   throw new InterpretError("selected attribute not exist");

	 rel = duplicate(true); // duplicate and load data
//System.out.println("in Relation select(..), index:" + index);
	 rel.select(index);
	 rel.name = relname; // set relation name
//System.out.println("in Relation select(..), relation name:" + relname + " numtuples " + numtuples + " numattrs " + num);
	 rel.numattrs = num; // set the number of attributes of result relation
	 // delete unused domains
	 Domain[] doms = new Domain[num];
	 for (int i = 0; i < num; i++)
	   doms[i] = rel.domains[i];
	 rel.domains = doms;
	 rel.save();
	 rel.unload(); // reclaim memory
	 return rel;
   }      
   /**
   * Select the rows whose value on a certain domain is true.
   *
   * @param   dom       domain to select, must be of BOOLEAN type
   * @param   relname   name of the returned relation
   */
   public Relation select(Domain dom, String relname)
   throws InterpretError
   {
	 return select(dom, numattrs, relname);
   }      
   public void setName(String name)
   {
	 this.name = name;
   } 
   
   //added by Yi Zheng April 23, 2002
   public void setEnv(Environment newEnv)
   {
   	this.myEnv = newEnv;
   	}     
   // Implicitly, sort is on the relations referenced by surrogates 
   // rather on the surrogate values themselves. If sort is on the
   // surrogate values themselves, build an array of flags and use
   // another sort method below.
   public void sort(Domain[] doms) 
   throws InterpretError
   {
	 boolean[] flags;

	 // implicitly, sort is on the relations referenced by surrogates 
	 if (doms == null || doms == domains) {
	   flags = new boolean[domains.length];
	   for (int i = 0; i < domains.length; i++) 
		 flags[i] = false;
	 } else {
   flags = new boolean[doms.length];
   for (int i = 0; i < doms.length; i++)
	  flags[i] = false;
	 }

	 heapSort(doms, flags);
   }      
   public void sort(Domain[] doms, boolean[] flags)
   throws InterpretError
   {
	 heapSort(doms, flags);
   }      
   // swap the domain and data array so that the domains in doms
   // come first
   public void swapColumns(Domain[] doms)
   {
	 int lena = doms.length;
	 int lenb = domains.length;
	 int len = lena + lenb;
	 Domain[] dom_tmp  = new Domain[len];
	 Object[] data_tmp = new Object[len];
	 for (int i = 0; i < lenb; i++) {
	   dom_tmp [i + lena] = domains[i];
	   data_tmp[i + lena] = data[i];
	 }
	 for (int i = 0; i < lena; i++) {
	   int index = indexOfDomain(doms[i]);
	   
	   dom_tmp [i] = doms[i];
	   data_tmp[i] = data[index];
	   dom_tmp [lena + index] = null;
	   data_tmp[lena + index] = null;
	 }
	 for (int i = 0, j = 0; i < len; i++) {
	   if (dom_tmp[i] != null) {
		 domains[j] = dom_tmp[i];
		 data   [j] = data_tmp[i]; j++;
	   }
	 }
   }      
   // swap two rows indexed by m and n respectively
   public void swapTwoRows(int m, int n)
   {
	 if (m == n) return;
	 for (int j = 0; j < numattrs; j++) {
	   switch(domains[j].type) {
	   case INTEGER : {
		 int[] ba = (int[]) data[j];
		 int tmp = ba[m]; ba[m] = ba[n]; ba[n] = tmp;
		 break; }
	   case UNIVERSAL ://from Andrey
	   case ATTRIBUTE ://from Andrey
	   case STRING :
	   case TEXT :
	   case STMT :
	   case EXPR : {
		 String[] ba = (String[]) data[j];
		 String tmp = ba[m]; ba[m] = ba[n]; ba[n] = tmp;
		 break; }
	   case NUMERIC:{ //from Yiyi
	       number[] ba = (number[])data[j];
	       number tmp = ba[m];
	       ba[m] = ba[n];
	       ba[n] = tmp;
	       break;
	   }
	   case BOOLEAN : {
		 byte[] ba = (byte[]) data[j];
		 byte tmp = ba[m]; ba[m] = ba[n]; ba[n] = tmp;
		 break; }
	   case SHORT : {
		 short[] ba = (short[]) data[j];
		 short tmp = ba[m]; ba[m] = ba[n]; ba[n] = tmp;
		 break; }
	   case LONG :
	   case COMP: //Yi Zheng May 1, 2002
	   case IDLIST : {
		 long[] ba = (long[]) data[j];
		 long tmp = ba[m]; ba[m] = ba[n]; ba[n] = tmp;
		 break; }
	   case FLOAT : {
		 float[] ba = (float[]) data[j];
		 float tmp = ba[m]; ba[m] = ba[n]; ba[n] = tmp;
		 break; }
	   case DOUBLE : {
		 double[] ba = (double[]) data[j];
		 double tmp = ba[m]; ba[m] = ba[n]; ba[n] = tmp;
		 break; }
	   }
	 }
   }    
    // return an array of attribute types.
   public int[] toTypes(Domain[] doms)
   {
	 int[] types = new int[doms.length]; // domain types
	 for (int i = 0; i < doms.length; i++) { 
	 	    types[i] = doms[i].type;
	 }
	 return types;
   }      

   // return an array of attribute types. -- Yi Zheng's version
   public int[] toTypes2(Domain[] doms)
   {
	 int[] types = new int[doms.length]; // domain types
	 for (int i = 0; i < doms.length; i++) { //modified by Yi Zheng March 21, 2002
	   if((this.name.compareTo(".rel") == 0) && (doms[i].name.compareTo(".rvc") ==0))
	   	types[i] = STRING; //Yi's addition to support pretty print of system relations
	   else if((this.name.compareTo(".dom") == 0) && (doms[i].name.compareTo(".type") == 0))
	    types[i] = STRING; //Yi's addition to support pretty print of system relations
	   else
	    types[i] = doms[i].type;
	 }
	 return types;
   }      
   public void unload()
   {
	 if (data != null) data = null;
   }      
   // recursive unload routine called by unloadAll() to unload all data
   protected void unload(Relation rel)
   throws InterpretError
   {
	 // unload all data for all nested domains it is defined on 
	 Domain[] doms = rel.domains;
	 for(int i = 0; i < doms.length; i++) { 
   String s = doms[i].name;
   if(doms[i].type == IDLIST && !s.equals(".id")) {
//PATCHANGE:
//       Relation r = reltable.getRel("." + s);
	  Relation r = myEnv.lookupRel("." + s, true);
	  r.unload();
	  unload(r);
   }
	 }
   }      
   // unload the data for relation rel and all the nested relations
   // it is defined on
   public void unloadAll(Relation rel)
   throws InterpretError
   {
	 rel.unload();
	 unload(rel);
   }      
   // write a tuple to BlockOutputStream, only attributes set by the 
   // corresponding flags are written
   /*
   private void writeAttributes(BlockOutputStream bout, Object[] dataArray, 
	 int index, int[] types, boolean[] flags)
   throws IOException
   {
	 int len = types.length;

	 for (int j = 0; j < len; j++) {
	   if (flags[j]) {
		 switch(types[j]) {
			case INTEGER :
			  bout.writeInt(((int[]) dataArray[j])[index]);
			  break;
                        case NUMERIC :
                            bout.writeInt(((number[]) dataArray[j])[index]);
			    break;

			case STRING :
			case STMT :
			case EXPR :
			  bout.writeString(((String[]) dataArray[j])[index]);
			  break;
			case BOOLEAN :
			  bout.writeByte(((byte[]) dataArray[j])[index]);
			  break;
			case SHORT :
			  bout.writeShort(((short[]) dataArray[j])[index]);
			  break;
			case LONG :
			case COMP: //May 1, 2002 Yi Zheng
			case IDLIST :
			  bout.writeLong(((long[]) dataArray[j])[index]);
			  break;
			case TEXT :
			  bout.writeText(((String[]) dataArray[j])[index]);
			  break;
			case FLOAT :
			  bout.writeFloat(((float[]) dataArray[j])[index]);
			  break;
			case DOUBLE :
			  bout.writeDouble(((double[]) dataArray[j])[index]);
			  break;
		 }
	   }
	 }
   }      
   */


   // write a tuple to BlockOutputStream in Ascii mode, only attributes set
   // by the corresponding flags are written
   //	-- modified by Weizhong Sun, Jun 14th 1999
   private void writeAttributes(BlockOutputStream bout, Object[] dataArray, 
	 int index, int[] types, boolean[] flags)
   throws IOException
   {
	 int len = types.length;
	 String s = new String("");

	 for (int j = 0; j < len; j++) {
	   if (flags[j]) {
		 switch(types[j]) {
			case INTEGER :
			  s = s.valueOf(((int[]) dataArray[j])[index]);
			  bout.writeMyBytes(INTEGER, s);
			  break;
		   case UNIVERSAL ://from Andrey
			case ATTRIBUTE ://from Andrey
			case STRING :
			case STMT :
			case EXPR :
			  s = ((String[]) dataArray[j])[index];
			  bout.writeString(s);
			  break;
			 case NUMERIC://from Yiyi
		     s = (((number[])dataArray[j])[index]).toString();
		     bout.writeString(s);
		     break;

			case BOOLEAN :
			  s = s.valueOf(((byte[]) dataArray[j])[index]);
			  bout.writeMyBytes(BOOLEAN, s);
			  break;
			case SHORT :
			  s = s.valueOf(((short[]) dataArray[j])[index]);
			  bout.writeMyBytes(SHORT, s);
			  break;
			case LONG :
			case COMP: //Yi Zheng May 1, 2002
			case IDLIST :
			  s = s.valueOf(((long[]) dataArray[j])[index]);
			  bout.writeMyBytes(LONG, s);
			  break;
			case TEXT :
			  s = ((String[]) dataArray[j])[index];
			  bout.writeText(s);
			  break;
			case FLOAT :
			  s = s.valueOf(((float[]) dataArray[j])[index]);
			  bout.writeMyBytes(FLOAT, s);
			  break;
			case DOUBLE :
			  s = s.valueOf(((double[]) dataArray[j])[index]);
			  bout.writeMyBytes(DOUBLE, s);
			  break;
		 }
//		bout.writeBytes("^F");
		bout.writeByte(6);	// Weizhong
	   }	// if
	 }	// for
	bout.writeBytes("\n");
   }      


   // write a dc's to BlockOutputStream, only attributes starting from 
   // offset and set by the corresponding flags are written
   /*
   private void writeDcAttributes(BlockOutputStream bout, int[] types, 
	 boolean[] flags, int offset)
   throws IOException
   {
	 int len = types.length;

	 for (int j = offset; j < len; j++) {
	   if (flags[j]) {
		 switch(types[j]) {
			case INTEGER :
			  bout.writeInt(INT_DC); break;
			case STRING :
			case STMT :
			case EXPR :
			  bout.writeString(STRING_DC); break;
			case BOOLEAN :
			  bout.writeByte(BOOL_DC); break;
			case SHORT :
			  bout.writeShort(SHORT_DC); break;
			case LONG :
			case COMP:
			case IDLIST :
			  bout.writeLong(LONG_DC); break;
			case TEXT :
			  bout.writeText(STRING_DC); break;
			case FLOAT :
			  bout.writeFloat(FLOAT_DC); break;
			case DOUBLE :
			  bout.writeDouble(DOUBLE_DC); break;
		 }
	   }
	 }
   }      
   */


   // write a dc's to BlockOutputStream in Ascii mode, only attributes
   // starting from offset and set by the corresponding flags are written
   // 	-- modified by Weizhong Sun, Jun 14th, 1999
   private void writeDcAttributes(BlockOutputStream bout, int[] types, 
	 boolean[] flags, int offset)
   throws IOException
   {
	 int len = types.length;
	 String s = new String("");

	 for (int j = offset; j < len; j++) {
	   if (flags[j]) {
		 switch(types[j]) {
			case INTEGER :
			  s = s.valueOf(INT_DC);
		 	  bout.writeMyBytes(INTEGER, s);
			  break;
			   case UNIVERSAL :
			case ATTRIBUTE :
			case STRING :
			case STMT :
			case EXPR :
			  s = s.valueOf(STRING_DC);
		 	  bout.writeString(s);
			  break;
			case BOOLEAN :
			  s = s.valueOf(BOOL_DC);
		 	  bout.writeMyBytes(BOOLEAN, s);
			  break;
			case SHORT :
			  s = s.valueOf(SHORT_DC);
		 	  bout.writeMyBytes(SHORT, s);
			  break;
			case LONG :
			case COMP: //Yi Zheng May 1, 2002
			case IDLIST :
			  s = s.valueOf(LONG_DC);
		 	  bout.writeMyBytes(LONG, s);
			  break;
			case TEXT :
			  s = s.valueOf(STRING_DC);
		 	  bout.writeText(s);
			  break;
			case FLOAT :
			  s = s.valueOf(FLOAT_DC);
		 	  bout.writeMyBytes(FLOAT, s);
			  break;
			case DOUBLE :
			  s = s.valueOf(DOUBLE_DC);
		 	  bout.writeMyBytes(DOUBLE, s);
			  break;
		 case NUMERIC://added by Yiyi
		     s = s.valueOf(NUMERIC_DC);
		     bout.writeMyBytes(NUMERIC,s);
		     break;

		 }	// switch
//		 bout.writeBytes("^F");
		 bout.writeByte(6);
	   }	// if
	 }	// for
	 bout.writeBytes("\n");
   }      
// for debug
public void printdata()
throws InterpretError
{
// print datas
load(true);
int[] types = toTypes(domains);  // domain types
for (int i = 0; i < numtuples; i++) {
 for (int j = 0; j < numattrs; j++) {
	switch(types[j]) {
	case INTEGER :
	  System.out.print(((int[])data[j])[i]); break;
        case NUMERIC :
//          System.out.print(((number [])data[j])[i].toString() ); break;
	    System.out.print(((number[])data[j])[i]);break; //from Yiyi

	   case UNIVERSAL ://from Andrey
	case ATTRIBUTE : //from Andrey
	case STRING :
	case STMT :
	case EXPR :
	  System.out.print(((String[])data[j])[i]); break;
	case BOOLEAN :
	  System.out.print(((byte[])data[j])[i]); break;
	case SHORT :
	  System.out.print(((short[]) data[j])[i]); break;
	case LONG :
	case IDLIST :
	case COMP :
	  System.out.print(((long[]) data[j])[i]); break;
	case TEXT :
	  System.out.print(((String[]) data[j])[i]); break;
	case FLOAT :
	  System.out.print(((float[]) data[j])[i]); break;
	case DOUBLE :
	  System.out.print(((double[]) data[j])[i]); break;
	}
	// insert the space after each field
	System.out.print(" ");
 } // loop on j
 System.out.println(";");
}
}
/*
	This function creates a relation 
    of all the attributes in the operand (relation).
    The output of this function is a relation on a single attribute 
	of type "ATTRIBUTE"  -- Andrey Rozenberg
*/
 public Relation AttribsOf(String relname)  throws InterpretError
   {
	 //creates a new domain "atr" of type ATTRIBUTE
	 // and adds it to the environment
	 Domain attr = new Domain("Attr",ATTRIBUTE,0,null);
	 myEnv.put(attr);
	 Domain[] domarray = new Domain[1];
	 domarray[0]=attr;
	 Relation rel = new Relation(relname,domarray,0,1,0,null);

	 //creates a data array 
	 Object[] dataArray = rel.createDataArray(domarray,numattrs);
	 for(int i=0;i<numattrs;i++)
	 {
		 String s = this.domains[i].name;// gets attribute names from the operand relation
	     ((String[]) dataArray[0])[i]=s;
	 }
	 rel.append(dataArray,numattrs); // appends data to the resulting relation
	
	 return rel;
   }      

   /**
   * jRelixNS - jRelix Nested Sigmajoin version
   * @author A. Chang  
   * ******** jRelixNS START OF CHANGES   *************
   */

   /**
   * Return the domains in the result of sigmaJoin to support type analysis.
   * 
   * @param   opcode   the join mode (i.e. SUP, ICOMP, etc.)
   * @param   relr     the other relation to join with 'this' relation.
   * @param   doml     domain array of 'this' relation to be joined on
   * @params  domr     domain array of relation rel to be joined on
   * @return  a Domain array of the result relation from a sigmaJoin
   * @since jRelixNS
   * @exception   InterpretError
   *
   * Note that if both doml and domr are null, the join is implicitly
   * on the common attributes of the two relations.
   */
   public Domain[] sigmaJoinType(int opcode, Domain[] doml,
          Relation relr, Domain[] domr)
   throws InterpretError
   {
	 Domain[] commonDomains;   // Domains that are common to rell and relr 
	 Domain[] leftJoinDomains; // join attributes from rell.
	 Domain[] rightJoinDomains;// join attributes from relr.

	 // Flags for special handling of case : R[X : sigmajoin : .id].X,
         // where X is a relation nested in R.
	 // true value of flattenNestRelSurrIdFlag means special handling is needed
         // to support flattening of nested relations.
	 boolean[] flattenNestRelSurrIdFlag;

         // Check for duplicate domains.
	 if (existDuplicateDomain(doml) 
            || existDuplicateDomain(domr))
	   throw new InterpretError("duplicate attributes in sigmajoin");

         // Find the common domains of the two relations.
	 commonDomains = commonDomains(this.domains, relr.domains);

	 // step 0. determine the join attributes leftJoinDomains 
         // and rightJoinDomains, check if there
	 // is any duplicate domain in the result
	 // check for existence of joined attributes ???
         // If the user specified domains are null, then
         // the common domains of both relations
         // are used as defualt domains to join on.
         // Otherwise, the specified domains are used as 
         // the domains to join on.
         //*****************************************************************
         // FORM:   L join R
         //*****************************************************************
	 if (doml == null && domr == null) {
	   leftJoinDomains  = new Domain[commonDomains.length];
	   rightJoinDomains = new Domain[commonDomains.length];
	   flattenNestRelSurrIdFlag = new boolean[commonDomains.length];
	   for (int i = 0; i < commonDomains.length; i++) {
	     leftJoinDomains[i]   = rightJoinDomains[i] = commonDomains[i];
	     flattenNestRelSurrIdFlag[i] = false; // nested attribute join .id not possible
	   }
	 }
	 else if (doml  != null && 
                  domr != null && 
                  doml.length == domr.length) 
         {
          //*****************************************************************
          // FORM:   L [X: join :Y] R
          //*****************************************************************
	   flattenNestRelSurrIdFlag = new boolean[doml.length];
	   for (int i = 0; i < doml.length; i++) {
             // check for type match 
	     if (doml[i].type != domr[i].type)
               throw new InterpretError("attribute type mismatch in sigma-join");
	     // default is no special handling for nested attribute join .id
	     flattenNestRelSurrIdFlag[i] = false;
  
             //*****************************************************************
             // FORM:  R[X: sigmajoin :.id].X  => 
             //          set flattenNestRelSurrIdFlag = true
             //*****************************************************************
	     // see if special handling for R[X : sigmajoin .id].X is
	     // necessary and set flattenNestRelSurrIdFlag is needed
	     // In this case, we are flattening a nested relation.
             // (Implemented as R[.id : sigmajoin : .id].X)
	     if (doml[i].type == IDLIST 
               && domr[i].type == IDLIST
               && (doml[i].name.equals(".id") 
                   || domr[i].name.equals(".id")))
               flattenNestRelSurrIdFlag[i] = true;

	     // if the joined on nested domains, the nested domains must
	     // defined on same set of attributes, not we don't force such
	     // requirement so that a nested domain can't join to the .id
	     // field in its corresponding dot relation
             // (Specified   as L[X  : sigmajoin :  Y]R)
             // (Implemented as L[.X : sigmajoin : .Y]R)
             //*****************************************************************
             // Restriction.  X(A,B,C) Y(A,B,C) L[W,X] R[Y,Z]: A,B,C is in order
             //*****************************************************************
	     if (doml[i].type == IDLIST
               && domr[i].type == IDLIST
	       && !doml[i].name.equals(".id") 
               && !domr[i].name.equals(".id")) 
             {
		Relation l = myEnv.lookupRel("." + doml[i].name, true);
		Relation r = myEnv.lookupRel("." + domr[i].name, true);
                // Next loop supercede this one for error checking,
                // but we get a more specific error detection here.
		if (!l.isDomainsIn(r.domains) || !r.isDomainsIn(l.domains))
	  throw new InterpretError(doml[i].name + " and " 
		 + domr[i].name + " not on same set of attributes");

		// A further and subtle restriction is that the order of the
		// attributes on which the two nested join attributes are
		// defined must be the same. This restriction is due the
		// sort order implicitly on the two nested join attributes
		// produced by the heapSort before the join process. It
		// is quite difficult to release this restriction if not
		// impossible.
		for (int j = 0; j < l.domains.length; j++) {
                  if (l.domains[j] != r.domains[j])
		    throw new InterpretError(doml[i].name + " and "
	            + domr[i].name + " not have same order of attributes");
		}
		 }
	   }


	   // check common attributes are joined in the same sequence.
           // If any user specified left domain that appear in the
           // commonDomains does not appear in the same order
           // in the user specified right domain, then it is an error.
           // If the user intends to change the sequence, relabelling
           // is required.
           // What happens if non-common domains are interspersed among
           // common domains? Would the effect not be similar as wrong
           // sequencing? 
	   for (int i = 0; i < doml.length; i++) {
                 if (isDomainIn(doml[i], commonDomains) 
                     && doml[i].type != domr[i].type)
			throw new 
		        InterpretError("relations not sigma-joined with common types");
           }

           // Passed all our checks here.  We can now use the specified domains.
	   leftJoinDomains  = doml;
           rightJoinDomains = domr;
	 } else {
	   throw new InterpretError("attributes number mismatch in mu-join");
	 }
         // After Step 0, we now know and have checked the domains to join on.

         // *********************************************************
         // VERIFIED:
         // 'this [ leftJoinDomains join rightJoinDomains ] relr'
         // where commonDomains is a subset of leftJoinDomains
         // and a subset of rightJoinDomains.
         // FORM:   L join R
	 //    leftJoinDomains[i]   = rightJoinDomains[i] = commonDomains[i];
         // FORM:   L [X: join :Y] R
	 //  leftJoinDomains  = doml;
         //  rightJoinDomains = domr;
         // *********************************************************
	 Domain[] leftGroupingDomains = nonJoinDomains(
	                 this.domains,leftJoinDomains);
	 Domain[] rightGroupingDomains = nonJoinDomains(
	                 relr.domains,rightJoinDomains);

	   // check if relations have unjoined domains in the other relation.
           // Does there exist some domains common to both relations that are
           // left out of the user specified domains?
           // If so, it is an error. This error will leave duplicates
           // in the domains after the sigma join is completed.
	   if (((leftGroupingDomains==null || leftGroupingDomains.length !=0) 
               && isDomainsIn(leftGroupingDomains, rightGroupingDomains)) ||
               ((rightGroupingDomains==null || rightGroupingDomains.length !=0)&& isDomainsIn(rightGroupingDomains, leftGroupingDomains)))
		 throw new 
		 InterpretError("relations not sigma-joined on all common attributes");

         // *********************************************************
         // VERIFIED:
         // 'this(leftJoinDomains, leftGroupingDomains)'
         // 'relr(rightJoinDomains, rightGroupingDomains)'
         // 'this [ leftJoinDomains : join : rightJoinDomains ] relr'
         // *********************************************************

	 // define the domain array for the result relation
	 Domain[] resultRelDomains =
                  new Domain[rightGroupingDomains.length+
                                    leftGroupingDomains.length];
	 int numTotalAttributesInResults = 0;
	 for (int i = 0; i < leftGroupingDomains.length; i++)
               resultRelDomains[numTotalAttributesInResults++] = 
                  leftGroupingDomains[i];
	 for (int i = 0; i < rightGroupingDomains.length; i++)
               resultRelDomains[numTotalAttributesInResults++] = 
                  rightGroupingDomains[i];

	 if (resultRelDomains.length==0) {
           Domain[] bool_doms = new Domain[1];
           bool_doms[0] = myEnv.lookupDom(".bool", true);
	   return bool_doms;
         } else
           return resultRelDomains;
   }

   /**
   * Sigma-join this relation with another relation on specified attributes and
   * return the result relation. The two relations participated in the
   * join are not changed. (Note that common attributes of the two
   * relations are used for null domain arrays). The existance of
   * domains in the relation is checked.
   *
   * @param   opcode   the join mode
   * @param   relr     the other relation to join with
   * @param   doml     domain array of this relation to be joined on
   * @params  domr     domain array of relation rel to be joined on
   * @params  relname  name the returned relation 
   * @since jRelixNS
   *
   * Note : Special handling for R[X : mujoin : .id].X is done, since in
   * this case, relation R must be sorted on nested attribute X according
   * to the surrogate values rather than the relations of these surrogates.
   * A special set of flags is used for methods heapSort and  
   * compareTwoRows.
   */
   public Relation sigmaJoin(int opcode, Domain[] doml,
          Relation relr, Domain[] domr, String relname)
   throws InterpretError
   {
//AC: DEBUG
//this.printdata();
//relr.printdata();
	 Domain[] commonDomains;   // Domains that are common to rell and relr 
	 Domain[] leftJoinDomains; // join attributes from rell.
	 Domain[] rightJoinDomains;// join attributes from relr.

	 // Flags for special handling of case : R[X : sigmajoin : .id].X,
         // where X is a relation nested in R.
	 // true value of flattenNestRelSurrIdFlag means special handling is needed
         // to support flattening of nested relations.
	 boolean[] flattenNestRelSurrIdFlag;

         // Check for duplicate domains.
	 if (existDuplicateDomain(doml) 
            || existDuplicateDomain(domr))
	   throw new InterpretError("duplicate attributes in sigmajoin");

         // Find the common domains of the two relations.
	 commonDomains = commonDomains(this.domains, relr.domains);

	 // step 0. determine the join attributes leftJoinDomains 
         // and rightJoinDomains, check if there
	 // is any duplicate domain in the result
	 // check for existence of joined attributes ???
         // If the user specified domains are null, then
         // the common domains of both relations
         // are used as defualt domains to join on.
         // Otherwise, the specified domains are used as 
         // the domains to join on.
         //*****************************************************************
         // FORM:   L join R
         //*****************************************************************
	 if (doml == null && domr == null) {
	   leftJoinDomains  = new Domain[commonDomains.length];
	   rightJoinDomains = new Domain[commonDomains.length];
	   flattenNestRelSurrIdFlag = new boolean[commonDomains.length];
	   for (int i = 0; i < commonDomains.length; i++) {
	     leftJoinDomains[i]   = rightJoinDomains[i] = commonDomains[i];
	     flattenNestRelSurrIdFlag[i] = false; // nested attribute join .id not possible
	   }
	 }
	 else if (doml  != null && 
                  domr != null && 
                  doml.length == domr.length) 
         {
          //*****************************************************************
          // FORM:   L [X: join :Y] R
          //*****************************************************************
	   flattenNestRelSurrIdFlag = new boolean[doml.length];
	   for (int i = 0; i < doml.length; i++) {
             // check for type match 
	     if (doml[i].type != domr[i].type)
               throw new InterpretError("attribute type mismatch in sigma-join");
	     // default is no special handling for nested attribute join .id
	     flattenNestRelSurrIdFlag[i] = false;
  
             //*****************************************************************
             // FORM:  R[X: sigmajoin :.id].X  => 
             //          set flattenNestRelSurrIdFlag = true
             //*****************************************************************
	     // see if special handling for R[X : sigmajoin .id].X is
	     // necessary and set flattenNestRelSurrIdFlag is needed
	     // In this case, we are flattening a nested relation.
             // (Implemented as R[.id : sigmajoin : .id].X)
	     if (doml[i].type == IDLIST 
               && domr[i].type == IDLIST
               && (doml[i].name.equals(".id") 
                   || domr[i].name.equals(".id")))
               flattenNestRelSurrIdFlag[i] = true;

	     // if the joined on nested domains, the nested domains must
	     // defined on same set of attributes, not we don't force such
	     // requirement so that a nested domain can't join to the .id
	     // field in its corresponding dot relation
             // (Specified   as L[X  : sigmajoin :  Y]R)
             // (Implemented as L[.X : sigmajoin : .Y]R)
             //*****************************************************************
             // Restriction.  X(A,B,C) Y(A,B,C) L[W,X] R[Y,Z]: A,B,C is in order
             //*****************************************************************
	     if (doml[i].type == IDLIST
               && domr[i].type == IDLIST
	       && !doml[i].name.equals(".id") 
               && !domr[i].name.equals(".id")) 
             {
		Relation l = myEnv.lookupRel("." + doml[i].name, true);
		Relation r = myEnv.lookupRel("." + domr[i].name, true);
                // Next loop supercede this one for error checking,
                // but we get a more specific error detection here.
		if (!l.isDomainsIn(r.domains) || !r.isDomainsIn(l.domains))
	  throw new InterpretError(doml[i].name + " and " 
		 + domr[i].name + " not on same set of attributes");

		// A further and subtle restriction is that the order of the
		// attributes on which the two nested join attributes are
		// defined must be the same. This restriction is due the
		// sort order implicitly on the two nested join attributes
		// produced by the heapSort before the join process. It
		// is quite difficult to release this restriction if not
		// impossible.
		for (int j = 0; j < l.domains.length; j++) {
                  if (l.domains[j] != r.domains[j])
		    throw new InterpretError(doml[i].name + " and "
	            + domr[i].name + " not have same order of attributes");
		}
		 }
	   }


	   // check common attributes are joined in the same sequence.
           // If any user specified left domain that appear in the
           // commonDomains does not appear in the same order
           // in the user specified right domain, then it is an error.
           // If the user intends to change the sequence, relabelling
           // is required.
           // What happens if non-common domains are interspersed among
           // common domains? Would the effect not be similar as wrong
           // sequencing? 
	   for (int i = 0; i < doml.length; i++) {
                 if (isDomainIn(doml[i], commonDomains) 
                     && doml[i].type != domr[i].type)
			throw new 
		        InterpretError("relations not sigma-joined with common types");
           }

           // Passed all our checks here.  We can now use the specified domains.
	   leftJoinDomains  = doml;
           rightJoinDomains = domr;
	 } else {
	   throw new InterpretError("attributes number mismatch in mu-join");
	 }
         // After Step 0, we now know and have checked the domains to join on.

         // *********************************************************
         // VERIFIED:
         // 'this [ leftJoinDomains join rightJoinDomains ] relr'
         // where commonDomains is a subset of leftJoinDomains
         // and a subset of rightJoinDomains.
         // FORM:   L join R
	 //    leftJoinDomains[i]   = rightJoinDomains[i] = commonDomains[i];
         // FORM:   L [X: join :Y] R
	 //  leftJoinDomains  = doml;
         //  rightJoinDomains = domr;
         // *********************************************************
	 Domain[] leftGroupingDomains = nonJoinDomains(
	                 this.domains,leftJoinDomains);
	 Domain[] rightGroupingDomains = nonJoinDomains(
	                 relr.domains,rightJoinDomains);

	   // check if relations have unjoined domains in the other relation.
           // Does there exist some domains common to both relations that are
           // left out of the user specified domains?
           // If so, it is an error. This error will leave duplicates
           // in the domains after the sigma join is completed.
	   if (((leftGroupingDomains==null || leftGroupingDomains.length !=0) 
               && isDomainsIn(leftGroupingDomains, rightGroupingDomains)) ||
               ((rightGroupingDomains==null || rightGroupingDomains.length !=0)&& isDomainsIn(rightGroupingDomains, leftGroupingDomains)))
		 throw new 
		 InterpretError("relations not sigma-joined on all common attributes");

         // *********************************************************
         // VERIFIED:
         // 'this(leftJoinDomains, leftGroupingDomains)'
         // 'relr(rightJoinDomains, rightGroupingDomains)'
         // 'this [ leftJoinDomains : join : rightJoinDomains ] relr'
         // *********************************************************
         // Order the left domains for sigma join.
	 Domain[] sigmaJoinOrderedLeftDomains =
                         orderSigmaJoinDomains(leftGroupingDomains,
	                 leftJoinDomains);
         // Order the right domains for sigma join.
	 Domain[] sigmaJoinOrderedRightDomains =
                         orderSigmaJoinDomains(rightGroupingDomains,
	                 rightJoinDomains);

         // Set the Left nested rel flattenNestRelSurrIdFlags for
         // the domains for sigma join.
         boolean[] sigmaJoinOrderedLeftNestRelSurrIdFlags = 
                  new boolean[sigmaJoinOrderedLeftDomains.length];
	 for (int i = 0; i < leftGroupingDomains.length; i++) 
                  sigmaJoinOrderedLeftNestRelSurrIdFlags[i] = false; 
	 for (int i = leftGroupingDomains.length; 
              i < sigmaJoinOrderedLeftDomains.length; i++)
                  sigmaJoinOrderedLeftNestRelSurrIdFlags[i] = 
                      flattenNestRelSurrIdFlag[i-leftGroupingDomains.length];
         // Set the Right flattenNestRelSurrIdFlags for the domains
         // for sigma join.
         boolean[] sigmaJoinOrderedRightNestRelSurrIdFlags = 
                   new boolean[sigmaJoinOrderedRightDomains.length];
	 for (int i = 0; i < rightGroupingDomains.length; i++) 
                  sigmaJoinOrderedRightNestRelSurrIdFlags[i] = false; 
	 for (int i = rightGroupingDomains.length; 
              i < sigmaJoinOrderedRightDomains.length; i++)
                  sigmaJoinOrderedRightNestRelSurrIdFlags[i] = 
                      flattenNestRelSurrIdFlag[i-rightGroupingDomains.length];

         boolean[] sigmaJoinOrderedRightInResultsFlags = 
                   new boolean[sigmaJoinOrderedRightDomains.length];
         boolean[] sigmaJoinOrderedLeftInResultsFlags = 
                   new boolean[sigmaJoinOrderedLeftDomains.length];
         boolean[] sigmaJoinOrderedRightNonGroupingFlags = 
                   new boolean[sigmaJoinOrderedRightDomains.length];
         boolean[] sigmaJoinOrderedLeftNonGroupingFlags = 
                   new boolean[sigmaJoinOrderedLeftDomains.length];

	 for (int i = 0; i < leftGroupingDomains.length; i++) {
                  sigmaJoinOrderedLeftInResultsFlags[i] = true;
                  sigmaJoinOrderedLeftNonGroupingFlags[i] = false;
         }
	 for (int i = leftGroupingDomains.length; 
              i < sigmaJoinOrderedLeftDomains.length; i++){
                  sigmaJoinOrderedLeftInResultsFlags[i] = false; 
                  sigmaJoinOrderedLeftNonGroupingFlags[i] = true;
         }
	 for (int i = 0; i < rightGroupingDomains.length; i++) {
                  sigmaJoinOrderedRightInResultsFlags[i] = true; 
                  sigmaJoinOrderedRightNonGroupingFlags[i] = false;
         }
	 for (int i = rightGroupingDomains.length; 
              i < sigmaJoinOrderedRightDomains.length; i++){
                  sigmaJoinOrderedRightInResultsFlags[i] = false; 
                  sigmaJoinOrderedRightNonGroupingFlags[i] = true;
         }

	 // step 1. duplicated and sort the copies of the relations to be joined
	 // then create the result relation
	 Relation LeftRel = this.duplicateLoadAll();
	 // note flattenNestRelSurrIdFlag tells if the sort is on the
         // surrogate values themselves
	 LeftRel.heapSort(sigmaJoinOrderedLeftDomains,
                            sigmaJoinOrderedLeftNestRelSurrIdFlags); 

	 Relation RightRel = relr.duplicateLoadAll(); 
	 RightRel.heapSort(sigmaJoinOrderedRightDomains,
                             sigmaJoinOrderedRightNestRelSurrIdFlags);
         // *********************************************************
         // VERIFIED:
         // 'LeftRel(leftJoinDomains, leftGroupingDomains)'
         // SORTED: by  sigmaJoinOrderedLeftDomains
         // 'RightRel(rightJoinDomains, rightGroupingDomains)'
         // SORTED: by  sigmaJoinOrderedRightDomains
         // 'LeftRel[ leftJoinDomains : join : rightJoinDomains ] RightRel'
         // *********************************************************

         // Set the number of tuples.
	 int numTuplesLeft  = LeftRel.numtuples;
         int numTuplesRight = RightRel.numtuples;

	 // length of joined attributes
         int numDomainsToJoin   = leftJoinDomains.length;      

         // Set up pointers to data array.
	 Object[] leftDataTuples = LeftRel.data;
         Object[] rightDataTuples = RightRel.data;
	 // Convert type array from domain array of joined attributes
         int[] leftRightJoinTypes  = toTypes(leftJoinDomains); 
	 int[] leftRelAllTypes  = toTypes(LeftRel.domains);
	 int[] rightRelAllTypes = toTypes(RightRel.domains);
         int[] rightGroupingTypes = toTypes(rightGroupingDomains);
         int[] leftGroupingTypes  = toTypes(leftGroupingDomains);

	 int numLeftRelDomains  = LeftRel.domains.length;
         int numRightRelDomains = RightRel.domains.length;

	 // define the domain array for the result relation
	 Domain[] resultRelDomains = new Domain[ leftGroupingDomains.length 
                                                +rightGroupingDomains.length];
	 int numTotalAttributesInResults = 0;
	 for (int i = 0; i < leftGroupingDomains.length; i++)
               resultRelDomains[numTotalAttributesInResults++] = 
                  sigmaJoinOrderedLeftDomains[i];
	 for (int i = 0; i < rightGroupingDomains.length; i++)
               resultRelDomains[numTotalAttributesInResults++] = 
                  sigmaJoinOrderedRightDomains[i];

         // initialize output relation
         // the number of tuples of the result relation
	 int resultRelNumTuples = 0;   
//REVISIT: Should we really use myEnv, or have an other env passed to us
      //Andy, May 12, 2002

         // Flag indicating that the sigma join grouping is changing
	 boolean leftGroupingIsChangingFlag;
	 boolean rightGroupingIsChangingFlag;
	 int currLeftTupleIndex          = 0;
	 int currRightTupleIndex         = 0;
	 int currLeftGroupingIndex       = 0;

         // Boolean flags showing if the set relationships have been violated
         // as we process the tuples.
	 boolean isSup, isSub, isSep;

         // Special handling for one or more EMPTY relations that are WITHOUT
         // any non-join or grouping attributes.
         // This special handling is needed because the while loop below
         // can processes tuple-by-tuple, and:
         // - An EMPTY relation that has some non-join or grouping attributes
         //   would imply that there exist no groupings of sets to be
         //   compared in the relation and the Cartesian product of the
         //   non-join attribute values from the left and the right relations
         //   would be empty; thus, the while loop below would 
         //   trivially yield the correct result relation that is empty.
         // - An EMPTY relation that is WITHOUT non-join or grouping attributes
         //   would imply that the entire relation is a single grouping/set
         //   to be compared, and that set is the empty set;
         //   the while loop below would ignore this empty group/set
         //   (as it has no-tuples) and yield the incorrect empty result
         //   relation.
         // If the empty relation have non-join or grouping attributes
         // processing will fall through the while loop below normally.
	 if ((numTuplesLeft ==0 && leftGroupingDomains.length ==0)
           ||(numTuplesRight==0 && rightGroupingDomains.length==0)) 
         {
             // Defaults to both relations (and therefore all sets/groupings)
             // are empty.
             isSup = isSub = isSep = true;
             // No more than one relation has tuples here.
	     if (numTuplesRight != 0) 
             {
	       isSup = false;
               if (isSigmaJoinTupleTrue(opcode,isSup,isSub,isSep))
               {
                 Relation temp_rel;
                 // We know there are no leftGroupingDomains here.
                 if (rightGroupingDomains.length!=0) 
                 {
                   // project on right domains only.
	           temp_rel = RightRel.project(rightGroupingDomains,relname);
                 }
                 else
                 {
                   Domain[] bool_doms = new Domain[1];
                   bool_doms[0] = myEnv.lookupDom(".bool", true);
	           temp_rel = RightRel.project(bool_doms,relname);
                 }
	         unloadAll(LeftRel);
	         unloadAll(RightRel);
                 return temp_rel;
               }
               // The result relation is going to be empty.
               // Just let it fall thru the while loop below.
             }
	     else if (numTuplesLeft != 0) 
             {
	       isSub = false;
               if (isSigmaJoinTupleTrue(opcode,isSup,isSub,isSep))
               {
                 Relation temp_rel;
                 // We know there are no rightGroupingDomains here.
                 if (leftGroupingDomains.length!=0)
                   // project on leftdomains only.
	           temp_rel = LeftRel.project(leftGroupingDomains,relname);
                 else
                 {
                   Domain[] bool_doms = new Domain[1];
                   bool_doms[0] = myEnv.lookupDom(".bool", true);
	           temp_rel = LeftRel.project(bool_doms,relname);
                 }
	         unloadAll(LeftRel);
	         unloadAll(RightRel);
                 return temp_rel;
               }
               // The result relation is going to be empty.
               // Just let it fall thru the while loop below.
             }
             // Both relations are empty.
	     else if ((numTuplesLeft ==0 && leftGroupingDomains.length ==0)
             &&(numTuplesRight==0 && rightGroupingDomains.length==0)) 
             {
               // The result relation is going to be nullary.
               Relation temp_rel;
               Domain[] bool_doms = new Domain[1];
               bool_doms[0] = myEnv.lookupDom(".bool", true);
               if (isSigmaJoinTupleTrue(opcode,isSup,isSub,isSep))
	          LeftRel.numtuples = 1;
               else
	          LeftRel.numtuples = 0;
	       temp_rel = LeftRel.project(bool_doms,relname);
	       unloadAll(LeftRel);
	       unloadAll(RightRel);
               return temp_rel;
             }
         }

         // The relation containing the sigma-join results.      
      //end of May 12, 2002   
	 Relation rel = new Relation(relname, resultRelDomains, 0, 
                                       resultRelDomains.length, 0, myEnv);

	 BlockOutputStream bout = null;
	 try {
	   bout = new BlockOutputStream(rel.name, BUFFERSIZE);
	 } catch (IOException e) {
	   try {
		 if (bout != null) bout.close();
	   } catch (IOException ex) {
	   } finally {
		 throw new InterpretError("IO error in writing sigma-join result");
	   }
	 }
/*
         // Flag indicating that the sigma join grouping is changing
	 boolean leftGroupingIsChangingFlag;
	 boolean rightGroupingIsChangingFlag;
	 int currLeftTupleIndex          = 0;
	 int currRightTupleIndex         = 0;
	 int currLeftGroupingIndex  = 0;
	 // define the domain array for the result relation

	 boolean isSup = true;
	 boolean isSub = true;
	 boolean isSep = true;
	 int ntuples = 0;   // the number of tuples of the result relation
 *///Andy, May 12, 2002
          // Prepare for the 1st Inner Loop
         // Flags default to both sets are empty.
         isSup = isSub = isSep = true;

     try {         //Start first Inner Loop
	 while (currLeftTupleIndex<numTuplesLeft && currRightTupleIndex<numTuplesRight) 
         {//This while loop contains both the Inner and the Outer Loop 
          //for the sigma-join implementation algorithm.
           int intraGroupingResult = compareTwoRowsWithDataOffset(
             leftDataTuples,  currLeftTupleIndex,
             rightDataTuples, currRightTupleIndex, 
             numDomainsToJoin,
             leftRightJoinTypes, flattenNestRelSurrIdFlag, 
             leftJoinDomains, rightJoinDomains,
             leftGroupingDomains.length,
             rightGroupingDomains.length);

	   if (currLeftTupleIndex+1<numTuplesLeft) {
 	     if (leftGroupingDomains.length != 0) {
	       leftGroupingIsChangingFlag = (0!=compareTwoRows(
               leftDataTuples,  currLeftTupleIndex,
               leftDataTuples,  currLeftTupleIndex+1,
               leftGroupingDomains.length, 
               leftGroupingTypes, sigmaJoinOrderedLeftNestRelSurrIdFlags, 
               leftGroupingDomains, leftGroupingDomains));
             }
             else
               leftGroupingIsChangingFlag = false;
           }
           else {
             leftGroupingIsChangingFlag = true;
           }

 	   if (currRightTupleIndex+1<numTuplesRight) {
 	     if (rightGroupingDomains.length != 0) {
	       rightGroupingIsChangingFlag = (0!=compareTwoRows(
                 rightDataTuples, currRightTupleIndex, 
                 rightDataTuples, currRightTupleIndex+1,
                 rightGroupingDomains.length, 
                 rightGroupingTypes, sigmaJoinOrderedRightNestRelSurrIdFlags, 
                 rightGroupingDomains, rightGroupingDomains));
             }
             else
               rightGroupingIsChangingFlag = false;
           }
           else {
             rightGroupingIsChangingFlag = true;
           }

           boolean advanceToNextGroup = false;
           if ( intraGroupingResult == EQDK ) {
             if (!leftGroupingIsChangingFlag) 
	       currLeftTupleIndex++;
             else 
               advanceToNextGroup = true;

             if (!rightGroupingIsChangingFlag) 
	       currRightTupleIndex++;
             else 
               advanceToNextGroup = true;
           } 
           if ( intraGroupingResult == EQ) {
             isSep = false; 
             if (!leftGroupingIsChangingFlag) 
	       currLeftTupleIndex++;
             else 
               advanceToNextGroup = true;

             if (!rightGroupingIsChangingFlag) 
	       currRightTupleIndex++;
             else 
               advanceToNextGroup = true;
           }
           else if (intraGroupingResult == LT) {
             isSub = false; 
             if (!leftGroupingIsChangingFlag) 
	       currLeftTupleIndex++;
             else 
               advanceToNextGroup = true;
           }
           else { // if (intraGroupingResult ==GT) 
             isSup = false; 
             if (!rightGroupingIsChangingFlag) 
	       currRightTupleIndex++;
             else 
               advanceToNextGroup = true;
           }

           if (advanceToNextGroup) {//Finish current Inner Loop then start next
             if (rightGroupingIsChangingFlag && leftGroupingIsChangingFlag ) {
               if (intraGroupingResult != 0)
               {
 	         if (numTuplesRight > 0)
                    isSup = false; 
 	         if (numTuplesLeft > 0)
                    isSub = false; 
               }
             }
             else if (rightGroupingIsChangingFlag && 
                      !leftGroupingIsChangingFlag ) {
               isSub = false; 
             }
             else if (!rightGroupingIsChangingFlag && 
                      leftGroupingIsChangingFlag ) {
               isSup = false; 
             }

               if (isSigmaJoinTupleTrue(opcode,isSup,isSub,isSep))
               {
	         if (leftGroupingDomains.length!=0){
		   writeAttributes(bout, leftDataTuples, currLeftGroupingIndex,
                    toTypes(sigmaJoinOrderedLeftDomains), 
                    sigmaJoinOrderedLeftInResultsFlags);
                 }
	         if (rightGroupingDomains.length!=0){
		   writeAttributes(bout, rightDataTuples, currRightTupleIndex,
                    toTypes(sigmaJoinOrderedRightDomains), 
                    sigmaJoinOrderedRightInResultsFlags);
                 }
	         resultRelNumTuples++;
               }
               // Flags default to both sets are empty.
               isSup = isSub = isSep = true;

                 while(!rightGroupingIsChangingFlag)
                 {
	           if (currRightTupleIndex+1<numTuplesRight) {
	             rightGroupingIsChangingFlag = (0!=compareTwoRows(
                      rightDataTuples,  currRightTupleIndex,
                      rightDataTuples,  currRightTupleIndex+1,
                      rightGroupingDomains.length, 
                      rightGroupingTypes, sigmaJoinOrderedRightNestRelSurrIdFlags, 
                      rightGroupingDomains, rightGroupingDomains));
                   }
                   else {
                     rightGroupingIsChangingFlag = true;
                   }
                   if (!rightGroupingIsChangingFlag)
                     currRightTupleIndex++;
                 }

                 while(!leftGroupingIsChangingFlag)
                 {
	           if (currLeftTupleIndex+1<numTuplesLeft) {
	             leftGroupingIsChangingFlag = (0!=compareTwoRows(
                      leftDataTuples,  currLeftTupleIndex,
                      leftDataTuples,  currLeftTupleIndex+1,
                      leftGroupingDomains.length, 
                      leftGroupingTypes, sigmaJoinOrderedLeftNestRelSurrIdFlags, 
                      leftGroupingDomains, leftGroupingDomains));
                   }
                   else {
                       leftGroupingIsChangingFlag = true;
                   }
                   if (!leftGroupingIsChangingFlag)
                     currLeftTupleIndex++;
                 }
               if (currRightTupleIndex+1 < numTuplesRight) {
	         currRightTupleIndex++;
                 currLeftTupleIndex = currLeftGroupingIndex;
               }
               else if (currLeftTupleIndex+1 < numTuplesLeft) {
	           currLeftTupleIndex++;
                   currLeftGroupingIndex= currLeftTupleIndex;
	           currRightTupleIndex = 0;
               } else {
	           currLeftTupleIndex++;
	           currRightTupleIndex++;
  //                 continue;
               }
                            //Start the next Inner Loop of the sigma-join 
             // implementation algorithm.

             }
           }
	 } catch (IOException e) {
	   try {
		 bout.close();
	   } catch (IOException ex) { 
	   }
	   throw new InterpretError("IO error in writing sigma-join result");
	 }

	 // step 5. set fields and return
	 rel.numtuples = resultRelNumTuples;
	 try {
	   bout.close();
	 } catch (IOException e) {
	   throw new InterpretError("IO error in writing mu-join result");
	 }
	 unloadAll(LeftRel);
	 unloadAll(RightRel);
	 if (resultRelDomains.length==0) {
           Domain[] bool_doms = new Domain[1];
           bool_doms[0] = myEnv.lookupDom(".bool", true);
	   return rel.project(bool_doms,relname);
         }
         else {
	   return rel;
         }
   }      

   /**
   * Returns a domain array whose elements are the grouping/non-join domains 
   * for sigma-join.
   *
   * @param allDomains a Domain array containing all domains for a relation.
   * @param joinDomains a Domain array with only join domains for a relation.
   * @return a Domain array of set of allDomains minus the set of joinDomains 
   * @since jRelixNS
   */
   private Domain[] nonJoinDomains(Domain[] allDomains,Domain[] joinDomains)
   {
     Domain[] dom = new Domain[allDomains.length - joinDomains.length]; 
     int j = 0; 
     for (int i = 0; i < allDomains.length; i++) { 
       if (!isDomainIn(allDomains[i], joinDomains))
         dom[j++] = allDomains[i];
     }
     return dom;
   }

   /**
   * Returns a domain array whose elements are ordered for sigmajoin,
   * with nonJoinDomains first followed by joinDomains.
   *
   * @param  nonJoinDomains
   * @param  joinDomains
   * @return a Domain array whose elements are ordered for sigmajoin
   * @since jRelixNS
   */
   protected Domain[] orderSigmaJoinDomains(
              Domain[] nonJoinDomains, Domain[] joinDomains)
   {
     int j = 0;
     Domain[] dom = new Domain[joinDomains.length + nonJoinDomains.length];
     for (int i = 0; i < nonJoinDomains.length; i++) 
       dom[j++] = nonJoinDomains[i];
     for (int i = 0; i < joinDomains.length; i++) 
       dom[j++] = joinDomains[i];
     return dom;
   }

   /**
   * Returns true if the row with the flags matches the sigma-join opcode.
   * Otherwise, returns false.
   * 
   * @param  sigmaJoinOpcode the sigma-join mode
   * @param  isSup    boolean flag based on the sigma-join algorithm
   * @param  isSub    boolean flag based on the sigma-join algorithm
   * @param  isSep    boolean flag based on the sigma-join algorithm
   * @return  a boolean true if sigma-join opcode matches, false otherwise.
   * @since jRelixNS
   * @exception   InterpretError
   */
   protected boolean isSigmaJoinTupleTrue(
              int sigmaJoinOpcode, boolean isSup, boolean isSub, boolean isSep)
   throws InterpretError
   {
     switch(sigmaJoinOpcode)
     {
       case OP_ICOMP :
       case OP_NOTIEJOIN :
         return(!isSep);
       case OP_EQJOIN :
         return(isSup&&isSub);
       case OP_GTJOIN :
         return(isSup&&!isSub);
       case OP_GEJOIN :
         return(isSup);
       case OP_LTJOIN :
         return(!isSup&&isSub);
       case OP_LEJOIN :
         return(isSub);
       case OP_IEJOIN :
       case OP_NOTICOMP :
         return(isSep);
       case OP_NOTEQJOIN :
         return(!isSub||!isSup);
       case OP_NOTGTJOIN :
         return(!isSup||isSub);
       case OP_NOTGEJOIN :
         return(!isSup);
       case OP_NOTLTJOIN :
         return(isSup||!isSub);
       case OP_NOTLEJOIN :
         return(!isSub);
       default:
         throw new InterpretError("isSigmaJoinTupleTrue: unknown sigma-join");
     }
   }

   /** 
    * Extends the compareTwoRows method so that
    * if the two rows are equal AND if they are both DK's,
    * then a EQDK value is returned instead of a regular EQ to indicate
    * this special case of DK equivalence.
    *
    * Compares the first len attributes of row m in datal with the first 
    * len attributes of row n in datar, the types of attributes are
    * given by types.  Handles DK and DC values.
    *
    * Note that flags is a boolean array tells if sort is on the surrogate
    * values of nested attributes. They are intended for special handling
    * of such join : R[X : ijoin : .id].X
    *
    * @param datal An Object array which contains the left data tuples.
    * @param m     index into the datal array
    * @param datar An Object array which contains the right data tuples.
    * @param n     index into the datar array
    * @param len   First len attributes of datal and datar are compared.
    * @param types array containing types of attributes being compared.
    * @param flags boolean array flagging nested relation surrogate domains.
    * @param domsl Domains array for the left data tuples. 
    * @param domsr Domains array for the right data tuples. 
    * @return LT   if ROW(m) < ROW(n)
    *         EQ   if ROW(m) = ROW(n)
    *         EQDK if ROW(m) = ROW(n) and ROW(m) = DK
    *         GT   if ROW(m) > ROW(n)
    * @exception   InterpretError from other called methods 
    * @see Relation#compareTwoRows 
    * @see TypeConstants
    * @since jRelixNS
    */
   public int compareTwoRowsWithDk(Object[] datal, int m, Object[] datar, int n,
	 int len, int[] types, boolean[] flags, Domain[] domsl, Domain[] domsr)
	 throws InterpretError
   {
     for (int j = 0; j < len; j++) {
       switch(types[j])
       {
         case INTEGER : {
           int bm = ((int[]) datal[j])[m];
           int bn = ((int[]) datar[j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else if (bm == bn && bm == INT_DK) return EQDK;
           else break; 
         }
         case STRING :
         case TEXT :
         case STMT :
         case EXPR : {
           String bm = ((String[]) datal[j])[m];
           String bn = ((String[]) datar[j])[n];
           int result = compareTwoStrings(bm, bn);
           if (result == GT || result == LT) return result;
           else if (result == GT || result == LT) return result;
           else if (result == EQ && bm.equals(STRING_DK)) return EQDK;
           break; 
         }
       case NUMERIC:{ //from Yiyi
	   number bm = ((number[])datal[j])[m];
	   number bn = ((number[])datar[j])[n];
	   int result = bm.compareTo(bn);
	   int bmcom = bm.compareTo(NUMERIC_DK);
	   if(result > 0)return GT;
	   else if(result < 0) return LT;
	   else if(result == 0 && bmcom == 0) return EQDK;
	   break;
       }

         case BOOLEAN : {
           byte bm = ((byte[]) datal[j])[m];
           byte bn = ((byte[]) datar[j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else if (bm == bn && bm == BOOL_DK) return EQDK;
           else break; 
         }
         case SHORT : {
           short bm = ((short[]) datal[j])[m];
           short bn = ((short[]) datar[j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else if (bm == bn && bm == SHORT_DK) return EQDK;
           else break; 
         }
         case LONG : {
           long bm = ((long[]) datal[j])[m];
           long bn = ((long[]) datar[j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else if (bm == bn && bm == LONG_DK) return EQDK;
           else break; 
         }
         case IDLIST : {
           long bm = ((long[]) datal[j])[m];
           long bn = ((long[]) datar[j])[n];
           if (flags[j] || domsl[j].name.equals(".id")
               || domsr[j].name.equals(".id")) {
             if (bm > bn) return GT;
             if (bm < bn) return LT;
             break;
           }
           if ((bm == bn) && bm == LONG_DK) 
             return EQDK;

           if (bm == bn) // equal if the surrogates are the same
             break;
           if (bm == LONG_DK) {
             if (bn != LONG_DK) return LT;
             else break;
           } else if (bm == LONG_DC) {
             if (bn == LONG_DK) return GT;
             else if (bn == LONG_DC) break;
             else return LT;
           }
           if (bn == LONG_DK) {
             if (bm != LONG_DK) return GT;
             else if (bm == LONG_DK) return EQDK;
             else break;
           } else if (bn == LONG_DC) {
             if (bm == LONG_DK) return LT;
             else if (bm == LONG_DC) break;
             else return GT;
           }
           Relation doml = myEnv.lookupRel("."+domsl[j].name, true);
           Relation rl = doml.getRelation(bm);
           Relation domr = myEnv.lookupRel("."+domsr[j].name, true);
           Relation rr = domr.getRelation(bn);
           int result = rl.compareRelation(rr);
           if (result != EQ) return result;
           else break;
         }
         case FLOAT : {
           float bm = ((float[]) datal[j])[m];
           float bn = ((float[]) datar[j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else if (bm == bn && bm == FLOAT_DK) return EQDK;
           else break;
         }
         case DOUBLE : {
           double bm = ((double[]) datal[j])[m];
           double bn = ((double[]) datar[j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else if (bm == bn && bm == DOUBLE_DK) return EQDK;
           else break;
         }
       } // end of switch
     }
     return EQ;
   }

   /** 
    * Extends the compareTwoRows method so that this
    * compares the len attributes, starting with
    * the leftJoinDataOffset, of row m in datal, 
    * with  len attributes, starting with the leftJoinDataOffset,
    * of row n in datar. The types of attributes are
    * given by types.
    *
    * Note that flags is a boolean array tells if sort is on the surrogate
    * values of nested attributes. They are intended for special handling
    * of such join : R[X : ijoin : .id].X
    *
    * @param datal An Object array which contains the left data tuples.
    * @param m     index into the datal array
    * @param datar An Object array which contains the right data tuples.
    * @param n     index into the datar array
    * @param len   First len attributes of datal and datar are compared.
    * @param types array containing types of attributes being compared.
    * @param flags boolean array flagging nested relation surrogate domains.
    * @param domsl Domains array for the left data tuples. 
    * @param domsr Domains array for the right data tuples. 
    * @param leftJoinDataOffset Domains array for the left data tuples. 
    * @param rightJoinDataOffset Domains array for the right data tuples. 
    * @return LT   if ROW(m) < ROW(n)
    *         EQ   if ROW(m) = ROW(n)
    *         GT   if ROW(m) > ROW(n)
    * @exception   InterpretError from other called methods 
    * @see Relation#compareTwoRows 
    * @see TypeConstants
    * @since jRelixNS
    */
   public int compareTwoRowsWithDataOffset(Object[] datal, int m, 
         Object[] datar, int n, 
	 int len, int[] types, boolean[] flags, Domain[] domsl, Domain[] domsr,
         int leftJoinDataOffset, int rightJoinDataOffset)
	 throws InterpretError
   {
     for (int j = 0; j < len; j++) {
       switch(types[j])
       {
         case INTEGER : {
           int bm = ((int[]) datal[leftJoinDataOffset+j])[m];
           int bn = ((int[]) datar[rightJoinDataOffset+j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else break; 
         }
         case STRING :
         case TEXT :
         case STMT :
         case EXPR : {
           String bm = ((String[]) datal[leftJoinDataOffset+j])[m];
           String bn = ((String[]) datar[rightJoinDataOffset+j])[n];
           int result = compareTwoStrings(bm, bn);
           if (result == GT || result == LT) return result;
           break; 
         }
         case BOOLEAN : {
           byte bm = ((byte[]) datal[leftJoinDataOffset+j])[m];
           byte bn = ((byte[]) datar[rightJoinDataOffset+j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else break; 
         }
         case SHORT : {
           short bm = ((short[]) datal[leftJoinDataOffset+j])[m];
           short bn = ((short[]) datar[rightJoinDataOffset+j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else break; 
         }
         case LONG : {
         long bm = ((long[]) datal[leftJoinDataOffset+j])[m];
         long bn = ((long[]) datar[rightJoinDataOffset+j])[n];
         if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else break; 
         }
         case IDLIST : { 
           long bm = ((long[]) datal[leftJoinDataOffset+j])[m];
           long bn = ((long[]) datar[rightJoinDataOffset+j])[n];
           if (flags[j] || domsl[j].name.equals(".id") 
               || domsr[j].name.equals(".id")) {
             if (bm > bn) return GT;
             if (bm < bn) return LT;
           break;
         }
         if (bm == bn) // equal if the surrogates are the same  
           break;
           if (bm == LONG_DK) {
               if (bn != LONG_DK) return LT;
               else break;
           } else if (bm == LONG_DC) {
             if (bn == LONG_DK) return GT;
             else if (bn == LONG_DC) break;
             else return LT;
           }
  
           if (bn == LONG_DK) {
             if (bm != LONG_DK) return GT;
             else break;
           } else if (bn == LONG_DC) {
             if (bm == LONG_DK) return LT;
             else if (bm == LONG_DC) break;
             else return GT;
           }
           Relation doml = myEnv.lookupRel("."+domsl[j].name, true);
           Relation rl = doml.getRelation(bm);
           Relation domr = myEnv.lookupRel("."+domsr[j].name, true);
           Relation rr = domr.getRelation(bn);
           int result = rl.compareRelation(rr);
           if (result != EQ) return result;
           else break; 
         }
         case FLOAT : {
           float bm = ((float[]) datal[leftJoinDataOffset+j])[m];
           float bn = ((float[]) datar[rightJoinDataOffset+j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else break; 
         }
         case DOUBLE : {
           double bm = ((double[]) datal[leftJoinDataOffset+j])[m];
           double bn = ((double[]) datar[rightJoinDataOffset+j])[n];
           if (bm > bn) return GT;
           else if (bm < bn) return LT;
           else break; 
         }
       } // end of switch
     }
     return EQ;
   }
   /**
   * for debug; same as printdata except this mothod does 
   * not call load() to load the data.
   *
   * @since jRelixNS
   */
public void noloadprintdata()
throws InterpretError
{
// print datas
int[] types = toTypes(domains);  // domain types
for (int i = 0; i < numtuples; i++) {
 for (int j = 0; j < numattrs; j++) {
	switch(types[j]) {
	case INTEGER :
	  System.out.print(((int[])data[j])[i]); break;
        case NUMERIC :
          System.out.print(((number [])data[j])[i].toString() ); break;
	case STRING :
	case STMT :
	case EXPR :
	  System.out.print(((String[])data[j])[i]); break;
	case BOOLEAN :
	  System.out.print(((byte[])data[j])[i]); break;
	case SHORT :
	  System.out.print(((short[]) data[j])[i]); break;
	case LONG :
	case IDLIST :
	case COMP :
	  System.out.print(((long[]) data[j])[i]); break;
	case TEXT :
	  System.out.print(((String[]) data[j])[i]); break;
	case FLOAT :
	  System.out.print(((float[]) data[j])[i]); break;
	case DOUBLE :
	  System.out.print(((double[]) data[j])[i]); break;
	}
	// insert the space after each field
	System.out.print(" ");
 } // loop on j
 System.out.println(";");
}
}

   /**
   * Returns true if the the relation returns a nullary relation
   * that is compabile with and can be converted to 
   * a scalar boolean.
   * 
   * @return  a boolean true if this is a nullary relation, false otherwise.
   * @since jRelixNS
   */
   public boolean isBoolCompatible() 
   {
     //if (this.domains.length == 1 && this.domains[0].name == ".bool")
     if (this.domains.length == 1 && (this.domains[0].name.compareTo(".bool") == 0))//hongyu modified it. Feb 2002
       return true; 
       else
          return false; 
   }

   /**
   * Returns boolean value converted from a
   * boolean compatible nullary relation/IDLIST.
   * 
   * @exception   InterpretError thrown if relation is not boolean compatible.
   * @return  a boolean true nullary relation contains true, false otherwise.
   * @since jRelixNS
   */
   public boolean boolCompatibleValue() 
   throws InterpretError
   {
       if (this.domains.length == 1 && this.domains[0].name.equals(".bool")){
           this.load(true);
           return((boolean)((((byte[]) this.data[0])[0]) == ((byte)1)));
       }
       else
          {
	  throw new InterpretError("boolCompatibleValue: relation/IDLIST " +
          this.name
          + " is not nullary/boolean compatible... expecting 1 domain with the domain name of '.bool', but found domain length of " +
          this.domains.length + " and the domain name of the first domain is " + this.domains[0].name);
          }
   }
   /**
   * jRelixNS - jRelix Nested Sigmajoin version
   * @author A. Chang  
   * ******** jRelixNS END OF CHANGES   *************
   */
}
