/**
 * Database interpreter (with semantic check)
 *
 * @author Biao Hao (bhao@cs.mcgill.ca)
 *
 * May 31, 1997 --- 
 *
 * TODO :
 * 1. mutual recursive view - semantics
 * 2. select, project, assignment, mujoin with virtual attributes
 * 3. initialization using another relation
 *    relation R(A,B) <- S; << S exists and defined on A and B
 * 4. update
 * 5. RAM version of evaluateExpression
 *
 * DONE :
 * 1. project with virtual attributes in the project list
 * 2. select with virtual attribute (October 20, 97)
 * 3. assignment with virtual attribute (October 21, 97)
 * 4. mujoin with virtual attribute (October 21, 97)
  *
 * Changes Since jRelCaNu
 *  - Special handling for OP_CAST.
 *    explicitely checked children for cast compatibility.
 *  - Also store the CastFromType in bits of the OP_TYPE node.
 *  - Allow select and project in let expressions.

 */

import java.lang.*;
import java.io.*;
import java.util.*;
import javax.swing.*;

public class Interpreter implements Constants, TypeConstants
//public class Interpreter extends Interpreter76
{
//debugging Yi Zheng
private boolean DEBUG=false;
private boolean DEBUG_09_22_2002 = false;

  public Environment env;   // the top level environment
  private CompTable comptable; // Relation system table
  private ExprTable exprtable;	// Expression system table
  private Parser parser;	// the parser
  
  private Stack instack;      //stack for old input streams
  private InputStream in;     //current parser input stream 
  
  // the prefix and current number for temporary file name
//  private static final String temp_name_prefix = "_temp_X9X_"; //changed by Yi Zheng March 27, 2002
  public static final String temp_name_prefix = "_temp_X9X_";
  private static final int MAX_TEMP_FILE = Integer.MAX_VALUE;
  private static int next_number = 0;
  
  private static final int MAX_INITIALIZE_SIZE = 10000;
  
//added by Yi Zheng June 24, 2002
public Relation destrel = null;
public int currow = -1;

  //   private int nested; 	// set the nested level for the relations
  
  public Interpreter () {
    this.env = Global.environment;
    this.comptable = Global.comptable;
    this.exprtable = Global.exprtable;
    //	nested = 0;	// no nested at first
  }
  private void assignLiteral(Relation rel, int i, int offset, SimpleNode node) 
       throws InterpretError
  {
    Domain[] doms = rel.domains;
    switch(doms[i + offset].type) 
      {
      case BOOLEAN : {
	if (node.opcode == OP_BOOLEAN) 
	  ((byte[]) rel.data[i+offset])[rel.capacity] 
	    = ((Byte) node.info).byteValue();
	else if (node.opcode == OP_DK)
	  ((byte[]) rel.data[i+offset])[rel.capacity] = BOOL_DK;
	else if (node.opcode == OP_DC)
	  ((byte[]) rel.data[i+offset])[rel.capacity] = BOOL_DC;
	else 
	  throw new InterpretError("invalid boolean literal");
	break;
      }
      case SHORT : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER 
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT 
	    || node.opcode == OP_DOUBLE) 
	  ((short[]) rel.data[i+offset])[rel.capacity] 
	    = ((Number) node.info).shortValue();
	else if (node.opcode == OP_DK)
	  ((short[]) rel.data[i+offset])[rel.capacity] = SHORT_DK;
	else if (node.opcode == OP_DC)
	  ((short[]) rel.data[i+offset])[rel.capacity] = SHORT_DC;
	else 
	  throw new InterpretError("invalid short literal");
	break;
      }
      case INTEGER : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER 
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT 
	    || node.opcode == OP_DOUBLE) 
	  ((int[]) rel.data[i+offset])[rel.capacity] 
	    = ((Number) node.info).intValue();
	else if (node.opcode == OP_DK)
	((int[]) rel.data[i+offset])[rel.capacity] = INT_DK;
	else if (node.opcode == OP_DC)
	  ((int[]) rel.data[i+offset])[rel.capacity] = INT_DC;
	else 
	  throw new InterpretError("invalid integer literal");
	break;
      }
      case LONG : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER 
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT 
	    || node.opcode == OP_DOUBLE) 
	  ((long[]) rel.data[i+offset])[rel.capacity] 
	    = ((Number) node.info).longValue();
	else if (node.opcode == OP_DK)
	  ((long[]) rel.data[i+offset])[rel.capacity] = LONG_DK;
	else if (node.opcode == OP_DC)
	  ((long[]) rel.data[i+offset])[rel.capacity] = LONG_DC;
	else 
	  throw new InterpretError("invalid long literal");
	break;
      }
  
      	case NUMERIC:{//case added by Yiyi
	  /*  if(node.opcode == OP_NUMERIC ||node.opcode == OP_STRING){
       		String s = (String) node.info;
		((number[])rel.data[i+offset])[rel.capacity] =
		    new number(s);
	    }*/
       if(node.opcode == OP_NUMERIC)
       	((number[])rel.data[i+offset])[rel.capacity] = (number)node.info;
       else if(node.opcode == OP_STRING){
       	String s = (String) node.info;
			((number[])rel.data[i+offset])[rel.capacity] =
		    new number(s);
       }
	    else if(node.opcode == OP_DK)
		((number[])rel.data[i+offset])[rel.capacity]
		    = NUMERIC_DK;
	    else if(node.opcode == OP_DC)
		((number[])rel.data[i+offset])[rel.capacity]
		    = NUMERIC_DC;
	    else
		throw new InterpretError("invalid numeric literal");
	    break;
	}
	      	 case UNIVERSAL ://from Andrey
	 case ATTRIBUTE://from Andrey
      case STRING :

      case TEXT :
      case STMT :
      case EXPR : {
	/*if( rel==null)
	  System.out.println( " NUll rel" );
	  else if ( rel.data[i+offset] == null )
	  System.out.println( " Null data" );
	  else if ( node.info == null )
	  System.out.println( " Null node info" );
	  else if ( rel.capacity == 0 )
	  System.out.println( " Null rel capacity" );
	  System.out.println( "rel.data size" +
	  (rel.data).length );
	  System.out.println( "node.info is " +
	  (node.info).toString() );*/
	
	
	if (node.opcode == OP_STRING) 
	  ((String[]) rel.data[i+offset])[rel.capacity] 
	    = (String) node.info;
	    //from Andrey
   else if ((node.opcode == OP_ATTRIBUTE)||(node.opcode == OP_TEXT)) 
	  ((String[]) rel.data[i+offset])[rel.capacity] 
		= (String) node.info;
	    //end from Andrey
//	else if (node.opcode == OP_NUMERIC) //from Yiyi
//	  ((number[]) rel.data[i+offset])[rel.capacity]
//	    = (number) node.info;
	
	else if (node.opcode == OP_DK)
	  ((String[]) rel.data[i+offset])[rel.capacity] = STRING_DK;
	else if (node.opcode == OP_DC)
	  ((String[]) rel.data[i+offset])[rel.capacity] = STRING_DC;
	else if (node.opcode == OP_NULL)
	  ((String[]) rel.data[i+offset])[rel.capacity] = "";
	else 
	  throw new InterpretError("invalid String literal");
	break;
      }
      case FLOAT : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER 
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT 
	    || node.opcode == OP_DOUBLE) 
	  ((float[]) rel.data[i+offset])[rel.capacity] 
	    = ((Number) node.info).floatValue();
	else if (node.opcode == OP_DK)
	  ((float[]) rel.data[i+offset])[rel.capacity] = FLOAT_DK;
	else if (node.opcode == OP_DC)
	  ((float[]) rel.data[i+offset])[rel.capacity] = FLOAT_DC;
	else 
	  throw new InterpretError("invalid float literal");
	break;
      }
      case DOUBLE : {
	if (node.opcode == OP_SHORT || node.opcode == OP_INTEGER 
	    || node.opcode == OP_LONG || node.opcode == OP_FLOAT 
	    || node.opcode == OP_DOUBLE) 
	  ((double[]) rel.data[i+offset])[rel.capacity] 
	    = ((Number) node.info).doubleValue();
	else if (node.opcode == OP_DK)
	  ((double[]) rel.data[i+offset])[rel.capacity] = DOUBLE_DK;
	else if (node.opcode == OP_DC)
	((double[]) rel.data[i+offset])[rel.capacity] = DOUBLE_DC;
	else 
	  throw new InterpretError("invalid double literal");
	break;
      }
      case IDLIST : { // dk and dc for nested relation
	if (node.opcode == OP_DK)
	  ((long[]) rel.data[i+offset])[rel.capacity] = LONG_DK;
	else if (node.opcode == OP_DC)
	  ((long[]) rel.data[i+offset])[rel.capacity] = LONG_DC;
	else 
	throw new InterpretError("invalid surrogate literal");
	break;
      }
      } // end of switch
  }      
  private Vector CreateDAG(String relname, Environment env)
       throws InterpretError
  {
    Vector vec = new Vector(64); //stack of names to process
    
    vec.addElement(relname); // add top level relation name
    
    int size = 0;
    while (size < vec.size()) {
      String s = (String) vec.elementAt(size);
      Domain[] doms = env.lookupRel(s, true).domains;
      for(int i = 0; i < doms.length; i++) {
	if(doms[i].type == IDLIST && !doms[i].name.equals(".id"))
	  vec.addElement("." + doms[i].name);
      }
      size++; 
      if (size > 1024) // has cycle or nested too deep
	throw new InterpretError("nested relation " + relname 
				 + " has cycle or nested too deep");
    }
    
    Hashtable ht = new Hashtable(64);
    
    for (int i = size - 1; i >= 0; i--) {
      String name = (String) vec.elementAt(i); 
      if (ht.containsKey(name))
	vec.removeElementAt(i);
      else
	ht.put(name, name);
    }
    
    return vec;
  }      
  // create data array for a relation and put the relaton into 
  // the hashtable
  private void CreateData(Hashtable ht, String name, Environment env)
       throws InterpretError
  {  
    Relation rel = env.lookupRel(name, true);
    rel.createDataArray(MAX_INITIALIZE_SIZE); 
    rel.capacity = 0;
    ht.put(name, rel);
  }       
  /**
   * Create all relations and their data arrays that need to be updated
   * (including nested relations) for initializing relation 'name'. Each
   * of these relations is stored in the resulting hash table with the
   * name of the relation as the key.
   */
  private Hashtable CreateDataHashTable(String name, Environment env)
       throws InterpretError
  {  
    Hashtable rels = new Hashtable();
    Vector relnames = new Vector(); //stack of names to process
    int size;
    
    // initialize the stack of names
    Domain[] doms = env.lookupRel(name, true).domains;
    for(int i = 0; i < doms.length; i++) {  
      if(doms[i].type == IDLIST) 
	relnames.addElement("." + doms[i].name);
    }
    
    // create top level relation and its data array, store in hashtable
    CreateData(rels, name, env);
    
    while (0 < (size = relnames.size())) {
      String nm = (String) relnames.lastElement();
      relnames.removeElementAt(size - 1);
      if (!rels.containsKey(nm)) {  
	CreateData(rels, nm, env);
	doms = env.lookupRel(nm, true).domains;
	for (int i = 0; i < doms.length; i++) {
	  String s = doms[i].name;
	  if (!rels.containsKey(s)) {  
	    if(doms[i].type == IDLIST && !s.equals(".id"))//.id of type IDLIST
	      relnames.addElement("." + doms[i].name);
	  }
	}
      }
    }
    return rels;
  }       
  // delete a data file on disk 
  public static void deleteFile(String filename)
  {
    File f = new File(filename);
    if (f.exists()) f.delete();
  }      
  // REVISIT: short cut for equality of two relations
  private boolean equals(Relation rl, Relation rr)
       throws InterpretError
  {
    String tmpname = nextTempName();
    
    // check if the two relations are defined on same set of attributes
    if (!rl.isDomainsIn(rr.domains) || !rr.isDomainsIn(rl.domains))
      throw new InterpretError("two relations are not comparable");
    if (rl.numtuples != rr.numtuples)
      return false;
    Relation rel = rl.mujoin(OP_SJOIN, null, rr, null, tmpname);
    if (rel.numtuples == 0) {
      deleteFile(tmpname);
      return true;
    } else {
      deleteFile(tmpname);
      return false;
    }
  }    
     /**
   * Processes the node for evaluating the join specified in the node's tree.
   *
   * JRelixNS modified/enhanced this method:
   *  Now handles Sigma-Joins in addition to mu-joins.
   */
  private Relation evaluateJoin(SimpleNode node, String filename,
				Hashtable htable, Environment env)
       throws InterpretError
  {
    int opcode = node.opcode;
    Domain[] domsl = null, domsr = null;
    Relation rl = null, rr = null;
    String lname = nextTempName();
    String rname = nextTempName();
    
    // step 1. get the join attributes and evaluate join expressions
    if (node.jjtGetNumChildren() == 2) { // no join attributes specified
      // evaluate left relational expressions
      rl = evaluateTLExpression((SimpleNode) node.jjtGetChild(0), 
				lname, htable, env);

      // evaluate right relational expressions
      rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1), 
				rname, htable, env);

    } else { // node.jjtGetNumChildren() == 4
      // get an array of domains for left ExpressionList
      SimpleNode nodel = (SimpleNode) node.jjtGetChild(1);
      domsl = ExpressionListToDomains(nodel);
      
      // get an array of domains for right ExpressionList
      SimpleNode noder = (SimpleNode) node.jjtGetChild(2);
      domsr = ExpressionListToDomains(noder);
      
      // evaluate left relational expressions
      rl = evaluateTLExpression((SimpleNode) node.jjtGetChild(0), 
				lname, htable, env);
      
      // evaluate right relational expressions
      rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(3), 
				rname, htable, env);
				
      // actualize if necessary
      if (rl.rvc == RELATION && !rl.isDomainsIn(domsl)) {
	Relation rltmp = new Actualizer(domsl, rl, env).actualize();
	rltmp.name = lname;
	rltmp.save();
	rltmp.unload();
	rl = rltmp;
      }
      if (rr.rvc == RELATION && !rr.isDomainsIn(domsr)) {
	Relation rrtmp = new Actualizer(domsr, rr, env).actualize();
	rrtmp.name = rname;
	rrtmp.save();
	rrtmp.unload();
	rr = rrtmp;
      }
    }

    // step 2. invoke the join method
    if(rl.rvc==COMPUTATION && rr.rvc==COMPUTATION ){
      throw new 
	InterpretError("join of two computations is not implemented.");
    }
    else if(rl.rvc == COMPUTATION){
    	if(opcode == OP_IJOIN) { //if switch added by Yi Zheng July 4, 2002
      Relation rel = 
	((Computation) rl).applyIjoin(rr, domsl, domsr, filename, env);
      deleteFile(lname);
      deleteFile(rname);
      return rel;
     }
     else
     	throw new InterpretError("join other than IJOIN between computation and relation not implemented");
    }else if(rr.rvc == COMPUTATION){
    	if (opcode == OP_IJOIN) {
      Relation rel = 
	((Computation) rr).applyIjoin(rl, domsr, domsl, filename, env);
      deleteFile(lname);
      deleteFile(rname);
      return rel;
      }
      else
     	throw new InterpretError("join other than IJOIN between computation and relation not implemented");
    }
    else
      {  //this is a regular join of two relations/views
	switch(opcode) 
	  {  case OP_IJOIN :
	  case OP_UJOIN :
	  case OP_SJOIN :
	  case OP_LJOIN :
	  case OP_RJOIN :
	  case OP_DLJOIN : // also OP_DJOIN
	  case OP_DRJOIN :
	    Relation rel = rl.mujoin(node.opcode, domsl, rr, domsr, filename);
	  deleteFile(lname);
	  deleteFile(rname);
	  return rel;
	  case OP_ICOMP : 
	  case OP_EQJOIN : 
	  case OP_GTJOIN : 
	  case OP_GEJOIN : 
	  case OP_LTJOIN : 
	  case OP_LEJOIN : 
	  case OP_IEJOIN : 
	  case OP_NOTICOMP : 
	  case OP_NOTEQJOIN : 
	  case OP_NOTGTJOIN : 
	  case OP_NOTGEJOIN : 
	  case OP_NOTLTJOIN : 
	  case OP_NOTLEJOIN : 
	  case OP_NOTIEJOIN : 
	    // START Since jRelixNS
	    // Process a sigma join only if it the optional flag 
	    // Relation.isSigmaJoinImplemented is set.
	    // Otherwise, issue a "TO BE IMPLEMENTED" message.
	  //  if (env.isSigmaJoinImplemented == true) {
	     if (Relation.isSigmaJoinImplemented == true) { //Andy, May 12, 2002
	      rel = rl.sigmaJoin(node.opcode, domsl, rr, domsr, filename);
	      deleteFile(lname);
	      deleteFile(rname);
	      return rel;
	    }
	    else
	      throw new InterpretError("TO BE IMPLEMENTED : sigma-join");
	    // END Since jRelixNS
	  case OP_NOP : {//case added by Yi Zheng July 4, 2002
	  	Relation noprel = null;
	  	if(node.jjtGetNumChildren() == 2)
	  	{
	  		boolean Ran = new Random().nextBoolean();
	  		if (!Ran)
	  		{
	  			noprel = rl.project(rl.domains, filename);
	  			deleteFile(lname);
	  			deleteFile(rname);
	  			return noprel;
	  		}
	  		else
	  		{
	  			noprel = rr.project(rr.domains, filename);
	  			deleteFile(lname);
	  			deleteFile(rname);
	  			return noprel;
	  		}
	  	}
	  	else 
	  		throw new InterpretError("invalid relational expression");
	  }	
	  default : 
	    throw new
	      InterpretError("invalid/not implemented relational expression");
	  } /* switch */
     }
  }      
  
  
  //YuLing
  private Relation evaluateGedit(SimpleNode node,String filename, Hashtable
				 htable, boolean flag, Environment env, Domain[] doms)
       throws InterpretError
  {
    int opcode=node.opcode;
    //	Domain[] doms = null;
    String tmpname = nextTempName();
    boolean projector2=false;
    SimpleNode noder=null;
    SimpleNode nodel1=null;
    SimpleNode nodel2=null;
    
    //get the first  porjector
    nodel1=(SimpleNode)node.jjtGetChild(0);
    
    //get the number of children
    if (node.jjtGetNumChildren() == 2)  // no second projector
      {
	//append Color, symbol, and layer to rel.
	noder=(SimpleNode)node.jjtGetChild(1);
      }
    else
      {
	nodel2=(SimpleNode)node.jjtGetChild(1);
	projector2=true;
	
	noder=(SimpleNode)node.jjtGetChild(2);
      }
    if (nodel1.isBitSet(0)) { // contains QT-Operator hongyu
      throw new InterpretError("TO BE IMPLEMENTED : QT-Operator");
    }
    // evaluate the relational expression (projection) it is projected on
    Relation rel = evaluateTLExpression(noder, tmpname, htable, env);
    
    if (nodel1.isBitSet(1)) { // contains ExpressionList in the first projector
      // get node containing ExpressionList in projector
      SimpleNode nd = (SimpleNode) nodel1.jjtGetChild(0);
      if(doms==null)
	doms=new Domain[8];
      
      doms=ExpressionListToGeditDomains(nd,doms,1);
    } else { // empty projector is not allowed
      throw new 
	InterpretError("empty projector is not allowed in gedit");
    }
    if(projector2==true)
      {
	if (nodel2.isBitSet(1)) { // contains ExpressionList in projector
	  // get node containing ExpressionList in projector
	  SimpleNode nd = (SimpleNode) nodel2.jjtGetChild(0);
	  doms = ExpressionListToGeditDomains(nd,doms,2); 
	} else { // empty projector not allowed
	  throw new
	    InterpretError("Empty projector is not allowed in gedit");
	}
      }
    Domain[] domains=null;
    
    if(doms[5]==null)
      {
        executeRelixCommand("dd .C;");
        executeRelixCommand("domain .C intg;");
        executeRelixCommand("relation COLOR(.C)<-{(204055055)};");
	
        domains=rel.domains;
	
	String attrs="";
	for(int i=0;i<domains.length;i++)
	  attrs=attrs+domains[i].name+",";
	//System.out.println("attrs in Interpret[doms5]="+attrs);
	
        executeRelixCommand(noder.name+"<-["+attrs+".C]("+noder.name+" ijoin COLOR);");
	doms[5]=env.lookupDom(".C",true);
	//System.out.println("doms[5].name="+doms[5].name);
	rel = getRelbyName(rel.name);
      }
    
    if(doms[6]==null)
      {
        executeRelixCommand("dd .Symb;");
        executeRelixCommand("domain .Symb intg;");
        executeRelixCommand("relation SYMBOL(.Symb)<-{(201)};");
	
        domains=rel.domains;
	
	String attrs="";
	for(int i=0;i<domains.length;i++)
	  attrs=attrs+domains[i].name+",";
	//System.out.println("attrs in Interpret doms[6]="+attrs);
	
        executeRelixCommand(noder.name+"<-[" +attrs+".Symb]("+noder.name+" ijoin SYMBOL);");
	doms[6]=env.lookupDom(".Symb",true);
	rel = getRelbyName(rel.name);
      }
    
    if(doms[7]==null)
      {
        executeRelixCommand("dd .L;");
        executeRelixCommand("domain .L string;");
        executeRelixCommand("relation LAYER(.L)<-{(\"DefautLayer\")};");
	
	domains=rel.domains;
	
	String attrs="";
	for(int i=0;i<domains.length;i++)
	  attrs=attrs+domains[i].name+",";
	//System.out.println("attrs in Interpret doms[7]="+attrs);
	
        executeRelixCommand(noder.name+"<-["+attrs+".L]("+noder.name+" ijoin LAYER);");
	doms[7]=env.lookupDom(".L",true);
	rel=getRelbyName(rel.name);
      }
    
    
    if (opcode == OP_WHENFROM) {
      
    } 
    
    if(flag){
      
      domains=rel.domains;
      Actualizer actualizer = new Actualizer(domains, rel, env);
      Relation rr = actualizer.actualize();
      
      actualizer.cleanup();
      
      deleteFile(tmpname);
      
      /*Domain[] sortdoms= new Domain[3];
	System.arraycopy(rr.domains,0,sortdoms,0,3);
	rr.sort(sortdoms);*/
      
      int S=((int[])(rr.data[2]))[0];
      System.out.println("Please wait a few minutes for the Geditor to start...");
      Geditor win=new Geditor(this,rr,doms);
      win.setVisible(true);
      
      return rr;
    }else{
      new Actualizer(doms, rel, env);
      return null;
    }
  }
  
  //YuLing's addition ends here
  
  private Relation evaluateProject(SimpleNode node, String filename,
				   Hashtable htable, boolean flag, Environment env)
       throws InterpretError
  {
    int opcode = node.opcode;
    Domain[] doms = null;
    String tmpname = nextTempName();
    
    // get the projector
    SimpleNode nodel = (SimpleNode) node.jjtGetChild(0);
    //get the selection
    SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
    
    //The follwing is added by hongyu. Feb 14, 02
    if(nodel.isBitSet(0)) { // contains QT-Operator
  //    System.out.println("here");
      SimpleNode nodeCopy = (SimpleNode)node.jjtDeepDuplicate();
      SimpleNode nodeCopyNodel = (SimpleNode)nodeCopy.jjtGetChild(0);
      SimpleNode nodeCopyNoder = (SimpleNode)nodeCopy.jjtGetChild(1);
      
      int num = nodeCopyNodel.jjtGetNumChildren();
      SimpleNode nodeQT=(SimpleNode)nodeCopyNodel.jjtGetChild(0);
 //     System.out.println("num "+num);
 //     System.out.println("node QT");
 //     printNode(nodeQT);
      nodeCopyNodel.jjtRemoveChild(0);
      int tempBits = nodeCopyNodel.bits;
      tempBits --;
      tempBits = tempBits | 0x10;
      if(num==1){
	switch (nodeQT.opcode){
	case OP_NUMBER:{
	  doms = new Domain[1];
	  doms[0] = env.lookupDom(".integer", true);
	  Relation r = new Relation(filename,doms,RELATION,1,1,1,null,env);
	  Object[] datal = r.createDataArray(doms,1);
	  ((int[])datal[0])[0] =1;
	  r.data = datal;
	  return r;
	}
	case OP_DOT:
	case OP_PERCENT:{
	  doms = new Domain[1];
	  doms[0] = env.lookupDom(".double", true);
	  Relation r = new Relation(filename,doms,RELATION,1,1,1,null,env);
	  Object[] datal = r.createDataArray(doms,1);
	  ((double[])datal[0])[0] =1.0;
	  r.data = datal;
	  return r;
	}
	}
      }
      else{
	nodeCopyNodel.bits = tempBits;// clear bits of QT-Operator option
	
	Relation R = evaluateProject(nodeCopy,tmpname,htable,flag, env);
	int nmuber =R.numtuples;
	if(nodeQT.opcode == OP_NUMBER){
	  deleteFile(tmpname);
	  doms = new Domain[1];
	  doms[0] = env.lookupDom(".integer", true);
	  Relation r = new Relation(filename,doms,RELATION,1,1,1,null,env);
	  Object[] datal = r.createDataArray(doms,1);
	  ((int[])datal[0])[0] =nmuber;
	  r.data = datal;
	  return r;
	}
	else{// case opcode == OP_DOT or opcode == OP_PERCENT
	  if((num=nodeCopyNoder.jjtGetNumChildren())>= 2){
	    nodeCopyNoder =(SimpleNode)nodeCopyNoder.jjtGetChild(num-1);
	  }
	  
	  Relation RR = evaluateTLExpression(nodeCopyNoder,filename, htable, env);// org relation
	  Actualizer actualizer = new Actualizer(R.domains, RR, env);
	  RR = actualizer.actualize();
	  RR.name = filename;
	  RR.save();
	  RR = RR.project(R.domains,tmpname);
	  actualizer.cleanup();
	  
	  int total = RR.numtuples;
	  if(total == 0){
	    throw new
	      InterpretError(" the total number is 0");
	  }
	  else{
	    deleteFile(tmpname);
	    deleteFile(filename);
	    doms = new Domain[1];
	    doms[0] = env.lookupDom(".double", true);
	    Relation r = new Relation(filename,doms,RELATION,1,1,1,null,env);
	    Object[] datal = r.createDataArray(doms,1);
	    ((double[])datal[0])[0] =(double)nmuber/(double)total;
	    r.data = datal;
	    return r;
	  }
	}
      }
    }
    //end of hongyu added. Feb 14,02 
    
    // evaluate the relational expression (projection) it is projected on
    //evaluateQSelect() is here.Hongyu
    Relation rel = evaluateTLExpression(noder, tmpname, htable, env);

    if (nodel.isBitSet(1)) { // contains ExpressionList in projector
      // get node containing ExpressionList in projector
      SimpleNode nd = (SimpleNode) nodel.jjtGetChild(0);
  //Modified by Yi Zheng April 30, 2002    doms = ExpressionListToDomains(nd); // REVISIT: generalization
	//doms = ExpressionListToDomains(nd, rel);//modified by Yi Zheng May 26, 2002, to integrate Andrey's code
	doms = ExpressionListToDomains(nd, rel, htable, env);
    } else { // empty projector, pass domain ".bool" 
      doms = new Domain[1];
      doms[0] = env.lookupDom(".bool", true);
      Relation rtmp = rel.project(doms, filename);
      deleteFile(tmpname);
      return rtmp;
    }
    
    
    if (opcode == OP_WHENFROM) {
    	throw new InterpretError("OP_WHENFROM not implemented");
     } 
    
    if(flag){
      Actualizer actualizer = new Actualizer(doms, rel, env);
      Relation rtmp = actualizer.actualize();
      rtmp.myEnv = env; //Yi Zheng June 28, 2002
      // Save the result relation to disk file. This is required by
      // other operations e.g. projection, selection etc.
      rtmp.name = tmpname;
      rtmp.save();
      
      //debugging...Sept 28, 2002
     // rtmp.print();
      
      // Do the projection..

 //     Relation rr = rtmp.project(doms, filename); //modified by Yi Zheng June 28, 2002
      
      //addition by Yi Zheng June 28, 2002
      //deal with red domain used for level lifting
      Relation rr = null;
      String rrTempName = nextTempName();
      
      Relation rrTemp = rtmp.project(doms, rrTempName);
      
      
      if((doms.length == 1) && (doms[0].isRedTempTop == 1))
      {
      	if(doms[0].type == IDLIST)
      	{
      		Relation dotRR = env.lookupRel("." + doms[0].name, true);
      		if(dotRR == null)
      			throw new InterpretError("relation ." + doms[0].name + " not found");
      	
      		Domain[] realDoms = new Domain[dotRR.domains.length - 1];
      		int jj = 0;
      		for(int i = 0; i < dotRR.domains.length; i++)
      		{
      			if(dotRR.domains[i].name.compareTo(".id") == 0)
      				continue;
      			realDoms[jj++] = dotRR.domains[i];
      		}
      		
      		rr = dotRR.project(realDoms, filename);
      		env.removeDom(doms[0].name, true);
      	}
      	else
      		rr = rrTemp.project(doms, filename); //this is only used in Computations: eg. Cnt <- [red + of 1] in customers;
      }
      else
      {
      	rr = rrTemp.project(doms, filename);
      }
      
      //end of addition June 28, 2002
      
      
      // System.out.println("rtmp.data in project ");
      //System.out.println(((int[])rtmp.data[3])[1]);
      // Do the clean-up.  Note, we have to put this procedure 
      // *after* all the operations on those relations which are
      // generated by the actualizer. This is too bad as OO scheme
      // seems to be broken here. However, this is a must, at least
      // for now. The reason is that when intermediate nested domains
      // are generated by the actualizer (e.g. in case 
      // "let x be S ijoin T ijoin U;", a temporary nested domain '0'
      // is generated.), their data contents are written to corresponding
      // disk files (e.g. '.0'). These intermediate disk files are 
      // necessary for the operations (such as the above-written
      // "rtmp.project(doms, filename);" since all relational
      // operations are based on disk files but not memory data. ;-(
      // As actualizer's cleanup() method will delete all the 
      // intermediate files, we have to make sure this happens *only*
      // after our major task (projection here) has finished.
      
      actualizer.cleanup();
      
      //Relation rtmp = rel.project(doms, filename);
      deleteFile(tmpname);
      deleteFile(rrTempName); //Yi Zheng June 28, 2002
      
//      System.out.println("rr to be returned is:");
//      rr.print();
      rr.load(true);
      return rr;
    }else{
      new Actualizer(doms, rel, env);
      return null;
    }
    
  }
  
  
  //The follwoing is added by  Hongyu ( Feb 14, 02)
  private Relation evaluateQSelect(SimpleNode orgNode, String filename,
				   Hashtable htable, Environment env)
       throws InterpretError
  { 
    Relation rel = null;
    switch(orgNode.opcode){
    case OP_WHEREIN: {
      
      /*STEP 1: 
	This is to get the original relation. (The after "in" part in a query)
	or if the query is without "in", to get the part after the "quant" list.
	*/
      SimpleNode node =  (SimpleNode)orgNode.jjtDeepDuplicate();
      SimpleNode qtCond = new SimpleNode(nextTempName()),orgRel = null;
      int numOfChildren = orgNode.jjtGetNumChildren();
      boolean cond = false,projIsSth = false,flag=true;
      String orgRelName = nextTempName();
      Domain d=null;
      Vector reOrderDomsV = new Vector();
      
      // case OP_WHEREFROM: the query is without "in" part
      if (numOfChildren == 2){
	orgRel = (SimpleNode)node.jjtGetChild(1);
	cond = true;
      }
      else {
	orgRel = (SimpleNode)node.jjtGetChild(numOfChildren-1);
	qtCond.jjtAddChild((SimpleNode)node.jjtGetChild(1));
	d = new Domain(nextTempName(), BOOLEAN, 0, qtCond);// define a temp boolean domain for "where" clause	
      }
      rel = evaluateTLExpression(orgRel,orgRelName,htable,env);// get the org relation
      //rel.load();
      //rel.print();
      
      /*STEP 2: get the project domain list */
      Vector allDoms = new Vector();
      allDoms = qtEvaluateProj(orgNode,env,allDoms);
      
      Domain[] projList = new Domain[allDoms.size()];
      allDoms.copyInto(projList);
      int p = projList.length;
      if(p == 1){
	if((projList[0].name.compareTo("0qtProjIsEmpty") == 0) ||
	   (projList[0].name.compareTo("0qtProjIsSth") == 0))
	  {
	    p = 0;
	    allDoms.removeAllElements();
	    if(projList[0].name.compareTo("0qtProjIsSth") == 0)
	      projIsSth = true; 
	  }
      }
      
     // for(int i=0;i<allDoms.size();i++)
//	System.out.println("Porj "+ i+ " "+((Domain)allDoms.elementAt(i)).name); 
      
      /*STEP 3: get the quantifier list */
      allDoms = qtEvaluateQuantDom(node,env,allDoms);
      
    //  for(int i=0;i<allDoms.size();i++)
//	System.out.println("Porj + Quant "+ i+ " "+((Domain)allDoms.elementAt(i)).name);
      
      /*STEP 4: if the project domain list is empty, we have to compare the reOrderDomsV with ori relation domain list */
      if (projList[0].name.compareTo("0qtProjIsEmpty") == 0){
	Vector reOrderDomsV1 =  new Vector();
	p=rel.domains.length;
	for (int i=0;i<p;i++)
	  reOrderDomsV1.addElement(rel.domains[i]);
	
	for (int i=0;i<allDoms.size();i++){
	  if(!(reOrderDomsV1.contains(allDoms.elementAt(i))))
	    reOrderDomsV1.addElement(allDoms.elementAt(i));
	  else{
	    p=0;
	    flag = false;
	    break;
	  }
	}
	if(flag){
	  allDoms= reOrderDomsV1;
	  flag=false;
	}
      }
      
      int q = allDoms.size() - p;
      
      /*STEP 5: get the "where" caluse if the query has it*/
      if(!cond)
	allDoms.addElement(d);
      
      /*STEP 6: get PredList */
      Domain[] predList = qtEvaluatePred(node,env);
      
      /*STEP 7: get domains inside each (# ) or (. ) or(% ) caluse. */
      for(int i=0;i<q;i++)
	allDoms = qtEvaluateDomInQuant(predList[i].tree,env,allDoms);
      
      
      /*STEP 8: get the reorder domain list */ 
      for(int i=0;i<allDoms.size();i++){
	if(!(reOrderDomsV.contains(allDoms.elementAt(i)))) // delete duplicate domains in allDoms list
	  reOrderDomsV.addElement(allDoms.elementAt(i));
      }
      
      
      /*STEP 9: get the domains in rel list */
      if(flag){
	for(int i=0;i<rel.domains.length;i++){
	  if(!(reOrderDomsV.contains(rel.domains[i])))
	    reOrderDomsV.addElement(rel.domains[i]);
	}
      }
      
      /*STEP 10: get the reorder domain list */	
      Domain[] doms = new Domain[reOrderDomsV.size()];
      reOrderDomsV.copyInto(doms);
      
//      for(int i=0;i<doms.length;i++){
//	System.out.println("allDoms "+ i +" " + doms[i].name );
//	System.out.println("allDoms type "+ i +" " + doms[i].type );
  //    }
      
      /*STEP 11: evaluating */
      Actualizer qtActualizer = new Actualizer(doms,rel,env);
      qtActualizer.qtActive = true;
      qtActualizer.p = p;
      qtActualizer.q = q;
      qtActualizer.predList = predList;
      qtActualizer.cond = cond;
      qtActualizer.qtCond = qtCond;
      qtActualizer.projIsSth = projIsSth;
      Relation r = qtActualizer.actualize();
      if(r.data== null)
	r.load();
      if(!flag){
	r.project(rel.domains);
      }
      qtActualizer.cleanup();
      rel=r;
      deleteFile(orgRelName);// might cause problems?
      
      //System.out.println("End");// End of evaluation.
      
      break;
    }
    
    case OP_WHENIN:{
      orgNode.opcode = OP_WHEREIN;
      rel =  evaluateQSelect(orgNode,filename,htable,env);
      orgNode.opcode = OP_WHENIN;
      break;
    }
    
    // End of Hongyu Added. July 15, 01.
    
    case OP_WHEREFROM : 
    case OP_WHENFROM :
      
      throw new InterpretError("TO BE IMPLEMENTED : QSelector"); 
    }
    return rel;
  }
  
  /* The following methods are for QSelector.
     Added by hongyu (June 25, 01) */
  private Vector qtEvaluateProj(SimpleNode node,Environment env, Vector allDoms)
       throws InterpretError 
  {  
    Domain[] doms = null;
    Domain d = null;
    SimpleNode nodeParent = (SimpleNode) node.jjtGetParent();
    //System.out.println(" nodeParent");
    //printNode(nodeParent);
    if(nodeParent.identifier.compareTo("Project")== 0){//case: a <- #[ITEM] in qaunt (#>2) COMP in SUPPLY
      d = new Domain("0qtProjIsEmpty",BOOLEAN,0,null);
      doms = new Domain[1];
      doms[0] = d;
      //System.out.println("--- Empty project list"+d.name);
    }
    else{
      SimpleNode orgGetExpresstion = (SimpleNode)nodeParent.jjtGetChild(0);
      
      //clone the orgGetExpresstion, and keep the orginal tree untouched.(recrusive view)
      SimpleNode getExpresstion =  (SimpleNode)orgGetExpresstion.jjtDeepDuplicate();
      
      if(getExpresstion.isBitSet(0)){
	//let s be # quant (#>2)COMP in SUPPLY; which is not a valid statement
	//a <- # quant (#>2)COMP in SUPPLY; which is not a valid statement either
	throw new  InterpretError(" QTselector: need to be implemented ");
      }
      else if(getExpresstion.isBitSet(1)){
	SimpleNode expresstionList =
	  (SimpleNode)getExpresstion.jjtGetChild(0);
	doms = ExpressionListToDomains(expresstionList);
      }
      
      else {// empty project list
	if(getExpresstion.type == OP_IDENTIFIER){//case: a<-  quant (#>2)COMP in SUPPLY;
	  //System.out.println("0qtProjIsEmpty 1  is set ");
	  d = new Domain("0qtProjIsEmpty",BOOLEAN,0,null);
	  doms = new Domain[1];
	  doms[0] = d;
	}
	else{
	  if(getExpresstion.jjtGetNumChildren() == 0){//case: a<- [] quant (#>2)COMP in SUPPLY;
	    //System.out.println(" getExpresstion.jjtGetNumChildren() == 0 ");
	    //System.out.println("0qtProjIsSth  is set ");
	    //System.out.println("type "+ getExpresstion.type);
	    d = new Domain("0qtProjIsSth",BOOLEAN,0,null);
	    doms = new Domain[1];
	    doms[0] = d;
	  }
	  else {
	    //System.out.println("loop ");
	    getExpresstion = (SimpleNode)getExpresstion.jjtGetChild(0);
	    doms = ExpressionListToDomains(getExpresstion);
	  }
	}
      }  
    }
    for(int i=0;i<doms.length;i++){
      if(!(allDoms.contains(doms[i])))
	allDoms.addElement(doms[i]);
      else
	throw new InterpretError(" duplicate attributes in project domain list. ");
    }
    return allDoms; 
  }
  
  private Vector qtEvaluateQuantDom(SimpleNode node,Environment env,Vector allDoms)
       throws InterpretError 
  {
    SimpleNode quantifierList = (SimpleNode)node.jjtGetChild(0);
    int n = quantifierList.jjtGetNumChildren();
    Vector doms= new Vector();
    Domain d = null;
    for(int i=0; i<n; i++){
      SimpleNode quantifier = 
	(SimpleNode)quantifierList.jjtGetChild(i).jjtGetChild(1); 
      d = env.lookupDom(quantifier.name,true);
      
      if (d == null)
	throw new  InterpretError("Domain " + quantifier.name + " is not declared.");
      else if(doms.contains(d))
	throw new InterpretError(" duplicate attributes in quantifier domain list. "); 
      else if (allDoms.contains(d))
	throw new InterpretError(" projected attribute appears in quantifier domain list.");
      else{
	doms.addElement(d);
	allDoms.addElement(d);
      }
    }	
    
    return allDoms;
  }
  
  private Domain[] qtEvaluatePred(SimpleNode node,Environment env)
       throws InterpretError 
  {
    SimpleNode quantifierList = (SimpleNode)node.jjtGetChild(0);
    int n = quantifierList.jjtGetNumChildren();
    Domain[] doms = new Domain[n];
    
    for(int i=0; i<n; i++){      
      SimpleNode quantifier = 
	(SimpleNode)quantifierList.jjtGetChild(i).jjtGetChild(0); 
      SimpleNode sn = new SimpleNode(" ");
      sn.list = new Vector();
      sn.list.addElement( new Vector()); // for #
      sn.list.addElement( new Vector()); // for " ." and " %"
      sn.jjtAddChild(quantifier);
      Domain d = new Domain(" ",BOOLEAN,0,sn);
      doms[i] = d;
      //System.out.println("Pred "+ i+ " "+d.name); 
    }
    return doms;
  }
  
  private Vector qtEvaluateDomInQuant(SimpleNode node, Environment env,Vector alldoms)
       throws InterpretError 
  { 
    if(node.type == OP_IDENTIFIER){
      //System.out.println("qtEvaluateDomInQuant opcode "+node.opcode);
      Domain d = env.lookupDom(node.name,true);
      
      /* Here is a defect.
	 for query:
	 a <- [ITEM] quant (# >= #[ITEM] in SUPPLY) COMP in  SUPPLY;
	 the parser will let it in, but actually #[ITEM] in SUPPLY returns A RELATIOPN WITH SINGLE ATTRIBUTE OF
	 TYPE .interger. So the query should be illegal.
	 When executes this, it will throw interpreter buildtree error.
	 */
      if (d == null){
	Relation r = env.lookupRel(node.name,true);
	if(r == null)
	  throw new  InterpretError(node.name + " is not declared.");
	else throw new  InterpretError(node.name + " is a relation. The query is invalid");
      }
      else if(!alldoms.contains(d))
	alldoms.addElement(d);
    }
    
    for(int i = 0; i < node.jjtGetNumChildren();i++)
      alldoms = qtEvaluateDomInQuant((SimpleNode)node.jjtGetChild(i),env,alldoms);
    
    return alldoms;
  }
  
  
  /*End of hongyu Added. Feb 14, 02 */ 
 
     /**
   * Returns relation only with tuples selected by the selector as
   *   specified by the tree of node.
   *
   * JRelixNS modified/enhanced this method:
   *   In jRelixSJ, nullary-Relation/IDLIST and boolean are now compatible.
   *   Now selector may contain IDLIST within and need cleanup
   *   of temporary relation files created for tuple-by-tuple
   *   actualization. 
   */
  private Relation evaluateSelect(SimpleNode node, String filename,
				  Hashtable htable, Environment env)
       throws InterpretError
  {
    int opcode = node.opcode;
    
    String tmpname = nextTempName();
    Relation rel = null;
    
    // get conditional expression for selection
    SimpleNode nodel = (SimpleNode) node.jjtGetChild(0);
    // get relational expression for selection
    SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
    //REVISIT:
    //This code is not computation aware.
    //Patrick added this next part to handle compuations
    //The whole code should be look over taking into  consideration
    //computations
    Relation rl = evaluateTLExpression(noder, tmpname, htable, env);
    if (rl.rvc == COMPUTATION)
      { 
       //return ((Computation) rl).applySelect(nodel,filename, env);
       //modified by Yi Zheng June 25, 2002
       if(this.destrel != null) //r1 is a nested attribute of destrel
       {
       	//System.out.println("nested computation");
       	Computation nr = ((Computation)rl).duplicate(); //each row get a new computation with new states
       	this.destrel.load(true); //make sure data is not null
       	return nr.applySelect(this.destrel, this.currow, nodel, filename, env);
       }
       else
       	return ((Computation) rl).applySelect(nodel,filename, env); //r1 is top level
      }
    deleteFile(tmpname);
    //End of code added by Patrick
    
    // check for special case for optimization
    //         where A in R; 
    // here A is in R and of boolean type, no actualization is needed
    if (nodel.type == OP_IDENTIFIER) {
      Domain d = env.lookupDom(nodel.name, true);
      if (d == null)
	throw new 
	  InterpretError("attribute " + nodel.name + " not defined");
      // evaluate the relational expression for selection
	  
	  //Yi Zheng May 13, 2002, following code is added to exclude state domains
	  //from being projected
	  if (d.isStateDom == 1)
	  	throw new InterpretError(nodel.name + " is a state domain");
	  
	  //end of Yi's Addition May 13, 2002
	  
      rel = evaluateTLExpression(noder, tmpname, htable, env);
      
 // START Since jRelixNS BUGFIX
// It can be either a boolean or an IDLIST.
// Can no longer assume it is a boolean here.
if (Relation.isSigmaJoinImplemented != true) {
     
      if (rel.isDomainIn(d)) { // domain d is in rel, no actualization
	Relation rr = rel.select(d, filename);
	deleteFile(tmpname);
	return rr;
      } 
      }
// END Since jRelixNS BUGFIX

    }
    // such two cases : 
    //    where A = 2 and B = 3 in R;
    //    <OR>
    //    let COND be A = 2 and B = 3;
    //    where COND in R;
    if (nodel.type != OP_IDENTIFIER) 
      rel = evaluateTLExpression(noder, tmpname, htable, env);//This seems redudancy as the one above hongyu
    //System.out.println(" after red getting org rel");
    //rel.print();
    SimpleNode sn = new SimpleNode("virtmp");
    sn.jjtAddChild(nodel);
    
    //added by Yi Zheng May 13, 2002, to prevent state domains from being used in condition
    traverseNode(nodel,0,env);
    //end of Yi Zheng's addition May 13, 2002
    
    Domain d = new Domain("0select0", BOOLEAN, 0, sn);
    Domain[] doms = new Domain[1];
    doms[0] = d;
    
    // the follwing is  added by Andy
    // START Since jRelixNS
    // added support to cleanup
 //   if (env.isSigmaJoinImplemented != true) {
 	 if (Relation.isSigmaJoinImplemented != true) {
      //System.out.println(" isSigmaJoinImplemented != true ");
      Relation rr = new Actualizer(doms, rel, env).actualize();
      rr.name = tmpname;
      rr.save();
      // select only the attributes in relation rel and with true value
      // in domain d. 
      Relation rrr = rr.select(d, rel.numattrs, filename);
      
      deleteFile(tmpname);
      
      return rrr;
    }
    else{
      // This is added for the boolean-IDLIST compatibility.
      // Now selector may contain IDLIST within and need cleanup
      // of temporary relation files created for tuple-by-tuple
      // actualization.
     // System.out.println(" isSigmaJoinImplemented == true ");
      //rel.print();
      Actualizer actualizer = new Actualizer(doms, rel, env);
      Relation rr = actualizer.actualize();
      rr.name = tmpname;
      rr.save();
      // select only the attributes in relation rel and with true value
      // in domain d. 
      Relation rrr = rr.select(d, rel.numattrs, filename);
      actualizer.cleanup();
      
      deleteFile(tmpname);
      
      return rrr;
    }
             // END Since jRelixNS

  }      
  // evaluate Top Level expression, which results in a relation
  // note that the hashtable is used to provide the initial value for
  // evaluating recursive view, and filename is the data file name for
  // the result of this expression
  public Relation evaluateTLExpression(SimpleNode node, 
				       String filename, Hashtable htable, Environment env)
       throws InterpretError
  {
    int type = node.type, opcode = node.opcode;
    String name = node.name;
   
   	//start Zongyan's addition, July 2002
	if (node.url != null ) {
	aldatpClient.request(node,env);  // need to check response and return it ^_^
	  if (aldatpClient.rel ==null)
		throw new InterpretError( aldatpClient.rspCode);
	  
	  if (aldatpClient.rel.rvc == 17) { // comp
		 Computation c = new Computation(aldatpClient.rel.tree,env);
		 c.name =aldatpClient.rel.name;
		 return c;
	  }
	  else 	// regular relation
		  return aldatpClient.rel;
	}
	
	else 
	// end Zongyan's addition, July 2002

    switch(type) {
    case OP_IDENTIFIER :
    case OP_EVENTNAME :
      {
	Relation rel = null;
	rel = (Relation) htable.get(name);
	if (rel != null) // there is initial for substitution
	  {    
	  return rel;
	  }
	rel = env.lookupRel(name, true);
	//REVISIT:This env lookup should be the first one
	//NREnvironment e2 = (NREnvironment) htable.get("#env2");
	//if (e2 != null){
	//   rel = e2.lookup(name);
	//}
	
	if (rel == null) {
	 // throw new InterpretError("no such relation : " + name);
		//bug fix by Yi Zheng Sept 28, 2002
		rel = env.lookupRel("." + name, true);
		if (rel == null)
	 		throw new InterpretError("no such relation : " + name);
		
		
		//end of bug fix Sept 28, 2002

	}
	// REVISIT : computation
	if (rel.rvc == COMPUTATION) {
	  return rel;
	}else if (rel.rvc == VIEW) {
	  return evaluateView(rel.tree, filename, htable, env);
	}else{ // ordinary relation
	  return rel;
	}
      }
    case OP_JOINOPERATOR : {
      return evaluateJoin(node, filename, htable, env);
    }
    case OP_PROJECT : {
      return evaluateProject(node, filename, htable, true, env);
    }
    
    //YuLing added
    case OP_GEDIT   : {   
      return evaluateGedit(node,filename,htable,true,env,null);
    }
    //end of YuLing added
    
    case OP_TSELECT : { // REVISIT: optomize TSelector in one pass
      // map this node to : projection + selection
      node.type = OP_PROJECT;
      node.opcode = OP_WHENIN;

      return evaluateTLExpression(node, filename, htable, env);
    }
    case OP_SELECT : {
      return evaluateSelect(node, filename, htable, env);
    }
    case OP_QSELECT : {
      return evaluateQSelect(node, filename, htable, env);
    }
    case OP_PICK : {
      SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
      String tmpname = nextTempName();
      Relation rel = evaluateTLExpression(nd, tmpname, htable, env);
      Relation rr = rel.pick(filename);
      deleteFile(tmpname);
      return rr;
    }
     case OP_ATTRIBSOF : { //case from Andrey
      SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
	  String tmpname = nextTempName();
	  Relation rel = evaluateTLExpression(nd, tmpname, htable, env);
	  Relation rr = rel.AttribsOf(filename);
	  deleteFile(tmpname);
	  return rr;
	 
 }

    case OP_ARRAY :
      //BUGS:
      //  Suppose Array has three attrs A,B,C.
      //  1) R <- Array[1,2,3,4]; 
      //     This will run, the 4 is of no consequence. This could be
      //     disallowed by looping through all 64 bits of the 'bits'
      //     field in the Simplenode ArrayList below. This loop is 
      //     necessary since the bits field is protected. 
      //  2) R <- Array[1,,,,,,,,];
      //     This will run. Furthermore, there is no way to fix this
      //     without changing the 'jjt' file and probably the also the
      //     SimpleNode class. Currently, the parser soed not provide me 
      //     with the information that there are so many commas in the call.
      //------------------------------------------------------------
      
      // Array syntax is syntactic sugar for a TSelector.
      // So convert syntax tree of Array to a syntax tree for a TSelector
      {  //Each domain contributes either to the projection list (ProjList)
	//or to the selection clause (compareNodes)
	SimpleNode ProjList = new SimpleNode("ExpressionList");
	ProjList.set(OP_EXPRLIST, OP_EXPRLIST);
	Vector compareNodes = new Vector();
	//Dissect parts of the actual array syntax tree structure
	//Identifier is the relation that array operator is applied to 
	SimpleNode Identifier = 
	  ((SimpleNode) node.jjtGetChild(0)).jjtDeepDuplicate();
	SimpleNode ArrayList = (SimpleNode) node.jjtGetChild(1);
	Relation r = env.lookupRel(Identifier.name, true);
	
	if (r==null)
	  {  throw new InterpretError("No such relation "+Identifier.name);
	  }
	//REVISIT: If r is a view, it must be evaluated first
	//--done March 19, 2002 Yi Zheng added the following
	 if (r.rvc == VIEW)
	 {
	   String tmpname = nextTempName();
	   r = evaluateView(r.tree, tmpname, htable, env);
	   deleteFile(tmpname);
	 }


	int ALindex=0; //index into the children of ArrayList
	for(int i=0; i<r.numattrs; i++)
	  {  if (ArrayList.isBitSet(i))
	    {  //becomes selection attribute
	      //Create this tree structure:
	      //    Compare:300:323:null:0
	      //     Identifier:230:230:I1:0
	      //     Literal:470:453:null:0    <- could be an expression etc.
	      SimpleNode Compare = new SimpleNode("Compare");
	      Compare.set(OP_BIOPERATOR, OP_EQ);
	      SimpleNode id = new SimpleNode("Identifier");
	      id.set(OP_IDENTIFIER, OP_IDENTIFIER,r.domains[i].name);
	      id.jjtSetParent(Compare); //link up Compare and id
	      Compare.jjtAddChild(id);  //link up Compare and id
	      //Graft the node from ArrayList onto new tree structure
	      SimpleNode nd = ((SimpleNode) ArrayList.jjtGetChild(ALindex))
		.jjtDeepDuplicate();
	      nd.jjtSetParent(Compare); //link up Compare and nd
	      Compare.jjtAddChild(nd);  //link up Compare and nd
	      ALindex++;
	      //Store this new 'compare' tree
	      compareNodes.addElement(Compare);
	    }
	  else
	    {  //becomes projection attribute
	      //Add the corresponding domain to the projection list
	      SimpleNode id = new SimpleNode("Identifier");
	      id.set(OP_IDENTIFIER, OP_IDENTIFIER,r.domains[i].name);
	      id.jjtSetParent(ProjList); //link up ProjList and id
	      ProjList.jjtAddChild(id);  //link up ProjList and id
	    }
	  }
	//Now take all compare nodes and 'AND' them together in pairs
	int numElements = compareNodes.size();
	//top points to the top of the synax tree as it is constructed
	SimpleNode top = null;
	if (numElements > 0)
	  {  top = (SimpleNode) compareNodes.lastElement();
	  compareNodes.removeElementAt(numElements-1);
	  numElements--;
	  for(; numElements>0; numElements--)
	    {  SimpleNode And = new SimpleNode("And");
	    And.set(OP_BIOPERATOR, OP_AND);
	    top.jjtSetParent(And); //link up And and top
	    And.jjtAddChild(top);  //link up And and top
	    SimpleNode Compare = (SimpleNode) compareNodes.lastElement();
	    compareNodes.removeElementAt(numElements-1);
	    Compare.jjtSetParent(And); //link up And and Compare
	    And.jjtAddChild(Compare);  //link up And and Compare
	    top = And;
	    }
	  }
	//Finally put everything together into a syntax tree
	if (ProjList.jjtGetNumChildren() == 0 && top == null)
	  {  //REVISIT: Impossible or view
	  }
	//else if(top == null) modified by Yi Zheng April 8, 2002 to fix COMPUTATION bug
	else if(top == null && r.rvc != COMPUTATION)
	  {  //Array is like a simple Project clause, not a TSelect
	    //Create following structure:
	    // Project:421:427:null:0
	    //  Projector:420:420:null:2
	    //   ExpressionList:592:592:null:0
	    //    Identifier:230:230:I1:0
	    //    Identifier:230:230:I2:0
	    //  Identifier:230:230:Array:0
	    SimpleNode Project = new SimpleNode("Project");
	    Project.set(OP_PROJECT, OP_WHENIN);
	    SimpleNode Projector = new SimpleNode("Projector");
	    Projector.set(OP_PROJECTOR, OP_PROJECTOR);
	    Projector.setBit(1);
	    Projector.jjtSetParent(Project); //link up Project and Projector
	    Project.jjtAddChild(Projector);  //link up Project and Projector
	    ProjList.jjtSetParent(Projector);//link up Projector and ProjList
	    Projector.jjtAddChild(ProjList); //link up Projector and ProjList
	    Identifier.jjtSetParent(Project); //link up Project and Identifier
	    Project.jjtAddChild(Identifier);  //link up Project and IDentifier
	    return evaluateTLExpression(Project, filename, htable, env);
	  }
	  //this else case added by Yi Zheng for cases like aComp[];
	  else if(top == null && r.rvc == COMPUTATION)
	  {//Array class is like a TSelect where Selector contains the literal TRUE
	  	SimpleNode TSelect = new SimpleNode("TSelect");
	  	TSelect.set(OP_TSELECT, OP_TSELECT);
	  	SimpleNode Projector = new SimpleNode("Projector");
	  	Projector.set(OP_PROJECTOR, OP_PROJECTOR);
	  	Projector.jjtSetParent(TSelect); //link up TSelect and Projector
	    TSelect.jjtAddChild(Projector);  //link up TSelect and Projector
	    if(ProjList.jjtGetNumChildren() > 0)
	      { Projector.setBit(1);
	      ProjList.jjtSetParent(Projector);//link up Projector and ProjList
	      Projector.jjtAddChild(ProjList); //link up Projector and ProjList
	      }
	    SimpleNode Selector = new SimpleNode("Selector");
	    Selector.set(OP_SELECT, OP_WHEREIN);
	    Selector.jjtSetParent(TSelect); //link up TSelect and Selector
	    TSelect.jjtAddChild(Selector);  //link up TSelect and Selector
	    SimpleNode ctop = new SimpleNode("Literal");
	    ctop.set(OP_LITERAL, OP_BOOLEAN);
	    ctop.setObject(new Byte((byte) 1));
	    ctop.jjtSetParent(Selector);
	    Selector.jjtAddChild(ctop);
	    Identifier.jjtSetParent(Selector);//link up Selector and Identifier
	    Selector.jjtAddChild(Identifier); //link up Selector and IDentifier
	    return evaluateTLExpression(TSelect, filename, htable, env);
	  }
	else
	  {  //Array call is like a TSelect
	    //Create following structure:
	    // TSelect:422:422:null:0
	    //  Projector:420:420:null:2
	    //   ExpressionList:592:592:null:0 <- (This is stored in ProjList)
	    //    Identifier:230:230:I3:0
	    //  Selector:423:425:null:0
	    //   Compare:300:323:null:0        -|
	    //     Identifier:230:230:I1:0      | (This portion is in 'top'
	    //     Literal:470:453:null:0      -|
	    //   Identifier:230:230:Array:0
	    SimpleNode TSelect = new SimpleNode("TSelect");
	    TSelect.set(OP_TSELECT, OP_TSELECT);
	    SimpleNode Projector = new SimpleNode("Projector");
	    Projector.set(OP_PROJECTOR, OP_PROJECTOR);
	    Projector.jjtSetParent(TSelect); //link up TSelect and Projector
	    TSelect.jjtAddChild(Projector);  //link up TSelect and Projector
	    if(ProjList.jjtGetNumChildren() > 0)
	      { Projector.setBit(1);
	      ProjList.jjtSetParent(Projector);//link up Projector and ProjList
	      Projector.jjtAddChild(ProjList); //link up Projector and ProjList
	      }
	    SimpleNode Selector = new SimpleNode("Selector");
	    Selector.set(OP_SELECT, OP_WHEREIN);
	    Selector.jjtSetParent(TSelect); //link up TSelect and Selector
	    TSelect.jjtAddChild(Selector);  //link up TSelect and Selector
	    top.jjtSetParent(Selector); //link up Selector and top
	    Selector.jjtAddChild(top);  //link up Selector and top
	    Identifier.jjtSetParent(Selector);//link up Selector and Identifier
	    Selector.jjtAddChild(Identifier); //link up Selector and IDentifier
	    return evaluateTLExpression(TSelect, filename, htable, env);
	  }
      }
    case OP_EDIT : 
    case OP_ZORDER : 
    case OP_EVAL : // ???
    case OP_RENAME : { // ??? 
      throw new InterpretError("TO BE IMPLEMENTED");
    }
    // many more other case ...
    
    // This case is added by hongyu (June 25, 01)
    // a <- [ ITEM, COMP] quant ( #  = 2) COMP , ( % > 0.1) DEPT where VOL1 > 2 in SUPPLY;
    case OP_BIOPERATOR : {
      // REVISIT: optomize TSelector in one pass
      // map this node to : projection + selection
      node.type = OP_PROJECT;
      node.opcode = OP_WHENIN;
      // System.out.println("here " + node.type + " " + node.opcode);
      return evaluateTLExpression(node, filename, htable, env);
    }
    case OP_FUNCTION:
    	throw new InterpretError("OP_FUNCTION is not implemented yet");
    default : {
      throw new 
	InterpretError("invalid/not implemented relational expression");
    }
    }
  }      
  // evaluate top level view and return the result relation
  /**
   * Given node is the tree of a view named node.name, evalutate it
   * and return the resulting relation.
   * @param node is the tree of the declared view.
   * @param filename is the name of the result relation.
   * @param htable is the null hash table.
   * @param env is the environment.
   *
   *         node
   *        /  |  \
   *       /   |   \
   *      /    |    \
   *    bit(0)    bit(1)  jjt optional bits
   *   [nodei] nd [nodet]
   * nd is the join expression.
   * 
   * jjtGetChild offset #:
   *           0    
   *      0    1 
   *      0    1     2
   *      0    1     2
   */
  private Relation evaluateView(SimpleNode node, String filename,
	 Hashtable htable, Environment env) 
       throws InterpretError
  {
    String name = node.name; // view name
   	String url = node.url;   // Zongyan   Aug 2002
    SimpleNode nodei = null, nodet = null, nd = null;
    Relation rel_init = null; // initial relation for recursive view
    Relation rel_target = null; // target relation for recursive view
    String initname = nextTempName(); // data file for rel_init
    String targetname = nextTempName(); // data file for rel_target
    
    if (node.isBitSet(0)) { // there is initial part
      nodei = (SimpleNode) node.jjtGetChild(0);
      nd = (SimpleNode) node.jjtGetChild(1); // body of view
      if (node.isBitSet(1)) // there is target part
	nodet = (SimpleNode) node.jjtGetChild(2);
    } else {
      nd = (SimpleNode) node.jjtGetChild(0); // body of view
      if (node.isBitSet(1)) // there is target part
     nodet = (SimpleNode) node.jjtGetChild(1);
    }
    
    if (isRecursiveView(nd, name, new Hashtable(), env)) { 
      // for recursive view
      
      
      // find start relation for recursion
      if (nodei != null) { // there is initial part
	// the initial part (if any) should not refer to itself
	if (isRecursiveView(nodei, name, new Hashtable(), env))
		throw new InterpretError("recursive view " + name
					 + " refers to itself in initial expression");
	rel_init = evaluateTLExpression(nodei, initname, htable, env);
      } else { // no initial part, see relational part under this name
	Relation rtmp = env.lookupRel(name, true);
	if (rtmp == null || rtmp.numattrs == 0) { // no relational part
	  throw new InterpretError("nowhere to start in recursive view");
	} else { // use relational part as initial
	  rel_init = rtmp.duplicate(false);
	}
      }
      
      // find if there is a target
      if (nodet != null) { // there is target part
	// the target part (if any) should not refer to itself
	if (isRecursiveView(nodet, name, new Hashtable(), env))
	  throw new InterpretError("recursive view " + name
				   + " refers to itself in target expression");
	rel_target = evaluateTLExpression(nodet, targetname, htable, env);
      }
      
      // check if rel_init == rel_target
      if (rel_target != null && equals(rel_init, rel_target)) {
	if (nodei == null) {
	  deleteFile(targetname);
	  return rel_init;
	} else {
	  Relation rel = rel_init;
	  deleteFile(initname);
	  deleteFile(targetname);
	  return rel;
	}
      } 
      
      int i = 0;
      while (true) { // loop for recursion until match target or unchanged
	// add (name, rel_init) pair to hash table to modify environment
	htable.put(name, rel_init);
	
	Relation rel = evaluateTLExpression(nd, filename, htable, env);
	
	
	if (rel_target != null && equals(rel, rel_target)) {
	  if (nodei != null) deleteFile(initname);
	  deleteFile(targetname);
	  htable.remove(name);
	  return rel;
	} else if (rel_target == null && equals(rel, rel_init)) {
	  if (nodei != null) deleteFile(initname);
	  htable.remove(name);
	  return rel;
	} else { // loop again
	  rel_init = rel;
		 }
      }
    } else { // for non recursive view, simply evaluate the expression
      //System.out.println("evaluating view : " + name);
      return evaluateTLExpression(nd, filename, htable, env);
    }
  }  
    // modified by Zongyan -- Aug 2002
  
  private void executeAssignment(SimpleNode node, Environment env)
       throws InterpretError
  {
    //printNode(node);
    int opcode = node.opcode;
    int type = node.type;

    switch(opcode) 
      {  case OP_ASSIGN : 
      case OP_ASSIGNADD :
		 Domain[] domsr = null, domsl = null;
      Relation rr = null, rl = null;
      String lname = ((SimpleNode) node.jjtGetChild(0)).name;
      //System.out.println( "ExecuteAssignment: "+lname );
      // data file name for right expression
      String tmpname = nextTempName(); 
      //rl = reltable.getRel(lname);
//      rl = env.lookupRel(lname, true);
  	  if ( type == OP_ASSIGNMENT)
	      rl = env.lookupRel(lname, true);
    
      if (node.jjtGetNumChildren() == 2) 
	{  
	  
	  rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1),
				    tmpname, new Hashtable(), env);
	} 
      else
	{  
	  
	  //System.out.println("four children");
	  // node.jjtGetNumChildren() == 4
	  // get the domain array for the left IDList
	  SimpleNode nodel = (SimpleNode) node.jjtGetChild(1);
	  domsl = IDListToDomains(nodel, env);
	  
	  // get the domain array for the right ExpressionList
	  SimpleNode noder = (SimpleNode) node.jjtGetChild(2);
	  domsr = ExpressionListToDomains(noder);
	  
	  rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(3),
				    tmpname, new Hashtable(), env);
	}
      // for case :
      // R <- R; or R <+ R; do nothing
      if (rl != null && rl == rr && domsl == null && domsr == null) 
	{  return; // for R <- R; do nothing
	} 
      // if domsr is not in relation rr, actualization is needed.
      if (!rr.isDomainsIn(domsr)) 
	{  
	
	//modified by Yi Zheng March 23, 2002
	//have to cleanup actualizer in the end
	Actualizer act = new Actualizer(domsr, rr, env);
	Relation rtmp = act.actualize(); 
	// can't use rr.name, since this may overwrite the
	// rr if it is ID rather than an expression
	rtmp.name = tmpname;
	rtmp.save(); rtmp.unload();
	rr = rtmp;
	}
		  Relation rel = null; //Zongyan's modification
	  String mytmp = nextTempName(); //Zongyan's modification

      if (opcode == OP_ASSIGN) 
	{  // REVISIT : better to assign first then delete
	  // if left side relation exists, delete it
	  	  if ( type == OP_ASSIGNMENT) {//added by Zongyan

	  if (rl != null) 
	    {  //Only remove rl if it is in the most deeply
	      //nested frame of the environment
	      env.removeRel(lname, false);
	    }
	  env.put(rr.assign(domsr, lname, domsl));
	  	  }
	  else if ( type == ALDATP_ASSIGNMENT) {
		 
		  rel = rr.assign(domsr, mytmp, domsl);
		 env.put(rel);
		
		 distributedStmt(rel.name,lname,node.url,OP_ASSIGN,env);
		 env.removeRel(rel.name, false);
	  }
	}
      else
	{  // OP_ASSIGNADD
		   if ( type == OP_ASSIGNMENT) {

	  if (rl == null) 
	    {  Global.println("Warning: cannot find relation " + lname);
	    env.put(rr.assign(domsr, lname, domsl));
	    }
	  else
	    {  // left relation does not exist
	      rr.assignAdd(domsr, rl, domsl);
	    }
	    	   } else  if ( type == ALDATP_ASSIGNMENT) {
		   rel = rr.assign(domsr, mytmp, domsl);
		   env.put(rel);
		   distributedStmt(rel.name,lname,node.url,OP_ASSIGNADD, env);
		 	 env.removeRel(rel.name, false);
	  
		  
	  }
	}
      //delete the temporary data file
   
      deleteFile(tmpname);
      break;
      
      default : // impossible to here
	System.out.println("unknown assignment statement");
	break;
      }
  } 
  
  /**
   * used in computation
   * @ author Yi Zheng
   * added April 8, 2002
   */
  public void executeCompCommand(SimpleNode node, Environment env)
  throws InterpretError
  {
  	executeCommand(node, env);
  }  
    
  private void executeCommand(SimpleNode node, Environment env) 
       throws InterpretError
  {
    int opcode = node.opcode; 
    String name = node.name; 
    
    RelTable reltable = env.reltable;
    DomTable domtable = env.domtable;
    
    switch(opcode) 
      {
      case OP_HELP : {
	help(name); 
	break;
      }
      case OP_QUIT : {
	// REVISIT : may be computation table
		Global.permit_start = false;

	Global.exit(0); 
	break;
      }
      case OP_TRACE : {
	Global.trace = Global.trace ? false : true;
	if (Global.trace && Global.traceoutstream == null) {
	  try {
	    Global.traceoutstream = new PrintStream(new FileOutputStream(".trace"), true);
	  } catch (IOException e) {
	    Global.trace = false;
	    Global.println("I/O Exception in creating file '.trace'."); 
	  }     
	}
	String s = Global.trace ? "on" : "off";
	Global.println("Note: trace is " + s);
	break;
      }
      
      case OP_INPUT : {
	String s = name.substring(1, name.length() - 1);
	FileInputStream fin = null;
	try {
	  fin = new FileInputStream(s);
	  Global.jin.resetIn(fin);
	  parser.ReInit(Global.jin);
	  instack.push(in);
	  in = fin;
	} catch (FileNotFoundException e) {
	  throw new InterpretError("Input : file " + name + " not found");
	}  
	break;
      }
      case OP_SSD : {
	Global.ssd = Global.ssd ? false : true;
	String s = Global.ssd ? "on" : "off";
	Global.println("Note: show system domain mode is " + s);
	break;
      }
      case OP_SSR : {
	Global.ssr = Global.ssr ? false : true;
	String s = Global.ssr ? "on" : "off";
	Global.println("Note: show system relation mode is " + s);
	break;
      }
      case OP_DEBUG : {
	Global.debug = Global.debug ? false : true;
	String s = Global.debug ? "on" : "off";
	Global.println("Note: debug mode is " + s);
	break;
      }
      case OP_BATCH : {
	Global.batch = Global.batch ? false : true;
	String s = Global.batch ? "on" : "off";
	Global.println("Note: batch mode is " + s);
	break;
      }
      case OP_EXPERT : {
	Global.expert = Global.expert ? false : true;
	String s = Global.expert ? "on" : "off";
	Global.println("Note: expert mode is " + s);
	break;
      }
      case OP_TIME : {
	Global.time = Global.time ? false : true;
	String s = Global.time ? "on" : "off";
	Global.println("Note: timer is " + s);
	break;
      }
      case OP_DELD : {
	SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
	for (int i = 0; i < nd.jjtGetNumChildren(); i++) {
	  //Can only remove a domain that is in most deeply 
	  //nested scope
	  env.domtable.rmDom(((SimpleNode) nd.jjtGetChild(i)).name);
	}
	break;
      }
      case OP_DELR : {
	SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
	String eaname = new String("");
	if (nd.opcode == OP_EVENTNAME) {
	  eaname = getEveAtt(node, env);  
	  env.reltable.rmRel(eaname); 
	} else {
	  for (int i = 0; i < nd.jjtGetNumChildren(); i++) {
	    env.reltable.rmRel(((SimpleNode) nd.jjtGetChild(i)).name);
	  }
	} 
	break;
      }
      case OP_PR : {
	// REVISIT : for COMP -- done. Weizhong. 
	// get the temporary file name for the result of expression
	String tmpname = nextTempName(); 
	
	SimpleNode firstchild = (SimpleNode) node.jjtGetChild(0);
	if (firstchild.opcode == OP_EVENTNAME) {
	  // event name
	  // firstchild.name = env.getEventName(firstchild);
	  firstchild.name = getEveAtt(node, env);
	} 
	
	// print trigger in the current environment.
	//first condition added by Yi Zheng, May 14, 2002
	if ((firstchild.name != null) && (firstchild.name.compareTo("trigger") == 0)) {
	  // handled in the overloaded executeCommand
	} else {
	  // deal with computation as well as relation.
	  Relation rel = evaluateTLExpression( firstchild, tmpname, new Hashtable(), env);
	  if (rel != null) {
	    rel.print(null, rel.name != tmpname); 
	  }
	}
	deleteFile(tmpname);
	break;
      }
      case OP_SC : {
		 //REVISIT: How about showing not top-level computations
	Relation c = env.lookupRel(node.name, true);
	if (c == null || c.rvc != COMPUTATION)
	  {  throw new InterpretError("No such computation: "+node.name);
	  }
	else
	  {  ((Computation) c).sc();
	  }
	break;
      }
      case OP_SD : {
	if(node.name != null){
	  Domain d = env.lookupDom(node.name, true);
	  if(d == null)
	    throw new InterpretError("domain '"+node.name+
				     "' does not exist.");
	  else
	    env.domtable.showDom(d);
	}else
	  env.domtable.showTable(Global.ssd);
	
	break;
      }
      case OP_SR : {
	if(node.name != null){
	  Relation r = env.lookupRel(node.name, true);
	  if(r == null)
	    throw new InterpretError("relation '"+node.name+
				     "' does not exist.");
	  else
	    env.reltable.showRel(r);
	}else
	  env.reltable.showTable(Global.ssr);
	
	break;
      }
      case OP_SRD : {//case modified by Yi Zheng Sept. 29, 2002 to allow srd <rel>

	if (((SimpleNode)node).name == null)
	{
		Relation rel = env.lookupRel(".rd", true);
		if (rel != null) 
		  rel.print();
		else // shouldn't be here
		  throw new InterpretError("relation .rd is not there");
	}
	else
	{
		
		Relation rel = env.lookupRel(node.name, true);
		if (rel == null)
			throw new InterpretError("relation " + node.name + " does not exist.");
		else
		{
			Global.environment.reltable.dump();

			executeJRelixCommand("pr (where .rel_name = \"" + node.name +
			"\" in .rd);");	

			
		}	
		
	}
	break;
      }
      case OP_PRINT : {
	Global.println(name.substring(1, name.length() - 1)); 
	break;
      }
      case OP_EVENTON : {
	SimpleNode firstchild = (SimpleNode) node.jjtGetChild(0);
	if (firstchild.opcode == OP_EVENTNAME) {
	  // event name
	  // firstchild.name = env.getEventName(firstchild);
	  firstchild.name = getEveAtt(node, env);
	  Relation temprel = env.lookupRel(firstchild.name, true);
	  if (temprel != null) { 
	    temprel.active = 1;
	    temprel.tree.active = 1;	
	  } else
	    throw new InterpretError("Event Handler not found!");
	} else {
	  throw new InterpretError ("No such Event Handler!");
	}
	break;
      }      
      case OP_EVENTOFF : {
	SimpleNode firstchild = (SimpleNode) node.jjtGetChild(0);
	if (firstchild.opcode == OP_EVENTNAME) {
	  // event name
	  // firstchild.name = env.getEventName(firstchild);
	  firstchild.name = getEveAtt(node, env);
	  Relation temprel = env.lookupRel(firstchild.name, true);
	  if (temprel != null) {
	    temprel.active = 0;
	    temprel.tree.active = 0;
	  } else
	    throw new InterpretError ("Event Handler not found!");
	} else {
	  throw new InterpretError ("No such Event Handler!");
	} 
	break;
      }
      case OP_UNDO:
	// handler in the overloaded executeCommand function
	break;
      default : {
	throw new InterpretError("No such command");
      }
      }
  }                                    
  
  // weizhong
  // get the eventname + attrlist
  private String getEveAtt(SimpleNode node, Environment env)
       throws InterpretError
  {
    SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
    String eaname = new String("");
    SimpleNode attrlist;
    
    eaname = env.getEventName(nd);
    if (node.jjtGetNumChildren() > 1) {
      attrlist = (SimpleNode) node.jjtGetChild(1);
    } else {
      attrlist = null;
    }
    if (attrlist != null) {
      int childnum = attrlist.jjtGetNumChildren();
      SimpleNode childatt;
      if (childnum != 0) {
	childatt = (SimpleNode)attrlist.jjtGetChild(0);
	eaname = eaname + "[" + childatt.name;
	for (int t = 1; t < childnum; t++) {
	  childatt = (SimpleNode)attrlist.jjtGetChild(t);
	  eaname = eaname + "," + childatt.name;
	}
	eaname = eaname + "]";
      }
    }
    return eaname; 
  }
  
  // Weizhong
  // overload the function to handle the different level's undo command 
  // in event handlers
  // except OP_UNDO, all commands still be dealed with in original executeCommand
  private void executeCommand(SimpleNode node, Environment env, String rlname, int opc)
       throws InterpretError
  {
    int opcode = node.opcode; 
    TriggerNode triggernode = env.triggerroot.getNode(rlname, opc);
    
    switch(opcode) 
      {
      case OP_UNDO:
	doTrigger(env, triggernode, 1);
	break;
      case OP_PR : 
	// to handle print trigger command
	String tmpname = nextTempName(); 
	
	SimpleNode firstchild = (SimpleNode) node.jjtGetChild(0);
	
	if (firstchild.opcode == OP_EVENTNAME) {
			// event name
	  firstchild.name = env.getEventName(firstchild);
	} 
	
	
	// print trigger in the current environment.
	if (firstchild.name.compareTo("trigger") == 0) {
	  int tid = printAllTrigger(triggernode);
	  if (tid == 0)
	    System.out.println("TRIGGER of " + rlname + " IS EMPTY.");
	} else {
	  // deal with computation as well as relation.
	  Relation rel = evaluateTLExpression( firstchild, tmpname, new Hashtable(), env);
	  if (rel != null) {
	    rel.print(null, rel.name != tmpname); 
	  }
	}
	deleteFile(tmpname);
	break;
      default : {
	executeCommand(node, env);
      }
      }
  }                                    
  
  
  // Weizhong
  private int printTrigger(CTrigger trigger) 
       throws InterpretError
  {
    if (trigger.relname.compareTo("") == 0)
      return 0;
	else {
	  System.out.println("Trigger of "+trigger.relname);
	  
	  String tempname1 = nextTempName();
	  String tempname2 = nextTempName();
	  Relation temp1 = trigger.oldp.mujoin(OP_IJOIN, null, trigger.newp, null, tempname1);
	  Relation temp2 = trigger.oldp.mujoin(OP_DJOIN, null, temp1, null, tempname2);
	  temp2.print(null, false);
	  
	}	
    return 1;
  }

  
  // Weizhong
  private int printAllTrigger(TriggerNode triggernode)
       throws InterpretError
  {	
    int tid = printTrigger(triggernode.trigger);
    int childnum = triggernode.jjtGetNumChildren();
    for (int i = 0; i < childnum; i++) {
      TriggerNode childnode = (TriggerNode) triggernode.jjtGetChild(i);
      tid = tid + printAllTrigger(childnode); 
    }
    return tid;
  }	
  
  //wrapper used in Computation.java
  //Yi Zheng May 20, 2002
  public void executeCompDeclaration(SimpleNode node, Environment env)
 	throws InterpretError
  {
  	executeDeclaration(node, env);
  }
  
     /**
   * JRelixNS modified/enhanced this method:
   *   Workaround for IDLIST update.
   */
  private void executeDeclaration(SimpleNode node, Environment env)
       throws InterpretError
   {
     int opcode = node.opcode;
     
     switch(opcode) 
       { 
       case OP_RELATION :
	 RelationalDeclaration(node, env); 
	 break;
	 
       case OP_VIEW :
	 // remove the identifier node representing the name of the view
	 SimpleNode sn = (SimpleNode) node.jjtRemoveChild(0);
	 // optionally set the name of the root to be the name of view
	 node.name = sn.name;
	 // check if a relation of this name already exists, if so
	 // simply append the tree for view and change the type to
	 // view, otherwise, create a new entry
	 //We allow views/relations of the same name to exist in
	 //different frames of the environment so we use false here
	 Relation rel = env.lookupRel(sn.name, false);
	 if (rel != null) 
	   {  // add view node and change type
	     rel.rvc = VIEW;
	     rel.tree = node;
	   } 
	 else
	   {  // add new view entry
	     Relation r=new Relation(node.name, null, VIEW,
				     0, 0, 0, node, env);
	     env.put(r);
	   }
	 break; 
	 
       case OP_DOMAIN :
	 Vector domlist = new Vector();
	 Domain[] domarry = null;
	 SimpleNode idList;
	 
	 // if domain is of type IDLIST, fill the domlist vector
	 SimpleNode type = (SimpleNode) node.jjtGetChild(1);
	 if((type.opcode == OP_IDLIST || type.opcode == OP_COMP) && type.jjtGetNumChildren() > 0) //last condition added by Yi Zheng April 23, 2002
	   {  idList = (SimpleNode) type.jjtGetChild(0); 
	   for(int i = 0; i < idList.jjtGetNumChildren(); i++)
	     {  String dname = ((SimpleNode) idList.jjtGetChild(i)).name;
	     
	     // Check validation of the dom name. But note, we can't
	     // increment the domain's numref at the moment; We have
	     // to complete _all_ validation check. If the check
	     // fails, an exception is thrown and nothing happens.
	     Domain oldD = env.lookupDom(dname, true);
	    // if(env.lookupDom(dname, true) == null) Yi Zheng May 9, 2002
	    if(oldD == null)
	       {  throw new 
		    InterpretError("domain "+ dname + " has not been declared");
	       }
	      else if (oldD.isStateDom == 1) //Yi Zheng May 9, 2002
	      	throw new InterpretError("domain " + dname + " is a state domain");
	     else
	       {  //domlist.addElement(env.lookupDom(dname,true)); Yi Zheng May 9, 2002
	       domlist.addElement(oldD);
	       }
	     }
	   domarry= new Domain[domlist.size()];
	   domlist.copyInto(domarry);
	   }
	
	//Yi Zheng April 23, 2002
	if (domarry == null)
		domarry = new Domain[0];
	 // get the identifiers.
	 idList = (SimpleNode) node.jjtGetChild(0);
	 for(int i = 0; i < idList.jjtGetNumChildren(); i++)
	   {  SimpleNode id = (SimpleNode) idList.jjtGetChild(i);
	   
	   //If domain is being used, just skip this single declaration.
	   //We allow domains of the same name to exist in different
	   //frames of the environment so we use false here
	   Domain d = env.lookupDom(id.name, false);
	   if(d != null && d.numref != 0)
	  //   {  Global.println("Error: domain '"+ id.name + 
		//	       "' is being used.");
	    // }
	    	     { throw new                // Zongyan Aug 2002
		    InterpretError ("Error: domain '"+ id.name + 
			       "' is being used.");
	     }

	   else if (type.opcode == OP_IDLIST)
	     {  Domain newDom = new Domain(id.name, IDLIST, domarry, env);
	     env.put(newDom);
	     }
	   else if (type.opcode == OP_COMP)
	     {  Domain newDom = new Domain(id.name, COMP, domarry, env);
	     env.put(newDom);
	     }
	   else 
	     {  // Need to domain validation check & type check.
	       Domain newDom = new Domain
		 (id.name, type.opcode - OP_BOOLEAN, 0, null);
	       env.put(newDom);
	     }
	   }
	 break; 
	 
       case OP_LET :
/*	 // Get the virtual domain name.
	 String domname = ((SimpleNode)node.jjtGetChild(0)).name;

	 // Remove the first child, the rest of the tree will be
	 // stored in the tree field of this domain.
	 //  Weizhong: keep the original node, and operate on the tempnode
	 SimpleNode tempnode = node.jjtDeepDuplicate();
	 tempnode.jjtRemoveChild(0);
	 // node.jjtRemoveChild(0);
	 
	 
	 // START Since 
	     // Enhancements to jRelix76 for newVirDomain.
		if (Relation.isSigmaJoinImplemented != true) { //Andy, May 12, 2002
//	 if (env.isSigmaJoinImplemented != true) {
	   // Add it to the domain table.
	   //  Domain newDom = newVirDomain(domname, node, env);
	   Domain newDom = newVirDomain(domname, tempnode, env);
	   env.put(newDom);
	 }else {
                    // Enhancements to jRelix76.
	   // Add it to the domain table.
	   // Domain newDom = newVirDomain(domname, node, env);
	   // Correct problem with IDLIST update.
	   // Old problem: newVirDomain already updated domain's rel structure.
	   // but 'put' of a domain deletes the relation and doesn't
	   // recreate another one.  The 'save' and 'put' combo will
	   // put in the correct relation so no run-time exceptions
	   // will occur after a IDLIST domain update here.
	   
	  // Domain  dom_saved = env.lookupDom(domname,false); modified by Yi Zheng March 20, 2002
	   Domain newDom = newVirDomain(domname, tempnode, env);
	   
	   Relation rel_saved = env.lookupRel("."+domname,true);
	   env.put(newDom);
	  // if ((rel_saved != null)&&(dom_saved != null)) modified by Yi Zheng March 20, 2002
	  if (rel_saved != null)
	     env.put(rel_saved);
	 }
	 // END Since jRelixNS
	 break;
	 
*/ //commented out to use Andrey's code, by Yi Zheng May 26, 2002
		SimpleNode tempnode = node.jjtDeepDuplicate();
		tempnode= (SimpleNode)tempnode.jjtGetChild(0);
		
		//This is a static evaluation of a "let" statement
		if(tempnode.opcode==OP_EVAL)
		{   
			Domain domList[] = ExpressionListToDomains(tempnode,new Hashtable(),env);
				
			for(int i=0;i<domList.length;i++)
			{
			//	tempnode = node.jjtDeepDuplicate();
				tempnode = (SimpleNode) node.jjtRemoveChild(0);
				SimpleNode newNode = SimpleNode.jjtCreateNode("Identifier",OP_IDENTIFIER,OP_IDENTIFIER,
														domList[i].name);
				node.jjtAddChildAtHead(newNode);
		        newNode.jjtSetParent(node);
		//		AddVirtDom2DomTable(node);
		AddVirtDom2DomTable(node, env); //modified by Yi Zheng May 26, 2002
			}
		}
		 
		else 
		//	AddVirtDom2DomTable(node);
		AddVirtDom2DomTable(node, env); //modified by Yi Zheng May 26, 2002
		break;
	
       case OP_COMPUTATION :
	 Computation c = new Computation(node, env); 
	 
	 Relation temprel = env.lookupRel(c.name, true);
	 if (temprel != null) {
	 	c = null;
	 	//added by Yi Zheng April 6, 2002
	 	//warning already issued in Global.java
	 } else {  
	   // env.put(c);
	   
	   if  (Global.saveCompCode(env) == 1){
	     // for a new event handler, default status is active
	     SimpleNode fchild = (SimpleNode) node.jjtGetChild(0);
	     if (fchild.opcode == OP_EVENTNAME) {
	       c.tree.active = 1;
	       c.active = 1;
	     }
	   }
	   else Global.println("Warning: error occurred while saving the computation code");	
	   
	   env.put(c);
	 }
	 break;
       }
   }    
   
  //Original code for processing Virtual Domain declaration
  // A. Rozenberg
  // enhanced with Andy's modification, May 26, 2002
  // parameter env added by Yi Zheng May 26, 2002
  private void AddVirtDom2DomTable(SimpleNode node,  Environment env) throws InterpretError
  {
	 // Get the virtual domain name.
	 String domname = ((SimpleNode)node.jjtGetChild(0)).name;

	 // Remove the first child, the rest of the tree will be
	 // stored in the tree field of this domain.
	 //  Weizhong: keep the original node, and operate on the tempnode
	 SimpleNode tempnode = node.jjtDeepDuplicate();
	 tempnode.jjtRemoveChild(0);
	 // node.jjtRemoveChild(0);
	 
	 
	 // START Since 
	     // Enhancements to jRelix76 for newVirDomain.
		if (Relation.isSigmaJoinImplemented != true) { //Andy, May 12, 2002
//	 if (env.isSigmaJoinImplemented != true) {
	   // Add it to the domain table.
	   //  Domain newDom = newVirDomain(domname, node, env);
	   Domain newDom = newVirDomain(domname, tempnode, env);
	   env.put(newDom);
	 }else {
                    // Enhancements to jRelix76.
	   // Add it to the domain table.
	   // Domain newDom = newVirDomain(domname, node, env);
	   // Correct problem with IDLIST update.
	   // Old problem: newVirDomain already updated domain's rel structure.
	   // but 'put' of a domain deletes the relation and doesn't
	   // recreate another one.  The 'save' and 'put' combo will
	   // put in the correct relation so no run-time exceptions
	   // will occur after a IDLIST domain update here.
	   
	  // Domain  dom_saved = env.lookupDom(domname,false); modified by Yi Zheng March 20, 2002
	   Domain newDom = newVirDomain(domname, tempnode, env);
	   //System.out.println("virtual domain " + domname + " is added to env");
	   Relation rel_saved = env.lookupRel("."+domname,true);
	   env.put(newDom);
	  // if ((rel_saved != null)&&(dom_saved != null)) modified by Yi Zheng March 20, 2002
	  if (rel_saved != null)
	     env.put(rel_saved);
	 }
	 // END Since jRelixNS
  }

   //wrapper used in Computation.java
   //Yi Zheng May 20, 2002
   public void executeCompStatement(SimpleNode node, Environment env)
   	throws InterpretError
   	{
   		executeStatement(node, env);
   	}     
  private void executeStatement(SimpleNode node, Environment env)
       throws InterpretError
  {  int opcode = node.opcode;
  
   switch(opcode) 
     {
     case OP_CONDITIONAL :
       SimpleNode ifNode, thenNode, elseNode;
       int numchildren = node.jjtGetNumChildren();
       ifNode = (SimpleNode) node.jjtGetChild(0);
       thenNode = (SimpleNode) node.jjtGetChild(1);
       
       if (EvalExpr.evalBooleanExpr(ifNode, env)) //short-circuit evaluation
	 {  
	   interpret(thenNode, env);
	 }
       else if(numchildren == 3)
	 {  
	   elseNode = (SimpleNode) node.jjtGetChild(2);
	   interpret(elseNode, env);
	 }
       break;
     case OP_STMTBLOCK :
     case OP_COMPBLOCK :	// Weizhong, for computation statements
       int numChildren = node.jjtGetNumChildren();
       for (int i=0; i<numChildren; i++)
	 {   interpret((SimpleNode) node.jjtGetChild(i), env);
	 }
       break;
     case OP_COMPCALL :
     	 // added by Zongyan Sep 2002
	  Relation r ;
	  SimpleNode compnode = ((SimpleNode) node.jjtGetChild(0));
	  if (compnode.url != null ) {
	  aldatpClient.request(compnode,env);  // need to check response and return it ^_^
	  if (aldatpClient.rel ==null)
		throw new InterpretError( aldatpClient.rspCode);
	  
	  if (aldatpClient.rel.rvc == 17) { // comp
		 Computation c = new Computation(aldatpClient.rel.tree,env);
		 c.name =aldatpClient.rel.name;	
		 r = c;
	  }
	  else 	// regular relation
		   throw new InterpretError(aldatpClient.rel.name  + "is not computation");
   	}
	else 
	 
	{
	   // end of Zongyan's addition
       String compName = ((SimpleNode) node.jjtGetChild(0)).name;
       r = env.lookupRel(compName, true);
       if (r == null || r.rvc != COMPUTATION)
	 {  throw new InterpretError("'"+compName+"' is not computation");
	 }
	 }
       String tmpname = nextTempName();
       ((Computation) r).applyInOut(node, tmpname, env);
       deleteFile(tmpname);
       break;
     case OP_SEQUENCE :
     case OP_PARALLEL :
     case OP_CHOICE :
     case OP_FORLOOP :
     case OP_WHILELOOP :
     case OP_EXIT : 
     case OP_DEADLOCK :
     case OP_EXEC :
       System.out.println("TO BE IMPLEMENTED"); 
       break;
     }
  }      

  // to get the '.id' col position
  // Weizhong
  private int getIDCol(Relation r)
       throws InterpretError
  {
    for(int i = 0; i < r.numattrs; i++) {
      if ((r.domains[i].name).compareTo(".id") == 0)
			return i;
    }
    
    // shouldn't be here
    return -1;
  }
  
  // compare the types of the two relations involved in assignment operation
  // if they are of the same type, return 1, else return 0.
  // Because the right relation is a top level relation, so it doesn't constain domain .id
  // and the left relation is a nested attribute, so it contains domain .id
  // Here we just need to make sure the domains except .id are the same.
  public static int CompareRelType(Relation left, Relation right)
       throws InterpretError
  { 
    if (left.isDomainsIn(right.domains) && (left.domains.length == right.domains.length + 1)) 
      return 1;
    else 
      return 0;
  }
  
  //Yi Zheng June 3, 2002
  //used for comparison between two relations:
  // 1. left and right are both top level
  // 2. left and right are both nested
  // 3. left is nested and right is top level
  public static int CompareRelTypeComp(Relation left, Relation right)
       throws InterpretError
  { 
    if (left.isDomainsIn(right.domains) && right.isDomainsIn(left.domains)) 
      return 1; //for the case when both are top level or both are nested
    else 
      return CompareRelType(left, right);
  }

  //wrapper used in Computation.java
  //Yi Zheng May 20, 2002
  public int lookCompUpdate(SimpleNode node, Environment env, Relation prtmp, TriggerNode triggernode) 
       throws InterpretError
  	{  return lookUpdate(node, env, prtmp, triggernode);
	}
  
  // to produce the three parts relation of trigger for the certain update
  // command in the certain environment -- each main update command in one
  // environment; execute this command but not make the result effective in
  // the system tables; trigger is only temperary for this environment.
  //  -- implemented by Weizhong, Aug. 1999
  private int lookUpdate(SimpleNode node, Environment env, Relation prtmp, TriggerNode triggernode) 
       throws InterpretError
  {  
    int opcode = node.opcode;
    String old_name = nextTempName();
    String rest_name = nextTempName();
    String new_name = nextTempName();
    int diff = 0;	// if oldtrigger != newtrigger, diff = 1;
    // else diff = 0, which means it's a void update
    // then Update command won't be activated.
    int assigntime = 0; 	// keep the times of assignments, for loading rtmp once
    int assignidlist = 0;   // if assign a relation to a nested attribute, set it to 1
    
    triggernode.setName(((SimpleNode) node.jjtGetChild(0)).name);
    triggernode.opcode = opcode;
    switch(opcode) 
	  { 
	  case OP_ADD :
	    
	    //This can be mapped to a '<+' stmt
	    //update R add expr;  ---> R <+ expr;
	    Relation add_rl = null;	// the relation to be updated
	    Relation add_rr = null;	// the relation to be added
	    Relation rrtemp = null; 	// a temp relation
	    SimpleNode add_idl = (SimpleNode) node.jjtGetChild(0);
	    
	    String rrtemp_name = nextTempName();

	    if ((add_rl = env.lookupRel(add_idl.name, true)) == null)
	    {
	      if ((add_rl = env.lookupNestedRel(add_idl.name, true)) == null) 		
			{	throw new InterpretError("Relation '"+add_idl.name+"' does not exist");
			}
		}
	    
	       
	    // add_rl ijoin prtmp (project on add_rl -> trigger.oldp
	    if (prtmp != null) {
	      Domain[] iddom = new Domain[1];
	 //     iddom[0] = env.domtable.getDom(".id");
	 	iddom[0] = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
	      Domain[] namedom = new Domain[1];
	 //     namedom[0] = env.domtable.getDom(add_idl.name);
	      namedom[0] = env.lookupDom(add_idl.name, true); //Yi Zheng June 14, 2002
	      
	//      Domain pid = env.domtable.getDom(".id");
			Domain pid = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
	      Relation temp_prtmp = prtmp.duplicate(false);
	      
	      int pidj = temp_prtmp.indexOfDomain(pid);
	      
	      if (pidj != ERROR) {
		Domain[] pdoms = new Domain[temp_prtmp.numattrs -1];
		int pdomid = 0;
		for (int pp = 0; pp < temp_prtmp.numattrs; pp++) {
		  if (pp != pidj) {
		    pdoms[pdomid] = temp_prtmp.domains[pp];
		    pdomid++;
		  }
		}
		temp_prtmp.load();
		temp_prtmp.project(pdoms);
	      }
	      
	      triggernode.trigger.oldp = add_rl.mujoin(OP_IJOIN, iddom, temp_prtmp, namedom, old_name); 
	      
	      
	      triggernode.trigger.oldp.load();
	      triggernode.trigger.oldp.project(add_rl.domains);
	      triggernode.trigger.oldp.save();
	      triggernode.trigger.relname = "." + add_idl.name;
	    } else {
	      // add_rl -> trigger.oldp
	      //			triggernode.trigger.oldp = add_rl.mujoin(OP_DJOIN, null, add_rl, null, old_name);
	      triggernode.trigger.oldp = add_rl.project(add_rl.domains, old_name);
	      triggernode.trigger.relname = add_idl.name;
	    }
	    
	    
	    // add_rl djoin trigger.oldp -> trigger.restp
	    triggernode.trigger.restp = add_rl.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, rest_name);

	    
	    
	    // trigger.oldp -> rrtemp which to be updated
	    //		  rrtemp = triggernode.trigger.oldp.mujoin(OP_UJOIN, null, triggernode.trigger.oldp, null, rrtemp_name);
	    
	    rrtemp = triggernode.trigger.oldp.project(triggernode.trigger.oldp.domains, rrtemp_name);
	    
	    // update rrtemp
	    String add_tempname = nextTempName();
	    add_rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1), add_tempname, new Hashtable(), env);

	    String tname4 = nextTempName();
	    if (prtmp != null) {
	      // nested!
	      String tname1 = nextTempName();
	      String tname2 = nextTempName();
	      String tname3 = nextTempName();
	      
	      // get the relation to be added with .id domain
	      // to add domain '.id' for add_rr
	      Relation utmp = rrtemp.mujoin(OP_UJOIN, null, add_rr, null, tname1);
	      add_rr = utmp.mujoin(OP_DJOIN, null, add_rl, null, tname2);
	      
	      //Relation addtemp = rrtemp.mujoin(OP_UJOIN, null, rrtemp, null, tname3);
	      Relation addtemp = rrtemp.project(rrtemp.domains, tname3);
	      int idcol = getIDCol(add_rr);
	      if (idcol == -1)
		throw new InterpretError(".id not found in getIDCol");
	      addtemp.load();
	      add_rr.load();
	      rrtemp.load();
	      long old_surr = 0;
	      for (int i = 0; i < rrtemp.numtuples; i++) {
		int rtid = getIDCol(rrtemp);
		if (rtid == -1)
		  throw new InterpretError(".id not found in getIDCol");
		long surr = ((long[])rrtemp.data[rtid])[i];
		if (old_surr != surr) {
		  old_surr = surr;
		  for (int j=0; j<add_rr.numtuples; j++) {
		    ((long[])add_rr.data[idcol])[j] = surr;
		  }
		  add_rr.assignAdd(null, addtemp, null);
		}
	      } 
	      //rrtemp = addtemp.mujoin(OP_UJOIN, null, addtemp, null, tname4);
	      rrtemp = addtemp.project(addtemp.domains, tname4);
	      deleteFile(tname1);
	      deleteFile(tname2);
	      deleteFile(tname3);
	    }
	    else {
	      // no nested
	      // rrtemp = add_rr.mujoin(OP_UJOIN, null, add_rr, null, tname4);
	      // rrtemp.assignAdd(null, add_rr, null);
	      // rrtemp = add_rr.project(add_rr.domains, tname4);
	      
	      rrtemp.load(true);
	      add_rr.load(true);
	      
	      add_rr.assignAdd(null, rrtemp, null);
	      
	      rrtemp.load(true);
	      
	    }
	    
	    // rrtemp -> trigger.newp
	    //		  triggernode.trigger.newp = rrtemp.mujoin(OP_UJOIN, null, rrtemp, null, new_name);
	    triggernode.trigger.newp = rrtemp.project(rrtemp.domains, new_name);

	    // compare the old and new trigger
	    Relation comp = triggernode.trigger.newp.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, tname4);
	    if (comp.numtuples != 0)
	      diff = 1;
	    
	    deleteFile(rrtemp_name);
	    deleteFile(add_tempname);
	    deleteFile(tname4);
	    // apply the trigger to get the current result for next statement
	    doTrigger(env, triggernode, 0);
	    
	    break;
	  case OP_DELETE :
	    // Weizhong, July 2 1999
	    //This cn be mapped to a 'djoin' stmt
	    //update R del expr;  ---> R <- R djoin expr;
	    //NOTE: that is not the same as R <- expr djoin expr; !!WRONG
	    
	    Relation del_rl = null;	// the relation to be updated
	    Relation del_rr = null;	// the relation to be deleted
	    Relation del_temp = null; 	// a temp relation
	    SimpleNode del_idl = (SimpleNode) node.jjtGetChild(0);
	    
	    String deltemp_name = nextTempName();
	    
	    if ((del_rl = env.lookupRel(del_idl.name, true)) == null)
	      if ((del_rl = env.lookupNestedRel(del_idl.name, true)) == null) 		
		{	
		  throw new InterpretError("Relation '"+del_idl.name+"' does not exist");
		}
	    
	    
	    // del_rl ijoin prtmp (project on del_rl) -> trigger.oldp
	    if (prtmp != null) {
	      Domain[] iddom = new Domain[1];
	   //   iddom[0] = env.domtable.getDom(".id");
	   iddom[0] = env.lookupDom(".id", true); //Yi Zheng June 14,2002
	      Domain[] namedom = new Domain[1];
	   //   namedom[0] = env.domtable.getDom(del_idl.name);
	      namedom[0] = env.lookupDom(del_idl.name, true); //Yi Zheng June 14, 2002
	      
	  //    Domain dpid = env.domtable.getDom(".id");
	  	Domain dpid = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
	      Relation dtemp_prtmp = prtmp.duplicate(false);
	      
	      
	      int dpidj = dtemp_prtmp.indexOfDomain(dpid);
	      
	      if (dpidj != ERROR) {
		Domain[] pdoms = new Domain[dtemp_prtmp.numattrs -1];
		int pdomid = 0;
		for (int pp = 0; pp < dtemp_prtmp.numattrs; pp++) {
		  if (pp != dpidj) {
				pdoms[pdomid] = dtemp_prtmp.domains[pp];
				pdomid++;
		  }
		}
		dtemp_prtmp.load();
		dtemp_prtmp.project(pdoms);
	      }
	      
	      triggernode.trigger.oldp = del_rl.mujoin(OP_IJOIN, iddom, dtemp_prtmp, namedom, old_name); 
	      
	      triggernode.trigger.oldp.load();
	      triggernode.trigger.oldp.project(del_rl.domains);
	      triggernode.trigger.oldp.save();
	      triggernode.trigger.relname = "." + del_idl.name;
	    } else {
	      // del_rl -> trigger.oldp
	      triggernode.trigger.oldp = del_rl.project(del_rl.domains, old_name);
	      triggernode.trigger.relname = del_idl.name;
	    }
	    
	    
	    // del_rl djoin trigger.oldp -> trigger.restp
	    triggernode.trigger.restp = del_rl.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, rest_name);
	    
	    
	    // trigger.oldp -> del_temp which to be updated
	    del_temp = triggernode.trigger.oldp.project(triggernode.trigger.oldp.domains, deltemp_name);
	    
	    // get del_rr which to be deleted
	    String del_tempname = nextTempName();
	    del_rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1), del_tempname, new Hashtable(), env);
	    
	    // del_temp djoin del_rr -> trigger.newp
	    String dtname1 = nextTempName();
	    
	    Relation del_rtmp = del_temp.mujoin(OP_DJOIN, null, del_rr, null, dtname1);
	    Domain[] iddoms = new Domain[1];
//	    iddoms[0] = env.domtable.getDom(".id"); //modified by Yi Zheng June 12, 2002
	 	iddoms[0] = env.lookupDom(".id", true);
	    if (iddoms != null) {
	      del_rtmp.load();
	      if (del_rtmp.isDomainIn(iddoms[0])) 
		del_rtmp.swapColumns(iddoms);
	    }
	    
	    triggernode.trigger.newp = del_rtmp.project(del_rtmp.domains, new_name);
	    
	    
	    // compare the old and new trigger
	    String dtname = nextTempName();
	    Relation dcomp = triggernode.trigger.newp.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, dtname);
	    if (dcomp.numtuples != 0 || triggernode.trigger.newp.numtuples != triggernode.trigger.oldp.numtuples) {
	      diff = 1;
	    }
	    
	    deleteFile(deltemp_name);
	    deleteFile(del_tempname);
	    deleteFile(dtname);
	    deleteFile(dtname1);
	    
	    // apply the trigger to get the current result for next statement
	    doTrigger(env, triggernode, 0);
	    
	    break;
	  case OP_CHANGE : 
	    Relation rel = null; //The relation to be updated
	    Relation real_old = null; // old version of subset of rel
	    Relation temp_real_old = null;
	    Relation rtmp = null;	// the relation for changing
	    Relation rrest = null;	// the rest of source relation
	    Relation rtmp_copy = null;	// a temp copy of rtmp before 
	    // discarding the joined domain
	    Relation rtmp_copy2 = null;  // a final copy of rtmp before 
	    // discarding the joined domain      
	    
	    SimpleNode id = (SimpleNode) node.jjtGetChild(0);
	    SimpleNode stmtList = (SimpleNode) node.jjtGetChild(1);
	    triggernode.trigger.relname = new String(""); // in case, no operate on current level
	    
	    String fname1 = nextTempName();
	    String fname2 = nextTempName();
	    String fname3 = nextTempName();
	    String fname4 = nextTempName();
	    String fname5 = nextTempName();
	    String fname6 = nextTempName();	// for rtmp_copy
	    String fname7 = nextTempName();      // for rtmp_copy2
	    String fname8 = nextTempName();
	    String fname9 = nextTempName();      // for real_old
	    String fname10 = nextTempName();
	    
	    SimpleNode usingExpr; //this is optional..
	    Relation reluser;
	    NREnvironment NRenv;
		
	    
	    if( (rel = env.lookupRel(id.name, true)) == null){
	      // extract the nested relation
		      if ((rel = env.lookupNestedRel(id.name, true)) == null) 		
			{	throw new InterpretError("Relation '"+id.name+"' does not exist");
			}
			else if(DEBUG)
			{
			}
	    }else if(DEBUG)
	    {
	    }
	    
	    //check if rel is extracted from a nested relation
	    //if so we can't just replace the current entry of rel in the environment
	    //instead, we need to append the updated result to the dot relation
	    //from which rel is extracted
	    //--Yi Zheng June 10, 2002
	    
	    //Handle using clause first..
	    usingExpr = null; //this is optional..
	    if(node.jjtGetNumChildren() == 3)
	      { 
		usingExpr = (SimpleNode) node.jjtGetChild(2);
		
		SimpleNode rrel;
		SimpleNode ldom;
		SimpleNode rdom;
		Domain[] ldomain;
		Domain[] rdomain;
		
		if (usingExpr.jjtGetNumChildren() == 3) 
		  { 
		    
		    ldom = (SimpleNode) usingExpr.jjtGetChild(0);
		    rdom = (SimpleNode) usingExpr.jjtGetChild(1);
		    rrel = (SimpleNode) usingExpr.jjtGetChild(2);
		    ldomain = ExpressionListToDomains(ldom);
		    rdomain = ExpressionListToDomains(rdom);
		    
		  } else {
		    ldomain = null;
		    rdomain = null;
		    rrel = (SimpleNode) usingExpr.jjtGetChild(0);
		  }
		
		reluser = null;
		reluser = evaluateTLExpression(rrel, fname5, new
					       Hashtable(), env);
		rtmp = rel.mujoin(usingExpr.opcode, ldomain, reluser,rdomain, fname1);
		rtmp.load();


		// make a copy to keep the rtmp with the joined domains
		// for virtual domain uses, e.g. Type <- NewType
		rtmp.swapColumns(rel.domains);
		rtmp.save();
		rtmp.load();



		Domain[] riddoms = new Domain[1];
		riddoms[0] = env.lookupDom(".id", true); //modified by Yi Zheng June 10, 2002
		if (riddoms != null) {
		  rtmp.load();
		  if (rtmp.isDomainIn(riddoms[0])) 
		    rtmp.swapColumns(riddoms);
		}
		rtmp.save();
		rtmp.load();
		
		rtmp_copy = rtmp.project(rtmp.domains, fname6);
		rtmp = rtmp_copy.project(rel.domains, fname8);
		
		// rtmp may contain new tuples beyond rel (after ujoin)
		// so we need to keep a real_old version of data
		real_old = rtmp.mujoin(OP_IJOIN, null, rel, null,
				       fname9);
		real_old.load();

	      } else {
		// no using option, copy the source relation to rtmp
		rtmp = rel.project(rel.domains, fname1);
		real_old = rel.project(rel.domains, fname9);
	      }
	    
	    // rtmp ijoin prtmp (project on rtmp) -> triggernode.trigger.oldp
	    if (prtmp != null) {
	      Domain[] iddom = new Domain[1];
	      Domain[] namedom = new Domain[1];
	  //    iddom[0] = env.domtable.getDom(".id");
	  		iddom[0] = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
	  //    namedom[0] = env.domtable.getDom(id.name); 
	  		namedom[0] = env.lookupDom(id.name, true); //Yi Zheng June 14, 2002
	      
	  //    Domain pcid = env.domtable.getDom(".id");
	  	Domain pcid = env.lookupDom(".id", true);//Yi Zheng June 14, 2002
	      Relation tempc_prtmp = prtmp.duplicate(false);
	      
	      
	      int pcidj = tempc_prtmp.indexOfDomain(pcid);
	      
	      if (pcidj != ERROR) {
		Domain[] pdoms = new Domain[tempc_prtmp.numattrs -1];
		int pdomid = 0;
		for (int pp = 0; pp < tempc_prtmp.numattrs; pp++) {
		  if (pp != pcidj) {
		    pdoms[pdomid] = tempc_prtmp.domains[pp];
		    pdomid++;
		  }
		}
		tempc_prtmp.load();
		tempc_prtmp.project(pdoms);
	      }
	      
	      triggernode.trigger.oldp = rtmp.mujoin(OP_IJOIN, iddom, tempc_prtmp, namedom, old_name); 

	      
	      triggernode.trigger.oldp.load();
	      
	      
	      triggernode.trigger.oldp.project(rtmp.domains);
	      
	      if (real_old != null){
		temp_real_old =
		  real_old.mujoin(OP_IJOIN,iddom,tempc_prtmp,namedom,fname10);
		temp_real_old.load();
		real_old =
		  temp_real_old.project(temp_real_old.domains,fname9);
		real_old.load();
	      }	
	    } else {
	      // rtmp -> trigger.oldp
	      triggernode.trigger.oldp = rtmp.project(rtmp.domains, old_name);
	    }
	    
	    // triggernode.trigger.oldp -> rtmp
	    rtmp = triggernode.trigger.oldp.project(triggernode.trigger.oldp.domains, fname1);
	    //		   rtmp.save();
	    rtmp.load();
	    
	    // if (rtmp_copy != null) then rtmp_copy ijoin rtmp -> rtmp_copy2 which includes virtual domains
	    if (rtmp_copy !=null) {
	      rtmp_copy.load();
	      rtmp_copy2 = rtmp_copy.mujoin(OP_IJOIN, null, rtmp, null, fname7);
	      
	    }
	    // rel djoin triggernode.trigger.oldp/rtmp -> triggernode.trigger.restp
	    triggernode.trigger.restp = rel.mujoin(OP_DJOIN, null, rtmp, null, rest_name);
	    
	    //Run each statement in 'stmtList'
	    int numStmts = stmtList.jjtGetNumChildren();
	    
	    for(int i = 0; i < numStmts; i++)
	      {	// for statements
		
		SimpleNode stmt = (SimpleNode) stmtList.jjtGetChild(i);
		
		switch(stmt.type)
		  {  // switch stmt type 
		  case OP_ASSIGNMENT :
		    // update rtmp
		    // set the current trigger under the current node
		    if (prtmp != null)
		      triggernode.trigger.relname = "." + id.name;
		    else
		      triggernode.trigger.relname = id.name;
		    
		    if (assigntime == 0) {
		      // switch rtmp and rtmp_copy2 if necessary
		      if (rtmp_copy2 != null) {
			Relation temp = null;
			String tempfname = nextTempName();
			// rtmp_copy2 -> temp
			temp = rtmp_copy2.project(rtmp_copy2.domains, tempfname);
			// rtmp -> rtmp_copy2
			rtmp_copy2 = rtmp.project(rtmp.domains, rtmp_copy2.name);
			// temp -> rtmp
			rtmp = temp.project(temp.domains, rtmp.name);
			
			deleteFile(tempfname);
		      }
		      rtmp.load();
		    }
		    
		    
		    assigntime++;
		    //Initialize the NREnvironment 
		    NRenv = env.addNRFrame(rtmp);
		    int nrow = rtmp.numtuples; // avoiding changing rtmp below
		    for(NRenv.row = 0; NRenv.row < nrow; NRenv.row++)
		      {	// for each row
			
			switch(stmt.opcode)
			  {  // switch stmt opcode
			  case OP_ASSIGN :
			    //REVISIT: I have only handled the case where numchildren == 2
			    // So: R <- S;
			    // But not: R [A: <- :B] S;
			    // Perhaps this last form should not be allowed in an update
			    if(stmt.jjtGetNumChildren() == 2)
			      {  // if 2
				SimpleNode left = (SimpleNode) stmt.jjtGetChild(0);
				SimpleNode expr = (SimpleNode) stmt.jjtGetChild(1);
				NRInfo nri = env.lookupNR(left.name, false);
				if (nri == null)
				  {  throw new InterpretError("'"+left.name+
					 "' is not a domain of "+rtmp.name);
				  }
				int col = nri.index;
				NREnvironment nre = nri.NRenv;
				
				switch(nri.type)
				  {  // switch 3
				  case BOOLEAN :
				    //REVISIT: implement this case..
				    if(EvalExpr.evalBooleanExpr(expr, env))
				      ((byte[]) nre.data[col])[nre.row] = 1;
				    else
				      ((byte[]) nre.data[col])[nre.row] = 0;
				    break;
				  case SHORT :
				    if (EvalExpr.evalShortExpr(expr,env) != SHORT_DC) {
				      ((short[]) nre.data[col])[nre.row] = 
					EvalExpr.evalShortExpr(expr, env);
				    }
				    break;
				  case INTEGER :  
				  //Yi Zheng April 18, 2002
				  EvalExpr.WEIZHONG = true;
				  int iv = EvalExpr.evalIntExpr(expr, env);
				  EvalExpr.WEIZHONG = false;
			 if (iv != INT_DC) {
				      ((int[]) nre.data[col])[nre.row] = 
				iv;
			//	    if (EvalExpr.evalIntExpr(expr, env) != INT_DC) {
			//	      ((int[]) nre.data[col])[nre.row] = 
			//		EvalExpr.evalIntExpr(expr, env);
		//		System.out.println("new value of " + left.name + " is: " + ((int[]) nre.data[col])[nre.row]);
				    }
				    break;
				  case LONG :
				    if (EvalExpr.evalLongExpr(expr, env) != LONG_DC) {
				      ((long[]) nre.data[col])[nre.row] = 
					EvalExpr.evalLongExpr(expr, env);
				    }
				    break;
				  case FLOAT :
				    if (EvalExpr.evalFloatExpr(expr, env) != FLOAT_DC) {
				      ((float[]) nre.data[col])[nre.row] = 
					EvalExpr.evalFloatExpr(expr, env);
				    } 
				    break;
				  case DOUBLE :
				    if (EvalExpr.evalDoubleExpr(expr, env) != DOUBLE_DC) {
				      ((double[]) nre.data[col])[nre.row] = 
					EvalExpr.evalDoubleExpr(expr, env);
				    }
				    break;
				    case NUMERIC ://from Yiyi
				      if(EvalExpr.evalNumExpr(expr,env).compareTo(NUMERIC_DC) != 0){
					  ((number[])nre.data[col])[nre.row] =
					      EvalExpr.evalNumExpr(expr,env);
				      }
				      break;

		 case UNIVERSAL ://from Andrey
			case ATTRIBUTE: //from Andrey
				  case STRING :
				    if (EvalExpr.evalStringExpr(expr, env) != STRING_DC) {
				      ((String[]) nre.data[col])[nre.row] = 
					EvalExpr.evalStringExpr(expr, env);
				    }
				    break;
				  case IDLIST :
				    // Remember these are nested relations.
				    // assign the new relation to the sub-relation got in 1.
				    
				    Relation nestrtmp = null;
				    String ttmpname1 = nextTempName();
				    String ttmpname2 = nextTempName();
				    String ttmpname3 = nextTempName();
				    String ttmpname4 = nextTempName();
				    long surrogate;
				    Domain[] iiddom = new Domain[1];
                     //              iiddom[0] = env.domtable.getDom(".id");
				   iiddom[0] = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
				   assignidlist = 1;
					
				   
				   // extract the nested relation
				   if ((rel = env.lookupNestedRel(left.name, true)) == null) 		
	//		{	throw new InterpretError("Nested Relation '"+id.name+"' does not exist");}
			{	throw new InterpretError("Nested Relation '"+left.name+"' does not exist");} //Yi Zheng June 2, 2002

				   if ((surrogate = env.getSurrogate(left.name)) == -1) {
				throw new InterpretError("Surrogate not found.");
				   }
				   if( (nestrtmp = env.lookupRelSurr(left.name, true)) == null)
				     {  throw new InterpretError(
						  "Relation '"+left.name+"' does not exist");//changed from id.name by Yi Zheng June 2, 2002
				     }
				   
				   rrest = rel.mujoin(OP_DJOIN, null, nestrtmp, null, ttmpname4);
				   
				   Relation rtmp_tmp1 = evaluateTLExpression(expr, ttmpname1, new Hashtable(), env); 
				   
				   // check whether nestrtmp and rtmp_tmp1 is the same type relation (attributes are of same type)	
				   // if not, throw an error
				   if (CompareRelType(nestrtmp, rtmp_tmp1) != 1) 
				     throw new InterpretError("The assignment is not allowed since the two operand relations are not of the same type!");
				   
                                   // project nestrtmp on .id
				   nestrtmp.project(iiddom);
				   
				   // (add domain .id to rtmp_tmp1) -> rtmp_tmp2	
				   Relation rtmp_tmp2  = nestrtmp.mujoin(OP_IJOIN, null, rtmp_tmp1, null, ttmpname2);
				   
				   if (nestrtmp.name == "")
				     nestrtmp.name = "Temp";
				   String nestrtmp_name = nestrtmp.name;
		   		   env.removeRel(nestrtmp_name, false);
		   		   // rename back to nestrtmp relation
		   		   env.put(rtmp_tmp2.assign(null, nestrtmp_name, null));
				   nestrtmp = env.lookupRel(nestrtmp_name, true);
				   nestrtmp.load();
				   if (triggernode.trigger.newp.numtuples == 0)
				     triggernode.trigger.newp = nestrtmp.project(nestrtmp.domains, new_name);
                                   else {
				     triggernode.trigger.newp = triggernode.trigger.newp.mujoin(OP_UJOIN, null, nestrtmp, null, new_name); 
				     //				  	 triggernode.trigger.newp.project(triggernode.trigger.newp.domains);	
				   };	
				   deleteFile(ttmpname1);
				   deleteFile(ttmpname2);
				   deleteFile(ttmpname3);
				   deleteFile(ttmpname4);
				   break;
				  case COMP :
				    // impossible to be here, Weizhong
				    throw new
				      InterpretError("Type 'comp' are not yet supported here!");
				  case TEXT :
				  case STMT :
				  case EXPR :
				    throw new
				      InterpretError("Types 'text', 'stmt', and 'expr' " +
						     "are not yet supported in computations");
				  default :
				    throw new InterpretError("Unknown type");
				  }
				// end of switch 3
			      } // end of if 2
			    break;
			  default :
			    throw new InterpretError ("Statement type not supported in an update");
			  }
			// end of switch stmt opcode
		      } // end of each row			
		    
			rtmp.save();
		    rtmp.save(); //Yi Zheng June 10, 2002
		    rtmp.load();
		    // reset NRenv.row for late use
		    env.removeNRFrame();
		    if (assignidlist == 0) {
		      
		      // adjust rtmp domains if necessary
		      if (rtmp_copy2 != null) {
			rtmp = rtmp.project(rtmp_copy2.domains, rtmp.name);
		      } else {
			rtmp.save();
					}
		      
		      // adjust new and old back, to keep the original old piece
		      if (assigntime >1)
			doTrigger(env, triggernode, 0);
		      
		      // rtmp -> newp
		      rtmp.load();
		      triggernode.trigger.newp = rtmp.project(rtmp.domains, new_name);

		    } else {
		      SimpleNode nestid = (SimpleNode) stmt.jjtGetChild(0);
		      Domain[] idom = new Domain[1];
		 //     idom[0] = env.domtable.getDom(".id");
		 		idom[0] = env.lookupDom(".id", true); //Yi Zheng June 14, 2002
		      Domain[] nndom = new Domain[1];
		 //     nndom[0] = env.domtable.getDom(nestid.name);
		      nndom[0] = env.lookupDom(nestid.name, true); //Yi Zheng June 14, 2002
		      triggernode.trigger.oldp = triggernode.trigger.oldp.mujoin(OP_IJOIN, nndom , rel, idom, old_name);
		      triggernode.trigger.oldp = triggernode.trigger.oldp.project(rel.domains, old_name);
		      triggernode.trigger.restp = rel.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, rest_name); 
		      assignidlist = 0;
		      triggernode.trigger.relname = "." + nestid.name; 
		      
		    };
		    
		    // compare the old and new trigger
		    String ctname4 = nextTempName();
		    Relation compc = triggernode.trigger.newp.mujoin(OP_DJOIN, null, triggernode.trigger.oldp, null, ctname4);
		    if (compc.numtuples != 0) {
		      diff = 1;
		    }
		    
		    deleteFile(ctname4);
		    		    // reset NRenv.row for late use
		//    env.removeNRFrame();

		    // apply the trigger to get the current result for next statement
		    doTrigger(env, triggernode, 0);
		    
		    break;
		  case OP_UPDATE:
		    // nested
		    
		    TriggerNode childnode = new TriggerNode(env, triggernode);
		    triggernode.jjtAddChild(childnode);
		    
		    int stmt_diff = lookUpdate(stmt, env, rtmp, childnode);
		    if (stmt_diff == 1 || diff == 1)
		      diff = 1;
		    else
		      diff = 0;
		    break;
		    
		    /**********************COMPUTATION*************************/
		  case OP_STATEMENT :
		    switch(stmt.opcode)
		      {  // switch 4
		      case OP_COMPCALL :
			triggernode.trigger.relname = id.name;
			
			String compName = ((SimpleNode) stmt.jjtGetChild(0)).name;
			
			Domain[] domss = new Domain[1];
			Domain  d = env.lookupDom(compName, true);

			if(d != null)
		 		domss[0] = d;
			else throw new InterpretError("'" + compName + "' is not a valid computation");
			
			String reltempname = nextTempName();
			Relation temprel = rel.project(domss, reltempname);
			temprel.load();
			long domsur = ((long[]) temprel.data[0])[0];
			
			Computation  r = comptable.lookup(domsur);
			if (r == null)
			  {  throw new InterpretError("'"+compName+"' is not a valid nested computation for the relation to be changed!");
			  }
			
			
			// To get the relation without domains of type COMP
			int j = 0;
	
			for(int ip = 0; ip < rel.domains.length; ip++)
			  {
			    if (rel.domains[ip].type !=  COMP){
			      j = j + 1;
			    }
			  }
	
			Domain[] realdomains = new Domain[j];
			Domain[] compdomains = new Domain[rel.domains.length - j];
			
			int q = 0;
			int z = 0;

			for (int ipp = 0; ipp < rel.domains.length; ipp++)
			  {
			    if (rel.domains[ipp].type != COMP){
			      realdomains[q] = rel.domains[ipp];
			      q = q + 1;
			    } else {
			      compdomains[z] = rel.domains[ipp];
			      z = z + 1;
			    }
			  }
	
			String reltempname1 = nextTempName();
			Relation temprel1 = rtmp.project(realdomains, reltempname1);
			temprel1.load();
			
			// Begin to apply this computation statement
			NREnvironment compNRenv;
			compNRenv = env.addNRFrame(temprel1);
			
			((Computation) r).applyInOut(temprel1, stmt, env);
			rtmp.load();
			
			String stemp1 = nextTempName();
			
			Relation rtmp_s1 = rtmp.project(compdomains, stemp1);
			rtmp_s1.load();
			
			//rtmp = temprel1.mujoin(OP_UJOIN, null, rtmp_s1, null, rtmp.name); Yi Zheng May 1, 2002
			rtmp = temprel1.mujoin(OP_IJOIN, null, rtmp_s1, null, rtmp.name);
			
			rtmp.load();
			
			triggernode.trigger.newp = rtmp.project(rtmp.domains, new_name);
			
			deleteFile(reltempname);
			deleteFile(reltempname1);
			deleteFile(stemp1);
			
			break;
		      default :
			throw new InterpretError
			  ("Statement type not supported in an update");
		      }
		    // end of switch 4
		    break;
		    /***************************************************************/
		    
		  default :
		    throw new InterpretError("Statement type not supported in an update");
		  }
		// end of switch stmt type
	      } // end of for statements

	    if (assigntime != 0) {
	      // reverse the original old piece
	      doTrigger(env, triggernode, 0);
	      
	      rtmp.save();
	      rtmp.load();
	      
	      // adjust oldp, newp, and rest 
	      // in case A<- if B then A else C;
	      
	      // restp = (newp ijoin oldp) ujoin restp
	      triggernode.trigger.oldp.load();
	      String riname = nextTempName();
	      Relation ri = null;
	      ri = triggernode.trigger.newp.mujoin(OP_IJOIN, null, triggernode.trigger.oldp, null, riname);
	      // ri = rtmp.mujoin(OP_IJOIN, null, triggernode.trigger.oldp, null, riname);
	      
	      triggernode.trigger.restp = triggernode.trigger.restp.mujoin(OP_UJOIN, null, ri, null, rest_name);
	      
	      deleteFile(riname);
	      
	      // oldp djoin restp = oldp
	      String tempoldname = nextTempName();
	      String tempoldname1 = nextTempName();
	      Relation temp_old = null;
	      Relation temp_old1 = null;
	      
	      temp_old = triggernode.trigger.oldp.mujoin(OP_DJOIN, null, triggernode.trigger.restp, null, tempoldname);
	      //temp_old.printdata();
	      //rel.printdata();
	      temp_old1 = temp_old.mujoin(OP_IJOIN, null,
					  real_old, null, tempoldname1);
	      triggernode.trigger.oldp = null;
	      triggernode.trigger.oldp = temp_old1.project(temp_old.domains, old_name);
	      
	      deleteFile(tempoldname);
	      deleteFile(tempoldname1);
	      
	      // newp djoin restp = newp
	      String tempnewname = nextTempName();
	      
	      Relation temp_new = null;
	      temp_new = triggernode.trigger.newp.mujoin(OP_DJOIN, null, triggernode.trigger.restp, null, tempnewname);
	      
	      triggernode.trigger.newp = temp_new.project(temp_new.domains, new_name);
	      deleteFile(tempnewname);
	      
	      // apply the change for this level's assignments
	      doTrigger(env, triggernode, 0);
	    }
	    
	    
	    // clear the temp relation
	    deleteFile(fname1);
	    deleteFile(fname2);
	    deleteFile(fname3);
	    deleteFile(fname4);
	    deleteFile(fname5);
	    deleteFile(fname6);
	    deleteFile(fname7);
	    deleteFile(fname8);
	    deleteFile(fname9);
	    deleteFile(fname10);
	    break;
	    // end of OP_CHANGE
	  default : // impossible to here
	    System.out.println("unknown update statement");
	    break;
	  }
    // end of switch
    
    
    // clear up empty trigger file
    if (triggernode.trigger.relname.compareTo("") == 0) {
      deleteFile(old_name);
		deleteFile(new_name);
		deleteFile(rest_name);
    }
    
    return diff;
    
  }   
  
  //wrapper used in Computation.java
  //Yi Zheng May 20, 2002
  public int executeCompUpdate(SimpleNode root, SimpleNode node, Environment env)
       throws InterpretError
  { return executeUpdate(root, node, env);
  }
  // implemented by Weizhong Sun, Aug 1999
  private int executeUpdate(SimpleNode root, SimpleNode node, Environment env)
       throws InterpretError
  {  int opcode = node.opcode;
  int assignDone = 0;	// to control the assignments in the same level
  // make sure only do trigger once for all the
  // assignments of one level 
  
  int preevent = executeEventH("pre", node, env);
  
  switch(opcode) 
    {  case OP_ADD :
    case OP_DELETE :
      // Weizhong, July 2 1999
      SimpleNode nid = (SimpleNode) node.jjtGetChild(0);
    String rname = nid.name;
    TriggerNode triggernode = env.triggerroot.getNode(rname, opcode);
    doTrigger(env, triggernode, 0);
    break;
    case OP_CHANGE : 
      SimpleNode id = (SimpleNode) node.jjtGetChild(0);
      SimpleNode stmtList = (SimpleNode) node.jjtGetChild(1);
      
      int numStmts = stmtList.jjtGetNumChildren();
      for(int i = 0; i < numStmts; i++)
	{  // for each statement 
	  SimpleNode stmt = (SimpleNode) stmtList.jjtGetChild(i);
	  switch(stmt.type)
	    {  // switch stmt type
	    case OP_ASSIGNMENT :
	      if (assignDone == 0) {
		TriggerNode trignode = env.triggerroot.getNode(id.name, opcode);
		doTrigger(env, trignode, 0);
		assignDone = 1;	// don't do trigger again on other assignments
		// of the same level update
	      }
	      break;
	    case OP_UPDATE:
	      // nested
	      int exeEventH = executeUpdate(root, stmt, env);
	      
	      break;
	    default :
	      throw new InterpretError("Statement type not supported in an update");
	    }
	  // end of switch stmt type
	} // end of each statements
      break;
      // end of OP_CHANGE
    default : // impossible to here
      System.out.println("unknown update statement");
      break;
    }
  // end of switch
  
  
  int postevent = executeEventH("post", node, env);
  
  if (preevent == 1 || postevent == 1)
    return 1;
  return 0;
  }   
  
	//wrapper used in Computation.java
	//Yi Zheng May 20, 2002
  public void doCompTrigger(Environment env, TriggerNode triggernode, int searchchild)
       throws InterpretError
	{ doTrigger(env, triggernode, searchchild);
	}	

  // use the env.trigger to produce the new updated relation, and switch
  // old/new relation in env.trigger -- for update/undo command
  // implemented by Weizhong, Aug. 1999
  private void doTrigger(Environment env, TriggerNode triggernode, int searchchild)
       throws InterpretError
  {
    
    if (triggernode.trigger.relname.compareTo("") != 0) {	
		
      
      String ctmpname = nextTempName();
	String utmpname1 = nextTempName();
	String utmpname2 = nextTempName();
	String utmpname3 = nextTempName();
	String tmpname1 = nextTempName();
	String tmpname2 = nextTempName();
	String tmpname3 = nextTempName();
	
	
	String old_name = triggernode.trigger.oldp.name;
	String new_name = triggernode.trigger.newp.name;
	String rest_name = triggernode.trigger.restp.name;
	
	String mytempnewname = nextTempName();
	String mytempoldname = nextTempName();
	String mytemprestname = nextTempName();
	
	// rest U new
	// if the relation contains computation as attribute, we project oldp, newp, restp to realdomains 
	
	int j = 0;
	for (int ip = 0; ip < triggernode.trigger.newp.domains.length; ip++)
	  {
	    if (triggernode.trigger.newp.domains[ip].type != COMP){
	      j = j + 1;
	    }
	  }
	
	Domain[] realdomains = new Domain[j];
	Domain[] compdomains = new Domain[triggernode.trigger.newp.domains.length - j];
	int q = 0;
	int z = 0;
	for (int ipp = 0; ipp < triggernode.trigger.newp.domains.length; ipp++)
	  {
		if (triggernode.trigger.newp.domains[ipp].type != COMP) {
		  realdomains[q] = triggernode.trigger.newp.domains[ipp];
		  q = q + 1;
		} else {
		  compdomains[z] = triggernode.trigger.newp.domains[ipp];
		  z = z + 1;
		}
	  }

if (z != 0) {
	  Relation mytempnew = triggernode.trigger.newp.project(realdomains, mytempnewname);
	  Relation mytempold = triggernode.trigger.oldp.project(realdomains, mytempoldname);
	  Relation mytemprest = triggernode.trigger.restp.project(realdomains, mytemprestname);
	  
	  //mytempnew.printdata();
	  //mytempold.printdata();
	  //mytemprest.printdata();
	  
	  triggernode.trigger.oldp.load();
	  triggernode.trigger.oldp.swapColumns(compdomains);
	
  Relation onlycomp = triggernode.trigger.oldp.project(compdomains, ctmpname);
	  onlycomp.load();
	  
	  
	  Relation utmp1 = mytempnew.mujoin(OP_UJOIN, null, mytemprest, null, utmpname1);
	  
	  //utmp1.printdata();
	  
	  utmp1.load();
//Yi Zheng May 1, 2002 Relation utmp2 = utmp1.mujoin(OP_UJOIN, null, onlycomp, null, utmpname2);
	  Relation utmp2 = utmp1.mujoin(OP_IJOIN, null, onlycomp, null, utmpname2);
	

	 //utmp2.printdata();
	  
	  utmp2.load();
	  utmp2.save();
	  
	  //restore the value of comp domains into utmp2
	  for (int it = 0; it < utmp2.numtuples; it++)
	    {  for (int jt = 0; jt < onlycomp.numattrs; jt++)
	      { 
		long tll = ((long[])onlycomp.data[jt])[0];
		int coll = q + jt;
		((long[])utmp2.data[coll])[it] = tll;
	      }
	    }  
	 
	 //May 1, 2002
	 utmp2.save();
	 utmp2.load();
		 


	  Relation utmp3 = utmp2.project(utmp2.domains, utmpname3);
	  
	  //utmp3.printdata();
	  
	  
	  
	  // update the relation in system.
	  	String rname = triggernode.trigger.relname;
	  	
	  	//addition by Yi Zheng June 11, 2002
	  IDInfo idi = env.lookup(rname, true);
	  if(idi == null)
	  	throw new InterpretError("relation " + rname + " not found"); //Yi Zheng June 14, 2002

	  if(!(idi instanceof NRInfo))
	  {//original code
	  	//env.removeRel(rname, false);
	  	//env.put(utmp3.assign(null, rname, null));
	  	
	  		if(idi instanceof RelInfo)
	  		{
	  			if(rname.startsWith("."))
	  			{
	  				Environment envFound = env.removeRel2(rname, true);
	 				if(envFound != null)
						env.put2(utmp3.assign(null, rname, null), envFound);
					else throw new InterpretError("no environment found for " + rname);
	  			}
	  			else
	  			{
	  				env.removeRel(rname, false);
	  				env.put(utmp3.assign(null, rname, null));
	  			}
	  		}else
	  		{
	  		//	env.removeRel(rname, false);
	  		//	env.put(utmp3.assign(null, rname, null));

	  			
	  		}

	  }else if(((NRInfo)idi).type == IDLIST)
	  {
	  	String relName = "." + rname;
	  	Relation dotRel = env.lookupRel(relName, true);
	  	if(dotRel != null)
	  	{
	  		long oldSur = ((NRInfo)idi).NRenv.getSurrogate(rname);
	  		if(oldSur != -1)
	  		{
	  				dotRel = dotRel.getRelExceptSurr(oldSur, dotRel.name);//CONTINUE HERE

	  			dotRel.insert(oldSur, utmp3);
	  			if(dotRel.data == null) 
	  				dotRel.load();
	  			dotRel.project(null);
	  			dotRel.save();

				//System.out.println("removing rel: "  + relName);
	  			env.removeRel(relName, true);
	  			env.put(relName, new RelInfo(dotRel), true);
	  			dotRel.save();
	  		}
	  	}
	  	else throw new InterpretError("relation " + relName + " not found");
	  	
	  }//end of addition June 11, 2002
	  	
	  	
	  	
	  // old -> new
	  
//May 1, 2002	  triggernode.trigger.newp = mytempold.mujoin(OP_UJOIN, null, onlycomp, null, new_name);
	  triggernode.trigger.newp = mytempold.mujoin(OP_IJOIN, null, onlycomp, null, new_name);
	  
	  // tempnew -> old
	  
//May 1, 2002	  triggernode.trigger.oldp = mytempnew.mujoin(OP_UJOIN, null, onlycomp, null, old_name);
	  triggernode.trigger.oldp = mytempnew.mujoin(OP_IJOIN, null, onlycomp, null, old_name);
	  
	} else { 
	  // no compdomain. use the usual way to doTrigger.

	  Relation utmp1 = triggernode.trigger.newp.mujoin(OP_UJOIN, null, triggernode.trigger.restp, null, utmpname1);

	  if(utmp1.isLoaded()) utmp1.save();
	  utmp1.load(true);
	  
	  String rname = triggernode.trigger.relname;
	  
	  //addition by Yi Zheng June 10, 2002
	  IDInfo idi = env.lookup(rname, true);
	  if(idi == null)
	  	throw new InterpretError("relation " + rname + " not found"); //Yi Zheng June 14, 2002
	  if(!(idi instanceof NRInfo))
	  {			  //original code
	  //		env.removeRel(rname, false);
	  //		env.put(utmp1.assign(null, rname, null));
	  		//end of original code

	  		if(idi instanceof RelInfo) 
	  		{
	  			
	  			if(rname.startsWith("."))
	  			{
	  				Environment envFound = env.removeRel2(rname, true);
	 				if(envFound != null)
						env.put2(utmp1.assign(null, rname, null), envFound);
					else throw new InterpretError("no environment found for " + rname);
	  			}
	  			else
	  			{
	  				env.removeRel(rname, false);
	  				env.put(utmp1.assign(null, rname, null));
	  			}
	  			
	  		}
	  		else{
	  		}
	  		
	  	
	  }else if(((NRInfo)idi).type == IDLIST)
	  {
	  	
	  	String relName = "." + rname;
	  	Relation dotRel = env.lookupRel(relName, true);
	  	if(dotRel != null)
	  	{
	  		long oldSur = ((NRInfo)idi).NRenv.getSurrogate(rname);
	  		
	  		//System.out.println("oldSur = " + oldSur);
	  		
	  		if(oldSur != -1)
	  		{
 			dotRel = dotRel.getRelExceptSurr(oldSur, dotRel.name);//CONTINUE HERE

	  			dotRel.insert(oldSur, utmp1);
	  			if(dotRel.data == null) 
	  				dotRel.load();
	  			dotRel.project(null);
	  			
	  			dotRel.save();

				//System.out.println("removing rel: " + relName);
	  			env.removeRel(relName, true);
	  			env.put(relName, new RelInfo(dotRel), true);
	  			dotRel.save();
	  		}
	  	}
	  	else throw new InterpretError("relation " + relName + " not found");
	  	
	  }//end of addition June 10, 2002
	  //utmp1.printdata();	
	  
	  // new -> tempnew
	  Relation tempnew = triggernode.trigger.newp.project(triggernode.trigger.newp.domains, tmpname1);
	  
	  // old -> new
	  triggernode.trigger.newp = triggernode.trigger.oldp.project(triggernode.trigger.oldp.domains,new_name);
	  
	  // tempnew -> old
	  triggernode.trigger.oldp = tempnew.project(tempnew.domains, old_name);
	  
	  
	}
	
	deleteFile(ctmpname);
	deleteFile(utmpname1);
	deleteFile(utmpname2);
	deleteFile(utmpname3);
	deleteFile(tmpname1);
	deleteFile(tmpname2);
	deleteFile(tmpname3);
	deleteFile(mytempnewname);
	deleteFile(mytempoldname);
	deleteFile(mytemprestname);
    }
    
    if (searchchild == 1) {
      // search children's trigger and do it.
      int childnum = triggernode.jjtGetNumChildren();
      for(int i = 0; i < childnum; i++) {
	TriggerNode childnode = (TriggerNode) triggernode.jjtGetChild(i);
	doTrigger(env, childnode, 1);
      }
    }
   }
  
  
  // YuLing added the following method
  //convert an ExpressionList node to an arrray of domains for Gedit
  private Domain[] ExpressionListToGeditDomains(SimpleNode node,Domain[] doms,int projectno) 
       throws InterpretError
  {
    int n=node.jjtGetNumChildren();
    if(projectno==1)
      { 
	if(n!=5) throw new InterpretError("Expecting five attribute in the first projection list of gedit");
	else
	  for(int i=0;i<n;i++)
	    {
	      SimpleNode nd=(SimpleNode)node.jjtGetChild(i);
	      String attrname=nd.name;
	      Domain d=env.lookupDom(attrname,true);
	      if(d==null){throw new InterpretError("domain "+attrname+"not found");}
	      doms[i]=d;
	    }
      }
    else if (projectno==2)
      {
	int Color=5,Symbol=6,Layer=7;
	for(int i=0;i<n;i++){
	  SimpleNode nd=(SimpleNode)node.jjtGetChild(i);
	  if(nd.type!=OP_BIOPERATOR)
	    {throw new
	  InterpretError("Expecting bioperator node instead of "+nd.type);
	    }

	  if (nd.jjtGetNumChildren()!=2)
	    {throw new
	       InterpretError("Expecting two children of comparison node instead of "+nd.jjtGetNumChildren());
	    }
	  
	SimpleNode ndl=(SimpleNode)nd.jjtGetChild(0);
	SimpleNode ndr=(SimpleNode)nd.jjtGetChild(1);
	if(ndl.name.compareTo("Color")!=0&&ndl.name.compareTo("Symbol")!=0&&ndl.name.compareTo("Layer")!=0)
	  {throw new
	     InterpretError("Expecting the following: Color , Symbol, Layer");
	}
	String s=ndr.name;
	Domain d=env.lookupDom(s,true);
	if(d==null){throw new InterpretError("domain "+s+"not found");}
	if(ndl.name.compareTo("Color")==0)
	  doms[Color]=d;
	else if(ndl.name.compareTo("Symbol")==0)
	  doms[Symbol]=d;
	else if(ndl.name.compareTo("Layer")==0)
	  doms[Layer]=d;
	}
      }
    else
      throw new
	InterpretError("Expection 1 or 2 as the projection no. in ExpressionListToGedit()");
    //System.out.println("doms[0].name before return"+doms[0].name);
    return doms;
 }
  //end of YuLing's addition

  // convert an ExpressionList node to an array of domains
  // REVISIT: generalization, may need another argument of type Relation 
  private Domain[] ExpressionListToDomains(SimpleNode node)
       throws InterpretError
  {
    int n = node.jjtGetNumChildren();
    Domain[] doms = new Domain[n];
    //System.out.println("children in the Expression List");
    
    for (int i = 0; i < n; i++) {
      SimpleNode nd = (SimpleNode) node.jjtGetChild(i);
      //printNode(nd);
      if (nd.type != OP_IDENTIFIER) {
	throw new 
	  InterpretError("only IDList implemented for ExpressionList");
      }
      String s = nd.name;
      Domain d = env.lookupDom(s, true);
      if (d == null)
	throw new InterpretError("domain : " + s + " not exist");
      else if (d.isStateDom == 1) //Yi Zheng May 6, 2002
      throw new InterpretError("domain: " + s + " is a stae domain");
      else
	doms[i] = d;
    }
    
    return doms;
  }  
  
   //Andrey Rozenberg    
   public Domain[] AttribsRel2DomList(Relation rel)
   throws InterpretError
   {
	   Domain[] domList = null;
		if(rel.domains[0].type==ATTRIBUTE) //modified by Yi Zheng June 24, 2002 to test for unary condition
//		if((rel.domains.length == 1) && (rel.domains[0].type == ATTRIBUTE))
		{
			 rel = rel.duplicateLoadAll();
			 domList = new Domain[rel.numtuples];
			 for(int i=0;i<domList.length;i++)
			 	 domList[i]= env.lookupDom(((String[]) rel.data[0])[i],true);
		}
		else
		{	
	//		throw new 
	//		InterpretError("Can not convert relation " + rel.name + " to domain list");
			
			//bug fix by Yi Zheng Sept 28, 2002
			domList = new Domain[rel.numattrs];
			for (int j = 0; j < rel.numattrs; j++)
				domList[j] = rel.domains[j];
			
			//end of bug fix Sept 28, 2002

		}
		
		return domList;
   }

 	/**
	 * wrapper function to combine the following version (v1) of ExpressionListToDomains
	 * with ExpressionListToDomains(SimpleNode node, Hashtable htable) (v2) from Andrey
	 * @author Yi Zheng May 26, 2002
	 */   
  private Domain[] ExpressionListToDomains(SimpleNode node, Relation rel, Hashtable htable, Environment env)
       throws InterpretError
	{ Domain[] domArray = ExpressionListToDomains(node, htable, env); //calls Andrey's code to allow relational expression in projection list
	 
	 //addition by Yi Zheng June 28, 2002
	 //add virtual domains if this is a red node intended for level lifting
	 domArray = handleRedDomain(domArray, node, rel, env);
	 
	 if((domArray.length == 1) && ((domArray[0].isRedTemp == 1) || (domArray[0].isRedTempTop == 1)))
	 {
	 	//System.out.println("return domArray directly");
	 	return domArray;
	 }
	 //end of addition June 28, 2002

	  return ExpressionListToDomains(domArray, rel); //calls Yi's modified code to test for state domain
	 }
	 
	  //this method adds hidden state 
  //domains from rel to the domains being returned
  //it also checks to make sure that no state is in the original
  //projection list   
  // Yi Zheng April, 2002
  // v1
  private Domain[] ExpressionListToDomains(Domain[] node, Relation rel)
       throws InterpretError
  {
    int n = node.length;
    int numDoms = 0;
    
    //find the number of state domains in rel
    for(int j = 0; j < rel.numattrs; j++)
    {
    	if (rel.domains[j].isStateDom == 1)
    		numDoms++;
    }
    Domain[] doms = new Domain[n + numDoms];
    //System.out.println("children in the Expression List");
    
    //check for state doms in the original list 
    //this is just superficial check, virtual domains which
    //are renamed from state doms will be caught in virdom declaration code
    int i;
    for (i = 0; i < n; i++) {
 		Domain d = node[i];
 		if (d.isStateDom == 1)
      throw new InterpretError("domain: " + d.name + " is a hidden state domain which cannot be projected upon");
		else
		doms[i] = d;
    }
    
    //finally, tack on state doms in rel
    	for (int k = 0; k < rel.numattrs; k++)
    	{
    		if (rel.domains[k].isStateDom == 0)
    			continue;
    		else
    			doms[i++] = rel.domains[k];
    	}
    
    return doms;
  }
	 
  //this method overloads adds hidden state 
  //domains from rel to the domains being returned
  //it also checks to make sure that no state is in the original
  //projection list   
  // Yi Zheng April, 2002
  // 
  private Domain[] ExpressionListToDomains(SimpleNode node, Relation rel)
       throws InterpretError
  {
    int n = node.jjtGetNumChildren();
    int numDoms = 0;
    
    //find the number of state domains in rel
    for(int j = 0; j < rel.numattrs; j++)
    {
    	if (rel.domains[j].isStateDom == 1)
    		numDoms++;
    }
    Domain[] doms = new Domain[n + numDoms];
    //System.out.println("children in the Expression List");
    
    //check for state doms in the original list 
    //this is just superficial check, virtual domains which
    //are renamed from state doms will be caught in virdom declaration code
    int i;
    for (i = 0; i < n; i++) {
      SimpleNode nd = (SimpleNode) node.jjtGetChild(i);
      //printNode(nd);
      if (nd.type != OP_IDENTIFIER) {
	throw new 
	  InterpretError("only IDList implemented for ExpressionList");
      }
      String s = nd.name;
      Domain d = env.lookupDom(s, true);
      if (d == null)
	throw new InterpretError("domain : " + s + " not exist");
      else if (d.isStateDom == 1)
      throw new InterpretError("domain: " + s + " is a hidden state domain which cannot be projected upon");
		else
		doms[i] = d;
    }
    
    //finally, tack on state doms in rel
    	for (int k = 0; k < rel.numattrs; k++)
    	{
    		if (rel.domains[k].isStateDom == 0)
    			continue;
    		else
    			doms[i++] = rel.domains[k];
    	}
    
    return doms;
  }  
  
     /*
   This function evaluates an expression and converts the resulting unary relation 
   with the attribute of type "attribute" to a list of domains.
   
   Input: a parsed tree =>   ExpressioList
							/  ...	| ... \	 
   
   Output: a list of domains
									A. Rozenberg
   * v2
   */
  // public Domain[] ExpressionListToDomains(SimpleNode node, Hashtable htable) //modified by Yi Zheng May 27, 2002

   public Domain[] ExpressionListToDomains(SimpleNode node, Hashtable htable, Environment env)
   throws InterpretError
   {
	 int n = node.jjtGetNumChildren();
	 Vector domList = new Vector();
	      
	 for (int i = 0; i < n; i++) 
	 {
		SimpleNode nd = (SimpleNode) node.jjtGetChild(i);
		//printNode(nd);
	    switch(nd.type)
		{	
			case OP_IDENTIFIER:	String s = nd.name;
					Relation rd = env.lookupRel(s, true);
   					if ((rd != null) && (rd.rvc == RELATION)) //modified by Yi Zheng May 31, 2002
					{ //Convert relation of type "attribute" to domain list
						Domain doms[] = AttribsRel2DomList(rd);
						domList.addElement(doms);
					}
					else
					{
						Domain d = env.lookupDom(s, true);
						if (d == null)
							throw new InterpretError("domain or relation: " + s + " does not exist");
						else
						{
							Domain[] doms = new Domain[1];
							doms[0] = d;
							domList.addElement(doms);
	 					}								
					}
					break;
					
			case OP_JOINOPERATOR:	String tmpname = nextTempName();
									Relation rel = evaluateTLExpression(nd, tmpname, htable, env);
									Domain doms[] = AttribsRel2DomList(rel);
									domList.addElement(doms);
									deleteFile(tmpname);
									break;
			
			case OP_EQUIV:
			case OP_PAR:
			case OP_RED: //Yi Zheng May 7, 2002
			case OP_FUN: //Yi Zheng May 7, 2002
	//				throw new InterpretError("Reduction and functional mapping in projection list: to be implemented");
			break; //Yi Zheng June 24, 2002, note this is temporay, I don't know how to cope with the following 2 cases yet:
				// case 1: let a be equiv + of 1 by (equiv ujoin of nestedAttrRel);
				//		if nestedAttrRel contains only one domain of type attr, then
				//		this let statement is legal, and we should allow it in actualization
				//	However, the challenge here is that we don't know what the type of the temp
				//	domain is (in this case equiv ujoin of nestedAttrRel), and we certainly 
				//	don't know if this temp domain will be IDLIST typed with one attribute of type attribute;
				//	secondly, it is not a good idea to try to actualize the temp domain here
				//	because this will not allow domain alias, for example,
				//	let b be a; in such case, actulizing on b will be problematic, as the actualizer will not know how to convert relation containing attr to attributes! 
				//	this is just another case of the shortcomings of early conversion in actualizer constructor by Andrey
				//case 2: let a be equiv + of 1 by (equiv + of c by d);
				//	This is also a legitimate case, however, we don't know the type of the temp
				//	domain equiv + of c by d, it could have been an IDLIST typed domain and which
				// 	doesn't have attr as its first attribute. 
				//	As a temporary compromise, I just let these cases pass with a break. But
				//	the danger is that case 1 will cause errors later on, if no code is installed for handling AttribsRel in actualization code other than the constructor
			case OP_PROJECT: //added by Yi Zheng June 24, 2002
			case OP_SELECT:
			case OP_QSELECT:
			case OP_TSELECT:
			case OP_PICK:
			case OP_ATTRIBSOF:
			case OP_ARRAY:
							tmpname = nextTempName();
							 
							 rel = evaluateTLExpression(nd, tmpname, htable, env);
						     
						     doms = AttribsRel2DomList(rel);
							domList.addElement(doms);
					
							deleteFile(tmpname);
							break;
							//end of addition June 24, 2002
			default:		
				/*			 tmpname = nextTempName();
							 
							 rel = evaluateTLExpression(nd, tmpname, htable, env);
						     doms = AttribsRel2DomList(rel);
							domList.addElement(doms);
							deleteFile(tmpname);
							*/ //commented out by Yi Zheng June 24, 2002
							break;
			 
	/*		throw new 
			InterpretError("only IDList and JoinExpression are implemented for ExpressionList");
	*/
		}
	 }
	 int domSize =0;
	 Domain[] doms=null;
	 for(int i=0;i<domList.size();i++)
	 {
		doms=(Domain[])domList.elementAt(i);
		domSize=domSize+doms.length;
	 }	
	 
	 Domain[] domsArray=new Domain[domSize];
	 int k=0;
	
	 for(int i=0;i<domList.size();i++)
	 {
		doms=(Domain[])domList.elementAt(i);
		for(int j=0;j<doms.length;j++)
		{
			domsArray[k]=doms[j];	
			k++;
		}
	 }
	
	 return domsArray;
   }  
   
   /**
    * generate a temporary domain for red node that is intended for level lifting
    * @author Yi Zheng June 28, 2002
    */
   private Domain[] handleRedDomain(Domain[] domArray, SimpleNode node, Relation rel, Environment env) 
	throws InterpretError
   {
 	 int n = node.jjtGetNumChildren();
 	 if(n != 1)
 	 	return domArray;
 	 	
	SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
	Domain newDom = null;
	switch(nd.type)
	{	
		case OP_RED: 
			SimpleNode ndCopy = (SimpleNode)nd.jjtDeepDuplicate();
			
			SimpleNode letNode = SimpleNode.jjtCreateNode("Declaration", OP_DECLARATION,
				OP_LET, null);
			letNode.jjtAddChild(ndCopy);
			ndCopy.jjtSetParent(letNode);
			String domname = "reddom" + Surrogate.next();
			newDom = newVirDomain(domname, letNode, env);
	   		Relation rel_saved = env.lookupRel("."+domname,true);
	   		env.put(newDom);
	   		newDom.isRedTempTop = 1;
			if (rel_saved != null)
	     		env.put(rel_saved);

			break;
		default:
			return domArray;		
	}
	 Domain[] doms = new Domain[1];
	 doms[0] = newDom;
	
	 return doms;  	
   	
   	
   	
   }    
  private void getDomsInCompList(SimpleNode node, Hashtable htable,
				 Environment env) 
       throws InterpretError
  {
    if(node.type == OP_IDENTIFIER){
      Domain d = env.lookupDom(node.name, true);
      if(d != null){   // Should never happen.
	htable.put(node.name, d);
	return;
      }else
	throw new InterpretError("domain '"+node.name+"' not found.");
    }else if(node.jjtGetNumChildren() == 2){
      SimpleNode left = (SimpleNode)node.jjtGetChild(0);
      SimpleNode right = (SimpleNode)node.jjtGetChild(1);
      
      switch(node.opcode){
      case OP_AND:
      case OP_OR:
	getDomsInCompList(left, htable, env);
	getDomsInCompList(right, htable, env);
	break;
      case OP_EQ:
      case OP_NEQ:
      case OP_GT:
      case OP_LT:
      case OP_GE:
      case OP_LE:
	Domain d;
	if(left.type == OP_IDENTIFIER){
	  d = env.lookupDom(left.name, true);
	  if(d != null) //Yi Zheng May 2, 2002
	  	htable.put(left.name, d);
	  else throw new InterpretError("domain " + left.name + " not found");
	}
	if(right.type == OP_IDENTIFIER){
	  d = env.lookupDom(right.name, true);
	  if(d != null) //Yi Zheng May 2, 2002
	  	htable.put(right.name, d);
	  else throw new InterpretError("domain " + right.name + " not found");
	}
	break;
      default:
	// START Since jRelixNS
//	if (env.isSigmaJoinImplemented != true) {
	  throw new InterpretError("getDomsInCompList: to be implemented.");
//	} else {
	  // This is added for the boolean-IDLIST compatibility.
	  // We will try to see if it is an IDLIST if all else
	  // fails for a boolean.
	  // when processing domains in a 'comparions' list
	  // for a OP_SELECT or OP_TSELECT. 
//	  if (getIDList(node,env) == null)
//	    throw new InterpretError("getDomsInCompList: BI-OP " + 
//				     node.opcode +" not supported; to be implemented.");
//	}
	// END Since jRelixNS
      }
      return;
    }else{
      // START Since jRelixNS
 //     if (env.isSigmaJoinImplemented != true) {
	throw new InterpretError("getDomsInCompList: to be implemented.");
  //    } else {
	// This is added for the boolean-IDLIST compatibility.
	// We will try to see if it is an IDLIST if all else
	// fails for a boolean.
	// when processing domains in a 'comparions' list
	// for a OP_SELECT or OP_TSELECT. 
//	if (getIDList(node,env) == null)
//	  throw new InterpretError("getDomsInCompList: Node type " 
//+node.type+ 
//"non-IDLIST/BIOP not supported; to be implemented.");
//      }
      // END Since jRelixNS
    }
  }      
  /**
   * Return a virtual relation which contains the domain array as the 
   * result of join operations etc. occurred in 'node'.
   * Added for jRelixNS:
   *    The relation could be real if a top-level relation is
   *    used in the domain algebra.
   * 
   * JRelixNS modified/enhanced this method:
   *   1) Handle Sigma-Joins in addition to mu-joins.
   *   2) Handle empty exprlist as it might not exist for an empty projector
   *      for nullary tuple results, more frequently seen for sigma-joins.
   */
   private Relation getIDList(SimpleNode node, Environment env)
       throws InterpretError
  {
 //   System.out.println("getIDlistnode type "+node.type);
    switch(node.type){
    case OP_IDENTIFIER:
      String rname = node.name;
      Relation r = env.lookupRel("."+rname, true);
      if(r == null){
	
	// REVISIT: should be looking into NREnvironment for top-level relation.
	// e.g. let x be S ijoin R; where 'R' is top-level rel.
	if((r = env.lookupRel(rname, true)) == null)
	  // should never happen..
	  throw new InterpretError("join relation not found.");
      }
      return r;
    case OP_JOINOPERATOR:  
    if (Relation.isSigmaJoinImplemented != true) { //Andy May 12, 2002, added this if case
	   Relation r1, r2;
	   Domain[] doms;

	   if(node.jjtGetNumChildren() == 2){
		 // No join attributes presented.
		 r1 = getIDList((SimpleNode)node.jjtGetChild(0), env);
		 r2 = getIDList((SimpleNode)node.jjtGetChild(1), env);
		 doms = r1.mujoinType(node.opcode, null, r2, null);
	   }else{
		 // Join attributes are given.
		 r1 = getIDList((SimpleNode)node.jjtGetChild(0), env);
		 r2 = getIDList((SimpleNode)node.jjtGetChild(3), env);
		 Domain[] doms1, doms2;
		 Vector domsv1 = new Vector();
		 Vector domsv2 = new Vector();

		 // Add .id to vectors.
		 domsv1.addElement(env.lookupDom(".id", true));
		 domsv2.addElement(env.lookupDom(".id", true));

		 SimpleNode expr1, expr2;
		 expr1 = (SimpleNode)node.jjtGetChild(1);
		 expr2 = (SimpleNode)node.jjtGetChild(2);

		 if(expr1.jjtGetNumChildren() != expr2.jjtGetNumChildren())
			// Should never happen...
			throw new InterpretError("different number of join attributes.");

		 for(int i = 0; i < expr1.jjtGetNumChildren(); i++){
			String name = ((SimpleNode)expr1.jjtGetChild(i)).name;
			Domain d = env.lookupDom(name, true);
			if(d != null) domsv1.addElement(d);
			else throw new InterpretError("domain '"+d+"' not found.");

			name = ((SimpleNode)expr2.jjtGetChild(i)).name;
			d = env.lookupDom(name, true);
			if(d != null) domsv2.addElement(d);
			else throw new InterpretError("domain '"+d+"' not found.");
		 }
		 doms1 = new Domain[domsv1.size()];
		 doms2 = new Domain[domsv2.size()];
		 domsv1.copyInto(doms1);
		 domsv2.copyInto(doms2);

		 doms = r1.mujoinType(node.opcode, doms1, r2, doms2);
	   }
	   return new Relation("", doms, 0, doms.length, 0, env);
	 }else{
      Relation r1, r2;
      boolean isSigma;	
      Domain[] doms;
      
      switch(node.opcode) 
	{  
	case OP_ICOMP : 
	case OP_EQJOIN : 
	case OP_GTJOIN : 
	case OP_GEJOIN : 
	case OP_LTJOIN : 
	case OP_LEJOIN : 
	case OP_IEJOIN : 
	case OP_NOTICOMP : 
	case OP_NOTEQJOIN : 
	case OP_NOTGTJOIN : 
	case OP_NOTGEJOIN : 
	case OP_NOTLTJOIN : 
	case OP_NOTLEJOIN : 
	case OP_NOTIEJOIN : 
	  isSigma = true;	
	  break;
	  
	case OP_IJOIN :
	case OP_UJOIN :
	case OP_SJOIN :
	case OP_LJOIN :
	case OP_DLJOIN :
	case OP_RJOIN :
	case OP_DRJOIN :
	default:
	  isSigma = false;	
	  break;
	}
      if(node.jjtGetNumChildren() == 2){
	// No join attributes presented.
	r1 = getIDList((SimpleNode)node.jjtGetChild(0), env);
	r2 = getIDList((SimpleNode)node.jjtGetChild(1), env);
	
	//added by Yi Zheng July 4, 2002 to handle let a be B nop C
	if(node.opcode == OP_NOP)
	{
		if((!r1.isDomainsIn(r2.domains)) || (!r2.isDomainsIn(r1.domains)))
			throw new InterpretError("type mismatch between domains");
		return new Relation("", r1.domains, 0, r1.domains.length, 0, env);

	}
	

	//end of addition July 4
	
	
	if (isSigma) 
	  doms = r1.sigmaJoinType(node.opcode, null, r2, null);
	else
	  doms = r1.mujoinType(node.opcode, null, r2, null);
	  
	if (!r1.isDomainIn(env.lookupDom(".id", true),doms))
	  {
	    Domain[] tmpdoms = doms;
	    doms = new Domain[tmpdoms.length+1];
	    doms[0]=env.lookupDom(".id", true);
	    for(int i = 0; i < tmpdoms.length; i++){
	      doms[i+1]=tmpdoms[i];
	    }
	  }
//	else if(doms.length == 1){
//	  doms = new Domain[2];
//	  doms[0]=env.lookupDom(".id", true);
//	  doms[1]=env.lookupDom(".bool", true);
//	}
      }else{
	// Join attributes are given.
	r1 = getIDList((SimpleNode)node.jjtGetChild(0), env);
	r2 = getIDList((SimpleNode)node.jjtGetChild(3), env);
	Domain[] doms1, doms2;
	Vector domsv1 = new Vector();
	Vector domsv2 = new Vector();
	
	// Add .id to vectors.
	domsv1.addElement(env.lookupDom(".id", true));
	domsv2.addElement(env.lookupDom(".id", true));
	
	SimpleNode expr1, expr2;
	expr1 = (SimpleNode)node.jjtGetChild(1);
	expr2 = (SimpleNode)node.jjtGetChild(2);
	
	if(expr1.jjtGetNumChildren() != expr2.jjtGetNumChildren())
	  // Should never happen...
	  throw new InterpretError("different number of join attributes.");
	
	for(int i = 0; i < expr1.jjtGetNumChildren(); i++){
	  String name = ((SimpleNode)expr1.jjtGetChild(i)).name;
	  Domain d = env.lookupDom(name, true);
	  if(d != null) domsv1.addElement(d);
	  else throw new InterpretError("domain '"+d+"' not found.");
	  
	  name = ((SimpleNode)expr2.jjtGetChild(i)).name;
	  d = env.lookupDom(name, true);
	  if(d != null) domsv2.addElement(d);
	  else throw new InterpretError("domain '"+d+"' not found.");
	}
	doms1 = new Domain[domsv1.size()];
	doms2 = new Domain[domsv2.size()];
	domsv1.copyInto(doms1);
	domsv2.copyInto(doms2);
	
	if (isSigma) {
	  Domain[] sigdoms = r1.sigmaJoinType(node.opcode, doms1, r2, doms2);
	  doms = new Domain[sigdoms.length+1];
	  doms[0]=env.lookupDom(".id", true);
	  for(int i = 0; i < sigdoms.length; i++){
	    doms[i+1]=sigdoms[i];
	  }
	}
	else
	  doms = r1.mujoinType(node.opcode, doms1, r2, doms2);
      }
      return new Relation("", doms, 0, doms.length, 0, env);
    } 
    
    case OP_PROJECT:
    case OP_TSELECT:
      // REVISIT: TSELECT also has a projection list at the first child.
      // Note, we are not checking for validality. Hopefully this will
      // be done at the actualization time.
      
      // Get the projection domain list.
      // START Since jRelixNS
                 // Support for empty projectors.
      SimpleNode exprlist;
 //     if (env.isSigmaJoinImplemented != true) { //Andy May 12, 2002
   if (Relation.isSigmaJoinImplemented != true) {
	exprlist = (SimpleNode)node.jjtGetChild(0);
      } 
      else {
	// exprlist might not exist for an empty projector 
	   // for nullary tuple results.
	exprlist = (SimpleNode)node.jjtGetChild(0);
	if(exprlist.jjtGetNumChildren() > 0) // expect 1 or 0 only.
	  exprlist = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
      }
      // END Since jRelixNS
      
      //added by Yi Zheng July 2, 2002
      //for implementation of level lifting
      if((exprlist.jjtGetNumChildren() == 1) &&
      	 ((((SimpleNode) exprlist.jjtGetChild(0)).type) == OP_RED))
      {
      	SimpleNode firstChild = null;
      	//if added by Yi Zheng July 9, 2002
      	if((((SimpleNode) exprlist.jjtGetChild(0)).opcode) != OP_REDFUNCALL)
      		firstChild = (SimpleNode) exprlist.jjtGetChild(0);
      	else
       		firstChild = (SimpleNode) exprlist.jjtGetChild(1);
     	//Aug 13
	//Utility.dump(firstChild, "firstChild ");
	
      	return getIDList(firstChild, env);
      }
      
      //end of Yi's addition July 2, 2002
      Vector domvec = new Vector();
      domvec.addElement(env.lookupDom(".id", true));
      for(int i = 0; i < exprlist.jjtGetNumChildren(); i++){
	String dname = ((SimpleNode)exprlist.jjtGetChild(i)).name;
	Domain d = env.lookupDom(dname, true);
	if(d == null) 
	  throw new InterpretError("domain "+dname+" not found.");
	domvec.addElement(d);
      }
                 // BEGIN Since jRelixNS BUGFIX 2
	   if (exprlist.jjtGetNumChildren() == 0)
	     domvec.addElement(env.lookupDom(".bool", true));
           // END Since jRelixNS BUGFIX 2

      Domain[] doms = new Domain[domvec.size()];
      domvec.copyInto(doms);
      return new Relation("", doms, 0, doms.length, 0, env);
    case OP_SELECT:
      SimpleNode srcrelnode = (SimpleNode)node.jjtGetChild(1);
      return getIDList(srcrelnode, env);
      
    case OP_QSELECT: //Added by hongyu Sept 30,01
      SimpleNode srcnode = null;
      Relation srcrel = null;
      if(node.jjtGetNumChildren() == 2)
	srcnode = (SimpleNode)node.jjtGetChild(1);
      else srcnode = (SimpleNode)node.jjtGetChild(2);
      switch(srcnode.type){
      case OP_IDENTIFIER:{
	String srcrelname = srcnode.name;
//	System.out.println("srcrelname "+srcrelname);
	srcrel = env.lookupRel("."+srcrelname, true);
	if(srcrel == null){
	  srcrel = env.lookupRel(srcrelname, true);
	  if(srcrel == null)// Should never happen. Hongyu
	    throw new InterpretError("source relation not found.");
	
	}
	return srcrel;
      }
      case OP_SELECT:
      case OP_PROJECT:
      case OP_TSELECT:
      case OP_QSELECT:
      case OP_JOINOPERATOR:{ 
	int t = traverseType(srcnode, env);
	if(t != IDLIST && t != COMP)
	  throw new InterpretError("mismatched type.");
	srcrel = getIDList(srcnode, env);
	return srcrel;
      }
      default:
	throw new InterpretError("Relation: to be implemented.");
      }
      // end of Hongyu Added.
    case OP_RED:
    case OP_EQUIV:
      //Check if OP_FUN and OP_PAR should be included. From sungsoo. Hongyu Feb 2002
      //         Vertical     for "let x be red ujoin of (S ijoin T);"
      //             \          
      //             Join
      //             /  \
      //            S    T
      //if added by Yi Zheng July 9, 2002
      if(node.opcode != OP_REDFUNCALL)
	{
//Aug 13
  //  Utility.dump((SimpleNode)node.jjtGetChild(0), "node.child(0) ");
      	return getIDList((SimpleNode)node.jjtGetChild(0), env);
	}
      else
      	return getIDList((SimpleNode)node.jjtGetChild(1), env);
      
    case OP_FUN: //From sungsoo. hongyu added Feb 2002
    case OP_PAR:
    //if added by Yi Zheng July 9, 2002
     if(node.opcode != OP_REDFUNCALL)
      return getIDList((SimpleNode)node.jjtGetChild(0),env);
     else
      return getIDList((SimpleNode)node.jjtGetChild(1),env);
     		
      //System.out.println("getIDList should be taken care of");
      // End of sungsoo's code
    case OP_ARRAY: {//Yi Zheng June 19, 2002
    	//1, get the identifier
    	SimpleNode id = (SimpleNode) node.jjtGetChild(0);
    	SimpleNode arrayIndex = (SimpleNode) node.jjtGetChild(1);
    	//2. look to see if id is a comp/idlist type domain
    	Domain idd = env.lookupDom(id.name, true);
    	if(idd != null)
    	{
    		if ((idd.type != IDLIST) && (idd.type != COMP))
    			throw new InterpretError(id.name + " is not a valid array identifier");
    		else
    		{
    			Relation rel = env.lookupRel("." + id.name, true);
    			if(rel == null)
    				throw new InterpretError("relation ." + id.name + " not found");
    			Domain dotid = env.lookupDom(".id", true);
    			if(!rel.isDomainIn(dotid))
    				return getProjListRel(rel, arrayIndex, env);
    			else //get rid of .id
    			{
    				Domain[] newD = new Domain[rel.domains.length - 1];
    				int j = 0;
    				for(int i = 0; i < rel.domains.length; i++)
    				{
    					if(rel.domains[i].name.compareTo(".id") != 0)
    						newD[j++] = rel.domains[i];
    				}
    				String tname = nextTempName();
    				Relation result = rel.project(newD, tname);
    				result.name = tname;
    				return getProjListRel(result, arrayIndex, env);
    			}
    		}
    	}
    	else //3. check if id is a top-level RELATION/COMPUTATION
    	{
    		Relation topR = env.lookupRel(id.name, true);
    		if(topR == null)
    			throw new InterpretError(id.name + " is not a valid array identifier");
    		return getProjListRel(topR, arrayIndex, env);
    	}
    	
    }
    default:
      throw new InterpretError ("getIDList: to be implemented.");
      
    }
  }      
  
  /**
   * method that returns a relation containing the domains to appear in the projection list
   * for array syntax check
   * @author Yi Zheng June 19, 2002
   */
   private Relation getProjListRel(Relation rel, SimpleNode arrayList, Environment env)
 	throws InterpretError
   {
   	
   	Vector domvec = new Vector();
   	for(int i = 0; i < rel.numattrs; i++)
   	{
   		if(!arrayList.isBitSet(i))
   			domvec.addElement(rel.domains[i]);
   	}
   	if(domvec.size() == rel.numattrs)
   		return rel;
 
   	Domain[] doms = new Domain[domvec.size()];
   	domvec.copyInto(doms);
   	String tname = nextTempName();
   	Relation ret = rel.project(doms, tname);
   	ret.name = tname;
   	return ret;
   }
   	
  private void help(String name)
  //added by Yi Zheng 
  throws InterpretError
  {
    //added by Yi Zheng March 18, 2002
    throw new InterpretError ("help: to be implementd");
  }      
  // convert an IDList node to an array of domains
  private Domain[] IDListToDomains(SimpleNode node, Environment env)
       throws InterpretError
  {
    int n = node.jjtGetNumChildren();
    Domain[] doms = new Domain[n];
    
    for (int i = 0; i < n; i++) {
      String s = ((SimpleNode) node.jjtGetChild(i)).name;
      Domain d = env.lookupDom(s, true);
      if (d == null)
	throw new InterpretError("domain : " + s + " not exist");
      else if (d.isStateDom == 1) //Yi Zheng May 6, 2002
      throw new InterpretError("domain : " + s + " is a state domain");
      else
	doms[i] = d;
    }
       
    return doms;
  }      
  public long interpret(SimpleNode node, Environment env)
       throws InterpretError
  {  
    //System.out.println("enter interpret");
    if (node == null) return 0;
    long ts = System.currentTimeMillis();
	int type = node.type;
	
	// start Zongyan's addition, July 22 2002
	
	if (node.url != null & type != ALDATP_ASSIGNMENT & type != ALDATP_UPDATE & type != ALDATP_DECLARATION) {
		if (Debug) System.out.println("call aldatpClient in interpret");
		aldatpClient.request(node,env);  // need to check response ^_^
		System.out.println(aldatpClient.rspCode);
		return System.currentTimeMillis() - ts;
	}
	// end Zongyan's addition,  July 22 2002
	
    
    switch(type) 
      {  case OP_COMMAND :
	executeCommand(node, env); 
      break;
      case OP_STATEMENT :
	executeStatement(node, env); 
	break;
      case OP_DECLARATION :
	executeDeclaration(node, env); 
	break;
		case ALDATP_DECLARATION: //Zongyan Aug 2002
	reverseView(node,env);   
	try{
    node = aldatpUtility.getRootNode();
	} catch (IOException e)
	{ throw new InterpretError(" IO error in rebuild tree");
	}
	interpret(node,env);
	break;

      case OP_ASSIGNMENT :
//	executeAssignment(node, env); 
	case ALDATP_ASSIGNMENT :  // Zongyan Aug 2002
	executeAssignment(node, env);   // local evaluate, remote assignment
	break;
	   case ALDATP_UPDATE:  //Zongyan Aug 2002
	   lefthandUpdate(node,env);

	break;
      case OP_UPDATE :
	Environment uenv = new Environment(env.reltable, env.domtable); 
//	Environment uenv = new Environment(env, env.getTable(), env.reltable, env.domtable); 
	// produce trigger tree
	int diftrig = lookUpdate(node, uenv, null, uenv.triggerroot);
	// reverse the result -- undo trigger
	doTrigger(env, uenv.triggerroot, 1);
	if (diftrig == 1)
	  executeUpdate(node, node, uenv); 
	RemoveTrigger(uenv.triggerroot);		// clear up
	break;
      default : // impossible to be here
	System.out.println("unknown statement");
	break;
      }
    return System.currentTimeMillis() - ts;
     }         
  
  // Weizhong
  // overload the original interpret to handle the UNDO command in 
   // event handlers.
  public long interpret(SimpleNode node, Environment env, String rlname, int opc)
       throws InterpretError
  {  
    if (node == null) return 0;
    int type = node.type;
    
    switch(type) 
      {  case OP_COMMAND :
	executeCommand(node, env, rlname, opc); 
      break;
      default :
	interpret(node, env);
	break;
      }
    return 1;
  }         
    //added by Zongyan Aug,2002
  private void lefthandUpdate(SimpleNode node, Environment env)
	  throws InterpretError
  {
	  String tempname = null;
	  Relation rr = null;
	  String url,name,rname;
	  Domain[] domsr = null ,domsl = null;
	  boolean flag = false;
	  
      switch (node.opcode) 
	  {
	  case OP_ADD:	
	  case OP_DELETE:
	       tempname = nextTempName();
	       rr = evaluateTLExpression((SimpleNode) node.jjtGetChild(1), tempname, new Hashtable(), env);
           String lname = ((SimpleNode) node.jjtGetChild(0)).name;
		   tempname = nextTempName();
		   Relation rel = rr.assign(domsr, tempname, domsl);
		   env.put(rel);
		   distributedStmt(rel.name,lname,node.url,node.opcode,env);
		   env.removeRel(rel.name, false);
		   break;
	  
	  case OP_CHANGE:
		   try{
		   File outputFile = new File("_tmp_aldatp_tmpFile.txt");
           FileOutputStream os = new FileOutputStream(outputFile);
           PrintWriter outFile = new PrintWriter(os);
		   if (Global.start_type == 0 )   
		    outFile.println( "StatementBlock" + "|" + node.url + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 2);
		   else
		    outFile.println( "StatementBlock" + "|" + node.url + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		   SimpleNode tmpnode = node.updateChangeDump(outFile);
		   
		   if (tmpnode != null){
			 if (tmpnode.opcode==OP_IDENTIFIER){
			   rname = tmpnode.name;
			 } else 
			 { 
		       tempname = nextTempName();
	           rr = evaluateTLExpression((SimpleNode) tmpnode, tempname, new Hashtable(), env);
               env.put(rr);
			   rname = rr.name;
			   flag = true;
			 }
			 
		  if (Global.start_type == 0 ) { // for stand alone JRelix, normal server can also go this way
			  name = aldatpClient.getRelName(node.url,env);	
			  
			 aldatpClient.putRel(rname,name,node.url,env);	// response ^_^
		  }
		  else { // for top-level server, normal server or prtoected server 
			name = rname;
			String path = System.getProperty("user.dir");
		    if( path.indexOf("public_aldatp") >= 0 ){
		    if (path.endsWith("public_aldatp"))
	    	  url = "aldatp://" + aldatpUtility.getHostname() +"/"  ;
		    else
			  url = "aldatp://" + aldatpUtility.getHostname() +"/" + path.substring(path.indexOf("public_aldatp") + 14) + System.getProperty("file.separator");
		    }
		    else url = "aldatProtect://" + aldatpUtility.getHostname() +"/" + path;
			 
		     
		    outFile.println( "ALDATP" + "|" + url + "|" + OP_ALDATP + "|" + OP_ALDATP + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		  } //server
		   outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + name + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 0);
		   if (Global.start_type == 0 ) { 
		   outFile.println( "Command" + "|" + null + "|" + OP_COMMAND + "|" + OP_DELR + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		   outFile.println( "IDList" + "|" + null + "|" + OP_IDLIST + "|" + OP_IDLIST + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		   outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + name + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 0);			  
		  }		  }//tmpnode!= null		   
		  else {
		   if (Global.start_type == 0 ) {		   outFile.println( "Command" + "|" + null + "|" + OP_COMMAND + "|" + OP_PR + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		   
		   outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + ((SimpleNode) node.jjtGetChild(0)).name + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 0);
			   }		   }
          outFile.close(); 
		  node = aldatpUtility.getRootNode();
	
	     } 
	     catch (IOException e){
	  	   Global.processThrowable(e);
		 }
	   
	     interpret(node,env);
	   
	     if(flag) env.removeRel(rr.name, false);
		   
	     break;
		   
	  default: //impossible
	      break;
	  }
  }
 
  //supose the following declaration is  made on mimi:mydir
  // aldatp://willy/dir1/(myView) is R;
  // reverseView() transfer the left hand remote declaration to the remote statement:
  // aldatp://willy/dir1{ myView is aldatp://mimi/mydir/(R)};
  //
  // Zongyan  Aug 2002
  private void reverseView(SimpleNode node, Environment env)
	  throws InterpretError
  {
	  String URL = node.url;
	   try{
		   File outputFile = new File("_tmp_aldatp_tmpFile.txt");
           FileOutputStream os = new FileOutputStream(outputFile);
           PrintWriter outFile = new PrintWriter(os);
		   outFile.println( "StatemnetBlock" + "|" + URL + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);		 ((SimpleNode) node.jjtGetChild(0)).url = URL;
		 if(Debug) System.out.println("child0.url:"+  ((SimpleNode) node.jjtGetChild(0)).url);			node.dump(outFile,URL,env);
            outFile.close(); 
		   
	   }
	   catch (IOException e){
		   Global.processThrowable(e);
		}
	}   
    public void distributedStmt(String rname, String lname, String URL, int op,Environment env)
	  throws InterpretError
  {
	  SimpleNode node = null;
	  String url = null;
	  String name = null;
	   try{
		   File outputFile = new File("_tmp_aldatp_tmpFile.txt");
           FileOutputStream os = new FileOutputStream(outputFile);
           PrintWriter outFile = new PrintWriter(os);
		   if(Global.start_type == 0 )  
		  outFile.println( "StatementBlock" + "|" + URL + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 2);
		   else
		   outFile.println( "StatementBlock" + "|" + URL + "|" + OP_STATEMENT + "|" + OP_STMTBLOCK + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		  switch (op)
		  {
		  case OP_ASSIGN:
		   outFile.println( "Assignment" + "|" + null + "|" + OP_ASSIGNMENT + "|" + OP_ASSIGN + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 2);
		   break;
		  case OP_ASSIGNADD:  
		   outFile.println( "Assignment" + "|" + null + "|" + OP_ASSIGNMENT + "|" + OP_ASSIGNADD + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 2);
		   break;
		   case OP_ADD:  
		   outFile.println( "Update" + "|" + null + "|" + OP_UPDATE + "|" + OP_ADD + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 2);
		   break;
		   case OP_DELETE:  
		   outFile.println( "Update" + "|" + null + "|" + OP_UPDATE + "|" + OP_DELETE + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 2);
		   break;
		  }
		  outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + lname + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 0);
		  
		 
		  if (Global.start_type == 0 ) { // for stand alone JRelix, server can also go this way
			 name = aldatpClient.getRelName(URL,env);	
			 
			 aldatpClient.putRel(rname,name,URL,env);	// response ^_^
		  }
		  else { // for top-level server, normal server or prtoected server
		  name = rname;	  
		  String path = System.getProperty("user.dir");
		  if( path.indexOf("public_aldatp") >= 0 ){
		  if (path.endsWith("public_aldatp"))
	    	  url = "aldatp://" + aldatpUtility.getHostname() +"/"  ;
		  else
			 url = "aldatp://" + aldatpUtility.getHostname() +"/" + path.substring(path.indexOf("public_aldatp") + 14) + System.getProperty("file.separator");
		  }
		  else url = "aldatProtect://" + aldatpUtility.getHostname() +"/" + path;
		  outFile.println( "ALDATP" + "|" + url + "|" + OP_ALDATP + "|" + OP_ALDATP + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		  }
		  outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + name + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 0);		  
		  if (Global.start_type == 0 ) { 
		   outFile.println( "Command" + "|" + null + "|" + OP_COMMAND + "|" + OP_DELR + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		   outFile.println( "IDList" + "|" + null + "|" + OP_IDLIST + "|" + OP_IDLIST + "|" + null + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 1);
		   outFile.println( "Identifier" + "|" + null + "|" + OP_IDENTIFIER + "|" + OP_IDENTIFIER + "|" + name + "|" + 0 + "|" 
					+ null + "|" + null + "|" + 0);			  
		  }		  
          outFile.close(); 
		  node = aldatpUtility.getRootNode();
		  
	   }
	   catch (IOException e){
		   Global.processThrowable(e);
		}
	   
	  interpret(node,env);
	   
  }
  // check if there is self-reference in a view expression, the semantic
  // check for the existence of operands in the view expression is also
  // done, the hash table contains names already checked to avoid
  // infinite recursive call for mutual recursion
  private boolean isRecursiveView(SimpleNode node, String name, 
				  Hashtable htable, Environment env)
       throws InterpretError
  {
    int type = node.type, opcode = node.opcode;
    boolean flag = false; // flag indicating reference to name thus cycle
    
    switch(type) {
    case OP_IDENTIFIER : {
      if (node.name.equals(name)) { // there is cycle
	flag = true;
      } else { // continue to do semantic check
	// check if this ID is already checked
	String s = (String) htable.get(node.name);
	if (s != null) // already checked
	  return flag;
//	  System.out.println("looking for: " + node.name);

	if (node.url == null){ // added by Zongyan
	// remote view definition will be shipped to remote site.
    // aldatp://mimi/(v) is aldatp://mimi/(V) ijoin R 
	// is converted to aldatp://mimi/{ V is V ijoin aldatp://myhost/mydir/(R)}
	// so interpreter will nerver encounter such situation : url!= null and is recursive

	Relation rel = env.lookupRel(node.name, true);
	if (rel == null) {
	  throw new InterpretError("no relation : " + node.name 
				   + " in view " + name);
	}
	// REVISIT : computation
	if (rel.rvc == VIEW) {
	  SimpleNode view = rel.tree, nd = null;
	  // get the expression body of the view
	  if (view.isBitSet(0)) // there is initial part
	    nd = (SimpleNode) view.jjtGetChild(1);
	  else 
	    nd = (SimpleNode) view.jjtGetChild(0);
	  
	  // add this ID to hash table so no check for it again
	  htable.put(node.name, node.name);
	  
	  // recursively check and set flag 
	  flag = flag || isRecursiveView(nd, name, htable, env);
	}
      }
      }
      return flag;
    }
    case OP_JOINOPERATOR : {
      SimpleNode nodel = null, noder = null;
      
      // get expression tree of left operand
      nodel = (SimpleNode) node.jjtGetChild(0);
      
      // get expression tree of right operand
      if (node.jjtGetNumChildren() == 2) { // no join attributes 
	noder = (SimpleNode) node.jjtGetChild(1);
      } else { // node.jjtGetNumChildren() == 4
	noder = (SimpleNode) node.jjtGetChild(3);
      }
      
      flag = flag || isRecursiveView(nodel, name, htable, env);
      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    }
    case OP_PROJECT : {
      // get expression tree of relational expression
      SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
      
      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    }
    case OP_TSELECT : { 
      // get expression tree of the selector
      SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
      
      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    }
    case OP_SELECT : {
      // get relational expression tree of the selector
      SimpleNode noder = (SimpleNode) node.jjtGetChild(1);
      
      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    }
    case OP_QSELECT : {
      //this is modified by hongyu. Sept 30, 01
      //throw new InterpretError("TO BE IMPLEMENTED : QSelector");
      // get relational expression tree of the selector

      SimpleNode noder = null;
      if(node.jjtGetNumChildren() == 2)
	noder= (SimpleNode)node.jjtGetChild(1);
      noder = (SimpleNode)node.jjtGetChild(2);
      
      flag = flag || isRecursiveView(noder, name, htable, env);
      return flag;
    } // end of hongyu modification

	case OP_ARRAY: //Yi Zheng May 31, 2002
    case OP_PICK : {
      // get relational expression tree
      SimpleNode nd = (SimpleNode) node.jjtGetChild(0);
      
      flag = flag || isRecursiveView(nd, name, htable, env);
            flag = flag || isRecursiveView(nd, name,  htable, env);
      return flag;
    }
	 case OP_ALDATP : { // Zongyan Aug 2002 ^_^
     
      flag = flag || false;
      return flag;
    }
    case OP_EDIT : 
    case OP_ZORDER : 
    case OP_EVAL : // ???
    case OP_RENAME : { // ??? 
      throw new InterpretError("TO BE IMPLEMENTED");
    }
    // many more other case ...
    default : {
      throw new 
	InterpretError("invalid/not implemented relational expression");
    }
    }
    
  }  
  
     /**
   * method that  traverse and output the dump info to screen 
   * for debug purpose
   * @author Zongyan Wang July 19, 2002
   */
	public void traverse(SimpleNode node) {	  System.out.println("identifer:" + node.identifier + " name:" + node.name + " opcode:" + node.opcode + " type:" + node.type);
	  System.out.println("numChildren:" + node.jjtGetNumChildren() +" info:" + node.info + " url:" + node.url);
															 	  for(int i = 0; i < node.jjtGetNumChildren(); i++) {
      traverse((SimpleNode) node.jjtGetChild(i));
      }  }  
    
  /**
   * The main loop for the interpretation.
   * @param args java.lang.String[]
   */
  public void mainLoop(String args[]) {
    
    in = Global.instream;
    instack = new Stack();
    parser = new JRelixParser(Global.jin);
    
    Global.print(">"); Global.flush();
    
    //testing, Sept 29, 2002
    //executeJRelixCommand("pr (where .rel_name = \".rel\" in .rd);");
    
    //end of testing
    
    while (true) {
      try {
	Global.beforeParse();
	Global.status = parser.Start();
	if (Global.status == OP_EOF) {
		
		
	  if ((in instanceof FileInputStream)||(in
						instanceof ByteArrayInputStream)) {
							
	
	    try {
	      Object o = instack.pop();
	      if (o instanceof FileInputStream) {
		in.close();
		in = (FileInputStream) o;
	       } else 
	       {
		 in.close();
		 in = (InputStream) o;
	       }
	      Global.jin.resetIn(in);
	      parser.ReInit(Global.jin);
	    }
	     catch (Exception e) {
	      Global.println("Exception during IO redirection: reset to System.in for input");
	      in = Global.instream; 
	      Global.jin.resetIn(in);
	      parser.ReInit(Global.jin);
	    }
	  }

	} else {
	  Global.root = (SimpleNode) parser.jjtree.rootNode();
	  Global.afterParse(env);
	  Global.beforeInterpret();
	  Global.interpretTime = interpret(Global.root, env);
	  
	  Global.afterInterpret();
	}
      } catch (ParseError e) {
	Global.processThrowable(e);
	//previous code:	parser.ReInit(new JRelixInputStream(in));

	//Bug fix added by Yi Zheng March 19, 2002
	Global.jin = new JRelixInputStream(in);
	parser.ReInit(Global.jin);
      } catch (InterpretError e) {
	Global.processThrowable(e);
	//Bug fix added by Yi Zheng March 19, 2002
	Global.jin = new JRelixInputStream(in);
	parser.ReInit(Global.jin);
      } catch (Exception e) {
	Global.processThrowable(e);
	//Bug fix added by Yi Zheng March 19, 2002
	Global.jin = new JRelixInputStream(in);
	parser.ReInit(Global.jin);
      } catch (Error e) {
	Global.processThrowable(e);
	//Bug fix added by Yi Zheng March 19, 2002
	Global.jin = new JRelixInputStream(in);
	parser.ReInit(Global.jin);
      } finally {
	parser.jjtree.reset();
       }
    }
  }
  /***************************************************************/
  /*          METHODS USED FOR DECLARING A VIRTUAL DOMAIN        */
  /***************************************************************/
  
   /**
    * This method is used for virtual domains.
    *   e.g: let X be A+B;
    */
  public Domain newVirDomain(String name, SimpleNode node, Environment env) 
       throws InterpretError
  {  
    int domtype;
    SimpleNode declNode;
    
    if(node.jjtGetNumChildren() == 1)
      // if(node.jjtGetNumChildren() == 2)
    {  
	declNode = (SimpleNode)node.jjtGetChild(0);
	//		declNode = (SimpleNode)node.jjtGetChild(1);
    }
    else
    {  // Something must be wrong.
	throw new InterpretError("Domain: syntax error: <INITIAL> not supported.");
    }
    
    // First make sure that all referenced domains in the syntax tree
    // exist, otherwise, throw an exception.
    
    // May 2, 2002 Yi Zheng: now also checks if any of the reference
    //domains has isStateDom == 1, if so, throw an exception
    traverseNode(declNode, 0, env);
    
    // If this domain already exists, we just insert/replace the
    // syntax tree to that domain's 'tree' field. Note, the reference
    // counters of those referenced domains are not incremented. 
    // However, before we do some actual work, we must make sure
    // that old/new domains' types are identical, even when we want
    // to overwrite a virtual domain since we can't differentiate a 
    // true virtual from a 2-meaning domain item (i.e. normal/virtual
    // domain).
    Domain doment = env.lookupDom(name, false);
    
    // Get the actualized type. Note, this type might be changed 
    // later, so must be checked again at run-time.
    domtype  = traverseType(declNode, env);
 //   System.out.println(" end of traverseType");
    if(doment != null)
    {  // domain exists.
		if(doment.type != domtype)
	  	{  throw new InterpretError("error: new type of '" + doment.name
				      + "' is '" + Utility.type2String(domtype) 
				      + "', but the old type is '"
						+ Utility.type2String(doment.type) + "'.");
	  	}
		else
	  	{  
	    	doment.tree = node;
	    // Weizhong
	    //		doment.tree.jjtRemoveChild(0);
	  	}
    }
    else
    {  // this is a new domain
	//	   doment = new Domain(name, domtype, 0, node);
	//	Node removenode = node.jjtGetChild(0);
	//	node.jjtRemoveChild(0);
		doment = new Domain(name, domtype, 0, node);
	//	node.jjtAddChildAtHead(removenode);
    }
    
    //COMP added by Yi Zheng, March 20, 2002
    if(domtype == COMP) //r is only used here for sd purpuse
    { Domain[] doms = getIDList(declNode, env).domains;
      // Add it to relTable without idlist info.
      Relation r = new Relation("."+name, doms, RELATION, 0, 
				doms.length, 0, null, env);
      env.put(r);
    }
    else if(domtype == IDLIST) //Yi Zheng June 17, 2002
    {
	//Aug 13
       //Utility.dump(declNode, "declNode ");

		Relation rdecl = getIDList(declNode, env);
      	Domain[] doms = rdecl.domains;
      	Domain iddom = env.lookupDom(".id", true);
		Domain[] newDoms = null;
		if(!rdecl.isDomainIn(iddom))
		{
	  		newDoms = new Domain[doms.length + 1];
	  		newDoms[0] = iddom;
	  		for(int i = 1; i < newDoms.length; i++)
	  			newDoms[i] = doms[i-1];
		}
		else
		{
			newDoms = new Domain[doms.length];
			for(int j = 0; j < newDoms.length; j++)
				newDoms[j] = doms[j];
		}
      	Relation r = new Relation ("." + name, newDoms, RELATION, 0, newDoms.length, 0, null, env);
      	env.put(r);
    }
    return doment;
  }      
  /***************************************************************/
  /*                  OTHER SHORT UTILITY METHODS                */
  /***************************************************************/
  
  // Return the a temporary name used as the file name for intermediate 
  // result in relational operation
  public static String nextTempName() //Yi Zheng June 11, 2002
//public String nextTempName()
  {
    int n = (next_number++)%MAX_TEMP_FILE;
    return temp_name_prefix + Integer.toString(n);
  }      
  //REVISIT: It seems this is never used, what is it for?
  private void parse(String sp)
  {
    StringBufferInputStream sin = new StringBufferInputStream(sp);
  }      
  /***************************************************************/
  /*            METHODS USED FOR DECLARING A RELATION            */
  /***************************************************************/
  
  /**
   * Code to handle parsing of a relational declaration with 
   * optional initialization
   */
  private void RelationalDeclaration(SimpleNode node, Environment env)
   throws InterpretError
  {  
    // type of the new relation(s) 
    SimpleNode type = (SimpleNode) node.jjtGetChild(1);
    // typeArray is the array of domains for the new relation(s)
    Domain[] typeArray = new Domain[type.jjtGetNumChildren()];
    for(int i = 0; i < typeArray.length; i++) {
      String dname = ((SimpleNode) type.jjtGetChild(i)).name;
      // check if this domain is already declared
      Domain d = env.lookupDom(dname, true);
      if(d == null) {
	throw new 
	  InterpretError("domain "+ dname + " has not been declared");
      }
      else if (d.isStateDom == 1) //Yi Zheng May 9, 2002
      throw new InterpretError("domain " + dname + " is a state domain");
      typeArray[i] = d;
    }
    
    // add the new relations to the reltable
    SimpleNode rels = (SimpleNode) node.jjtGetChild(0);
    for(int i = 0; i < rels.jjtGetNumChildren(); i++) {
      String relname = ((SimpleNode) rels.jjtGetChild(i)).name;
      Relation r = new Relation(relname, typeArray, RELATION,
				0, typeArray.length, 0, null, env);
      env.put(r);
    }
    
    
    
    // perform initialization of the relation(s) if necessary
    if (node.jjtGetNumChildren() == 3) { 
      // node of tuples
      SimpleNode nd = (SimpleNode) node.jjtGetChild(2).jjtGetChild(0);
      
      // initialize the first relation
      String name = ((SimpleNode) rels.jjtGetChild(0)).name;
      Hashtable relsHT = CreateDataHashTable(name, env);
      RelationalInitialization(nd, relsHT, name, 0, 0);
      
      // append data for nested domains, note that there is no need to
      // check for duplicate in appending data for nested domains
      
      Vector dag = CreateDAG(name, env);
      
      for (int i = dag.size() - 1; i >= 0; i--) {
	String s = (String) dag.elementAt(i);
	Relation r = (Relation) relsHT.get(s);
	int tmp = r.numtuples;
	r.numtuples = r.capacity;
	r.capacity = tmp;
	r.project(null); // project to eliminate duplicate
	//System.out.println("processing duplicate in " + s);
      }
      
      for (int i = dag.size() - 1; i > 0; i--) {
	String s = (String) dag.elementAt(i);
	Relation r = (Relation) relsHT.get(s);
	int tmp = r.numtuples;
	r.numtuples = r.capacity;
	r.append(r.data, tmp);

	r.data = null; r.capacity = 0;
	r.numsortattrs = 0;
//	if (Global.debug) {
//	  Global.println(name);
//	  r.print();
//	}
	//	  Global.println("appending in " + s);
      }
      
      Relation rel = env.lookupRel(name, true);
      Object[] dataArray = rel.data;
      int tuples = rel.numtuples;
      int sort = rel.numsortattrs;
      rel.numtuples = 0;
      
      // append top level data to each relation, note the need to
      // check for duplicate tuples in initialization
      for (int i = 0; i < rels.jjtGetNumChildren(); i++) {
	name = ((SimpleNode) rels.jjtGetChild(i)).name;
	Relation r = env.lookupRel(name, true);
	r.append(dataArray, tuples);
	r.numsortattrs = sort;
	

	r.data = null; r.capacity = 0;
	if (Global.debug) {
	  Global.println(name);
	  r.print();
	}
      }
    }
	 //Global.syncAllToDisk();
  }                 
   /**
    * recursive routine called only by RelationalDeclaration()
    * to initialize a relation using the curly bracket syntax
   * START Since jRelixNS
   * Support empty nested relations.
   * Empty nested relations are all assigned with the Surrogate of value 0.
    */
  private long RelationalInitialization(SimpleNode node, Hashtable relsHT, 
					String name, long surr,  int tupleNum) //modified by Yi Zheng Sept 28, 2002
   throws InterpretError
  {  
    int type = node.type, opcode = node.opcode;
    
	 switch(type) {
	 case OP_CONSTANT :
	   switch(opcode) {
	   case OP_TUPLE :
	     //System.out.println("OP_TUPLE starts");
	     Relation rel = (Relation) relsHT.get(name);
	     if ( rel == null) { // should never happen
	       throw new 
		 InterpretError("Error: look up of relation " + name + " failed");
	     }
	     Domain[] doms = rel.domains;
	     
	     // do a semantic check to see if user gave right number
	     // of entries for a tuple (careful for ".id" field)
	     int numchld = node.jjtGetNumChildren();
	     if( (!name.startsWith(".") && (doms.length != numchld)) ||
		 (name.startsWith(".") && (doms.length != numchld+1)) ) 
	       {  
		 throw new InterpretError("Tuple " + tupleNum + " for relation " + name 
					  + " contains wrong number of attributes: " + numchld + 
					  ". The expected number of attributes is: " + 
					  ((!name.startsWith("."))?doms.length:(doms.length-1)));
	       }
	     
	     // find the entries of the tuple
	     // first, see if we have to put a ".id" entry
	     // if there is a ".id" field, then other domains are offset by 1
	     int offset = 0;
	     if(name.startsWith(".")) {
	       ((long[]) rel.data[0])[rel.capacity] = surr;
	       offset = 1;
	     }
	     
	     // now get all other entries of this tuple
	     for(int i = 0; i < node.jjtGetNumChildren(); i++) {
	       SimpleNode child = (SimpleNode) node.jjtGetChild(i);
	       if(child.type == OP_LITERAL) {
		 //System.out.println("start assignLiteral");
		 try{
		 assignLiteral(rel, i, offset, child);
		}catch (InterpretError ee)
		{
			throw new InterpretError("Tuple " + tupleNum + " of relation " + rel.name +
			" has invalid value for attribute " + (i+1) + "; message: " + ee.getMessage());	
		}
	       } else { // child is a nested domain
		 long result;
		 String newrel = (doms[i+offset].name.startsWith(".")) ?
		   doms[i+offset].name : "." + doms[i+offset].name; 
		 result = RelationalInitialization(child, relsHT, newrel, 0, 0);
		 ((long[]) rel.data[i + offset])[rel.capacity] = result;
	       }
	     }
	     rel.capacity++;
	     return 0;
	   case OP_TUPLELIST :
	     //System.out.println("OP_TUPLELIST starts");
	     // get a surrogate number if we are a nested domain
	     long s = 0;
//	     if (name.startsWith(".")) 
//	       s = Surrogate.next(); //get a surrogate
                 // START Since jRelixNS
                 // Support empty nested relations.
                 // Let surrogate remain 0 so they are always smaller
                 // than any other relation.
                 if (Relation.isSigmaJoinImplemented == true) {
		   if (name.startsWith("."))
                     if (node.jjtGetNumChildren() > 0)
		       s = Surrogate.next(); //get a surrogate
                 }
                 else
		   if (name.startsWith(".")) 
		     s = Surrogate.next(); //get a surrogate
                 // END Since jRelixNS //Andy, May 12, 2002

	     //System.out.println("num of children="+node.jjtGetNumChildren());
	     for(int i=0; i < node.jjtGetNumChildren(); i++) {
	       SimpleNode child = (SimpleNode) node.jjtGetChild(i);
	       RelationalInitialization(child, relsHT, name, s, i + 1);
	     }
	     return s;
	   } // end of second switch
	   
	   //YuLing added
	 case OP_FILEPATH:
	   String inputfilename=node.name.substring(1,node.name.length()-1);
	   //System.out.println("inputfilename="+inputfilename);
	   FileInputStream fin=null;
	   Relation rel = null;
	   try{
	     rel = (Relation) relsHT.get(name);
	     
	     fin=new FileInputStream(inputfilename);
	     DataInputStream dis=new DataInputStream(fin);
	     
	     char[] ch={06}; //separate character of relation file
	     String s=String.valueOf(ch);
	     
	     String str=null;
	     
	     //create the tuplelist node
	     SimpleNode tplistnd=new SimpleNode("Tuplelist");
	     tplistnd.set(OP_CONSTANT,OP_TUPLELIST,null);
	     tplistnd.jjtSetParent(node);
	     node.jjtAddChild(tplistnd);
	     
	     Domain[] doms = rel.domains;
	     int childnum=0;
	     int i = -1;
	     while((str=dis.readLine())!=null)//read the lines of the relation file
	       { 
	       childnum++;
	       try {
		 SimpleNode tpnd=new SimpleNode("Tuple");
		 tpnd.set(OP_CONSTANT,OP_TUPLE,null);
		 tpnd.jjtSetParent(tplistnd);
		 tplistnd.jjtAddChild(tpnd);
		 
	//	 StringTokenizer st= new StringTokenizer(str,s); 
	//modified by Yi Zheng as a bug fix: to allow empty string
	//Sept 22, 2002
	     
	     StringTokenizer st = new StringTokenizer(str,s,true);
	     boolean prevIsTok = true;
	 
	 	 
	 //end of bug fix Sept 22, 2002
	 
		 i=-1;
		 while (st.hasMoreTokens())
		   { 
		     String stringval=null;
		     SimpleNode literalnd=new SimpleNode("Literal");
		     String val=st.nextToken();
		     
		     //bug fix Sept 22, 2002
		     if (DEBUG_09_22_2002)
		     	System.out.println("val = " + val);
		     	
		     if (val.compareTo(s) == 0)
		     {
		     	if (!prevIsTok)
		     	{
		     		prevIsTok = true;
		     		continue;
		     	}
		     	else //empty string encountered
		     	{
		     		val = "";
		     		i++;
		     		prevIsTok = true;
		     	}	
		     }
		     else
		     {
		     	prevIsTok = false;
		     	i++;
		     }
		     //end of bug fix Sept 22, 2002
		     
   		     switch(doms[i].type)
		       { 
		       case INTEGER:
			 if(val.charAt(0)=='-')
			   stringval="-"+val.substring(1);
			 else
			   stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			 literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else 
			   {
			     literalnd.set(OP_LITERAL,OP_INTEGER,null);
			     literalnd.setObject(new Integer(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
		       case  BOOLEAN:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else 
			   {literalnd.set(OP_LITERAL,OP_BOOLEAN,null);
			   if(stringval.compareTo("1")==0)
			     literalnd.setObject(new Byte((byte)1));
			   else
			     literalnd.setObject(new Byte((byte)0));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
		       case SHORT:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else 
			   {literalnd.set(OP_LITERAL,OP_SHORT,null);
			   literalnd.setObject(new Short(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
		       case LONG:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else 
			   {literalnd.set(OP_LITERAL,OP_LONG,null);
			   literalnd.setObject(new Long(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
			  case NUMERIC://from Yiyi
          stringval = val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else
			   {literalnd.set(OP_LITERAL,OP_NUMERIC,null);
			   literalnd.setObject(new number(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;

		       case FLOAT:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else 
			   {literalnd.set(OP_LITERAL,OP_FLOAT,null);
			   literalnd.setObject(new Float(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
		       case DOUBLE:
			 stringval=val.substring(0);
			 if(((int)stringval.charAt(0))==05)//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if(((int)stringval.charAt(0))==02)
			   literalnd.set(OP_LITERAL,OP_DK,null);
			 else 
			   { literalnd.set(OP_LITERAL,OP_DOUBLE,null);
			   literalnd.setObject(new Double(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
			 tpnd.jjtAddChild(literalnd);
			 break;
			 case UNIVERSAL: //from Andrey
			case ATTRIBUTE: //from Andrey
		       case STRING:
		       case TEXT:
		       
		       //stringval = val.substring(0);
		       //bug fix Sept 22, 2002
		       if (val.length() > 0)
			 		stringval=val.substring(0);
			 	else
			 		stringval = new String("");
			 		//end of bug fix Sept 22, 2002
			 
			 //debugging Sept 22, 2002:
			 if (DEBUG_09_22_2002)
			 	System.out.println("stringval = " + stringval);
			 	
	//		 if(((int)stringval.charAt(0))==05)//dc
	//		   literalnd.set(OP_LITERAL,OP_DC,null);
	//		 else if(((int)stringval.charAt(0))==02)
	//		   literalnd.set(OP_LITERAL,OP_DK,null);
	//bug fix Sept 22, 2002
			   if((stringval.length() > 0) && (((int)stringval.charAt(0))==05))//dc
			   literalnd.set(OP_LITERAL,OP_DC,null);
			 else if((stringval.length() > 0) && (((int)stringval.charAt(0))==02))
			   literalnd.set(OP_LITERAL,OP_DK,null);
	//end of bug fix Sept 22, 2002
			 else 
			   {literalnd.set(OP_LITERAL,OP_STRING,null);
			   literalnd.setObject(new String(stringval));
			   }
			 literalnd.jjtSetParent(tpnd);
		        tpnd.jjtAddChild(literalnd);
			break;
		       default: 
			 System.out.println("unkown domain type:"+doms[i].type);
			 
			 throw new InterpretError("Error: Unexpected domain type");
		       }//end of switch
		   //  i++; //bug fix Sept 22, 2002
		     
		     //debuggin Sept 22, 2002
		     if (DEBUG_09_22_2002)
		     	System.out.println("tokens = " + i);
		   }
		}
		   catch (ArrayIndexOutOfBoundsException aie)
		   {
		   	 throw new InterpretError("Tuple " + (childnum + 1) + " contains too many attributes;" + 
		   	  " offending attribute position: " + (i + 1) + ". Expected number of attributes: " + doms.length);
		   }
	       }
	     
	     RelationalInitialization(tplistnd,relsHT,name,surr,0);
	   }
	   catch(IOException e){System.out.println("IOException in RelationalInitialization");}
	   //added by Yi Zheng Oct 2, 2002
	   if (rel != null)
	   	{	
	   		env.put(rel);
	   		rel.save();
	   	}
	   //end of addition Oct 2, 2002
	   	
	   return 0;
	   //end of YuLing added
	 case OP_IDENTIFIER://added by Yi Zheng March 23, 2002
	 throw new InterpretError("Initialization by Identifier not implemented yet.");
	 default :
	   //System.out.println("the type is"+type);
	   throw new InterpretError("Error: Unexpected SimpleNode");
	 }
  }      
 
  /**
   * Traverse the syntax tree of a virtual domain.
   * @param node root of the syntax tree
   * @param flag    0: don't change domains' reference counters
   *                1: increment domains' reference counters
   *                2: decrement domains' reference counters
   */
  private void traverseNode(SimpleNode node, int flag, Environment env)
       throws InterpretError
  {
    // REVISIT: have to check other cases e.g. node.name is
    // Rel name etc.
    
    if(node == null) return;
    
    if(node.type == OP_IDENTIFIER){
      Domain d = env.lookupDom(node.name, true);
      
      
      if(d != null){
	// if it's a declared domain...
	      //Yi Zheng May 2, 2002
			if (d.isStateDom == 1)
				throw new InterpretError("virtual domain defined on a state domain:" +
				d.name);
	
	if(flag == 1)
	  d.numref++;      // increment domain's numref
	else if(flag == 2)
	  d.numref--;      // decrement domain's numref
      }else{
	Relation r = env.lookupRel(node.name, true);
	if(r == null)
	  throw new InterpretError(
				   "domain '"+node.name +"' has not been declared.");
      }
    }
    
    // continue traversing...
    int numchildren = node.jjtGetNumChildren();
    if(numchildren != 0){
      // the source relation in projection and selection should not
      // be counted as domain. 
      //-- this is no longer true once we allow
      // state and local vars to be of type IDLIST, so the following 2 
      //lines are commented out by Yi Zheng May 2, 2002
//      if(node.type == OP_PROJECT)
//	numchildren--;
      
      // Traverse into the children.
      for(int i = 0; i < numchildren; i++){
	SimpleNode n = (SimpleNode)node.jjtGetChild(i);
	
	// For join operations, the operand might not be a domain.
	// For example, let x be S ijoin R; where R is a top-level
	// relation. Hence, we postpone the check alittle later.
	//-- for the same reason as above, the following if switch is commented
	//out by Yi Zheng, May 2, 2002
//	if(!(node.type == OP_JOINOPERATOR && n.type == OP_IDENTIFIER))
			traverseNode(n, flag, env);
      }
    }
  }      
  
  
  // Get the actualized type. Note, this type might be changed 
  // later, so must be checked again at run-time.
     /**
   * JRelixNS modified/enhanced this method:
   *  1) Check for empty exprlist.
   *  2) Support processing for non-scalars in OP_SELECT.
   *  3) Support processing of stand-alone top-level RELATION for sigma-join.
   */
  private int traverseType(SimpleNode node, Environment env)
       throws InterpretError
  {
    int i = 0;
    int[] typeArray;                    // An array of types
    int size = node.jjtGetNumChildren();
    boolean cond = false; //Added by hongyu. Sept 30,01
    SimpleNode  qtCond = null;//Added by hongyu. Sept 30,01
    
    switch(node.type){
   case OP_QUOTE://******************** QUOTE OPERATION ********************** //from Andrey
	return ATTRIBUTE;		 //????????????????????????
    case OP_RED:
      // There is only only child for Red operations.
    case OP_EQUIV:
      // For Equiv operations, we only need to take care of the first
      // child for type information.
     //Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
	if(node.opcode != OP_REDFUNCALL) 
      		return(traverseType((SimpleNode)node.jjtGetChild(0), env));
        else
		return(traverseType((SimpleNode)node.jjtGetChild(1), env));
    case OP_FUN://From sungsoo. hongyu added Feb 2002
      //System.out.println("OP_FUN: traverseType1");
	//Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
	if(node.opcode != OP_REDFUNCALL)
	      return (traverseType((SimpleNode)node.jjtGetChild(0),env));
	else
		return(traverseType((SimpleNode)node.jjtGetChild(1), env));
    case OP_PAR:
      //System.out.println("OP_PAR: traverseType1");
	//Yi Zheng July 8, 2002: if opcode is OP_REDFUNCALL, check second child
	if(node.opcode != OP_REDFUNCALL)
	      return (traverseType((SimpleNode)node.jjtGetChild(0),env));
	else
		return (traverseType((SimpleNode)node.jjtGetChild(1), env));
      //End of sungsoo's code. hongyu added Feb 2002
      
    case OP_IFTHENELSE:
      if(BOOLEAN == traverseType((SimpleNode)node.jjtGetChild(0),env)){
	int type = traverseType((SimpleNode)node.jjtGetChild(1), env);
	if(type == traverseType((SimpleNode)node.jjtGetChild(2), env))
	  return type;
	else
	  throw new InterpretError("type mismatch between 'then' clause"+
				   " and 'else' clause.");
      }else
	throw new InterpretError("type of 'if' clause is not boolean.");
    case OP_SELECT:
    case OP_PROJECT:
    case OP_TSELECT:
    case OP_QSELECT: // this is added by hongyu a <- quant (#>2) COMP in SUPPLY;  
      // This branch was for OP_SELECT & OP_PROJECT only. To fit for
      // OP_TSELECT, both SELECT & PROJECT are used. As for TSELECT,
      // the following tree structure is presented for
      //        let x be [A,B] where A=3 or B=2 in (S ijoin T);
      //                  TSelect
      //                  /     \
      //            Projector   Selector
      //                /        /    \
      //           ExprList    Or      Join
      //             /  \     /  \     /  \
      //            A    B  Comp Comp  S  T
      //                    /  \ /  \
      //                    A  3 B  2
      // First, get the source relation.
      Domain[] darray = null;
      Relation srcrel = null;
      //SimpleNode srcnode = (SimpleNode)node.jjtGetChild(1);
      SimpleNode srcnode = null; //revised by hongyu
      if((node.type == OP_PROJECT) ||(node.type == OP_SELECT)){
	srcnode = (SimpleNode)node.jjtGetChild(1);
      }
      else if(node.type == OP_TSELECT)// here is case of OP_TSELECT
	srcnode = (SimpleNode)node.jjtGetChild(1);

      else{//OP_QSELECT  a <- quant (#>2) COMP in SUPPLY; 
	if(node.jjtGetNumChildren()==2){
	  cond = true;
	  srcnode = (SimpleNode)node.jjtGetChild(1);
	}
	else{
	  qtCond = (SimpleNode)node.jjtGetChild(1);
	  srcnode = (SimpleNode)node.jjtGetChild(2);
	}
      }
   //   System.out.println(" srcnode.type "  + srcnode.type);
   //   System.out.println(" srcnode.name "  + srcnode.name); // End of Hongyu modification
      
      switch(srcnode.type){
      case OP_IDENTIFIER:
	String srcrelname = "."+srcnode.name;
	srcrel = env.lookupRel(srcrelname, true);
	
	// REVISIT: use NREnvironment to find the top-level rel.
	if(srcrel == null){
	  srcrel = env.lookupRel(srcnode.name, true);
	  if(srcrel == null)
	    throw new InterpretError("source relation" + srcnode.name + " not found.");
	}
	break;
      case OP_SELECT:
      case OP_PROJECT:
      case OP_TSELECT:
      case OP_JOINOPERATOR:
      case OP_QSELECT:
	// Firstly we go through the source relation tree (might be
	// a tree of join operations.
	int t = traverseType(srcnode, env);
	if(t != IDLIST && t != COMP)
	  // should never happen..
	  throw new InterpretError("mismatched type.");
	srcrel = getIDList(srcnode, env);
//	System.out.println("end of getIDList");
	break;
		   case OP_TRANSPOSE: //?????????????????????
		 return IDLIST; //from Andrey
		 //added by Yi Zheng July 7, 2002
	case OP_ARRAY:
	//	return IDLIST;
            break; //modified August 13, 2002
		//end of Yi's addition July 7, 2002
      default:
	throw new InterpretError("traverseType1: to be implemented.");
      }
   //   System.out.println("other.type "  + node.type);
      if(node.type == OP_PROJECT || node.type == OP_TSELECT)//Revised by hongyu
	{
	  // Get the projection domain list.
	  // START Since jRelixNS
	                   //  o Check for empty exprlist.
	  SimpleNode exprlist;
//	  if (env.isSigmaJoinImplemented != true) { //Andy, May 12, 2002
           if (Relation.isSigmaJoinImplemented != true) {
	    exprlist = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
	  } 
	  else {
                    // This is an enhancement to jRelix76.
	    // The exprlist might not exist for an empty projector, [],
	    // used for nullary tuple results.
	    exprlist = (SimpleNode)node.jjtGetChild(0);
	    if(exprlist.jjtGetNumChildren() > 0) // expect 1 or 0 only.
	      exprlist = (SimpleNode)node.jjtGetChild(0).jjtGetChild(0);
	  }
	  // END Since jRelixNS
	  Vector domvec = new Vector();
	  for(i = 0; i < exprlist.jjtGetNumChildren(); i++){
//In order to to allow for the following:
//let domB be [red + of a] in A;
//  we have to generate a temporary domain for red + of a
//  set domB's type to the temporary domain's type
//	this temporary domain is generated when actualized
// -- Yi Zheng July 1, 2002

		if((i == 0) && (exprlist.jjtGetNumChildren() == 1) 
			&& (((SimpleNode) exprlist.jjtGetChild(0)).type == OP_RED))
		{
			SimpleNode firstChild = null;
			if(((SimpleNode)exprlist.jjtGetChild(0)).opcode != OP_REDFUNCALL)
				firstChild = (SimpleNode) exprlist.jjtGetChild(0);
			else
				firstChild = (SimpleNode) exprlist.jjtGetChild(1);
	
			int redType = traverseType(firstChild, env);
			return redType;
		}
		else
		{
			String dname = ((SimpleNode)exprlist.jjtGetChild(i)).name;
		    Domain d = env.lookupDom(dname, true);
		    if(d == null) 
		      throw new InterpretError("domain "+dname+" not found.");
			
		    domvec.addElement(d);
		}
		/* original code //commented out July 1, 2002
		
	    String dname = ((SimpleNode)exprlist.jjtGetChild(i)).name;
	    Domain d = env.lookupDom(dname, true);
	    if(d == null) 
	      throw new InterpretError("domain "+dname+" not found.");
		
	    	domvec.addElement(d);
	    	*/
	    	
	  }
	  darray = new Domain[domvec.size()];
	  domvec.copyInto(darray);
	  
	  //Yi Zheng May 2, 2002
	  // "Newing" an actualizer will report any errors involved..
	  new Actualizer(darray, srcrel, env);

	  return IDLIST;
	}
      else if(node.type == OP_SELECT)// revised by hongyu
	{  
	 
	  // node.type == OP_SELECT
	                // START Since jRelixNS
              // o Support processing for non-scalars in OP_SELECT.
              if (Relation.isSigmaJoinImplemented != true) { //Andy, May 12, 2002
	  SimpleNode complist = (SimpleNode)node.jjtGetChild(0);
	  Hashtable htbl = new Hashtable();
	  getDomsInCompList(complist, htbl, env);
	  
	  darray = new Domain[htbl.size()];
	  int count = 0;
	  for(Enumeration e = htbl.keys(); e.hasMoreElements();)
	    darray[count++] = (Domain) htbl.get(e.nextElement());
	  		  // "Newing" an actualizer will report any errors involved..
	  new Actualizer(darray, srcrel, env);
	  
	  // Always return IDLIST type at this stage.
	  return IDLIST;

	                }else{
	         SimpleNode sn = new SimpleNode("virtmp");
	         sn.jjtAddChild((SimpleNode)node.jjtGetChild(0));
	         Domain d = new Domain("0select0", BOOLEAN, 0, sn);
	         Domain[] doms = new Domain[1];
	         doms[0] = d;
	         // "Newing" an actualizer will report any errors involved..
	         new Actualizer(doms, srcrel, env);
	         // Always return IDLIST type at this stage.
	         return IDLIST;
              }
              // END Since jRelixNS
	}


      else if(node.type == OP_QSELECT){// revised by hongyu
	return IDLIST;
      }
      else
	throw new InterpretError("node type is not decalred ");
    case OP_JOINOPERATOR:
      typeArray = new int[2];
      typeArray[0] = traverseType((SimpleNode)node.jjtGetChild(0), env);
      if(node.jjtGetNumChildren() == 2)
	typeArray[1] = traverseType((SimpleNode)node.jjtGetChild(1), env);
      else
	typeArray[1] = traverseType((SimpleNode)node.jjtGetChild(3), env);
      return(Utility.typeCheck(typeArray, node.opcode));
	case OP_TRANSPOSE: //?????????????????????
	 return IDLIST; //from Andrey
    case OP_FUNCTION: //Yi Zheng May 3, 2002
	 SimpleNode argument = (SimpleNode) node.jjtGetChild(0);
	 int argType = traverseType(argument, env);
	 if(argType == NUMERIC) return argType;//from Yiyi

     switch (node.opcode)
     {
		case OP_ABS:
			if(!Utility.isNumeric(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return argType;


		case OP_SQRT:
		case OP_SIN:
		case OP_ASIN:
		case OP_COS:
		case OP_ACOS:
		case OP_TAN:
		case OP_ATAN:		
		case OP_LOG:		
		case OP_LOG10:
			if(!Utility.isNumeric(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return DOUBLE;

		case OP_ROUND:
		case OP_CEIL:
		case OP_FLOOR:
			if(!Utility.isNumeric(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return INTEGER;
		case OP_ORD:
		
			if (!Utility.isText(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return INTEGER; //elevate to highest level of precision
		case OP_CHR:
			if (! Utility.isNumeric(argType))
				throw new InterpretError("function argument type mismatch");
			else
				return STRING;
				
		
		case OP_ISKNOWN:
		case OP_SINH:
		case OP_COSH:
		case OP_TANH:
			throw new InterpretError("Not implemented");
		default:
			throw new InterpretError("Unknown function operator");
     }
    case OP_ARRAY:
    	return IDLIST; //Yi Zheng June 19, 2002
    default:
      // START Since jRelixNS
  //    if (env.isSigmaJoinImplemented == true) {
	// Allow processing for sigma-join.
       if (Relation.isSigmaJoinImplemented == true) { //Andy, May 12, 2002
  
               // START Since jRelCaNu a.
             // OP_CAST is a OP_BIOPERATOR, but
             // the 2 children nodes should be explicitely 
             // checked for cast compatibility.
             // Also store the CastFromType in bits of the OP_TYPE node.
             if(node.opcode == OP_CAST) {// also (node.type == OP_BIOPERATOR)
                int CastToType = Utility.typeConvert(((SimpleNode)node.jjtGetChild(0)).opcode);
                int CastFromType = traverseType((SimpleNode)node.jjtGetChild(1), env);
                ((SimpleNode)node.jjtGetChild(0)).bits = CastFromType;
                if (!Utility.domsCastCompatible(CastToType, CastFromType)) 
                {
                   Utility.dump(node,"traverseType CAST Error ->");
                   throw new InterpretError("traverseType: Casting incompatible types:" 
                   + " CastToType: " + Utility.type2String(CastToType)
                   + " CastFromType: " + Utility.type2String(CastFromType));
                }
                return CastToType;
             }
             // END Since jRelCaNu a.

       // Support processing of stand-alone top-level RELATION for sigma-join
	Domain IdNodeDom;
	Relation IdNodeRel;
	
	if(size != 0){
	  typeArray = new int[size]; 
	  for (java.util.Enumeration e = node.children.elements();
	       e.hasMoreElements();) {
	    SimpleNode n = (SimpleNode)e.nextElement();
	    switch(n.type){
	    case OP_LITERAL:
	      typeArray[i++] = Utility.typeConvert(n.opcode);
	      break;
	    case OP_IDENTIFIER:
	      // Handle the relation case:
	      // REVISIT: need use 'true' for lookup?
	      //          and how about relation case?
	      // Old code below will cause a java null pointer exception
	      // in case we have allowd a top level relation node.
	      //typeArray[i++] = env.lookupDom(n.name, true).type;
	      if ((IdNodeDom = env.lookupDom(n.name, true)) != null) 
		typeArray[i++] = IdNodeDom.type; 
	      //typeArray[i++] = env.lookupDom(n.name, true).type;
	      else if ((IdNodeRel = env.lookupRel(n.name, true)) != null) 
		typeArray[i++] = RELATION; 
	      break;
	    case OP_BIOPERATOR: 

	    case OP_UOPERATOR:
	    case OP_JOINOPERATOR:
	                      //START Since JRelCaNu b.
                  // - to allow project and select in
                  //   nested relational expressions;
                  //   for example, in a OP_LET expression.
                  // Should have been in jRelixNS.
                  case OP_SELECT:
                  case OP_PROJECT:
                  case OP_TSELECT:
                  //End Since JRelCaNu b.

	      typeArray[i++] = traverseType(n, env);
	      break;
	    case OP_RED:
	    case OP_EQUIV:
		//Yi Zheng July 8, 2002, added if switch
		if(n.opcode != OP_REDFUNCALL)
	      typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(0), env);
		else
		typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(1), env);

	      break;
	    case OP_FUN: //sungsoo's code hongyu added Feb 2002
	    case OP_PAR:
	      //System.out.println("traverseType2");
		//Yi Zheng July 8, 2002, added if switch
		if(n.opcode != OP_REDFUNCALL)
	      typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(0),env);
		else
		typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(1),env);

	      break;
		
	    case OP_IFTHENELSE:
	      typeArray[i++] = traverseType(n, env);
	      break;
	    case OP_FUNCTION: //Yi Zheng May 6, 2002
	     typeArray[i++] = traverseType(n, env);
	      break;
	    default:
	      throw new InterpretError("traverseType2 4991: to be implemented.");
	    }
	  }
	  return(Utility.typeCheck(typeArray, node.opcode));
	}
	else{
	  if(node.type == OP_LITERAL)
	    // Node with children e.g. let b be 1;
	    return(Utility.typeConvert(node.opcode));
	  else 
	    // Node without children e.g. let b be a;
	    // REVISIT: need use 'true' for lookup?
	    // Added to handle the relation case:
	    {
	      if ((IdNodeDom = env.lookupDom(node.name, true)) != null) 
		return(IdNodeDom.type);
	      else if ((IdNodeRel = env.lookupRel(node.name, true)) != null)
	      { if (IdNodeRel.rvc == RELATION || IdNodeRel.rvc == VIEW)
				return(IDLIST);
			else return (COMP);//Yi Zheng May 2, 2002
		}
	      else
		throw new InterpretError("traverseType: OP_IDENTIFIER not LITERAL, DOM, or REL/COMPUTATION.");
	    }
	}
      } 
      else {
	if(size != 0){
	  typeArray = new int[size]; 
	  for (java.util.Enumeration e = node.children.elements();
	       e.hasMoreElements();) {
	    SimpleNode n = (SimpleNode)e.nextElement();
	    switch(n.type){
	    case OP_LITERAL:
	      typeArray[i++] = Utility.typeConvert(n.opcode);
	      break;
	    case OP_IDENTIFIER:
	      // REVISIT: need use 'true' for lookup?
	      //          and how about relation case?
	      typeArray[i++] = env.lookupDom(n.name, true).type;
	      break;
	    case OP_BIOPERATOR:
	    case OP_UOPERATOR:
	    case OP_JOINOPERATOR:
	      typeArray[i++] = traverseType(n, env);
	      break;
	    case OP_RED:
	    case OP_EQUIV:
		//Yi Zheng July 8, 2002 added if switch
		if(n.opcode != OP_REDFUNCALL)
	      typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(0), env);
		else
		typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(1), env);

	      break;
	      
	    case OP_FUN://sungsoo's code. hongyu added Feb 2002
	    case OP_PAR:
	      //System.out.println("traverseType2");
		//Yi Zheng July 8, 2002 added if switch
		if(n.opcode != OP_REDFUNCALL)
	      typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(0),env);
		else
		typeArray[i++] = traverseType((SimpleNode)n.jjtGetChild(1),env);

	      break;
	    case OP_IFTHENELSE:
	      typeArray[i++] = traverseType(n, env);
	      break;
	    default:
	      throw new InterpretError("traverseType2: to be implemented.");
	    }
	  }
	  return(Utility.typeCheck(typeArray, node.opcode));
	}else{
	  if(node.type == OP_LITERAL)
	    // Node with children e.g. let b be 1;
	    return(Utility.typeConvert(node.opcode));
	  else
	    // Node without children e.g. let b be a;
	    // REVISIT: need use 'true' for lookup?
	    //System.out.println(node.name + " of type " + Utility.type2String(env.lookupDom(node.name, true).type) + " about to be returned from traverseType()");
	    return(env.lookupDom(node.name, true).type);
	}
      }
    }      
  }
  
  // implemented by Weizhong Sun, August 1999
  private int executeEventH(String prefixname, SimpleNode enode, Environment env)
       throws InterpretError
  {
    String relname = new String("");
    String actionname = new String("");
    String attrlist = new String("");
    
    RelTable reltable = env.reltable;
    Hashtable reltb = reltable.getRelTable();
    //	int resetnested = 0;
    int ret = 0;
    
    // get the event query
    int opcode = enode.opcode;
    actionname = env.opcodeToString(opcode);
    relname = ((SimpleNode)enode.jjtGetChild(0)).name;
    SimpleNode stmtList = (SimpleNode) enode.jjtGetChild(1);
    int numStmts = stmtList.jjtGetNumChildren();
    // for Add/Delete, no attrlist
    if (opcode == OP_CHANGE) {
      for(int i = 0; i < numStmts; i++)
	{
	  SimpleNode stmt = (SimpleNode) stmtList.jjtGetChild(i);
	  String attr = ((SimpleNode) stmt.jjtGetChild(0)).name;
	  if(attrlist.compareTo("") == 0)
	    attrlist = attr;
	  else
	    attrlist = attrlist+","+attr;
	}
    }
    
    // and select the right event handler to execute
    for(Enumeration e = reltb.keys(); e.hasMoreElements();)
      {  
	Relation rel = (Relation)reltb.get(e.nextElement());
	String name = rel.name;
	
	if (rel.tree != null && rel.rvc == COMPUTATION && rel.active == 1) {
	  SimpleNode root = (SimpleNode) rel.tree;
	  
	  if (root.jjtGetNumChildren() != 0) {
	    // check the first child's opcode 
	    SimpleNode firstchild = (SimpleNode) root.jjtGetChild(0);
	    if (firstchild.opcode == OP_EVENTNAME) {
	      // cut name into ename and eattr
	      int bindex = name.indexOf('[');
	      int eindex;
	      String ename = new String("");
	      String eattr = new String("");
	      if (bindex != -1) {
		eindex = name.indexOf(']');
		ename = name.substring(0, bindex);
		eattr = name.substring(bindex+1, eindex);
	      } else 
		ename = name;
	      
	      // compare event name with prefixname, actionname, 
	      // relname, attrlist
	      String iname = prefixname+":"+actionname+":"+relname;
	      if (ename.compareTo(iname) == 0) {

		// check semantic of add/delete event handler
		if (((actionname.compareTo("add") == 0) ||
		     (actionname.compareTo("delete") == 0)) &&
		    (eattr.compareTo("") != 0))
		  throw new InterpretError("add/delete event handler semantic error.");
		
		// check the attribute list
		if (IsEmbody(eattr, attrlist)) {
		  
		  ret=1;
		  SimpleNode secondchild = (SimpleNode)root.jjtGetChild(1);
		  // execute Declaration body whose opcode is OP_COMPVARDELA
		  SimpleNode declnode = (SimpleNode) secondchild.jjtGetChild(0);
		  int declnum = declnode.jjtGetNumChildren();
		  for (int i = 0; i < declnum; i++) {
		    SimpleNode declstmt = (SimpleNode) declnode.jjtGetChild(i);
		    executeDeclaration(declstmt, env);
		  }
		  
		  // execute Computation body whose opcode is OP_COMPBLOCK
		  SimpleNode blocknode = (SimpleNode) secondchild.jjtGetChild(1);
		  int blocknum = blocknode.jjtGetNumChildren();
		  
		  for (int i = 0; i < blocknum; i++) {
		    SimpleNode blockstmt = (SimpleNode) blocknode.jjtGetChild(i);
		    
		    interpret(blockstmt, env, relname, opcode);
		  }
		  
		  /*
		    if (resetnested != 0) {
		    nested = resetnested; 
		    resetnested = 0;
		    }
		    */
		}
	      }
	    } // else skip
	  } else {
	    // never should be here
	    throw new InterpretError("computation root error.");
	  }
	}
	
      }
    
    return ret;
   }
  
  
  //written by YuLing Chen, Oct 28, 2000
  public boolean executeEventH( String actionname, Environment env)
       throws InterpretError
  {
    
    RelTable reltable = env.reltable;
    Hashtable reltb = reltable.getRelTable();
    boolean exist = false;//whether the event handler exists or not
    int opcode=env.stringToOpcode(actionname);
    
    // and select the right event handler to execute
    for(Enumeration e = reltb.keys(); e.hasMoreElements();)
      {  
	Relation rel = (Relation)reltb.get(e.nextElement());
	String name = rel.name;
	
	if (rel.tree != null && rel.rvc == COMPUTATION && rel.active == 1) {
	  SimpleNode root = (SimpleNode) rel.tree;
	  
	  if (root.jjtGetNumChildren() != 0) {
	    // check the first child's opcode 
	    SimpleNode firstchild = (SimpleNode) root.jjtGetChild(0);
	    if (firstchild.opcode == OP_EVENTNAME) {
	      // cut name into ename and eattr
	      int bindex = name.indexOf('[');
	      int eindex;
	      String ename = new String("");
	      String eattr = new String("");
	      if (bindex != -1) {
		eindex = name.indexOf(']');
		ename = name.substring(0, bindex);
		eattr = name.substring(bindex+1, eindex);
	      } else 
		ename = name;
	      
	      // compare event name with  actionname, 
	      String iname = actionname+":";
	      if (ename.compareTo(iname) == 0) {
		
		// check semantic of add/delete event handler
		if (((actionname.compareTo("add") == 0) ||
		     (actionname.compareTo("delete") == 0)) &&
		    (eattr.compareTo("") != 0))
		  throw new InterpretError("add/delete event handler semantic error.");
		
		// check the attribute list
			SimpleNode secondchild = (SimpleNode)root.jjtGetChild(1);
			// execute Declaration body whose opcode is OP_COMPVARDELA
			SimpleNode declnode = (SimpleNode) secondchild.jjtGetChild(0);
			int declnum = declnode.jjtGetNumChildren();
			for (int i = 0; i < declnum; i++) {
			  SimpleNode declstmt = (SimpleNode) declnode.jjtGetChild(i);
			  executeDeclaration(declstmt, env);
			}
			
			// execute Computation body whose opcode is OP_COMPBLOCK
			SimpleNode blocknode = (SimpleNode) secondchild.jjtGetChild(1);
			int blocknum = blocknode.jjtGetNumChildren();
			
			for (int i = 0; i < blocknum; i++) {
			  SimpleNode blockstmt = (SimpleNode) blocknode.jjtGetChild(i);
		//	  System.out.println("execute the computation");
			  
			  interpret(blockstmt, env );
			  exist=true;
			}
			
			/*
			  if (resetnested != 0) {
			  nested = resetnested; 
			  resetnested = 0;
			  }
			  */
	      }
	    } // else skip
	  } else {
	    // never should be here
	    throw new InterpretError("computation root error.");
	  }
	}
	
      }
    
    return exist;
  }
  // Weizhong, Aug. 1999
  // check if handAtt <= queAtt, for event handler's attribute list checking
  // 	handAtt: the attrlist of event handler
  //	queAtt: the attrlist of event query
  // example 1:
  // attrlist in event handler: [A,B]
  // attrlist in event query: [B]
  // then return false, so the event handler won't be active.
  // example 2:
  // attrlist in event handler: []
  // attrlist in event query: [B]
  // then return true, so the event handler will be active.
  public boolean IsEmbody(String handAtt, String queAtt) 
  {
    
    if (handAtt.compareTo("") == 0)
      return true;
    
    int bindex = handAtt.indexOf(',');
    if (bindex == -1) {
      // just one attribute in handAtt list
      if (queAtt.indexOf(handAtt) != -1 )
	return true;
      else 
	return false;
    } else {
      // several attribute in handAtt list
      // get first attribute in handAtt list
      String att = handAtt.substring(0, bindex);
      if (queAtt.indexOf(att) == -1)
	return false;
      bindex++;
      int nindex = handAtt.indexOf(',', bindex);
      while(nindex != -1) {
	// get more attributes
	att = handAtt.substring(bindex, nindex);
	if (queAtt.indexOf(att) == -1)
	  return false;
	bindex = nindex + 1;
	nindex = handAtt.indexOf(',', bindex);
      }
      // get the last one
      att = handAtt.substring(bindex);
      if (queAtt.indexOf(att) == -1)
	return false;
    }
    
    return true;
  }
  
  // Weizhong, to clear up trigger
  public void RemoveTrigger(TriggerNode triggernode) 
  {
    
    // remove the current level's trigger
    if (triggernode.trigger.relname.compareTo("") != 0) {
      triggernode.trigger.relname = new String("");
      if (triggernode.trigger.oldp.name.compareTo("") != 0) {
	deleteFile(triggernode.trigger.oldp.name);
	triggernode.trigger.oldp = null;
      }
      if (triggernode.trigger.newp.name.compareTo("") != 0) {
	deleteFile(triggernode.trigger.newp.name);
	triggernode.trigger.newp = null;
      }
      if (triggernode.trigger.restp.name.compareTo("") != 0) {
	deleteFile(triggernode.trigger.restp.name);
	triggernode.trigger.restp = null;
      }
    }
    
    // remove children's trigger
    for (int i = 0; i < triggernode.jjtGetNumChildren(); i++)
      RemoveTrigger((TriggerNode)triggernode.jjtGetChild(i));
    
    triggernode = null;
  }
  
  public void printNode(SimpleNode Node) 
  {
    System.out.println("Node.type="+Node.type);
    System.out.println("Node.name="+Node.name);
    System.out.println("Node.identifier="+Node.identifier);
    System.out.println("Node.opcode="+Node.opcode);
  }
  
  public  int executeRelixCommand(String commd)
  {
    byte[] barray=commd.getBytes();
    ByteArrayInputStream binstream=new ByteArrayInputStream(barray);
    InputStream bin=(InputStream)(binstream);
    Global.jin.resetIn(bin);
    parser.ReInit(Global.jin);
    if(in instanceof BufferedInputStream)instack.push(in);
    in=bin;
    try {
      Global.beforeParse();
      Global.status = parser.Start();
      if (Global.status == OP_EOF) {
	System.out.println("OP_EOF in executeRelix");
	if (in instanceof FileInputStream) {
	  try {
	    Object o = instack.pop();
	    if (o instanceof FileInputStream) {
	      in.close();
				in = (FileInputStream) o;
	    } else {
	      in.close();
	      in = (InputStream) o;
	    }
	    Global.jin.resetIn(in);
	    parser.ReInit(Global.jin);
	  } catch (Exception e) {
	    Global.println("Exception during IO redirection: reset to System.in for input");
	    in = Global.instream;
	    Global.jin.resetIn(in);
	    parser.ReInit(Global.jin);
	  }
	}
      } else {
	Global.root = (SimpleNode) parser.jjtree.rootNode();
	Global.afterParse(env);
	Global.beforeInterpret(); //nothing
	Global.interpretTime = interpret(Global.root, env);
	
	Global.afterInterpret();
      }
    } catch (ParseError e) {
      Global.processThrowable(e);
      parser.ReInit(new JRelixInputStream(in));
      return(-1);
    } catch (InterpretError e) {
      Global.processThrowable(e);
      parser.ReInit(new JRelixInputStream(in));
      return(-2);
    } catch (Exception e) {
      Global.processThrowable(e);
      parser.ReInit(new JRelixInputStream(in));
      return(-3);
    } catch (Error e) {
      Global.processThrowable(e);
      parser.ReInit(new JRelixInputStream(in));
      return(-4);
    } finally {
      parser.jjtree.reset();
    }
    return(0);
  }//end of executeRelixCommand
  
  //method added by Yi Zheng Sept 29, 2002 for srd <rel>
   public  void executeJRelixCommand(String commd)
  {
    byte[] barray=commd.getBytes();
    ByteArrayInputStream binstream=new ByteArrayInputStream(barray);
    InputStream bin=(InputStream)(binstream);
    Global.jin.resetIn(bin);
    parser.ReInit(Global.jin);
    instack.push(in);
    in=bin;
    try {
      Global.beforeParse();
      Global.status = parser.Start();
      if (Global.status == OP_EOF) {
	System.out.println("OP_EOF in executeJRelixCommand");

      } else {
	Global.root = (SimpleNode) parser.jjtree.rootNode();
	Global.afterParse(env);
	Global.beforeInterpret(); //nothing
	Global.interpretTime = interpret(Global.root, env);
	
	Global.afterSRDInterpret();
      }
    } catch (ParseError e) {
      Global.processThrowable(e);
    } catch (InterpretError e) {
      Global.processThrowable(e);
    } catch (Exception e) {
      Global.processThrowable(e);
    } catch (Error e) {
      Global.processThrowable(e);
    } finally {
    InputStream origIn = (InputStream) instack.pop();	
    Global.jin.resetIn(origIn);
    parser.ReInit(Global.jin);
    in = origIn;
      parser.jjtree.reset();
    }
  }//end of executeJRelixCommand 
  
  public Relation getRelbyName(String name) throws InterpretError
  {
    Relation rel = null;
    rel = env.lookupRel(name, true);
    rel.load();
    return rel;
  }
}













